@@ -1,17 +1,30 @@
 /** @file
- * innotek Portable Runtime - Types.
+ * IPRT - Types.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * The contents of this file may alternatively be used under the terms
+ * of the Common Development and Distribution License Version 1.0
+ * (CDDL) only, as it comes in the "COPYING.CDDL" file of the
+ * VirtualBox OSE distribution, in which case the provisions of the
+ * CDDL are applicable instead of those of the GPL.
+ *
+ * You may elect to license modified versions of this file under the
+ * terms and conditions of either the GPL or the CDDL or both.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 #ifndef ___iprt_types_h
@@ -27,7 +40,7 @@
 
 # if defined(RT_OS_DARWIN) && defined(KERNEL)
     /*
-     * Klugde for the darwin kernel:
+     * Kludge for the darwin kernel:
      *  stddef.h is missing IIRC.
      */
 #  ifndef _PTRDIFF_T
@@ -39,8 +52,8 @@
 # elif defined(RT_OS_FREEBSD) && defined(_KERNEL)
     /*
      * Kludge for the FreeBSD kernel:
-     *  stddef.h and sys/types.h has sligtly different offsetof definitions
-     *  when compiling in kernel mode. This is just to make GCC keep shut.
+     *  stddef.h and sys/types.h have slightly different offsetof definitions
+     *  when compiling in kernel mode. This is just to make GCC shut up.
      */
 #  ifndef _STDDEF_H_
 #   undef offsetof
@@ -91,12 +104,12 @@
 # endif
 
 #else /* no crt */
-# include <iprt/nocrt/compiler/gcc.h>
+# include <iprt/nocrt/compiler/compiler.h>
 #endif /* no crt */
 
 
 
-/** @defgroup grp_rt_types  innotek Portable Runtime Base Types
+/** @defgroup grp_rt_types  IPRT Base Types
  * @{
  */
 
@@ -164,7 +177,7 @@ typedef struct int128_s
 
 
 /**
- * 16-bit unsigned interger union.
+ * 16-bit unsigned integer union.
  */
 typedef union RTUINT16U
 {
@@ -182,14 +195,14 @@ typedef union RTUINT16U
         uint16_t    Hi;
     } s;
 } RTUINT16U;
-/** Pointer to a 16-bit unsigned interger union. */
+/** Pointer to a 16-bit unsigned integer union. */
 typedef RTUINT16U *PRTUINT16U;
-/** Pointer to a const 32-bit unsigned interger union. */
+/** Pointer to a const 32-bit unsigned integer union. */
 typedef const RTUINT16U *PCRTUINT16U;
 
 
 /**
- * 32-bit unsigned interger union.
+ * 32-bit unsigned integer union.
  */
 typedef union RTUINT32U
 {
@@ -215,14 +228,14 @@ typedef union RTUINT32U
     /** 8-bit view. */
     uint8_t     au8[4];
 } RTUINT32U;
-/** Pointer to a 32-bit unsigned interger union. */
+/** Pointer to a 32-bit unsigned integer union. */
 typedef RTUINT32U *PRTUINT32U;
-/** Pointer to a const 32-bit unsigned interger union. */
+/** Pointer to a const 32-bit unsigned integer union. */
 typedef const RTUINT32U *PCRTUINT32U;
 
 
 /**
- * 64-bit unsigned interger union.
+ * 64-bit unsigned integer union.
  */
 typedef union RTUINT64U
 {
@@ -258,19 +271,19 @@ typedef union RTUINT64U
     /** 8-bit view. */
     uint8_t     au8[8];
 } RTUINT64U;
-/** Pointer to a 64-bit unsigned interger union. */
+/** Pointer to a 64-bit unsigned integer union. */
 typedef RTUINT64U *PRTUINT64U;
-/** Pointer to a const 64-bit unsigned interger union. */
+/** Pointer to a const 64-bit unsigned integer union. */
 typedef const RTUINT64U *PCRTUINT64U;
 
 
 /**
- * 128-bit unsigned interger union.
+ * 128-bit unsigned integer union.
  */
 typedef union RTUINT128U
 {
     /** Natural view.
-     * WARNING! This member depends on compiler supporing 128-bit stuff. */
+     * WARNING! This member depends on the compiler supporting 128-bit stuff. */
     uint128_t   u;
     /** Hi/Low view. */
     struct
@@ -314,9 +327,9 @@ typedef union RTUINT128U
     /** 8-bit view. */
     uint8_t     au8[16];
 } RTUINT128U;
-/** Pointer to a 64-bit unsigned interger union. */
+/** Pointer to a 64-bit unsigned integer union. */
 typedef RTUINT128U *PRTUINT128U;
-/** Pointer to a const 64-bit unsigned interger union. */
+/** Pointer to a const 64-bit unsigned integer union. */
 typedef const RTUINT128U *PCRTUINT128U;
 
 
@@ -393,6 +406,10 @@ typedef const RTUINT   *PCRTUINT;
 typedef int64_t         RTFOFF;
 /** Pointer to a file offset / size. */
 typedef RTFOFF         *PRTFOFF;
+/** The max value for RTFOFF. */
+#define RTFOFF_MAX      INT64_MAX
+/** The min value for RTFOFF. */
+#define RTFOFF_MIN      INT64_MIN
 
 /** File mode (see iprt/fs.h). */
 typedef uint32_t        RTFMODE;
@@ -438,6 +455,8 @@ typedef uint16_t        RTSEL;
 typedef RTSEL          *PRTSEL;
 /** Pointer to const selector. */
 typedef const RTSEL    *PCRTSEL;
+/** Max selector value. */
+#define RTSEL_MAX       UINT16_MAX
 
 /** Far 16-bit pointer. */
 #pragma pack(1)
@@ -486,18 +505,24 @@ typedef const RTFAR64 *PCRTFAR64;
  * @{
  */
 
-/** HC Natural signed integer. */
+/** HC Natural signed integer.
+ * @deprecated silly type. */
 typedef int32_t         RTHCINT;
-/** Pointer to HC Natural signed integer. */
+/** Pointer to HC Natural signed integer.
+ * @deprecated silly type. */
 typedef RTHCINT        *PRTHCINT;
-/** Pointer to const HC Natural signed integer. */
+/** Pointer to const HC Natural signed integer.
+ * @deprecated silly type. */
 typedef const RTHCINT  *PCRTHCINT;
 
-/** HC Natural unsigned integer. */
+/** HC Natural unsigned integer.
+ * @deprecated silly type. */
 typedef uint32_t        RTHCUINT;
-/** Pointer to HC Natural unsigned integer. */
+/** Pointer to HC Natural unsigned integer.
+ * @deprecated silly type. */
 typedef RTHCUINT       *PRTHCUINT;
-/** Pointer to const HC Natural unsigned integer. */
+/** Pointer to const HC Natural unsigned integer.
+ * @deprecated silly type. */
 typedef const RTHCUINT *PCRTHCUINT;
 
 
@@ -509,9 +534,9 @@ typedef int64_t         RTHCINTPTR;
 #else
 #  error Unsupported HC_ARCH_BITS value.
 #endif
-/** Pointer to signed interger which can contain a HC pointer. */
+/** Pointer to signed integer which can contain a HC pointer. */
 typedef RTHCINTPTR     *PRTHCINTPTR;
-/** Pointer to const signed interger which can contain a HC pointer. */
+/** Pointer to const signed integer which can contain a HC pointer. */
 typedef const RTHCINTPTR *PCRTHCINTPTR;
 
 /** Signed integer which can contain a HC ring-3 pointer. */
@@ -522,9 +547,9 @@ typedef int64_t         RTR3INTPTR;
 #else
 #  error Unsupported R3_ARCH_BITS value.
 #endif
-/** Pointer to signed interger which can contain a HC ring-3 pointer. */
+/** Pointer to signed integer which can contain a HC ring-3 pointer. */
 typedef RTR3INTPTR     *PRTR3INTPTR;
-/** Pointer to const signed interger which can contain a HC ring-3 pointer. */
+/** Pointer to const signed integer which can contain a HC ring-3 pointer. */
 typedef const RTR3INTPTR *PCRTR3INTPTR;
 
 /** Signed integer which can contain a HC ring-0 pointer. */
@@ -535,9 +560,9 @@ typedef int64_t         RTR0INTPTR;
 #else
 #  error Unsupported R0_ARCH_BITS value.
 #endif
-/** Pointer to signed interger which can contain a HC ring-0 pointer. */
+/** Pointer to signed integer which can contain a HC ring-0 pointer. */
 typedef RTR0INTPTR     *PRTR0INTPTR;
-/** Pointer to const signed interger which can contain a HC ring-0 pointer. */
+/** Pointer to const signed integer which can contain a HC ring-0 pointer. */
 typedef const RTR0INTPTR *PCRTR0INTPTR;
 
 
@@ -549,9 +574,9 @@ typedef uint64_t        RTHCUINTPTR;
 #else
 #  error Unsupported HC_ARCH_BITS value.
 #endif
-/** Pointer to unsigned interger which can contain a HC pointer. */
+/** Pointer to unsigned integer which can contain a HC pointer. */
 typedef RTHCUINTPTR    *PRTHCUINTPTR;
-/** Pointer to unsigned interger which can contain a HC pointer. */
+/** Pointer to unsigned integer which can contain a HC pointer. */
 typedef const RTHCUINTPTR *PCRTHCUINTPTR;
 
 /** Unsigned integer which can contain a HC ring-3 pointer. */
@@ -562,9 +587,9 @@ typedef uint64_t        RTR3UINTPTR;
 #else
 #  error Unsupported R3_ARCH_BITS value.
 #endif
-/** Pointer to unsigned interger which can contain a HC ring-3 pointer. */
+/** Pointer to unsigned integer which can contain a HC ring-3 pointer. */
 typedef RTR3UINTPTR    *PRTR3UINTPTR;
-/** Pointer to unsigned interger which can contain a HC ring-3 pointer. */
+/** Pointer to unsigned integer which can contain a HC ring-3 pointer. */
 typedef const RTR3UINTPTR *PCRTR3UINTPTR;
 
 /** Unsigned integer which can contain a HC ring-0 pointer. */
@@ -575,15 +600,13 @@ typedef uint64_t        RTR0UINTPTR;
 #else
 #  error Unsupported R0_ARCH_BITS value.
 #endif
-/** Pointer to unsigned interger which can contain a HC ring-0 pointer. */
+/** Pointer to unsigned integer which can contain a HC ring-0 pointer. */
 typedef RTR0UINTPTR    *PRTR0UINTPTR;
-/** Pointer to unsigned interger which can contain a HC ring-0 pointer. */
+/** Pointer to unsigned integer which can contain a HC ring-0 pointer. */
 typedef const RTR0UINTPTR *PCRTR0UINTPTR;
 
 
-/** Host Physical Memory Address.
- * @todo    This should be configurable at compile time too...
- */
+/** Host Physical Memory Address. */
 typedef uint64_t        RTHCPHYS;
 /** Pointer to Host Physical Memory Address. */
 typedef RTHCPHYS       *PRTHCPHYS;
@@ -594,11 +617,11 @@ typedef const RTHCPHYS *PCRTHCPHYS;
  * NIL_RTHCPHYS is used to signal an invalid physical address, similar
  * to the NULL pointer.
  */
-#define NIL_RTHCPHYS     ((RTHCPHYS)~0U) /** @todo change this to (~(VBOXHCPHYS)0) */
+#define NIL_RTHCPHYS     (~(RTHCPHYS)0)
 
 
 /** HC pointer. */
-#ifndef IN_GC
+#ifndef IN_RC
 typedef void *          RTHCPTR;
 #else
 typedef RTHCUINTPTR     RTHCPTR;
@@ -690,18 +713,32 @@ typedef const RTR0UINTREG  *PCRTR0UINTREG;
  * @{
  */
 
-/** Natural signed integer in the GC. */
+/** Natural signed integer in the GC.
+ * @deprecated silly type. */
+#if GC_ARCH_BITS == 32
 typedef int32_t         RTGCINT;
-/** Pointer to natural signed interger in GC. */
+#elif GC_ARCH_BITS == 64 /** @todo this isn't right, natural int is 32-bit, see RTHCINT. */
+typedef int64_t         RTGCINT;
+#endif
+/** Pointer to natural signed integer in GC.
+ * @deprecated silly type. */
 typedef RTGCINT        *PRTGCINT;
-/** Pointer to const natural signed interger in GC. */
+/** Pointer to const natural signed integer in GC.
+ * @deprecated silly type. */
 typedef const RTGCINT  *PCRTGCINT;
 
-/** Natural signed uninteger in the GC. */
+/** Natural unsigned integer in the GC.
+ * @deprecated silly type. */
+#if GC_ARCH_BITS == 32
 typedef uint32_t        RTGCUINT;
-/** Pointer to natural unsigned interger in GC. */
+#elif GC_ARCH_BITS == 64 /** @todo this isn't right, natural int is 32-bit, see RTHCUINT. */
+typedef uint64_t        RTGCUINT;
+#endif
+/** Pointer to natural unsigned integer in GC.
+ * @deprecated silly type. */
 typedef RTGCUINT       *PRTGCUINT;
-/** Pointer to const natural unsigned interger in GC. */
+/** Pointer to const natural unsigned integer in GC.
+ * @deprecated silly type. */
 typedef const RTGCUINT *PCRTGCUINT;
 
 /** Signed integer which can contain a GC pointer. */
@@ -709,12 +746,10 @@ typedef const RTGCUINT *PCRTGCUINT;
 typedef int32_t         RTGCINTPTR;
 #elif GC_ARCH_BITS == 64
 typedef int64_t         RTGCINTPTR;
-#else
-#  error Unsupported GC_ARCH_BITS value.
 #endif
-/** Pointer to signed interger which can contain a GC pointer. */
+/** Pointer to signed integer which can contain a GC pointer. */
 typedef RTGCINTPTR     *PRTGCINTPTR;
-/** Pointer to const signed interger which can contain a GC pointer. */
+/** Pointer to const signed integer which can contain a GC pointer. */
 typedef const RTGCINTPTR *PCRTGCINTPTR;
 
 /** Unsigned integer which can contain a GC pointer. */
@@ -725,13 +760,27 @@ typedef uint64_t        RTGCUINTPTR;
 #else
 #  error Unsupported GC_ARCH_BITS value.
 #endif
-/** Pointer to unsigned interger which can contain a GC pointer. */
+/** Pointer to unsigned integer which can contain a GC pointer. */
 typedef RTGCUINTPTR     *PRTGCUINTPTR;
-/** Pointer to unsigned interger which can contain a GC pointer. */
+/** Pointer to unsigned integer which can contain a GC pointer. */
 typedef const RTGCUINTPTR *PCRTGCUINTPTR;
 
+/** Unsigned integer which can contain a 32 bits GC pointer. */
+typedef uint32_t        RTGCUINTPTR32;
+/** Pointer to unsigned integer which can contain a 32 bits GC pointer. */
+typedef RTGCUINTPTR32   *PRTGCUINTPTR32;
+/** Pointer to unsigned integer which can contain a 32 bits GC pointer. */
+typedef const RTGCUINTPTR32 *PCRTGCUINTPTR32;
+
+/** Unsigned integer which can contain a 64 bits GC pointer. */
+typedef uint64_t        RTGCUINTPTR64;
+/** Pointer to unsigned integer which can contain a 32 bits GC pointer. */
+typedef RTGCUINTPTR64   *PRTGCUINTPTR64;
+/** Pointer to unsigned integer which can contain a 32 bits GC pointer. */
+typedef const RTGCUINTPTR64 *PCRTGCUINTPTR64;
+
 /** Guest Physical Memory Address.*/
-typedef RTGCUINTPTR     RTGCPHYS;
+typedef uint64_t        RTGCPHYS;
 /** Pointer to Guest Physical Memory Address. */
 typedef RTGCPHYS       *PRTGCPHYS;
 /** Pointer to const Guest Physical Memory Address. */
@@ -739,35 +788,112 @@ typedef const RTGCPHYS *PCRTGCPHYS;
 /** @def NIL_RTGCPHYS
  * NIL GC Physical Address.
  * NIL_RTGCPHYS is used to signal an invalid physical address, similar
- * to the NULL pointer.
+ * to the NULL pointer. Note that this value may actually be valid in
+ * some contexts.
  */
-#define NIL_RTGCPHYS     ((RTGCPHYS)~0U) /** @todo change this to (~(RTGCPHYS)0) */
+#define NIL_RTGCPHYS     (~(RTGCPHYS)0U)
 
 
+/** Guest Physical Memory Address; limited to 32 bits.*/
+typedef uint32_t        RTGCPHYS32;
+/** Pointer to Guest Physical Memory Address. */
+typedef RTGCPHYS32     *PRTGCPHYS32;
+/** Pointer to const Guest Physical Memory Address. */
+typedef const RTGCPHYS32 *PCRTGCPHYS32;
+/** @def NIL_RTGCPHYS32
+ * NIL GC Physical Address.
+ * NIL_RTGCPHYS32 is used to signal an invalid physical address, similar
+ * to the NULL pointer. Note that this value may actually be valid in
+ * some contexts.
+ */
+#define NIL_RTGCPHYS32     (~(RTGCPHYS32)0)
+
+
+/** Guest Physical Memory Address; limited to 64 bits.*/
+typedef uint64_t        RTGCPHYS64;
+/** Pointer to Guest Physical Memory Address. */
+typedef RTGCPHYS64     *PRTGCPHYS64;
+/** Pointer to const Guest Physical Memory Address. */
+typedef const RTGCPHYS64 *PCRTGCPHYS64;
+/** @def NIL_RTGCPHYS64
+ * NIL GC Physical Address.
+ * NIL_RTGCPHYS64 is used to signal an invalid physical address, similar
+ * to the NULL pointer. Note that this value may actually be valid in
+ * some contexts.
+ */
+#define NIL_RTGCPHYS64     (~(RTGCPHYS64)0)
+
+/** Guest context pointer, 32 bits.
+ * Keep in mind that this type is an unsigned integer in
+ * HC and void pointer in GC.
+ */
+typedef RTGCUINTPTR32   RTGCPTR32;
+/** Pointer to a guest context pointer. */
+typedef RTGCPTR32      *PRTGCPTR32;
+/** Pointer to a const guest context pointer. */
+typedef const RTGCPTR32 *PCRTGCPTR32;
+/** @def NIL_RTGCPTR32
+ * NIL GC pointer.
+ */
+#define NIL_RTGCPTR32   ((RTGCPTR32)0)
+
+/** Guest context pointer, 64 bits.
+ */
+typedef RTGCUINTPTR64   RTGCPTR64;
+/** Pointer to a guest context pointer. */
+typedef RTGCPTR64      *PRTGCPTR64;
+/** Pointer to a const guest context pointer. */
+typedef const RTGCPTR64 *PCRTGCPTR64;
+/** @def NIL_RTGCPTR64
+ * NIL GC pointer.
+ */
+#define NIL_RTGCPTR64   ((RTGCPTR64)0)
+
 /** Guest context pointer.
  * Keep in mind that this type is an unsigned integer in
  * HC and void pointer in GC.
  */
-#ifdef IN_GC
-typedef void           *RTGCPTR;
-#else
-typedef RTGCUINTPTR     RTGCPTR;
-#endif
+#if GC_ARCH_BITS == 64
+typedef RTGCPTR64       RTGCPTR;
 /** Pointer to a guest context pointer. */
-typedef RTGCPTR        *PRTGCPTR;
+typedef PRTGCPTR64      PRTGCPTR;
 /** Pointer to a const guest context pointer. */
-typedef const RTGCPTR  *PCRTGCPTR;
+typedef PCRTGCPTR64    PCRTGCPTR;
 /** @def NIL_RTGCPTR
  * NIL GC pointer.
  */
-#define NIL_RTGCPTR     ((RTGCPTR)0)
-
+#define NIL_RTGCPTR     NIL_RTGCPTR64
+#elif GC_ARCH_BITS == 32
+typedef RTGCPTR32       RTGCPTR;
+/** Pointer to a guest context pointer. */
+typedef PRTGCPTR32      PRTGCPTR;
+/** Pointer to a const guest context pointer. */
+typedef PCRTGCPTR32    PCRTGCPTR;
+/** @def NIL_RTGCPTR
+ * NIL GC pointer.
+ */
+#define NIL_RTGCPTR     NIL_RTGCPTR32
+#else
+# error "Unsupported GC_ARCH_BITS!"
+#endif
 
 /** Unsigned integer register in the guest context. */
-#if GC_ARCH_BITS == 32
-typedef uint32_t            RTGCUINTREG;
-#elif GC_ARCH_BITS == 64
-typedef uint64_t            RTGCUINTREG;
+typedef uint32_t              RTGCUINTREG32;
+/** Pointer to an unsigned integer register in the guest context. */
+typedef RTGCUINTREG32        *PRTGCUINTREG32;
+/** Pointer to a const unsigned integer register in the guest context. */
+typedef const RTGCUINTREG32  *PCRTGCUINTREG32;
+
+typedef uint64_t              RTGCUINTREG64;
+/** Pointer to an unsigned integer register in the guest context. */
+typedef RTGCUINTREG64        *PRTGCUINTREG64;
+/** Pointer to a const unsigned integer register in the guest context. */
+typedef const RTGCUINTREG64  *PCRTGCUINTREG64;
+
+#if GC_ARCH_BITS == 64
+typedef RTGCUINTREG64       RTGCUINTREG;
+#elif GC_ARCH_BITS == 32
+typedef RTGCUINTREG32       RTGCUINTREG;
 #else
 # error "Unsupported GC_ARCH_BITS!"
 #endif
@@ -778,6 +904,53 @@ typedef const RTGCUINTREG  *PCRTGCUINTREG;
 
 /** @} */
 
+/** @defgroup grp_rt_types_rc  Raw mode Context Basic Types
+ * @ingroup grp_rt_types
+ * @{
+ */
+
+/** Raw mode context pointer; a 32 bits guest context pointer.
+ * Keep in mind that this type is an unsigned integer in
+ * HC and void pointer in RC.
+ */
+#ifdef IN_RC
+typedef void *          RTRCPTR;
+#else
+typedef uint32_t        RTRCPTR;
+#endif
+/** Pointer to a raw mode context pointer. */
+typedef RTRCPTR        *PRTRCPTR;
+/** Pointer to a const raw mode context pointer. */
+typedef const RTRCPTR  *PCRTRCPTR;
+/** @def NIL_RTGCPTR
+ * NIL RC pointer.
+ */
+#define NIL_RTRCPTR    ((RTRCPTR)0)
+/** @def RTRCPTR_MAX
+ * The maximum value a RTRCPTR can have. Mostly used as INVALID value.
+ */
+#define RTRCPTR_MAX    ((RTRCPTR)UINT32_MAX)
+
+/** Raw mode context pointer, unsigned integer variant. */
+typedef int32_t         RTRCINTPTR;
+/** @def RTRCUINPTR_MAX
+ * The maximum value a RTRCUINPTR can have.
+ */
+#define RTRCUINTPTR_MAX ((RTRCUINTPTR)UINT32_MAX)
+
+/** Raw mode context pointer, signed integer variant. */
+typedef uint32_t        RTRCUINTPTR;
+/** @def RTRCINPTR_MIN
+ * The minimum value a RTRCINPTR can have.
+ */
+#define RTRCINTPTR_MIN ((RTRCINTPTR)INT32_MIN)
+/** @def RTRCINPTR_MAX
+ * The maximum value a RTRCINPTR can have.
+ */
+#define RTRCINTPTR_MAX ((RTRCINTPTR)INT32_MAX)
+
+/** @} */
+
 
 /** @defgroup grp_rt_types_cc  Current Context Basic Types
  * @ingroup grp_rt_types
@@ -785,7 +958,7 @@ typedef const RTGCUINTREG  *PCRTGCUINTREG;
  */
 
 /** Current Context Physical Memory Address.*/
-#ifdef IN_GC
+#ifdef IN_RC
 typedef RTGCPHYS RTCCPHYS;
 #else
 typedef RTHCPHYS RTCCPHYS;
@@ -799,7 +972,7 @@ typedef const RTCCPHYS *PCRTCCPHYS;
  * NIL_RTCCPHYS is used to signal an invalid physical address, similar
  * to the NULL pointer.
  */
-#ifdef IN_GC
+#ifdef IN_RC
 # define NIL_RTCCPHYS   NIL_RTGCPHYS
 #else
 # define NIL_RTCCPHYS   NIL_RTHCPHYS
@@ -807,23 +980,27 @@ typedef const RTCCPHYS *PCRTCCPHYS;
 
 /** Unsigned integer register in the current context. */
 #if ARCH_BITS == 32
-typedef uint32_t            RTCCUINTREG;
+typedef uint32_t                RTCCUINTREG;
 #elif ARCH_BITS == 64
-typedef uint64_t            RTCCUINTREG;
+typedef uint64_t                RTCCUINTREG;
 #else
 # error "Unsupported ARCH_BITS!"
 #endif
 /** Pointer to an unsigned integer register in the current context. */
-typedef RTCCUINTREG        *PRTCCUINTREG;
+typedef RTCCUINTREG            *PRTCCUINTREG;
 /** Pointer to a const unsigned integer register in the current context. */
-typedef const RTCCUINTREG  *PCRTCCUINTREG;
+typedef RTCCUINTREG const      *PCRTCCUINTREG;
 
-/** @deprecated */
-typedef RTCCUINTREG         RTUINTREG;
-/** @deprecated */
-typedef RTCCUINTREG        *PRTUINTREG;
-/** @deprecated */
-typedef const RTCCUINTREG  *PCRTUINTREG;
+/** Signed integer register in the current context. */
+#if ARCH_BITS == 32
+typedef int32_t                 RTCCINTREG;
+#elif ARCH_BITS == 64
+typedef int64_t                 RTCCINTREG;
+#endif
+/** Pointer to a signed integer register in the current context. */
+typedef RTCCINTREG             *PRTCCINTREG;
+/** Pointer to a const signed integer register in the current context. */
+typedef RTCCINTREG const       *PCRTCCINTREG;
 
 /** @} */
 
@@ -836,7 +1013,7 @@ typedef RTFILE                                     *PRTFILE;
 #define NIL_RTFILE                                  (~(RTFILE)0)
 
 /** Loader module handle. */
-typedef HCPTRTYPE(struct RTLDRMODINTERNAL *)        RTLDRMOD;
+typedef R3PTRTYPE(struct RTLDRMODINTERNAL *)        RTLDRMOD;
 /** Pointer to a loader module handle. */
 typedef RTLDRMOD                                   *PRTLDRMOD;
 /** Nil loader module handle. */
@@ -872,7 +1049,7 @@ typedef RTR0PROCESS                                *PRTR0PROCESS;
 
 /** @typedef RTSEMEVENT
  * Event Semaphore handle. */
-typedef HCPTRTYPE(struct RTSEMEVENTINTERNAL *)      RTSEMEVENT;
+typedef R3R0PTRTYPE(struct RTSEMEVENTINTERNAL *)    RTSEMEVENT;
 /** Pointer to an event semaphore handle. */
 typedef RTSEMEVENT                                 *PRTSEMEVENT;
 /** Nil event semaphore handle. */
@@ -880,7 +1057,7 @@ typedef RTSEMEVENT                                 *PRTSEMEVENT;
 
 /** @typedef RTSEMEVENTMULTI
  * Event Multiple Release Semaphore handle. */
-typedef HCPTRTYPE(struct RTSEMEVENTMULTIINTERNAL *) RTSEMEVENTMULTI;
+typedef R3R0PTRTYPE(struct RTSEMEVENTMULTIINTERNAL *) RTSEMEVENTMULTI;
 /** Pointer to an event multiple release semaphore handle. */
 typedef RTSEMEVENTMULTI                            *PRTSEMEVENTMULTI;
 /** Nil multiple release event semaphore handle. */
@@ -888,7 +1065,7 @@ typedef RTSEMEVENTMULTI                            *PRTSEMEVENTMULTI;
 
 /** @typedef RTSEMFASTMUTEX
  * Fast mutex Semaphore handle. */
-typedef HCPTRTYPE(struct RTSEMFASTMUTEXINTERNAL *)  RTSEMFASTMUTEX;
+typedef R3R0PTRTYPE(struct RTSEMFASTMUTEXINTERNAL *) RTSEMFASTMUTEX;
 /** Pointer to a mutex semaphore handle. */
 typedef RTSEMFASTMUTEX                             *PRTSEMFASTMUTEX;
 /** Nil fast mutex semaphore handle. */
@@ -896,7 +1073,7 @@ typedef RTSEMFASTMUTEX                             *PRTSEMFASTMUTEX;
 
 /** @typedef RTSEMMUTEX
  * Mutex Semaphore handle. */
-typedef HCPTRTYPE(struct RTSEMMUTEXINTERNAL *)      RTSEMMUTEX;
+typedef R3R0PTRTYPE(struct RTSEMMUTEXINTERNAL *)    RTSEMMUTEX;
 /** Pointer to a mutex semaphore handle. */
 typedef RTSEMMUTEX                                 *PRTSEMMUTEX;
 /** Nil mutex semaphore handle. */
@@ -904,14 +1081,14 @@ typedef RTSEMMUTEX                                 *PRTSEMMUTEX;
 
 /** @typedef RTSEMRW
  * Read/Write Semaphore handle. */
-typedef HCPTRTYPE(struct RTSEMRWINTERNAL *)         RTSEMRW;
+typedef R3R0PTRTYPE(struct RTSEMRWINTERNAL *)       RTSEMRW;
 /** Pointer to a read/write semaphore handle. */
 typedef RTSEMRW                                    *PRTSEMRW;
 /** Nil read/write semaphore handle. */
 #define NIL_RTSEMRW                                 0
 
 /** Spinlock handle. */
-typedef HCPTRTYPE(struct RTSPINLOCKINTERNAL *)      RTSPINLOCK;
+typedef R3R0PTRTYPE(struct RTSPINLOCKINTERNAL *)    RTSPINLOCK;
 /** Pointer to a spinlock handle. */
 typedef RTSPINLOCK                                 *PRTSPINLOCK;
 /** Nil spinlock handle. */
@@ -925,29 +1102,87 @@ typedef RTSOCKET *PRTSOCKET;
 #define NIL_RTSOCKET                                (~(RTSOCKET)0)
 
 /** Thread handle.*/
-typedef HCPTRTYPE(struct RTTHREADINT *)             RTTHREAD;
+typedef R3R0PTRTYPE(struct RTTHREADINT *)           RTTHREAD;
 /** Pointer to thread handle. */
 typedef RTTHREAD                                   *PRTTHREAD;
 /** Nil thread handle. */
 #define NIL_RTTHREAD                                0
 
+/** A TLS index. */
+typedef RTHCINTPTR                                  RTTLS;
+/** Pointer to a TLS index. */
+typedef RTTLS                                      *PRTTLS;
+/** Pointer to a const TLS index. */
+typedef RTTLS const                                *PCRTTLS;
+/** NIL TLS index value. */
+#define NIL_RTTLS                                   (-1)
+
 /** Handle to a simple heap. */
-typedef HCPTRTYPE(struct RTHEAPSIMPLEINTERNAL *)    RTHEAPSIMPLE;
+typedef R3R0PTRTYPE(struct RTHEAPSIMPLEINTERNAL *)  RTHEAPSIMPLE;
 /** Pointer to a handle to a simple heap. */
 typedef RTHEAPSIMPLE                               *PRTHEAPSIMPLE;
 /** NIL simple heap handle. */
 #define NIL_RTHEAPSIMPLE                            ((RTHEAPSIMPLE)0)
 
 /** Handle to an environment block. */
-typedef HCPTRTYPE(struct RTENVINTERNAL *)           RTENV;
+typedef R3PTRTYPE(struct RTENVINTERNAL *)           RTENV;
 /** Pointer to a handle to an environment block. */
 typedef RTENV                                      *PRTENV;
 /** NIL simple heap handle. */
 #define NIL_RTENV                                   ((RTENV)0)
 
+/** A CPU identifier.
+ * @remarks This doesn't have to correspond to the APIC ID (intel/amd). Nor
+ *          does it have to correspond to the bits in the affinity mask, at
+ *          least not until we've sorted out Windows NT. */
+typedef uint32_t                                    RTCPUID;
+/** Pointer to a CPU identifier. */
+typedef RTCPUID                                    *PRTCPUID;
+/** Pointer to a const CPU identifier. */
+typedef RTCPUID const                              *PCRTCPUID;
+/** Nil CPU Id. */
+#define NIL_RTCPUID                                 ((RTCPUID)~0)
+
+/** A CPU set.
+ * Treat this as an opaque type and always use RTCpuSet* for manupulating it.
+ * @remarks Subject to change. */
+typedef uint64_t                                    RTCPUSET;
+/** Pointer to a CPU set. */
+typedef RTCPUSET                                   *PRTCPUSET;
+/** Pointer to a const CPU set. */
+typedef RTCPUSET const                             *PCRTCPUSET;
+
+/** A handle table handle. */
+typedef struct RTHANDLETABLEINT                    *RTHANDLETABLE;
+/** A pointer to a handle table handle. */
+typedef RTHANDLETABLE                              *PRTHANDLETABLE;
+/** @def NIL_RTHANDLETABLE
+ * NIL handle table handle. */
+#define NIL_RTHANDLETABLE                           ((RTHANDLETABLE)0)
+
+/** A handle to a low resolution timer. */
+typedef struct RTTIMERLRINT                        *RTTIMERLR;
+/** A pointer to a low resolution timer handle. */
+typedef RTTIMERLR                                  *PRTTIMERLR;
+/** @def NIL_RTTIMERLR
+ * NIL low resolution timer handle value. */
+#define NIL_RTTIMERLR                               ((RTTIMERLR)0)
+
+/** Handle to a random number generator. */
+typedef struct RTRANDINT                           *RTRAND;
+/** Pointer to a random number generator handle. */
+typedef RTRAND                                     *PRTRAND;
+/** NIL random number genrator handle value. */
+#define NIL_RTRAND                                  ((RTRAND)0)
+
 
 /**
  * UUID data type.
+ *
+ * @note IPRT defines that the first three integers in the @c Gen struct
+ * interpretation are in little endian representation. This is different to
+ * many other UUID implementation, and requires conversion if you need to
+ * achieve consistent results.
  */
 typedef union RTUUID
 {
@@ -959,17 +1194,16 @@ typedef union RTUUID
     uint32_t    au32[4];
     /** 64-bit view. */
     uint64_t    au64[2];
-    /** The way the UUID is declared by the ext2 guys. */
+    /** The way the UUID is declared by the DCE specification. */
     struct
     {
         uint32_t    u32TimeLow;
         uint16_t    u16TimeMid;
         uint16_t    u16TimeHiAndVersion;
-        uint16_t    u16ClockSeq;
+        uint8_t     u8ClockSeqHiAndReserved;
+        uint8_t     u8ClockSeqLow;
         uint8_t     au8Node[6];
     } Gen;
-    /** @deprecated */
-    unsigned char aUuid[16];
 } RTUUID;
 /** Pointer to UUID data. */
 typedef RTUUID *PRTUUID;
@@ -1003,9 +1237,9 @@ typedef const RTUNICP  *PCRTUNICP;
  * UTF-16 character.
  * @remark  wchar_t is not usable since it's compiler defined.
  * @remark  When we use the term character we're not talking about unicode code point, but
- *          the basic unit of the string encoding. Thus cuc - count of unicode chars - means
- *          count of RTUTF16. And cch means count of the typedef 'char', which is assumed
- *          to be an octet.
+ *          the basic unit of the string encoding. Thus cwc - count of wide chars - means
+ *          count of RTUTF16; cuc - count of unicode chars - means count of RTUNICP;
+ *          and cch means count of the typedef 'char', which is assumed to be an octet.
  */
 typedef uint16_t        RTUTF16;
 /** Pointer to a UTF-16 character. */
@@ -1013,22 +1247,6 @@ typedef RTUTF16        *PRTUTF16;
 /** Pointer to a const UTF-16 character. */
 typedef const RTUTF16  *PCRTUTF16;
 
-/**
- * UCS-2 character.
- * @remark wchar_t is not usable since it's compiler defined.
- * @deprecated  Use RTUTF16!
- */
-typedef RTUTF16         RTUCS2;
-/** Pointer to UCS-2 character.
- * @deprecated  Use PRTUTF16!
- */
-typedef PRTUTF16        PRTUCS2;
-/** Pointer to const UCS-2 character.
- * @deprecated  Use PCRTUTF16!
- */
-typedef PCRTUTF16       PCRTUCS2;
-
-
 
 /**
  * Wait for ever if we have to.
@@ -1067,6 +1285,28 @@ typedef RTRECT *PRTRECT;
 /** Pointer to a const rectangle. */
 typedef const RTRECT *PCRTRECT;
 
+
+/**
+ * Ethernet MAC address.
+ *
+ * The first 24 bits make up the Organisationally Unique Identifier (OUI),
+ * where the first bit (little endian) indicates multicast (set) / unicast,
+ * and the second bit indicates locally (set) / global administered. If all
+ * bits are set, it's a broadcast.
+ */
+typedef union RTMAC
+{
+    /** @todo add a bitfield view of this stuff. */
+    /** 8-bit view. */
+    uint8_t     au8[6];
+    /** 16-bit view. */
+    uint16_t    au16[3];
+} RTMAC;
+/** Pointer to a MAC address. */
+typedef RTMAC *PRTMAC;
+/** Pointer to a readonly MAC address. */
+typedef const RTMAC *PCRTMAC;
+
 /** @} */
 
 #endif
