@@ -3,20 +3,27 @@
  * VirtualBox Windows NT/2000/XP guest video driver
  *
  * VRDP graphics orders pipeline.
- *
- * Copyright (C) 2006-2007 innotek GmbH
+ */
+
+/*
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 #include "driver.h"
 #include "vrdpbmp.h"
+#include <VBox/VRDPOrders.h>
 
 #define VRDP_MAKE_OP(__c) (__c)
 
@@ -40,13 +47,6 @@ typedef struct _VRDPBRUSH
    } u;
 } VRDPBRUSH;
 
-typedef struct _VRDPCLIPRECTS
-{
-    RECTL rclDstOrig; /* Original bounding rectancle. */
-    RECTL rclDst;     /* Bounding rectangle of all rects. */
-    CLIPRECTS rects;  /* Rectangles to update. */
-} VRDPCLIPRECTS;
-
 #define VRDP_CLIP_OK              0
 #define VRDP_CLIP_NO_INTERSECTION 1
 #define VRDP_CLIP_TOO_MANY_RECTS  2
@@ -168,7 +168,7 @@ void vrdpAdjustRect (SURFOBJ *pso, RECTL *prcl)
     int w;
     int h;
 
-    DISPDBG((1, "vrdpAdjustRect: %d-%d %d-%d\n", prcl->left, prcl->right, prcl->top, prcl->bottom));
+    DISPDBG((1, "vrdpAdjustRect: %d-%d %d-%d on %dx%d\n", prcl->left, prcl->right, prcl->top, prcl->bottom, pso->sizlBitmap.cx, pso->sizlBitmap.cy));
 
     if (prcl->left <= prcl->right)
     {
@@ -230,6 +230,8 @@ void vrdpAdjustRect (SURFOBJ *pso, RECTL *prcl)
     prcl->top    = y;
     prcl->right  = x + w;
     prcl->bottom = y + h;
+
+    DISPDBG((1, "vrdpAdjustRect: result %d-%d %d-%d\n", prcl->left, prcl->right, prcl->top, prcl->bottom));
 }
 
 static int vrdpGetIntersectingClipRects (VRDPCLIPRECTS *pClipRects, SURFOBJ *pso, RECTL *prcl, CLIPOBJ *pco, POINTL *pptlSrc)
@@ -435,7 +437,7 @@ void vrdpReportDirtyRect (PPDEV ppdev, RECTL *prcl)
 
     if (bytesPerPixel > 0)
     {
-        bRc = vrdpWriteHdr (ppdev, VRDP_MAKE_OP(VBVA_VRDP_DIRTY_RECT));
+        bRc = vrdpWriteHdr (ppdev, VRDP_MAKE_OP(VRDP_ORDER_DIRTY_RECT));
 
         if (bRc)
         {
@@ -518,7 +520,7 @@ static BOOL vrdpReportBounds (PPDEV ppdev,
     bounds.pt2.x = (int16_t)(prcl->right);
     bounds.pt2.y = (int16_t)(prcl->bottom);
 
-    return vrdpReportOrder (ppdev, &bounds, sizeof (bounds), VBVA_VRDP_BOUNDS);
+    return vrdpReportOrder (ppdev, &bounds, sizeof (bounds), VRDP_ORDER_BOUNDS);
 }
 
 static BOOL vrdpReportRepeat (PPDEV ppdev,
@@ -541,7 +543,7 @@ static BOOL vrdpReportRepeat (PPDEV ppdev,
             repeat.bounds.pt2.x = (int16_t)(prcl->right);
             repeat.bounds.pt2.y = (int16_t)(prcl->bottom);
 
-            bRc = vrdpReportOrder (ppdev, &repeat, sizeof (repeat), VBVA_VRDP_REPEAT);
+            bRc = vrdpReportOrder (ppdev, &repeat, sizeof (repeat), VRDP_ORDER_REPEAT);
 
             if (!bRc)
             {
@@ -627,11 +629,11 @@ static BOOL vrdpGetIntersectingRects (CLIPRECTS *pRects,
     return fReportOrder;
 }
 
-static BOOL vrdpReportOrderGeneric (PPDEV ppdev,
-                                    const VRDPCLIPRECTS *pClipRects,
-                                    const void *pvOrder,
-                                    unsigned cbOrder,
-                                    unsigned code)
+BOOL vrdpReportOrderGeneric (PPDEV ppdev,
+                             const VRDPCLIPRECTS *pClipRects,
+                             const void *pvOrder,
+                             unsigned cbOrder,
+                             unsigned code)
 {
     BOOL bRc;
 
@@ -688,7 +690,7 @@ static void vrdpReportSolidRect (PPDEV ppdev,
     order.h     = (uint16_t)(prclTrg->bottom - prclTrg->top);
     order.rgb   = rgb;
 
-    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VBVA_VRDP_SOLIDRECT);
+    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VRDP_ORDER_SOLIDRECT);
 }
 
 static void vrdpReportSolidBlt (PPDEV ppdev,
@@ -706,7 +708,7 @@ static void vrdpReportSolidBlt (PPDEV ppdev,
     order.rgb   = rgb;
     order.rop   = rop3;
 
-    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VBVA_VRDP_SOLIDBLT);
+    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VRDP_ORDER_SOLIDBLT);
 }
 
 static void vrdpReportPatBlt (PPDEV ppdev,
@@ -756,7 +758,7 @@ static void vrdpReportPatBlt (PPDEV ppdev,
 
     memcpy (order.pattern, pBrush->u.pat.au8Pattern, sizeof (order.pattern));
 
-    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VBVA_VRDP_PATBLTBRUSH);
+    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VRDP_ORDER_PATBLTBRUSH);
 }
 
 static void vrdpReportDstBlt (PPDEV ppdev,
@@ -772,7 +774,7 @@ static void vrdpReportDstBlt (PPDEV ppdev,
     order.h     = (uint16_t)(prclTrg->bottom - prclTrg->top);
     order.rop   = rop3;
 
-    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VBVA_VRDP_DSTBLT);
+    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VRDP_ORDER_DSTBLT);
 }
 
 static void vrdpReportScreenBlt (PPDEV ppdev,
@@ -791,7 +793,7 @@ static void vrdpReportScreenBlt (PPDEV ppdev,
     order.ySrc  = (int16_t)pptlSrc->y;
     order.rop   = rop3;
 
-    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VBVA_VRDP_SCREENBLT);
+    vrdpReportOrderGeneric (ppdev, pClipRects, &order, sizeof (order), VRDP_ORDER_SCREENBLT);
 }
 
 static void vrdpReportMemBltRect (PPDEV ppdev,
@@ -814,7 +816,7 @@ static void vrdpReportMemBltRect (PPDEV ppdev,
     VBVA_ASSERT(sizeof (*phash) == sizeof (order.hash));
     memcpy (order.hash, phash, sizeof (*phash));
 
-    vrdpReportOrder (ppdev, &order, sizeof (order), VBVA_VRDP_MEMBLT);
+    vrdpReportOrder (ppdev, &order, sizeof (order), VRDP_ORDER_MEMBLT);
 }
 
 static void vrdpReportMemBlt (PPDEV ppdev,
@@ -858,7 +860,7 @@ static void vrdpReportCachedBitmap (PPDEV ppdev,
     VBVA_ASSERT(sizeof (*phash) == sizeof (order.hash));
     memcpy (order.hash, phash, sizeof (*phash));
 
-    bRc = vrdpReportOrder (ppdev, &order, sizeof (order), VBVA_VRDP_CACHED_BITMAP);
+    bRc = vrdpReportOrder (ppdev, &order, sizeof (order), VRDP_ORDER_CACHED_BITMAP);
 
     if (bRc)
     {
@@ -883,7 +885,7 @@ static void vrdpReportDeletedBitmap (PPDEV ppdev,
     VBVA_ASSERT(sizeof (*phash) == sizeof (order.hash));
     memcpy (order.hash, phash, sizeof (*phash));
 
-    vrdpReportOrder (ppdev, &order, sizeof (order), VBVA_VRDP_DELETED_BITMAP);
+    vrdpReportOrder (ppdev, &order, sizeof (order), VRDP_ORDER_DELETED_BITMAP);
 }
 
 
@@ -1172,19 +1174,33 @@ void vrdpTextOut(
     else if (   pstro->pwszOrg == NULL
              || prclExtra != NULL
              || (pfo->flFontType & FO_TYPE_RASTER) == 0
-             || pstro->cGlyphs > 256)
+             || pstro->cGlyphs > VRDP_TEXT_MAX_GLYPHS
+             || (pboOpaque && pboOpaque->iSolidColor == 0xFFFFFFFF)
+             || pfo->iUniq == 0
+            )
     {
         /* Unknown/unsupported parameters. */
-        DISPDBG((1, "VRDP::vrdpTextOut: unsupported: pstro->pwszOrg=%p, prclExtra=%p, pfo->flFontType & FO_TYPE_RASTER = 0x%08X, pstro->cGlyphs = %d\n",
-                    pstro->pwszOrg, prclExtra, pfo->flFontType & FO_TYPE_RASTER, pstro->cGlyphs));
+        DISPDBG((1, "VRDP::vrdpTextOut: unsupported: pstro->pwszOrg=%p, prclExtra=%p, pfo->flFontType & FO_TYPE_RASTER = 0x%08X, pstro->cGlyphs = %d, pboOpaque->iSolidColor %p, pfo->iUniq = %p\n",
+                    pstro->pwszOrg, prclExtra, pfo->flFontType & FO_TYPE_RASTER, pstro->cGlyphs, pboOpaque? pboOpaque->iSolidColor: 0, pfo->iUniq));
         vrdpReportDirtyRects (ppdev, &clipRects);
     }
     else
     {
 #if 0
+        /* Testing: report a red rectangle for the text area. */
         vrdpReportSolidRect (ppdev, &clipRects, 0x0000FF);
 #else
-        vrdpReportDirtyRects (ppdev, &clipRects);
+        /* Try to report the text order. */
+        ULONG ulForeRGB = pboFore? vrdpColor2RGB (pso, pboFore->iSolidColor): 0;
+        ULONG ulBackRGB = pboOpaque? vrdpColor2RGB (pso, pboOpaque->iSolidColor): 0;
+
+        DISPDBG((1, "VRDP::vrdpTextOut: calling vboxReportText fg %x bg %x\n",
+                    ulForeRGB, ulBackRGB));
+
+        if (!vboxReportText (ppdev, &clipRects, pstro, pfo, prclOpaque, ulForeRGB, ulBackRGB))
+        {
+            vrdpReportDirtyRects (ppdev, &clipRects);
+        }
 #endif
     }
 
@@ -1256,7 +1272,7 @@ void vrdpLineTo(
                     x1, y1, x2, y2, order.mix, order.rgb,
                     prclBounds->left, prclBounds->right, prclBounds->top, prclBounds->bottom, clipRects.rects.c));
 
-        vrdpReportOrderGeneric (ppdev, &clipRects, &order, sizeof (order), VBVA_VRDP_LINE);
+        vrdpReportOrderGeneric (ppdev, &clipRects, &order, sizeof (order), VRDP_ORDER_LINE);
     }
 }
 
@@ -1328,7 +1344,7 @@ static void vrdpPointFX2Point (const POINTFIX *pptfx, VRDPORDERPOINT *ppt)
 
 static void vrdpPolyPointsAdd (VRDPORDERPOLYPOINTS *pPoints, const VRDPORDERPOINT *ppt)
 {
-    VBVA_ASSERT(pPoints->c < ELEMENTS(pPoints->a));
+    VBVA_ASSERT(pPoints->c < RT_ELEMENTS(pPoints->a));
 
     pPoints->a[pPoints->c] = *ppt;
 
@@ -1419,7 +1435,7 @@ void vrdpStrokePath(
     }
     else if (ppo->fl & PO_ELLIPSE)
     {
-        if (vboxOrderSupported (ppdev, VBVA_VRDP_ELLIPSE))
+        if (vboxOrderSupported (ppdev, VRDP_ORDER_ELLIPSE))
         {
             VRDPORDERELLIPSE order;
 
@@ -1432,7 +1448,7 @@ void vrdpStrokePath(
             order.fillMode = 0;
             order.rgb = vrdpColor2RGB (pso, pbo->iSolidColor);
 
-            vrdpReportOrderGeneric (ppdev, &clipRects, &order, sizeof (order), VBVA_VRDP_ELLIPSE);
+            vrdpReportOrderGeneric (ppdev, &clipRects, &order, sizeof (order), VRDP_ORDER_ELLIPSE);
         }
         else
         {
@@ -1500,12 +1516,12 @@ void vrdpStrokePath(
                 vrdpPolyPointsAdd     (&order.points, &pt);
                 vrdpExtendOrderBounds (&bounds, &pt);
 
-                if (order.points.c == ELEMENTS(order.points.a))
+                if (order.points.c == RT_ELEMENTS(order.points.a))
                 {
                     /* Flush the order and start a new order. */
                     DISPDBG((1, "vrdpStrokePath: Report order, points overflow.\n"));
 
-                    vrdpReportOrderGenericBounds (ppdev, &clipRects, &bounds, &order, sizeof (order), VBVA_VRDP_POLYLINE);
+                    vrdpReportOrderGenericBounds (ppdev, &clipRects, &bounds, &order, sizeof (order), VRDP_ORDER_POLYLINE);
 
                     order.points.c = 0;
                     order.ptStart = pt;
@@ -1521,7 +1537,7 @@ void vrdpStrokePath(
                 if (   ptStart.x != pt.x
                     || ptStart.y != pt.y)
                 {
-                    VBVA_ASSERT(order.points.c < ELEMENTS(order.points.a));
+                    VBVA_ASSERT(order.points.c < RT_ELEMENTS(order.points.a));
 
                     vrdpPolyPointsAdd     (&order.points, &ptStart);
                     vrdpExtendOrderBounds (&bounds, &ptStart);
@@ -1535,7 +1551,7 @@ void vrdpStrokePath(
 
                 if (order.points.c > 0)
                 {
-                    vrdpReportOrderGenericBounds (ppdev, &clipRects, &bounds, &order, sizeof (order), VBVA_VRDP_POLYLINE);
+                    vrdpReportOrderGenericBounds (ppdev, &clipRects, &bounds, &order, sizeof (order), VRDP_ORDER_POLYLINE);
                 }
 
                 order.points.c = 0;
@@ -1554,6 +1570,24 @@ void vrdpStrokePath(
     return;
 }
 
+void vrdpStrokeAndFillPath(
+    SURFOBJ   *pso,
+    PATHOBJ   *ppo,
+    CLIPOBJ   *pco,
+    XFORMOBJ  *pxo,
+    BRUSHOBJ  *pboStroke,
+    LINEATTRS *plineattrs,
+    BRUSHOBJ  *pboFill,
+    POINTL    *pptlBrushOrg,
+    MIX        mixFill,
+    FLONG      flOptions
+    )
+{
+    PPDEV ppdev = (PPDEV)pso->dhpdev;
+
+    vrdpReportDirtyPath (ppdev, pco, ppo);
+}
+
 BOOL vrdpRealizeBrush(
     BRUSHOBJ *pbo,
     SURFOBJ  *psoTarget,
@@ -1702,7 +1736,7 @@ void vrdpSaveScreenBits(
             order.ident = (uint8_t)ident;
             order.restore = 0;
 
-            vrdpReportOrderGeneric (ppdev, NULL, &order, sizeof (order), VBVA_VRDP_SAVESCREEN);
+            vrdpReportOrderGeneric (ppdev, NULL, &order, sizeof (order), VRDP_ORDER_SAVESCREEN);
         } break;
 
         case SS_RESTORE:
@@ -1717,7 +1751,7 @@ void vrdpSaveScreenBits(
             order.ident = (uint8_t)ident;
             order.restore = 1;
 
-            if (vrdpReportOrderGeneric (ppdev, NULL, &order, sizeof (order), VBVA_VRDP_SAVESCREEN))
+            if (vrdpReportOrderGeneric (ppdev, NULL, &order, sizeof (order), VRDP_ORDER_SAVESCREEN))
             {
                 uint8_t *pu8Bits;
                 int32_t lDelta;
