@@ -4,15 +4,19 @@
  *
  * Based on a Microsoft DDK sample
  *
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 // #define LOG_ENABLED
@@ -114,26 +118,32 @@ typedef struct _VBoxGlobalContext
     volatile LONG fVBGLInited;
     volatile LONG fVBGLInitFailed;
     volatile LONG fHostInformed;
+    volatile LONG fHostMouseFound;
 } VBoxGlobalContext;
 
-VBoxGlobalContext g_ctx = { 0, FALSE, FALSE, FALSE };
+static VBoxGlobalContext g_ctx = { 0, FALSE, FALSE, FALSE, FALSE };
 
-BOOLEAN vboxIsVBGLInited (void)
+static BOOLEAN vboxIsVBGLInited (void)
 {
    return InterlockedCompareExchange (&g_ctx.fVBGLInited, TRUE, TRUE) == TRUE;
 }
 
-BOOLEAN vboxIsVBGLInitFailed (void)
+static BOOLEAN vboxIsVBGLInitFailed (void)
 {
    return InterlockedCompareExchange (&g_ctx.fVBGLInitFailed, TRUE, TRUE) == TRUE;
 }
 
-BOOLEAN vboxIsHostInformed (void)
+static BOOLEAN vboxIsHostInformed (void)
 {
-   return InterlockedCompareExchange (&g_ctx.fVBGLInitFailed, TRUE, TRUE) == TRUE;
+   return InterlockedCompareExchange (&g_ctx.fHostInformed, TRUE, TRUE) == TRUE;
 }
 
-void vboxDeviceAdded (PDEVICE_EXTENSION devExt)
+static BOOLEAN vboxIsHostMouseFound (void)
+{
+   return InterlockedCompareExchange (&g_ctx.fHostMouseFound, TRUE, TRUE) == TRUE;
+}
+
+static void vboxDeviceAdded (PDEVICE_EXTENSION devExt)
 {
     LONG c = InterlockedIncrement (&g_ctx.cDevicesStarted);
 
@@ -145,32 +155,151 @@ void vboxDeviceAdded (PDEVICE_EXTENSION devExt)
             /*  Do one time initialization. */
             int vboxRC = VbglInit ();
 
-            if (VBOX_SUCCESS (vboxRC))
+            if (RT_SUCCESS (vboxRC))
             {
                 /* Atomically set the flag. */
                 InterlockedExchange (&g_ctx.fVBGLInited, TRUE);
-                dprintf(("VBoxMouse::vboxDeviceStarted: guest library initialization OK\n"));
-
-                /* Mark the first device as the Host one, that is the emulated mouse.
-                 * For this device the filter will query absolute mouse coords from the host.
-                 * @todo: Better would be to query the device information and detect the emulated device.
-                 */
-                devExt->HostMouse = TRUE;
+                dprintf(("VBoxMouse::vboxDeviceAdded: Guest library initialization OK\n"));
             }
             else
             {
                 InterlockedExchange (&g_ctx.fVBGLInitFailed, TRUE);
-                dprintf(("VBoxMouse::vboxDeviceStarted: guest library initialization failed\n"));
+                dprintf(("VBoxMouse::vboxDeviceAdded: Guest library initialization failed\n"));
+            }
+        }
+    }
+    if (!vboxIsHostMouseFound ())
+    {
+        NTSTATUS status;
+        UCHAR Property[512];
+        ULONG ResultLength;
+
+        ResultLength = 0;
+        status = IoGetDeviceProperty(devExt->PDO, DevicePropertyBootConfiguration,
+                                     sizeof (Property),
+                                     &Property,
+                                     &ResultLength);
+
+        dprintf(("VBoxMouse::vboxDeviceAdded: looking for host mouse: Len is %d, status %x\n",
+                 ResultLength, status));
+
+        if (status == STATUS_SUCCESS)
+        {
+            /* Check whether the device claims the IO port 0x60 or interrupt 12. */
+            BOOLEAN bDetected = FALSE;
+
+            CM_RESOURCE_LIST *pResourceList = (CM_RESOURCE_LIST *)&Property[0];
+
+            dprintf(("VBoxMouse::vboxDeviceAdded: Configuration: Number of descriptors = %d\n",
+                      pResourceList->Count));
+
+            ULONG iDescriptor = 0;
+            for (; iDescriptor < pResourceList->Count; iDescriptor++)
+            {
+                CM_FULL_RESOURCE_DESCRIPTOR *pFullDescriptor = &pResourceList->List[iDescriptor];
+
+                dprintf(("VBoxMouse::vboxDeviceAdded: Descriptor %d: InterfaceType %d, BusType %d, list ver %d, list rev %d, count %d\n",
+                         iDescriptor,
+                         pFullDescriptor->InterfaceType,
+                         pFullDescriptor->BusNumber,
+                         pFullDescriptor->PartialResourceList.Version,
+                         pFullDescriptor->PartialResourceList.Revision,
+                         pFullDescriptor->PartialResourceList.Count));
+
+                ULONG iPartialDescriptor = 0;
+                for (; iPartialDescriptor < pFullDescriptor->PartialResourceList.Count; iPartialDescriptor++)
+                {
+                    CM_PARTIAL_RESOURCE_DESCRIPTOR *pPartialDescriptor = &pFullDescriptor->PartialResourceList.PartialDescriptors[iPartialDescriptor];
+
+                    dprintf(("VBoxMouse::vboxDeviceAdded: PartialDescriptor %d: type %d, ShareDisposition %d, Flags 0x%04X, Start 0x%llx, length 0x%x\n",
+                             iPartialDescriptor,
+                             pPartialDescriptor->Type, pPartialDescriptor->ShareDisposition, pPartialDescriptor->Flags,
+                             pPartialDescriptor->u.Generic.Start.QuadPart, pPartialDescriptor->u.Generic.Length));
+
+                    switch(pPartialDescriptor->Type)
+                    {
+
+                    case CmResourceTypePort:
+
+                        dprintf(("VBoxMouse::vboxDeviceAdded: PartialDescriptor %d: Port 0x%x\n", iPartialDescriptor, pPartialDescriptor->u.Port.Start.QuadPart));
+                        if (pPartialDescriptor->u.Port.Start.QuadPart == 0x60)
+                        {
+                            bDetected = TRUE;
+                        }
+                        break;
+
+                    case CmResourceTypeInterrupt:
+
+                        dprintf(("VBoxMouse::vboxDeviceAdded: PartialDescriptor %d: Int %ld\n", iPartialDescriptor, pPartialDescriptor->u.Interrupt.Vector));
+                        if (pPartialDescriptor->u.Interrupt.Vector == 0x12)
+                        {
+                            bDetected = TRUE;
+                        }
+                        break;
+
+                    default:
+                        break;
+                    }
+                }
+            }
+
+            if (bDetected)
+            {
+                /* It's the emulated 8042 PS/2 mouse/kbd device, so mark it as the Host one.
+                 * For this device the filter will query absolute mouse coords from the host.
+                 */
+                InterlockedExchange (&g_ctx.fHostMouseFound, TRUE);
+                devExt->HostMouse = TRUE;
+                dprintf(("VBoxMouse::vboxDeviceAdded: Host mouse found.\n"));
             }
         }
     }
 }
 
-void vboxDeviceRemoved (PDEVICE_EXTENSION devExt)
+static void vboxDeviceRemoved (PDEVICE_EXTENSION devExt)
 {
     dprintf(("VBoxMouse::vboxDeviceRemoved\n"));
 
-    LONG c = InterlockedIncrement (&g_ctx.cDevicesStarted);
+    /* Save the allocated request pointer and clear the devExt. */
+    VMMDevReqMouseStatus *reqSC = (VMMDevReqMouseStatus *)InterlockedExchangePointer ((PVOID volatile *)&devExt->reqSC, NULL);
+
+    if (devExt->HostMouse && vboxIsHostInformed ())
+    {
+        // tell the VMM that from now on we can't handle absolute coordinates anymore
+        VMMDevReqMouseStatus *req = NULL;
+
+        int vboxRC = VbglGRAlloc ((VMMDevRequestHeader **)&req, sizeof (VMMDevReqMouseStatus), VMMDevReq_SetMouseStatus);
+
+        if (RT_SUCCESS(vboxRC))
+        {
+            req->mouseFeatures = 0;
+            req->pointerXPos = 0;
+            req->pointerYPos = 0;
+
+            vboxRC = VbglGRPerform (&req->header);
+
+            if (RT_FAILURE(vboxRC) || RT_FAILURE(req->header.rc))
+            {
+                dprintf(("VBoxMouse::vboxDeviceRemoved: ERROR communicating new mouse capabilities to VMMDev.\n"
+                         "rc = %d, VMMDev rc = %Rrc\n", vboxRC, req->header.rc));
+            }
+
+            VbglGRFree (&req->header);
+        }
+        else
+        {
+            dprintf(("VBoxMouse::vboxDeviceRemoved: the request allocation has failed.\n"));
+        }
+
+        InterlockedExchange (&g_ctx.fHostInformed, FALSE);
+    }
+
+    if (reqSC)
+    {
+        VbglGRFree (&reqSC->header);
+    }
+
+    LONG c = InterlockedDecrement (&g_ctx.cDevicesStarted);
 
     if (c == 0)
     {
@@ -179,60 +308,28 @@ void vboxDeviceRemoved (PDEVICE_EXTENSION devExt)
             /* Set the flag to prevent reinitializing of the VBGL. */
             InterlockedExchange (&g_ctx.fVBGLInitFailed, TRUE);
 
-            /* Save the allocated request pointer and clear the devExt. */
-            VMMDevReqMouseStatus *reqSC = devExt->reqSC;
-            devExt->reqSC = NULL;
-
-            // tell the VMM that from now on we can't handle absolute coordinates anymore
-            VMMDevReqMouseStatus *req = NULL;
-
-            int vboxRC = VbglGRAlloc ((VMMDevRequestHeader **)&req, sizeof (VMMDevReqMouseStatus), VMMDevReq_SetMouseStatus);
-
-            if (VBOX_SUCCESS(vboxRC))
-            {
-                req->mouseFeatures = 0;
-                req->pointerXPos = 0;
-                req->pointerYPos = 0;
-
-                vboxRC = VbglGRPerform (&req->header);
-
-                if (VBOX_FAILURE(vboxRC) || VBOX_FAILURE(req->header.rc))
-                {
-                    dprintf(("VBoxMouse::vboxDeviceRemoved: ERROR communicating new mouse capabilities to VMMDev.\n"
-                             "rc = %d, VMMDev rc = %Vrc\n", vboxRC, req->header.rc));
-                }
-
-                VbglGRFree (&req->header);
-            }
-
-            if (reqSC)
-            {
-                VbglGRFree (&reqSC->header);
-            }
-
             VbglTerminate ();
 
             /* The VBGL is now in the not initialized state. */
-            InterlockedExchange (&g_ctx.fHostInformed, FALSE);
             InterlockedExchange (&g_ctx.fVBGLInited, FALSE);
             InterlockedExchange (&g_ctx.fVBGLInitFailed, FALSE);
         }
     }
 }
 
-void vboxInformHost (PDEVICE_EXTENSION devExt)
+static void vboxInformHost (PDEVICE_EXTENSION devExt)
 {
-    dprintf (("VBoxMouse::vboxInformHost: %p\n", devExt));
+    dprintf (("VBoxMouse::vboxInformHost: DevExt=%p, HostMouse=%d\n", devExt, devExt->HostMouse));
 
     if (vboxIsVBGLInited ())
     {
-        if (!vboxIsHostInformed ())
+        if (devExt->HostMouse && !vboxIsHostInformed ())
         {
             VMMDevReqMouseStatus *req = NULL;
 
             int vboxRC = VbglGRAlloc ((VMMDevRequestHeader **)&req, sizeof (VMMDevReqMouseStatus), VMMDevReq_SetMouseStatus);
 
-            if (VBOX_SUCCESS(vboxRC))
+            if (RT_SUCCESS(vboxRC))
             {
                 /* Inform host that we support absolute */
                 req->mouseFeatures = VBOXGUEST_MOUSE_GUEST_CAN_ABSOLUTE;
@@ -241,10 +338,10 @@ void vboxInformHost (PDEVICE_EXTENSION devExt)
 
                 vboxRC = VbglGRPerform (&req->header);
 
-                if (VBOX_FAILURE(vboxRC) || VBOX_FAILURE(req->header.rc))
+                if (RT_FAILURE(vboxRC) || RT_FAILURE(req->header.rc))
                 {
                     dprintf(("VBoxMouse::vboxInformHost: ERROR communicating new mouse capabilities to VMMDev."
-                              "rc = %d, VMMDev rc = %Vrc\n", vboxRC, req->header.rc));
+                              "rc = %d, VMMDev rc = %Rrc\n", vboxRC, req->header.rc));
                 }
                 else
                 {
@@ -262,9 +359,9 @@ void vboxInformHost (PDEVICE_EXTENSION devExt)
 
             int vboxRC = VbglGRAlloc ((VMMDevRequestHeader **)&req, sizeof (VMMDevReqMouseStatus), VMMDevReq_GetMouseStatus);
 
-            if (VBOX_SUCCESS(vboxRC))
+            if (RT_SUCCESS(vboxRC))
             {
-                devExt->reqSC = req;
+                InterlockedExchangePointer ((PVOID volatile *)&devExt->reqSC, req);
                 dumpDevExt (devExt);
             }
             else
@@ -293,6 +390,20 @@ VBoxMouse_AddDevice(
 
     dprintf(("VBoxMouse::AddDevice Driver %p, PDO %p\n", Driver, PDO));
 
+#ifdef LOG_ENABLED
+    WCHAR wszProperty[512];
+    ULONG ResultLength = 0;
+    wszProperty[0] = 0;
+    status = IoGetDeviceProperty(PDO, DevicePropertyDeviceDescription,
+                                 sizeof (wszProperty),
+                                 &wszProperty,
+                                 &ResultLength);
+    if (status == STATUS_SUCCESS)
+    {
+        dprintf(("VBoxMouse::AddDevice %ls Len is %d\n", wszProperty, ResultLength));
+    }
+#endif /* LOG_ENABLED */
+
     status = IoCreateDevice(Driver,
                             sizeof(DEVICE_EXTENSION),
                             NULL,
@@ -332,7 +443,7 @@ VBoxMouse_AddDevice(
     device->Flags |= (DO_BUFFERED_IO | DO_POWER_PAGABLE);
     device->Flags &= ~DO_DEVICE_INITIALIZING;
 
-    dprintf(("DevExt = %p, returning from AddDevice with rc = 0x%x\n", devExt, status));
+    dprintf(("VBoxMouse::AddDevice: DevExt = %p, returning from AddDevice with rc = 0x%x\n", devExt, status));
     return status;
 }
 
@@ -396,12 +507,12 @@ Routine Description:
 
     switch (irpStack->MajorFunction) {
     case IRP_MJ_CREATE:
-        dprintf(("IRP_MJ_CREATE\n"));
+        dprintf(("VBoxMouse::CreateClose: IRP_MJ_CREATE\n"));
         if (NULL == devExt->UpperConnectData.ClassService) {
             //
             // No Connection yet.  How can we be enabled?
             //
-            dprintf(("VBoxMouse: Not connected, returning STATUS_INVALID_DEVICE_STATE\n"));
+            dprintf(("VBoxMouse::CreateClose: Not connected, returning STATUS_INVALID_DEVICE_STATE\n"));
             status = STATUS_INVALID_DEVICE_STATE;
         }
         else if ( 1 >= InterlockedIncrement(&devExt->EnableCount)) {
@@ -419,7 +530,7 @@ Routine Description:
 
     case IRP_MJ_CLOSE:
 
-        dprintf(("IRP_MJ_CLOSE\n"));
+        dprintf(("VBoxMouse::CreateClose: IRP_MJ_CLOSE\n"));
         ASSERT(0 < devExt->EnableCount);
 
         if (0 >= InterlockedDecrement(&devExt->EnableCount)) {
@@ -706,8 +817,8 @@ Return Value:
     switch (irpStack->MinorFunction) {
     case IRP_MN_START_DEVICE: {
 
-        dprintf(("IRP_MN_START_DEVICE\n"));
-        
+        dprintf(("VBoxMouse_PnP: IRP_MN_START_DEVICE\n"));
+
         //
         // The device is starting.
         //
@@ -738,7 +849,7 @@ Return Value:
                NULL); // No timeout
         }
 
-        dprintf(("status: %x, irp status: %x\n", Irp->IoStatus.Status));
+        dprintf(("VBoxMouse_PnP: Status: %x, irp status: %x\n", Irp->IoStatus.Status));
 
         if (NT_SUCCESS(status) && NT_SUCCESS(Irp->IoStatus.Status)) {
             devExt->Started = TRUE;
@@ -758,9 +869,9 @@ Return Value:
     }
 
     case IRP_MN_SURPRISE_REMOVAL:
-    
-        dprintf(("IRP_MN_SURPRISE_REMOVAL\n"));
-        
+
+        dprintf(("VBoxMouse_PnP: IRP_MN_SURPRISE_REMOVAL\n"));
+
         //
         // Same as a remove device, but don't call IoDetach or IoDeleteDevice
         //
@@ -774,7 +885,7 @@ Return Value:
 
     case IRP_MN_REMOVE_DEVICE:
     {
-        dprintf(("IRP_MN_REMOVE_DEVICE\n"));
+        dprintf(("VBoxMouse_PnP: IRP_MN_REMOVE_DEVICE\n"));
 // VBOX start
         vboxDeviceRemoved (devExt);
 // VBOX end
@@ -1005,7 +1116,7 @@ Return Value:
     {
         int rc = VbglGRPerform (&req->header);
 
-        if (VBOX_SUCCESS(rc) && VBOX_SUCCESS(req->header.rc))
+        if (RT_SUCCESS(rc) && RT_SUCCESS(req->header.rc))
         {
             if (req->mouseFeatures & VBOXGUEST_MOUSE_HOST_CAN_ABSOLUTE)
             {
@@ -1023,7 +1134,7 @@ Return Value:
         else
         {
             dprintf(("VBoxMouse::VBoxMouse_ServiceCallback: ERROR querying mouse capabilities from VMMDev."
-                     "rc = %Vrc, VMMDev rc = %Vrc\n", rc, req->header.rc));
+                     "rc = %Rrc, VMMDev rc = %Rrc\n", rc, req->header.rc));
         }
     }
 // VBOX end
