@@ -2,15 +2,19 @@
  *
  * VBoxGuest -- VirtualBox Win32 guest support driver PnP code
  *
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 // enable backdoor logging
@@ -153,7 +157,7 @@ NTSTATUS VBoxGuestPnP(PDEVICE_OBJECT pDevObj, PIRP pIrp)
             if (NT_SUCCESS(rc) && NT_SUCCESS(pIrp->IoStatus.Status))
             {
                 dprintf(("VBoxGuest::START_DEVICE: pStack->Parameters.StartDevice.AllocatedResources = %p\n", pStack->Parameters.StartDevice.AllocatedResources));
-                
+
                 if (!pStack->Parameters.StartDevice.AllocatedResources)
                 {
                     dprintf(("VBoxGuest::START_DEVICE: no resources, pDevExt = %p, nextLowerDriver = %p!!!\n", pDevExt, pDevExt? pDevExt->nextLowerDriver: NULL));
@@ -169,7 +173,7 @@ NTSTATUS VBoxGuestPnP(PDEVICE_OBJECT pDevObj, PIRP pIrp)
                     /** @todo cleanup and merging codepath with NT */
                     int rcVBox;
                     rcVBox = VbglInit (pDevExt->startPortAddress, pDevExt->pVMMDevMemory);
-                    if (!VBOX_SUCCESS(rcVBox))
+                    if (!RT_SUCCESS(rcVBox))
                     {
                         dprintf(("VBoxGuest::START_DEVICE: VbglInit failed. rcVBox = %d\n", rcVBox));
                         rc = STATUS_UNSUCCESSFUL;
@@ -178,9 +182,19 @@ NTSTATUS VBoxGuestPnP(PDEVICE_OBJECT pDevObj, PIRP pIrp)
                     if (NT_SUCCESS(rc))
                     {
                         rcVBox = VbglGRAlloc ((VMMDevRequestHeader **)&pDevExt->irqAckEvents, sizeof (VMMDevEvents), VMMDevReq_AcknowledgeEvents);
-                        if (!VBOX_SUCCESS(rc))
+                        if (!RT_SUCCESS(rc))
                         {
-                            dprintf(("VBoxGuest::START_DEVICE: VbglAlloc failed. rcVBox = %d\n", rcVBox));
+                            dprintf(("VBoxGuest::START_DEVICE: VbglAlloc failed for irqAckEvents. rcVBox = %d\n", rcVBox));
+                            rc = STATUS_UNSUCCESSFUL;
+                        }
+                    }
+
+                    if (NT_SUCCESS(rc))
+                    {
+                        rcVBox = VbglGRAlloc ((VMMDevRequestHeader **)&pDevExt->powerStateRequest, sizeof (VMMDevPowerStateRequest), VMMDevReq_SetPowerStatus);
+                        if (!RT_SUCCESS(rc))
+                        {
+                            dprintf(("VBoxGuest::START_DEVICE: VbglAlloc failed for powerStateRequest. rcVBox = %d\n", rcVBox));
                             rc = STATUS_UNSUCCESSFUL;
                         }
                     }
@@ -235,6 +249,10 @@ NTSTATUS VBoxGuestPnP(PDEVICE_OBJECT pDevObj, PIRP pIrp)
                 // initialize the event notification semaphore
                 KeInitializeEvent(&pDevExt->keventNotification, NotificationEvent, FALSE);
 
+                /* Preallocated constant timeout 250ms for HGCM async waiter. */
+                pDevExt->HGCMWaitTimeout.QuadPart  = 250;
+                pDevExt->HGCMWaitTimeout.QuadPart *= -10000;     /* relative in 100ns units */
+
                 VBoxInitMemBalloon(pDevExt);
 
                 // ready to rumble!
@@ -265,7 +283,7 @@ NTSTATUS VBoxGuestPnP(PDEVICE_OBJECT pDevObj, PIRP pIrp)
             pIrp->IoStatus.Information = 0;
             IoCompleteRequest(pIrp, IO_NO_INCREMENT);
             rc = STATUS_UNSUCCESSFUL;
-            
+
             dprintf(("VBoxGuest::VBoxGuestPnp: refuse with rc = %p\n", pIrp->IoStatus.Status));
 #else
             pIrp->IoStatus.Status = STATUS_SUCCESS;
@@ -314,6 +332,12 @@ NTSTATUS VBoxGuestPnP(PDEVICE_OBJECT pDevObj, PIRP pIrp)
             {
                 IoDetachDevice(pDevExt->nextLowerDriver);
             }
+
+#ifdef VBOX_WITH_HGCM
+            if (pDevExt->SessionSpinlock != NIL_RTSPINLOCK)
+                RTSpinlockDestroy(pDevExt->SessionSpinlock);
+#endif
+
             UNICODE_STRING win32Name;
             RtlInitUnicodeString(&win32Name, VBOXGUEST_DEVICE_NAME_DOS);
             IoDeleteSymbolicLink(&win32Name);
@@ -526,7 +550,7 @@ NTSTATUS VBoxGuestPower(PDEVICE_OBJECT pDevObj, PIRP pIrp)
 
                             int rc = VbglGRAlloc ((VMMDevRequestHeader **)&req, sizeof (VMMDevReqMouseStatus), VMMDevReq_SetMouseStatus);
 
-                            if (VBOX_SUCCESS(rc))
+                            if (RT_SUCCESS(rc))
                             {
                                 req->mouseFeatures = 0;
                                 req->pointerXPos = 0;
@@ -534,10 +558,10 @@ NTSTATUS VBoxGuestPower(PDEVICE_OBJECT pDevObj, PIRP pIrp)
 
                                 rc = VbglGRPerform (&req->header);
 
-                                if (VBOX_FAILURE(rc) || VBOX_FAILURE(req->header.rc))
+                                if (RT_FAILURE(rc) || RT_FAILURE(req->header.rc))
                                 {
                                     dprintf(("VBoxGuest::PowerStateRequest: error communicating new power status to VMMDev."
-                                             "rc = %d, VMMDev rc = %Vrc\n", rc, req->header.rc));
+                                             "rc = %d, VMMDev rc = %Rrc\n", rc, req->header.rc));
                                 }
 
                                 VbglGRFree (&req->header);
@@ -552,23 +576,21 @@ NTSTATUS VBoxGuestPower(PDEVICE_OBJECT pDevObj, PIRP pIrp)
                             if (powerState.SystemState >= PowerSystemShutdown)
                             {
                                 dprintf(("VBoxGuest::VBoxGuestPower: Telling the VMMDev to close the VM...\n"));
-                                VMMDevPowerStateRequest *req = NULL;
 
-                                int rc = VbglGRAlloc ((VMMDevRequestHeader **)&req, sizeof (VMMDevPowerStateRequest), VMMDevReq_SetPowerStatus);
-
-                                if (VBOX_SUCCESS(rc))
+                                if (pDevExt && pDevExt->powerStateRequest)
                                 {
+                                    VMMDevPowerStateRequest *req = pDevExt->powerStateRequest;
+
+                                    req->header.requestType = VMMDevReq_SetPowerStatus;
                                     req->powerState = VMMDevPowerState_PowerOff;
 
-                                    rc = VbglGRPerform (&req->header);
+                                    int rc = VbglGRPerform (&req->header);
 
-                                    if (VBOX_FAILURE(rc) || VBOX_FAILURE(req->header.rc))
+                                    if (RT_FAILURE(rc) || RT_FAILURE(req->header.rc))
                                     {
                                         dprintf(("VBoxGuest::PowerStateRequest: error communicating new power status to VMMDev."
-                                                 "rc = %d, VMMDev rc = %Vrc\n", rc, req->header.rc));
+                                                 "rc = %d, VMMDev rc = %Rrc\n", rc, req->header.rc));
                                     }
-
-                                    VbglGRFree (&req->header);
                                 }
                             }
                             break;
