@@ -5,17 +5,23 @@
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
+#include "vfsmod.h"
+
 static int
 sf_dir_open (struct inode *inode, struct file *file)
 {
@@ -31,35 +37,38 @@ sf_dir_open (struct inode *inode, struct file *file)
         BUG_ON (!sf_i);
 
         if (file->private_data) {
-                elog ("dir_open called on already opened directory %s\n",
-                      sf_i->path->String.utf8);
+                LogFunc(("dir_open called on already opened directory %s\n",
+                         sf_i->path->String.utf8));
                 return 0;
         }
 
         sf_d = sf_dir_info_alloc ();
 
         if (!sf_d) {
-                elog ("could not allocate directory info for %s\n",
-                      sf_i->path->String.utf8);
+                LogRelFunc(("could not allocate directory info for %s\n",
+                            sf_i->path->String.utf8));
                 return -ENOMEM;
         }
 
         params.CreateFlags = 0
                 | SHFL_CF_DIRECTORY
                 | SHFL_CF_ACT_OPEN_IF_EXISTS
+                | SHFL_CF_ACT_FAIL_IF_NEW
                 | SHFL_CF_ACCESS_READ
                 ;
 
+        LogFunc(("sf_dir_open: calling vboxCallCreate, folder %s, flags %#x\n",
+                 sf_i->path->String.utf8, params.CreateFlags));
         rc = vboxCallCreate (&client_handle, &sf_g->map, sf_i->path, &params);
-        if (VBOX_FAILURE (rc)) {
-                elog ("vboxCallCreate(%s) failed rc=%d\n",
-                      sf_i->path->String.utf8, rc);
+        if (RT_FAILURE (rc)) {
+                LogFunc(("vboxCallCreate(%s) failed rc=%Rrc\n",
+                         sf_i->path->String.utf8, rc));
                 sf_dir_info_free (sf_d);
                 return -EPERM;
         }
 
         if (params.Result != SHFL_FILE_EXISTS) {
-                elog ("directory %s does not exist\n", sf_i->path->String.utf8);
+                LogFunc(("directory %s does not exist\n", sf_i->path->String.utf8));
                 sf_dir_info_free (sf_d);
                 return -ENOENT;
         }
@@ -67,9 +76,9 @@ sf_dir_open (struct inode *inode, struct file *file)
         err = sf_dir_read_all (sf_g, sf_i, sf_d, params.Handle);
         if (err) {
                 rc = vboxCallClose (&client_handle, &sf_g->map, params.Handle);
-                if (VBOX_FAILURE (rc)) {
-                        elog ("vboxCallClose(%s) after err=%d failed rc=%d\n",
-                              sf_i->path->String.utf8, err, rc);
+                if (RT_FAILURE (rc)) {
+                        LogFunc(("vboxCallClose(%s) after err=%d failed rc=%Rrc\n",
+                                 sf_i->path->String.utf8, err, rc));
                 }
                 sf_dir_info_free (sf_d);
                 return err;
@@ -77,9 +86,9 @@ sf_dir_open (struct inode *inode, struct file *file)
 
 
         rc = vboxCallClose (&client_handle, &sf_g->map, params.Handle);
-        if (VBOX_FAILURE (rc)) {
-                elog ("vboxCallClose(%s) failed rc=%d\n",
-                      sf_i->path->String.utf8, rc);
+        if (RT_FAILURE (rc)) {
+                LogFunc(("vboxCallClose(%s) failed rc=%Rrc\n",
+                          sf_i->path->String.utf8, rc));
         }
 
         file->private_data = sf_d;
@@ -108,7 +117,7 @@ sf_dir_release (struct inode *inode, struct file *file)
    0       all ok
    1       end reached
    -errno  some form of error*/
-int
+static int
 sf_getdent (struct file *dir, char d_name[NAME_MAX])
 {
         loff_t cur;
@@ -195,7 +204,7 @@ sf_dir_read (struct file *dir, void *opaque, filldir_t filldir)
                         case -1:
                         default:
                                 /* skip erroneous entry and proceed */
-                                elog ("sf_getdent error %d\n", err);
+                                LogFunc(("sf_getdent error %d\n", err));
                                 dir->f_pos += 1;
                                 continue;
                 }
@@ -205,14 +214,14 @@ sf_dir_read (struct file *dir, void *opaque, filldir_t filldir)
                 sanity = dir->f_pos + 0xbeef;
                 fake_ino = sanity;
                 if (sanity - fake_ino) {
-                        elog2 ("can not compute ino\n");
+                        LogRelFunc(("can not compute ino\n"));
                         return -EINVAL;
                 }
 
                 err = filldir (opaque, d_name, strlen (d_name),
                                dir->f_pos, fake_ino, DT_UNKNOWN);
                 if (err) {
-                        DBGC elog ("filldir returned error %d\n", err);
+                        LogFunc(("filldir returned error %d\n", err));
                         /* Rely on the fact that filldir returns error
                            only when it runs out of space in opaque */
                         return 0;
@@ -224,7 +233,7 @@ sf_dir_read (struct file *dir, void *opaque, filldir_t filldir)
         BUG ();
 }
 
-static struct file_operations sf_dir_fops = {
+struct file_operations sf_dir_fops = {
         .open    = sf_dir_open,
         .readdir = sf_dir_read,
         .release = sf_dir_release,
@@ -280,15 +289,19 @@ sf_lookup (struct inode *parent, struct dentry *dentry
         else {
                 sf_new_i = kmalloc (sizeof (*sf_new_i), GFP_KERNEL);
                 if (!sf_new_i) {
-                        elog2 ("could not allocate memory for new inode info\n");
+                        LogRelFunc(("could not allocate memory for new inode info\n"));
                         err = -ENOMEM;
                         goto fail1;
                 }
 
                 ino = iunique (parent->i_sb, 1);
+#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 4, 25)
+                inode = iget_locked (parent->i_sb, ino);
+#else
                 inode = iget (parent->i_sb, ino);
+#endif
                 if (!inode) {
-                        elog2 ("iget failed\n");
+                        LogFunc(("iget failed\n"));
                         err = -ENOMEM;          /* XXX: ??? */
                         goto fail2;
                 }
@@ -296,6 +309,10 @@ sf_lookup (struct inode *parent, struct dentry *dentry
                 SET_INODE_INFO (inode, sf_new_i);
                 sf_init_inode (sf_g, inode, &fsinfo);
                 sf_new_i->path = path;
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 4, 25)
+                unlock_new_inode(inode);
+#endif
         }
 
         sf_i->force_restat = 0;
@@ -331,16 +348,19 @@ sf_instantiate (const char *caller, struct inode *parent,
 
         sf_new_i = kmalloc (sizeof (*sf_new_i), GFP_KERNEL);
         if (!sf_new_i) {
-                elog3 ("%s: %s: could not allocate inode info\n",
-                       caller, __func__);
+                LogRelFunc(("could not allocate inode info.  caller=%s\n", caller));
                 err = -ENOMEM;
                 goto fail0;
         }
 
         ino = iunique (parent->i_sb, 1);
+#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 4, 25)
+        inode = iget_locked (parent->i_sb, ino);
+#else
         inode = iget (parent->i_sb, ino);
+#endif
         if (!inode) {
-                elog3 ("%s: %s: iget failed\n", caller, __func__);
+                LogFunc(("iget failed.  caller=%s\n", caller));
                 err = -ENOMEM;
                 goto fail1;
         }
@@ -354,6 +374,11 @@ sf_instantiate (const char *caller, struct inode *parent,
         sf_new_i->force_restat = 1;
 
         d_instantiate (dentry, inode);
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 4, 25)
+                unlock_new_inode(inode);
+#endif
+
         return 0;
 
  fail1:
@@ -383,7 +408,7 @@ sf_create_aux (struct inode *parent, struct dentry *dentry, int dirop)
 
         params.CreateFlags = 0
                 | SHFL_CF_ACT_CREATE_IF_NEW
-                | SHFL_CF_ACT_OVERWRITE_IF_EXISTS
+                | SHFL_CF_ACT_FAIL_IF_EXISTS
                 | SHFL_CF_ACCESS_READWRITE
                 | (dirop ? SHFL_CF_DIRECTORY : 0)
                 ;
@@ -402,31 +427,37 @@ sf_create_aux (struct inode *parent, struct dentry *dentry, int dirop)
                 goto fail0;
         }
 
+        LogFunc(("calling vboxCallCreate, folder %s, flags %#x\n",
+                 path->String.utf8, params.CreateFlags));
         rc = vboxCallCreate (&client_handle, &sf_g->map, path, &params);
-        if (VBOX_FAILURE (rc)) {
+        if (RT_FAILURE (rc)) {
+                if (rc == VERR_WRITE_PROTECT) {
+                        err = -EROFS;
+                        goto fail0;
+                }
                 err = -EPROTO;
-                elog ("(%d): vboxCallCreate(%s) failed rc=%d\n", dirop,
-                      sf_i->path->String.utf8, rc);
+                LogFunc(("(%d): vboxCallCreate(%s) failed rc=%Rrc\n", dirop,
+                         sf_i->path->String.utf8, rc));
                 goto fail0;
         }
 
         if (params.Result != SHFL_FILE_CREATED) {
                 err = -EPERM;
-                elog ("(%d): could not create file %s result=%d\n", dirop,
-                      sf_i->path->String.utf8, params.Result);
+                LogFunc(("(%d): could not create file %s result=%d\n", dirop,
+                         sf_i->path->String.utf8, params.Result));
                 goto fail0;
         }
 
         err = sf_instantiate (__func__, parent, dentry, path, &params.Info);
         if (err) {
-                elog ("(%d): could not instantiate dentry for %s err=%d\n",
-                      dirop, sf_i->path->String.utf8, err);
+                LogFunc(("(%d): could not instantiate dentry for %s err=%d\n",
+                         dirop, sf_i->path->String.utf8, err));
                 goto fail1;
         }
 
         rc = vboxCallClose (&client_handle, &sf_g->map, params.Handle);
-        if (VBOX_FAILURE (rc)) {
-                elog ("(%d): vboxCallClose failed rc=%d\n", dirop, rc);
+        if (RT_FAILURE (rc)) {
+                LogFunc(("(%d): vboxCallClose failed rc=%Rrc\n", dirop, rc));
         }
 
         sf_i->force_restat = 1;
@@ -434,8 +465,8 @@ sf_create_aux (struct inode *parent, struct dentry *dentry, int dirop)
 
  fail1:
         rc = vboxCallClose (&client_handle, &sf_g->map, params.Handle);
-        if (VBOX_FAILURE (rc)) {
-                elog ("(%d): vboxCallClose failed rc=%d\n", dirop, rc);
+        if (RT_FAILURE (rc)) {
+                LogFunc(("(%d): vboxCallClose failed rc=%Rrc\n", dirop, rc));
         }
 
  fail0:
@@ -479,25 +510,10 @@ sf_unlink_aux (struct inode *parent, struct dentry *dentry, int dirop)
 
         rc = vboxCallRemove (&client_handle, &sf_g->map, path,
                              dirop ? SHFL_REMOVE_DIR : SHFL_REMOVE_FILE);
-        if (VBOX_FAILURE (rc)) {
-                DBGC elog ("(%d): vboxCallRemove(%s) failed rc=%d\n", dirop,
-                           path->String.utf8, rc);
-
-                switch (rc) {
-                        case VERR_PATH_NOT_FOUND:
-                                err = -ENOENT;
-                                break;
-
-                        case VERR_DIR_NOT_EMPTY:
-                                err = -ENOTEMPTY;
-                                break;
-
-                        default:
-                                err = -EPROTO;
-                                elog ("(%d): vboxCallRemove(%s) failed rc=%d\n", dirop,
-                                      path->String.utf8, rc);
-                                break;
-                }
+        if (RT_FAILURE (rc)) {
+                LogFunc(("(%d): vboxCallRemove(%s) failed rc=%Rrc\n", dirop,
+                         path->String.utf8, rc));
+                         err = -RTErrConvertToErrno (rc);
                 goto fail1;
         }
 
@@ -528,71 +544,55 @@ static int
 sf_rename (struct inode *old_parent, struct dentry *old_dentry,
            struct inode *new_parent, struct dentry *new_dentry)
 {
-        int err, rc;
+        int err = 0, rc = VINF_SUCCESS;
         struct sf_glob_info *sf_g = GET_GLOB_INFO (old_parent->i_sb);
-        struct sf_inode_info *sf_old_i = GET_INODE_INFO (old_parent);
-        struct sf_inode_info *sf_new_i = GET_INODE_INFO (new_parent);
-        struct sf_inode_info *sf_file_i = GET_INODE_INFO (old_dentry->d_inode);
-        SHFLSTRING *old_path;
-        SHFLSTRING *new_path;
-        SHFLSTRING *temp_path;
 
         TRACE ();
-        BUG_ON (!sf_old_i);
-        BUG_ON (!sf_new_i);
-        BUG_ON (!sf_file_i);
 
         if (sf_g != GET_GLOB_INFO (new_parent->i_sb)) {
-                elog2 ("rename with different roots\n");
-                return -EINVAL;
-        }
+                LogFunc(("rename with different roots\n"));
+                err = -EINVAL;
+        } else {
+                struct sf_inode_info *sf_old_i = GET_INODE_INFO (old_parent);
+                struct sf_inode_info *sf_new_i = GET_INODE_INFO (new_parent);
+                /* As we save the relative path inside the inode structure, we need to change
+                  this if the rename is successful. */
+                struct sf_inode_info *sf_file_i = GET_INODE_INFO (old_dentry->d_inode);
+                SHFLSTRING *old_path;
+                SHFLSTRING *new_path;
 
-        err = sf_path_from_dentry (__func__, sf_g, sf_old_i,
-                                   old_dentry, &old_path);
-        if (err) {
-                elog2 ("failed to create old path\n");
-                return err;
-        }
+                BUG_ON (!sf_old_i);
+                BUG_ON (!sf_new_i);
+                BUG_ON (!sf_file_i);
 
-        err = sf_path_from_dentry (__func__, sf_g, sf_new_i,
-                                   new_dentry, &new_path);
-        if (err) {
-                elog2 ("failed to create new path\n");
-                goto fail0;
-        }
+                old_path = sf_file_i->path;
+                err = sf_path_from_dentry (__func__, sf_g, sf_new_i,
+                                          new_dentry, &new_path);
+                if (err) {
+                        LogFunc(("failed to create new path\n"));
+                } else {
+                        int is_dir = ((old_dentry->d_inode->i_mode & S_IFDIR) != 0);
 
-        rc = vboxCallRename (&client_handle, &sf_g->map, old_path,
-                             new_path, SHFL_RENAME_FILE);
-        if (VBOX_FAILURE (rc)) {
-                switch (rc) {
-                        case VERR_FILE_NOT_FOUND:
-                        case VERR_PATH_NOT_FOUND:
-                                err = -ENOENT;
-                                goto fail1;
-
-                        default:
-                                err = -EPROTO;
-                                elog ("vboxCallRename failed rc=%d\n", rc);
-                                goto fail1;
+                        rc = vboxCallRename (&client_handle, &sf_g->map, old_path,
+                                             new_path, is_dir ? 0 : SHFL_RENAME_FILE | SHFL_RENAME_REPLACE_IF_EXISTS);
+                        if (RT_SUCCESS(rc)) {
+                                kfree (old_path);
+                                sf_new_i->force_restat = 1;
+                                sf_old_i->force_restat = 1; /* XXX: needed? */
+                                /* Set the new relative path in the inode. */
+                                sf_file_i->path = new_path;
+                        } else {
+                                LogFunc(("vboxCallRename failed rc=%Rrc\n", rc));
+                                err = -RTErrConvertToErrno (rc);
+                        }
+                        if (0 != err)
+                                kfree (new_path);
                 }
         }
-
-        err = 0;
-
-        sf_new_i->force_restat = 1;
-        sf_old_i->force_restat = 1; /* XXX: needed? */
-        /* We want to free the old saved path, not the new one. */
-        temp_path = sf_file_i->path;
-        sf_file_i->path = new_path;
-        new_path = temp_path;
- fail1:
-        kfree (old_path);
- fail0:
-        kfree (new_path);
         return err;
 }
 
-static struct inode_operations sf_dir_iops = {
+struct inode_operations sf_dir_iops = {
         .lookup     = sf_lookup,
         .create     = sf_create,
         .mkdir      = sf_mkdir,
