@@ -1,7 +1,6 @@
+/* $Id: VBoxBFE.cpp 18645 2009-04-02 15:38:31Z vboxsync $ */
 /** @file
- *
- * VBox frontends: Basic Frontend (BFE):
- * VBoxBFE main routines
+ * Basic Frontend (BFE): VBoxBFE main routines.
  *
  * VBoxBFE is a limited frontend that sits directly on the Virtual Machine
  * Manager (VMM) and does _not_ use COM to communicate.
@@ -11,15 +10,19 @@
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2009 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 /*******************************************************************************
@@ -42,15 +45,16 @@ using namespace com;
 #ifdef VBOXBFE_WITH_USB
 # include <VBox/vusb.h>
 #endif
-#ifdef VBOX_HGCM
+#ifdef VBOX_WITH_HGCM
 # include <VBox/shflsvc.h>
 #endif
+#include <iprt/alloc.h>
 #include <iprt/alloca.h>
 #include <iprt/assert.h>
 #include <iprt/ctype.h>
 #include <iprt/file.h>
 #include <iprt/path.h>
-#include <iprt/runtime.h>
+#include <iprt/initterm.h>
 #include <iprt/semaphore.h>
 #include <iprt/stream.h>
 #include <iprt/string.h>
@@ -159,7 +163,9 @@ static bool        g_fAudio  = false;
 static bool        g_fUSB    = false;
 #endif
 static char       *g_pszHdaFile   = NULL;
+static bool        g_fHdaSpf = false;
 static char       *g_pszHdbFile   = NULL;
+static bool        g_fHdbSpf = false;
 static char       *g_pszCdromFile = NULL;
 static char       *g_pszFdaFile   = NULL;
        const char *g_pszStateFile = NULL;
@@ -175,6 +181,7 @@ static bool        g_fCSAM  = true;
 static bool        g_fRestoreState = false;
 static const char *g_pszShareDir[MaxSharedFolders];
 static const char *g_pszShareName[MaxSharedFolders];
+static bool        g_fShareReadOnly[MaxSharedFolders];
 static unsigned    g_uNumShares;
 static bool        g_fPreAllocRam = false;
 static int         g_iBootMenu = 2;
@@ -198,10 +205,15 @@ typedef struct BFENetworkDevice
     }           enmType;    /**< The type of network driver. */
     bool        fSniff;     /**< Set if the network sniffer should be installed. */
     const char *pszSniff;   /**< Output file for the network sniffer. */
-    PDMMAC      Mac;        /**< The mac address for the device. */
-    const char *pszName;     /**< The device name of a HIF device. The name of the internal network. */
+    RTMAC       Mac;        /**< The mac address for the device. */
+    const char *pszName;    /**< The device name of a HIF device. The name of the internal network. */
+#ifdef RT_OS_OS2
+    bool        fHaveConnectTo; /**< Whether fConnectTo is set. */
+    int32_t     iConnectTo; /**< The lanX to connect to (bridge with). */
+#elif 1//defined(RT_OS_LINUX)
     bool        fHaveFd;    /**< Set if fd is valid. */
     int32_t     fd;         /**< The file descriptor of a HIF device.*/
+#endif
 } BFENETDEV, *PBFENETDEV;
 
 /** Array of network device configurations. */
@@ -231,9 +243,9 @@ static int networkArg2Index(const char *pszArg, int cchRoot)
 {
     uint32_t n;
     int rc = RTStrToUInt32Ex(&pszArg[cchRoot], NULL, 10, &n);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
     {
-        RTPrintf("Error: invalid network device option (rc=%Vrc): %s\n", rc, pszArg);
+        RTPrintf("Error: invalid network device option (rc=%Rrc): %s\n", rc, pszArg);
         return -1;
     }
     if (n < 1 || n > NetworkAdapterCount)
@@ -264,7 +276,7 @@ int GenerateMACAddress(char pszAddress[MAC_STRING_LEN + 1])
     int rc = RTUuidCreate(&uuid);
     if (RT_FAILURE(rc))
     {
-        LogFlowFunc(("RTUuidCreate failed, returning %Vrc\n", rc));
+        LogFlowFunc(("RTUuidCreate failed, returning %Rrc\n", rc));
         return rc;
     }
     RTStrPrintf(pszAddress, MAC_STRING_LEN + 1, "080027%02X%02X%02X",
@@ -354,7 +366,8 @@ static void show_usage()
              "  -statefile <file>  Define the file name for VM save/restore\n"
              "  -restore           Restore the VM if the statefile exists, normal start otherwise\n"
              "  -nofstoggle        Forbid switching to/from fullscreen mode\n"
-             "  -share <dir><name> Share directory <dir> as name <name>\n"
+             "  -share <dir> <name> [readonly]\n"
+             "                     Share directory <dir> as name <name>. Optionally read-only.\n"
              "  -nohostkey         Disable hostkey\n"
              "  -[no]acpi          Enable or disable ACPI (default: enabled)\n"
              "  -[no]ioapic        Enable or disable the IO-APIC (default: disabled)\n"
@@ -373,10 +386,13 @@ static void show_usage()
 #if 0
              "  -netsniff<1-N>     Enable packet sniffer\n"
 #endif
+#ifdef RT_OS_OS2
+             "  -brdev<1-N> lan<X> Bridge network adaptor <N> with the 'lanX' device.\n"
+#endif
 #ifdef RT_OS_LINUX
              "  -tapfd<1-N> <fd>   Use existing TAP device, don't allocate\n"
 #endif
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
              "  -vrdp [port]       Listen for VRDP connections on port (default if not specified)\n"
 #endif
 #ifdef VBOX_SECURELABEL
@@ -399,36 +415,10 @@ static void show_usage()
 
 
 /** entry point */
-int main(int argc, char **argv)
+extern "C" DECLEXPORT(int) TrustedMain (int argc, char **argv, char **envp)
 {
-#ifdef RT_OS_L4
-#ifndef L4API_l4v2onv4
-    /* clear Fiasco kernel trace buffer */
-    fiasco_tbuf_clear();
-#endif
-    /* set the environment.  Must be done before the runtime is
-       initialised.  Yes, it really must. */
-    for (int i = 0; i < argc; i++)
-        if (strcmp(argv[i], "-env") == 0)
-        {
-            if (++i >= argc)
-                return SyntaxError("missing argument to -env (format: var=value)!\n");
-            /* add it to the environment */
-            if (putenv(argv[i]) != 0)
-                return SyntaxError("Error setting environment string %s.\n", argv[i]);
-        }
-#endif /* RT_OS_L4 */
-
-    /*
-     * Before we do *anything*, we initialize the runtime.
-     */
-    int rc = RTR3Init();
-    if (VBOX_FAILURE(rc))
-        return FatalError("RTR3Init failed rc=%Vrc\n", rc);
-
-
     bool fFullscreen = false;
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     int32_t portVRDP = -1;
 #endif
 #ifdef VBOX_SECURELABEL
@@ -439,8 +429,9 @@ int main(int argc, char **argv)
 #ifdef RT_OS_L4
     uint32_t u32MaxVRAM;
 #endif
+    int rc = VINF_SUCCESS;
 
-    RTPrintf("VirtualBox Simple SDL GUI built %s %s\n", __DATE__, __TIME__);
+    RTPrintf("Sun VirtualBox Simple SDL GUI built %s %s\n", __DATE__, __TIME__);
 
     // less than one parameter is not possible
     if (argc < 2)
@@ -500,8 +491,8 @@ int main(int argc, char **argv)
             if (++curArg >= argc)
                 return SyntaxError("missing argument for memory size!\n");
             rc = RTStrToUInt32Ex(argv[curArg], NULL, 0, &g_u32MemorySizeMB);
-            if (VBOX_FAILURE(rc))
-                return SyntaxError("bad memory size: %s (error %Vrc)\n",
+            if (RT_FAILURE(rc))
+                return SyntaxError("bad memory size: %s (error %Rrc)\n",
                                    argv[curArg], rc);
         }
         else if (strcmp(pszArg, "-vram") == 0)
@@ -511,8 +502,8 @@ int main(int argc, char **argv)
             uint32_t uVRAMMB;
             rc = RTStrToUInt32Ex(argv[curArg], NULL, 0, &uVRAMMB);
             g_u32VRamSize = uVRAMMB * _1M;
-            if (VBOX_FAILURE(rc))
-                return SyntaxError("bad video ram size: %s (error %Vrc)\n",
+            if (RT_FAILURE(rc))
+                return SyntaxError("bad video ram size: %s (error %Rrc)\n",
                                    argv[curArg], rc);
         }
         else if (strcmp(pszArg, "-statefile") == 0)
@@ -533,6 +524,11 @@ int main(int argc, char **argv)
             if (++curArg >= argc)
                 return SyntaxError("missing 2nd argument for share!\n");
             g_pszShareName[g_uNumShares] = argv[curArg];
+            if (curArg < argc-1 && strcmp(argv[curArg+1], "readonly") == 0)
+            {
+                g_fShareReadOnly[g_uNumShares] = true;
+                curArg++;
+            }
             g_uNumShares++;
         }
         else if (strcmp(pszArg, "-fullscreen") == 0)
@@ -569,6 +565,10 @@ int main(int argc, char **argv)
             if (!g_pszHdaFile)
                 return SyntaxError("The path to the specified harddisk, '%s', could not be resolved.\n", argv[curArg]);
         }
+        else if (strcmp(pszArg, "-hdaspf") == 0)
+        {
+            g_fHdaSpf = true;
+        }
         else if (strcmp(pszArg, "-hdb") == 0)
         {
             if (++curArg >= argc)
@@ -580,6 +580,10 @@ int main(int argc, char **argv)
             if (!g_pszHdbFile)
                 return SyntaxError("The path to the specified harddisk, '%s', could not be resolved.\n", argv[curArg]);
         }
+        else if (strcmp(pszArg, "-hdbspf") == 0)
+        {
+            g_fHdbSpf = true;
+        }
         else if (strcmp(pszArg, "-fda") == 0)
         {
             if (++curArg >= argc)
@@ -648,7 +652,7 @@ int main(int argc, char **argv)
             {
                 rc = GenerateMACAddress(szMacGen);
                 if (RT_FAILURE(rc))
-                    return SyntaxError("failed to generate a hardware address for network device %d (error %Vrc)\n",
+                    return SyntaxError("failed to generate a hardware address for network device %d (error %Rrc)\n",
                                        i, rc);
                 pszMac = szMacGen;
             }
@@ -669,25 +673,47 @@ int main(int argc, char **argv)
         }
         else if (strncmp(pszArg, "-netsniff", 9) == 0)
         {
-            int i = networkArg2Index(pszArg, 7);
-            if (rc < 0)
+            int i = networkArg2Index(pszArg, 9);
+            if (i < 0)
                 return 1;
             g_aNetDevs[i].fSniff = true;
             /** @todo filename */
         }
+#ifdef RT_OS_OS2
+        else if (strncmp(pszArg, "-brdev", 6) == 0)
+        {
+            int i = networkArg2Index(pszArg, 6);
+            if (i < 0)
+                return 1;
+            if (g_aNetDevs[i].enmType != BFENETDEV::HIF)
+                return SyntaxError("%d is not a hif device! Make sure you put the -hifdev argument first.\n", i);
+            if (++curArg >= argc)
+                return SyntaxError("missing argument for %s!\n", pszArg);
+            if (    strncmp(argv[curArg], "lan", 3)
+                ||  argv[curArg][3] < '0'
+                ||  argv[curArg][3] >= '8'
+                ||  argv[curArg][4])
+                return SyntaxError("bad interface name '%s' specified with '%s'. Expected 'lan0', 'lan1' and similar.\n",
+                                   argv[curArg], pszArg);
+            g_aNetDevs[i].iConnectTo = argv[curArg][3] - '0';
+            g_aNetDevs[i].fHaveConnectTo = true;
+        }
+#endif
 #ifdef RT_OS_LINUX
         else if (strncmp(pszArg, "-tapfd", 6) == 0)
         {
-            int i = networkArg2Index(pszArg, 7);
+            int i = networkArg2Index(pszArg, 6);
+            if (i < 0)
+                return 1;
             if (++curArg >= argc)
                 return SyntaxError("missing argument for %s!\n", pszArg);
             rc = RTStrToInt32Ex(argv[curArg], NULL, 0, &g_aNetDevs[i].fd);
-            if (VBOX_FAILURE(rc))
-                return SyntaxError("bad tap file descriptor: %s (error %VRc)\n", argv[curArg], rc);
+            if (RT_FAILURE(rc))
+                return SyntaxError("bad tap file descriptor: %s (error %Rrc)\n", argv[curArg], rc);
             g_aNetDevs[i].fHaveFd = true;
         }
 #endif /* RT_OS_LINUX */
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
         else if (strcmp(pszArg, "-vrdp") == 0)
         {
             // -vrdp might take a port number (positive).
@@ -695,13 +721,13 @@ int main(int argc, char **argv)
             if (curArg + 1 < argc && argv[curArg + 1][0] != '-')
             {
                 rc = RTStrToInt32Ex(argv[curArg], NULL, 0, &portVRDP);
-                if (VBOX_FAILURE(rc))
-                    return SyntaxError("cannot vrpd port: %s (%VRc)\n", argv[curArg], rc);
+                if (RT_FAILURE(rc))
+                    return SyntaxError("cannot vrpd port: %s (%Rrc)\n", argv[curArg], rc);
                 if (portVRDP < 0 || portVRDP >= 0x10000)
                     return SyntaxError("vrdp port number is out of range: %RI32\n", portVRDP);
             }
         }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 #ifdef VBOX_SECURELABEL
         else if (strcmp(pszArg, "-securelabel") == 0)
         {
@@ -788,10 +814,9 @@ int main(int argc, char **argv)
      * the meantime, we have to service the SDL event loop.
      */
 
-
     RTTHREAD thread;
     rc = RTThreadCreate(&thread, VMPowerUpThread, 0, 0, RTTHREADTYPE_MAIN_WORKER, 0, "PowerUp");
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
     {
         RTPrintf("Error: Thread creation failed with %d\n", rc);
         return -1;
@@ -805,7 +830,7 @@ int main(int argc, char **argv)
     /* loop until the powerup processing is done */
     do
     {
-#if defined(RT_OS_LINUX) && defined(USE_SDL)
+#if defined(VBOXBFE_WITH_X11) && defined(USE_SDL)
         if (   machineState == VMSTATE_CREATING
             || machineState == VMSTATE_LOADING)
         {
@@ -867,7 +892,6 @@ int main(int argc, char **argv)
     }
 #endif
 
-
     /*
      * Main event loop
      */
@@ -941,7 +965,7 @@ leave:
         {
             /* Power off VM */
             PVMREQ pReq;
-            rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)VMR3PowerOff, 1, pVM);
+            rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)VMR3PowerOff, 1, pVM);
         }
 
         /* And destroy it */
@@ -958,10 +982,45 @@ leave:
     delete gMachineDebugger;
 
     RTLogFlush(NULL);
-    return VBOX_FAILURE (rc) ? 1 : 0;
+    return RT_FAILURE (rc) ? 1 : 0;
 }
 
 
+#ifndef VBOX_WITH_HARDENING
+/**
+ * Main entry point.
+ */
+int main(int argc, char **argv)
+{
+# ifdef RT_OS_L4
+# ifndef L4API_l4v2onv4
+    /* clear Fiasco kernel trace buffer */
+    fiasco_tbuf_clear();
+# endif
+    /* set the environment.  Must be done before the runtime is
+       initialised.  Yes, it really must. */
+    for (int i = 0; i < argc; i++)
+        if (strcmp(argv[i], "-env") == 0)
+        {
+            if (++i >= argc)
+                return SyntaxError("missing argument to -env (format: var=value)!\n");
+            /* add it to the environment */
+            if (putenv(argv[i]) != 0)
+                return SyntaxError("Error setting environment string %s.\n", argv[i]);
+        }
+# endif /* RT_OS_L4 */
+
+    /*
+     * Before we do *anything*, we initialize the runtime.
+     */
+    int rc = RTR3Init();
+    if (RT_FAILURE(rc))
+        return FatalError("RTR3Init failed rc=%Rrc\n", rc);
+
+    return TrustedMain(argc, argv, NULL);
+}
+#endif /* !VBOX_WITH_HARDENING */
+
 
 /**
  * VM state callback function. Called by the VMM
@@ -1028,14 +1087,14 @@ DECLCALLBACK(void) setVMErrorCallback(PVM pVM, void *pvUser, int rc, RT_SRC_POS_
                                       const char *pszFormat, va_list args)
 {
     /** @todo accessing shared resource without any kind of synchronization */
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
         szError[0] = '\0';
     else
     {
         va_list va2;
         va_copy(va2, args); /* Have to make a copy here or GCC will break. */
         RTStrPrintf(szError, sizeof(szError),
-                    "%N!\nVBox status code: %d (%Vrc)", pszFormat, &va2, rc, rc);
+                    "%N!\nVBox status code: %d (%Rrc)", pszFormat, &va2, rc, rc);
         RTPrintf("%s\n", szError);
         va_end(va2);
     }
@@ -1047,19 +1106,22 @@ DECLCALLBACK(void) setVMErrorCallback(PVM pVM, void *pvUser, int rc, RT_SRC_POS_
  *
  * @param   pVM         The VM handle.
  * @param   pvUser      The user argument.
- * @param   fFata       Wheather it is a fatal error or not.
+ * @param   fFlags      The action flags. See VMSETRTERR_FLAGS_*.
  * @param   pszErrorId  Error ID string.
  * @param   pszError    Error message format string.
- * @param   args        Error message arguments.
+ * @param   va          Error message arguments.
  * @thread EMT.
  */
-DECLCALLBACK(void) setVMRuntimeErrorCallback(PVM pVM, void *pvUser, bool fFatal,
+DECLCALLBACK(void) setVMRuntimeErrorCallback(PVM pVM, void *pvUser, uint32_t fFlags,
                                              const char *pszErrorId,
-                                             const char *pszFormat, va_list args)
+                                             const char *pszFormat, va_list va)
 {
     va_list va2;
-    va_copy(va2, args); /* Have to make a copy here or GCC will break. */
-    RTPrintf("%s: %s!\n%N!\n", fFatal ? "Error" : "Warning", pszErrorId, pszFormat, &va2);
+    va_copy(va2, va); /* Have to make a copy here or GCC/AMD64 will break. */
+    RTPrintf("%s: %s!\n%N!\n",
+             fFlags & VMSETRTERR_FLAGS_FATAL ? "Error" : "Warning",
+             pszErrorId, pszFormat, &va2);
+    RTStrmFlush(g_pStdErr);
     va_end(va2);
 }
 
@@ -1079,9 +1141,9 @@ DECLCALLBACK(int) VMPowerUpThread(RTTHREAD Thread, void *pvUser)
         static char szError[RTPATH_MAX + 128] = "";
         PRTLOGGER pLogger;
         rc2 = RTLogCreateEx(&pLogger, RTLOGFLAGS_PREFIX_TIME_PROG, "all",
-                            "VBOX_RELEASE_LOG", ELEMENTS(s_apszGroups), s_apszGroups,
+                            "VBOX_RELEASE_LOG", RT_ELEMENTS(s_apszGroups), s_apszGroups,
                             RTLOGDEST_FILE, szError, sizeof(szError), "./VBoxBFE.log");
-        if (VBOX_SUCCESS(rc2))
+        if (RT_SUCCESS(rc2))
         {
             /* some introductory information */
             RTTIMESPEC TimeSpec;
@@ -1108,24 +1170,25 @@ DECLCALLBACK(int) VMPowerUpThread(RTTHREAD Thread, void *pvUser)
     /*
      * Create empty VM.
      */
-    rc = VMR3Create(setVMErrorCallback, NULL, vboxbfeConfigConstructor, NULL, &pVM);
-    if (VBOX_FAILURE(rc))
+    rc = VMR3Create(1, setVMErrorCallback, NULL, vboxbfeConfigConstructor, NULL, &pVM);
+    if (RT_FAILURE(rc))
     {
-        RTPrintf("Error: VM creation failed with %Vrc.\n", rc);
+        RTPrintf("Error: VM creation failed with %Rrc.\n", rc);
         goto failure;
     }
 
+
     /*
      * Register VM state change handler
      */
     rc = VMR3AtStateRegister(pVM, vmstateChangeCallback, NULL);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
     {
-        RTPrintf("Error: VMR3AtStateRegister failed with %Vrc.\n", rc);
+        RTPrintf("Error: VMR3AtStateRegister failed with %Rrc.\n", rc);
         goto failure;
     }
 
-#ifdef VBOX_HGCM
+#ifdef VBOX_WITH_HGCM
     /*
      * Add shared folders to the VM
      */
@@ -1133,49 +1196,52 @@ DECLCALLBACK(int) VMPowerUpThread(RTTHREAD Thread, void *pvUser)
     {
         for (unsigned i=0; i<g_uNumShares; i++)
         {
-            VBOXHGCMSVCPARM  parms[2];
+            VBOXHGCMSVCPARM  parms[SHFL_CPARMS_ADD_MAPPING];
             SHFLSTRING      *pFolderName, *pMapName;
             int              cbString;
-            PRTUCS2          aHostPath, aMapName;
+            PRTUTF16         aHostPath, aMapName;
             int              rc;
 
-            rc = RTStrUtf8ToUcs2(&aHostPath, g_pszShareDir[i]);
+            rc = RTStrToUtf16(g_pszShareDir[i], &aHostPath);
             AssertRC(rc);
-            rc = RTStrUtf8ToUcs2(&aMapName, g_pszShareName[i]);
+            rc = RTStrToUtf16(g_pszShareName[i], &aMapName);
             AssertRC(rc);
 
-            cbString = (RTStrUcs2Len (aHostPath) + 1) * sizeof (RTUCS2);
+            cbString = (RTUtf16Len (aHostPath) + 1) * sizeof (RTUTF16);
             pFolderName = (SHFLSTRING *) RTMemAllocZ (sizeof (SHFLSTRING) + cbString);
             Assert (pFolderName);
             memcpy (pFolderName->String.ucs2, aHostPath, cbString);
 
             pFolderName->u16Size   = cbString;
-            pFolderName->u16Length = cbString - sizeof(RTUCS2);
+            pFolderName->u16Length = cbString - sizeof(RTUTF16);
 
             parms[0].type = VBOX_HGCM_SVC_PARM_PTR;
             parms[0].u.pointer.addr = pFolderName;
             parms[0].u.pointer.size = sizeof (SHFLSTRING) + cbString;
 
-            cbString = (RTStrUcs2Len (aMapName) + 1) * sizeof (RTUCS2);
+            cbString = (RTUtf16Len (aMapName) + 1) * sizeof (RTUTF16);
             pMapName = (SHFLSTRING *) RTMemAllocZ (sizeof(SHFLSTRING) + cbString);
             Assert (pMapName);
             memcpy (pMapName->String.ucs2, aMapName, cbString);
 
             pMapName->u16Size   = cbString;
-            pMapName->u16Length = cbString - sizeof (RTUCS2);
+            pMapName->u16Length = cbString - sizeof (RTUTF16);
 
             parms[1].type = VBOX_HGCM_SVC_PARM_PTR;
             parms[1].u.pointer.addr = pMapName;
             parms[1].u.pointer.size = sizeof (SHFLSTRING) + cbString;
 
+            parms[2].type = VBOX_HGCM_SVC_PARM_32BIT;
+            parms[2].u.uint32 = !g_fShareReadOnly[i];
+
             rc = gVMMDev->hgcmHostCall ("VBoxSharedFolders",
-                                        SHFL_FN_ADD_MAPPING, 2, &parms[0]);
+                                        SHFL_FN_ADD_MAPPING, SHFL_CPARMS_ADD_MAPPING, &parms[0]);
             AssertRC(rc);
             LogRel(("Added share %s: (%s)\n", g_pszShareName[i], g_pszShareDir[i]));
             RTMemFree (pFolderName);
             RTMemFree (pMapName);
-            RTStrUcs2Free (aHostPath);
-            RTStrUcs2Free (aMapName);
+            RTUtf16Free (aHostPath);
+            RTUtf16Free (aMapName);
         }
     }
 #endif
@@ -1201,7 +1267,7 @@ DECLCALLBACK(int) VMPowerUpThread(RTTHREAD Thread, void *pvUser)
     /*
      * Power on the VM (i.e. start executing).
      */
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
     {
         PVMREQ pReq;
 
@@ -1211,15 +1277,15 @@ DECLCALLBACK(int) VMPowerUpThread(RTTHREAD Thread, void *pvUser)
             && RTPathExists(g_pszStateFile))
         {
             startProgressInfo("Restoring");
-            rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT,
-                             (PFNRT)VMR3Load, 4, pVM, g_pszStateFile, &callProgressInfo, NULL);
+            rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT,
+                             (PFNRT)VMR3Load, 4, pVM, g_pszStateFile, &callProgressInfo, (uintptr_t)NULL);
             endProgressInfo();
-            if (VBOX_SUCCESS(rc))
+            if (RT_SUCCESS(rc))
             {
                 VMR3ReqFree(pReq);
-                rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT,
+                rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT,
                                  (PFNRT)VMR3Resume, 1, pVM);
-                if (VBOX_SUCCESS(rc))
+                if (RT_SUCCESS(rc))
                 {
                     rc = pReq->iStatus;
                     VMR3ReqFree(pReq);
@@ -1227,26 +1293,26 @@ DECLCALLBACK(int) VMPowerUpThread(RTTHREAD Thread, void *pvUser)
                 gDisplay->setRunning();
             }
             else
-                AssertMsgFailed(("VMR3Load failed, rc=%Vrc\n", rc));
+                AssertMsgFailed(("VMR3Load failed, rc=%Rrc\n", rc));
         }
         else
         {
-            rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)VMR3PowerOn, 1, pVM);
-            if (VBOX_SUCCESS(rc))
+            rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)VMR3PowerOn, 1, pVM);
+            if (RT_SUCCESS(rc))
             {
                 rc = pReq->iStatus;
                 AssertRC(rc);
                 VMR3ReqFree(pReq);
             }
             else
-                AssertMsgFailed(("VMR3PowerOn failed, rc=%Vrc\n", rc));
+                AssertMsgFailed(("VMR3PowerOn failed, rc=%Rrc\n", rc));
         }
     }
 
     /*
      * On failure destroy the VM.
      */
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         goto failure;
 
     return 0;
@@ -1279,24 +1345,24 @@ DECLCALLBACK(int) VBoxDriversRegister(PCPDMDRVREGCB pCallbacks, uint32_t u32Vers
 
     rc = pCallbacks->pfnRegister(pCallbacks, &Mouse::DrvReg);
     AssertRC(rc);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         return rc;
     rc = pCallbacks->pfnRegister(pCallbacks, &Keyboard::DrvReg);
     AssertRC(rc);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         return rc;
 
     rc = pCallbacks->pfnRegister(pCallbacks, &VMDisplay::DrvReg);
     AssertRC(rc);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         return rc;
     rc = pCallbacks->pfnRegister(pCallbacks, &VMMDev::DrvReg);
     AssertRC(rc);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         return rc;
 
     rc = pCallbacks->pfnRegister(pCallbacks, &VMStatus::DrvReg);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         return rc;
 
     return VINF_SUCCESS;
@@ -1314,7 +1380,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
     int rcAll = VINF_SUCCESS;
     int rc;
 
-#define UPDATE_RC() do { if (VBOX_FAILURE(rc) && VBOX_SUCCESS(rcAll)) rcAll = rc; } while (0)
+#define UPDATE_RC() do { if (RT_FAILURE(rc) && RT_SUCCESS(rcAll)) rcAll = rc; } while (0)
 
     /*
      * Root values.
@@ -1322,6 +1388,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
     PCFGMNODE pRoot = CFGMR3GetRoot(pVM);
     rc = CFGMR3InsertString(pRoot,  "Name",           "Default VM");                UPDATE_RC();
     rc = CFGMR3InsertInteger(pRoot, "RamSize",        g_u32MemorySizeMB * _1M);     UPDATE_RC();
+    rc = CFGMR3InsertInteger(pRoot, "RamHoleSize",    512U * _1M);                  UPDATE_RC();
     if (g_fPreAllocRam)
     {
         rc = CFGMR3InsertInteger(pRoot, "RamPreAlloc",    1);                       UPDATE_RC();
@@ -1373,6 +1440,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
     rc = CFGMR3InsertInteger(pInst, "Trusted",              1);     /* boolean */   UPDATE_RC();
     rc = CFGMR3InsertNode(pInst,    "Config",         &pCfg);                       UPDATE_RC();
     rc = CFGMR3InsertInteger(pCfg,  "RamSize",        g_u32MemorySizeMB * _1M);     UPDATE_RC();
+    rc = CFGMR3InsertInteger(pCfg,  "RamHoleSize",    512U * _1M);                  UPDATE_RC();
     rc = CFGMR3InsertString(pCfg,   "BootDevice0",    g_pszBootDevice);             UPDATE_RC();
     rc = CFGMR3InsertString(pCfg,   "BootDevice1",    "NONE");                      UPDATE_RC();
     rc = CFGMR3InsertString(pCfg,   "BootDevice2",    "NONE");                      UPDATE_RC();
@@ -1384,15 +1452,6 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
     RTUuidClear(&Uuid);
     rc = CFGMR3InsertBytes(pCfg,    "UUID", &Uuid, sizeof(Uuid));                   UPDATE_RC();
 
-    /* Default: no bios logo. */
-    rc = CFGMR3InsertInteger(pCfg,  "FadeIn",         1);                           UPDATE_RC();
-    rc = CFGMR3InsertInteger(pCfg,  "FadeOut",        0);                           UPDATE_RC();
-    rc = CFGMR3InsertInteger(pCfg,  "LogoTime",       0);                           UPDATE_RC();
-    rc = CFGMR3InsertString(pCfg,   "LogoFile",       "");                          UPDATE_RC();
-
-    /* Boot menu */
-    rc = CFGMR3InsertInteger(pCfg,  "ShowBootMenu",   g_iBootMenu);                 UPDATE_RC();
-
     /*
      * ACPI
      */
@@ -1403,6 +1462,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
         rc = CFGMR3InsertInteger(pInst, "Trusted",        1);       /* boolean */   UPDATE_RC();
         rc = CFGMR3InsertNode(pInst,    "Config",         &pCfg);                   UPDATE_RC();
         rc = CFGMR3InsertInteger(pCfg,  "RamSize",        g_u32MemorySizeMB * _1M); UPDATE_RC();
+        rc = CFGMR3InsertInteger(pCfg,  "RamHoleSize",    512U * _1M);              UPDATE_RC();
         rc = CFGMR3InsertInteger(pCfg,  "IOAPIC",         g_fIOAPIC);               UPDATE_RC();
         rc = CFGMR3InsertInteger(pInst, "PCIDeviceNo",    7);                       UPDATE_RC();
         rc = CFGMR3InsertInteger(pInst, "PCIFunctionNo",  0);                       UPDATE_RC();
@@ -1569,6 +1629,15 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
     rc = CFGMR3InsertNode(pInst,    "Config",         &pCfg);                       UPDATE_RC();
     rc = CFGMR3InsertInteger(pCfg,  "VRamSize",       g_u32VRamSize);               UPDATE_RC();
 
+    /* Default: no bios logo. */
+    rc = CFGMR3InsertInteger(pCfg,  "FadeIn",         1);                           UPDATE_RC();
+    rc = CFGMR3InsertInteger(pCfg,  "FadeOut",        0);                           UPDATE_RC();
+    rc = CFGMR3InsertInteger(pCfg,  "LogoTime",       0);                           UPDATE_RC();
+    rc = CFGMR3InsertString(pCfg,   "LogoFile",       "");                          UPDATE_RC();
+
+    /* Boot menu */
+    rc = CFGMR3InsertInteger(pCfg,  "ShowBootMenu",   g_iBootMenu);                 UPDATE_RC();
+
 #ifdef RT_OS_L4
     /* XXX hard-coded */
     rc = CFGMR3InsertInteger(pCfg,  "HeightReduction", 18);                         UPDATE_RC();
@@ -1599,12 +1668,6 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
 
     if (g_pszHdaFile)
     {
-        const char *szDriver;
-        if (0 == strcmp(RTPathExt(g_pszHdaFile), ".vdi"))
-            szDriver = "VBoxHDD";  /* .vdi */
-        else
-            szDriver = "VD";       /* .vmdk */
-
         rc = CFGMR3InsertNode(pInst,    "LUN#0",          &pLunL0);                 UPDATE_RC();
         rc = CFGMR3InsertString(pLunL0, "Driver",         "Block");                 UPDATE_RC();
         rc = CFGMR3InsertNode(pLunL0,   "Config",         &pCfg);                   UPDATE_RC();
@@ -1612,19 +1675,30 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
         rc = CFGMR3InsertInteger(pCfg,  "Mountable",      0);                       UPDATE_RC();
 
         rc = CFGMR3InsertNode(pLunL0,   "AttachedDriver", &pDrv);                   UPDATE_RC();
-        rc = CFGMR3InsertString(pDrv,   "Driver",         szDriver);                UPDATE_RC();
+        rc = CFGMR3InsertString(pDrv,   "Driver",         "VD");                    UPDATE_RC();
         rc = CFGMR3InsertNode(pDrv,     "Config",         &pCfg);                   UPDATE_RC();
         rc = CFGMR3InsertString(pCfg,   "Path",           g_pszHdaFile);            UPDATE_RC();
+
+        if (g_fHdaSpf)
+        {
+            rc = CFGMR3InsertString(pCfg, "Format",       "SPF");                   UPDATE_RC();
+        }
+        else
+        {
+            char *pcExt = RTPathExt(g_pszHdaFile);
+            if ((pcExt) && (!strcmp(pcExt, ".vdi")))
+            {
+                rc = CFGMR3InsertString(pCfg, "Format",       "VDI");               UPDATE_RC();
+            }
+            else
+            {
+                rc = CFGMR3InsertString(pCfg, "Format",       "VMDK");              UPDATE_RC();
+            }
+        }
     }
 
     if (g_pszHdbFile)
     {
-        const char *szDriver;
-        if (0 == strcmp(RTPathExt(g_pszHdbFile), ".vdi"))
-            szDriver = "VBoxHDD";  /* .vdi */
-        else
-            szDriver = "VD";       /* .vmdk */
-
         rc = CFGMR3InsertNode(pInst,    "LUN#1",          &pLunL1);                 UPDATE_RC();
         rc = CFGMR3InsertString(pLunL1, "Driver",         "Block");                 UPDATE_RC();
         rc = CFGMR3InsertNode(pLunL1,   "Config",         &pCfg);                   UPDATE_RC();
@@ -1632,9 +1706,26 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
         rc = CFGMR3InsertInteger(pCfg,  "Mountable",      0);                       UPDATE_RC();
 
         rc = CFGMR3InsertNode(pLunL1,   "AttachedDriver", &pDrv);                   UPDATE_RC();
-        rc = CFGMR3InsertString(pDrv,   "Driver",         szDriver);                UPDATE_RC();
+        rc = CFGMR3InsertString(pDrv,   "Driver",         "VD");                    UPDATE_RC();
         rc = CFGMR3InsertNode(pDrv,     "Config",         &pCfg);                   UPDATE_RC();
         rc = CFGMR3InsertString(pCfg,   "Path",           g_pszHdbFile);            UPDATE_RC();
+
+        if (g_fHdbSpf)
+        {
+            rc = CFGMR3InsertString(pCfg, "Format",       "SPF");                   UPDATE_RC();
+        }
+        else
+        {
+            char *pcExt = RTPathExt(g_pszHdbFile);
+            if ((pcExt) && (!strcmp(pcExt, ".vdi")))
+            {
+                rc = CFGMR3InsertString(pCfg, "Format",       "VDI");               UPDATE_RC();
+            }
+            else
+            {
+                rc = CFGMR3InsertString(pCfg, "Format",       "VMDK");              UPDATE_RC();
+            }
+        }
     }
 
     if (g_pszCdromFile)
@@ -1668,7 +1759,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
                                              !ulInstance ? 3 : ulInstance - 1 + 8); UPDATE_RC();
             rc = CFGMR3InsertInteger(pInst, "PCIFunctionNo",  0);                   UPDATE_RC();
             rc = CFGMR3InsertNode(pInst,    "Config",         &pCfg);               UPDATE_RC();
-            rc = CFGMR3InsertBytes(pCfg,    "MAC",            &g_aNetDevs[ulInstance].Mac, sizeof(PDMMAC));
+            rc = CFGMR3InsertBytes(pCfg,    "MAC",            &g_aNetDevs[ulInstance].Mac, sizeof(RTMAC));
                                                                                     UPDATE_RC();
 
             /*
@@ -1720,15 +1811,13 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
 #if defined(RT_OS_LINUX)
                 if (g_aNetDevs[ulInstance].fHaveFd)
                 {
-                    rc = CFGMR3InsertString(pCfg, "Device", g_aNetDevs[ulInstance].pszName);
-                                                                                    UPDATE_RC();
-                    rc = CFGMR3InsertInteger(pCfg, "FileHandle", g_aNetDevs[ulInstance].fd);
-                                                                                    UPDATE_RC();
+                    rc = CFGMR3InsertString(pCfg, "Device", g_aNetDevs[ulInstance].pszName); UPDATE_RC();
+                    rc = CFGMR3InsertInteger(pCfg, "FileHandle", g_aNetDevs[ulInstance].fd); UPDATE_RC();
                 }
                 else
 #endif
                 {
-#if defined (RT_OS_LINUX) || defined (RT_OS_L4)
+#if defined(RT_OS_LINUX) || defined(RT_OS_L4)
                     /*
                      * Create/Open the TAP the device.
                      */
@@ -1736,9 +1825,9 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
                     rc = RTFileOpen(&tapFD, "/dev/net/tun",
                                     RTFILE_O_READWRITE | RTFILE_O_OPEN |
                                     RTFILE_O_DENY_NONE | RTFILE_O_INHERIT);
-                    if (VBOX_FAILURE(rc))
+                    if (RT_FAILURE(rc))
                     {
-                        FatalError("Failed to open /dev/net/tun: %Vrc\n", rc);
+                        FatalError("Failed to open /dev/net/tun: %Rrc\n", rc);
                         return rc;
                     }
 
@@ -1762,7 +1851,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
                     if (rc)
                     {
                         int rc2 = RTErrConvertFromErrno(errno);
-                        FatalError("ioctl TUNSETIFF '%s' failed: errno=%d rc=%d (%Vrc)\n",
+                        FatalError("ioctl TUNSETIFF '%s' failed: errno=%d rc=%d (%Rrc)\n",
                                    IfReq.ifr_name, errno, rc, rc2);
                         return rc2;
                     }
@@ -1771,7 +1860,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
                     if (rc)
                     {
                         int rc2 = RTErrConvertFromErrno(errno);
-                        FatalError("fcntl F_SETFL/O_NONBLOCK '%s' failed: errno=%d rc=%d (%Vrc)\n",
+                        FatalError("fcntl F_SETFL/O_NONBLOCK '%s' failed: errno=%d rc=%d (%Rrc)\n",
                                    IfReq.ifr_name, errno, rc, rc2);
                         return rc2;
                     }
@@ -1779,6 +1868,24 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
                     rc = CFGMR3InsertString(pCfg, "Device", g_aNetDevs[ulInstance].pszName);        UPDATE_RC();
                     rc = CFGMR3InsertInteger(pCfg, "FileHandle", (RTFILE)tapFD);                    UPDATE_RC();
 
+#elif defined(RT_OS_SOLARIS)
+                    rc = CFGMR3InsertString(pCfg, "Device", g_aNetDevs[ulInstance].pszName); UPDATE_RC();
+# ifdef VBOX_WITH_CROSSBOW
+                    rc = CFGMR3InsertBytes(pCfg, "MAC", &g_aNetDevs[ulInstance].Mac, sizeof(g_aNetDevs[ulInstance].Mac));
+                    UPDATE_RC();
+# endif
+
+#elif defined(RT_OS_OS2)
+                    /*
+                     * The TAP driver does all the opening and setting up,
+                     * as it was originally was ment to be (stupid fork() problems).
+                     */
+                    rc = CFGMR3InsertString(pCfg, "Device", g_aNetDevs[ulInstance].pszName); UPDATE_RC();
+                    if (g_aNetDevs[ulInstance].fHaveConnectTo)
+                    {
+                        rc = CFGMR3InsertInteger(pCfg, "ConnectTo", g_aNetDevs[ulInstance].iConnectTo);
+                        UPDATE_RC();
+                    }
 #elif defined(RT_OS_WINDOWS)
                     /*
                      * We need the GUID too here...
@@ -1788,7 +1895,7 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
                     rc = CFGMR3InsertString(pCfg, "GUID", g_aNetDevs[ulInstance].pszName /*pszGUID*/);  UPDATE_RC();
 
 
-#else /* !RT_OS_LINUX && !RT_OS_L4 */
+#else
                     FatalError("Name based HIF devices not implemented yet for this host platform\n");
                     return VERR_NOT_IMPLEMENTED;
 #endif
@@ -1875,3 +1982,4 @@ static DECLCALLBACK(int) vboxbfeConfigConstructor(PVM pVM, void *pvUser)
 
     return rc;
 }
+
