@@ -1,38 +1,45 @@
+/* $Id: ConsoleVRDPServer.cpp 17669 2009-03-11 09:56:29Z vboxsync $ */
+
 /** @file
  *
  * VBox Console VRDP Helper class
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2008 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 #include "ConsoleVRDPServer.h"
 #include "ConsoleImpl.h"
 #include "DisplayImpl.h"
-#ifdef VRDP_NO_COM
 #include "KeyboardImpl.h"
 #include "MouseImpl.h"
-#include <VBox/VRDPOrders.h>
-#endif /* VRDP_NO_COM */
 
 #include "Logging.h"
 
 #include <iprt/asm.h>
 #include <iprt/ldr.h>
+#include <iprt/param.h>
+#include <iprt/path.h>
 #include <iprt/alloca.h>
 
 #include <VBox/err.h>
+#ifdef VBOX_WITH_VRDP
+#include <VBox/VRDPOrders.h>
+#endif /* VBOX_WITH_VRDP */
 
-#ifdef VRDP_NO_COM
 class VRDPConsoleCallback : public IConsoleCallback
 {
 public:
@@ -133,6 +140,11 @@ public:
         return S_OK;
     }
 
+    STDMETHOD(OnStorageControllerChange) ()
+    {
+        return S_OK;
+    }
+
     STDMETHOD(OnVRDPServerChange)()
     {
         return S_OK;
@@ -154,7 +166,7 @@ public:
         return S_OK;
     }
 
-    STDMETHOD(OnRuntimeError)(BOOL fatal, INPTR BSTR id, INPTR BSTR message)
+    STDMETHOD(OnRuntimeError)(BOOL fatal, IN_BSTR id, IN_BSTR message)
     {
         return S_OK;
     }
@@ -411,10 +423,16 @@ static void mousePointerGenerateANDMask (uint8_t *pu8DstAndMask, int cbDstAndMas
     }
 }
 
-STDMETHODIMP VRDPConsoleCallback::OnMousePointerShapeChange (BOOL visible, BOOL alpha, ULONG xHot, ULONG yHot,
-                                                             ULONG width, ULONG height, BYTE *shape)
+STDMETHODIMP VRDPConsoleCallback::OnMousePointerShapeChange (
+    BOOL visible,
+    BOOL alpha,
+    ULONG xHot,
+    ULONG yHot,
+    ULONG width,
+    ULONG height,
+    BYTE *shape)
 {
-    LogSunlover(("VRDPConsoleCallback::OnMousePointerShapeChange: %d, %d, %dx%d, @%d,%d\n", visible, alpha, width, height, xHot, yHot));
+    LogSunlover(("VRDPConsoleCallback::OnMousePointerShapeChange: %d, %d, %lux%lu, @%lu,%lu\n", visible, alpha, width, height, xHot, yHot));
 
     if (m_server)
     {
@@ -568,16 +586,14 @@ STDMETHODIMP VRDPConsoleCallback::OnMousePointerShapeChange (BOOL visible, BOOL
 
     return S_OK;
 }
-#endif /* VRDP_NO_COM */
 
 
 // ConsoleVRDPServer
 ////////////////////////////////////////////////////////////////////////////////
 
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
 RTLDRMOD ConsoleVRDPServer::mVRDPLibrary;
 
-#ifdef VRDP_NO_COM
 PFNVRDPCREATESERVER ConsoleVRDPServer::mpfnVRDPCreateServer = NULL;
 
 VRDPENTRYPOINTS_1 *ConsoleVRDPServer::mpEntryPoints = NULL;
@@ -598,23 +614,7 @@ VRDPCALLBACKS_1 ConsoleVRDPServer::mCallbacks =
     ConsoleVRDPServer::VRDPCallbackInput,
     ConsoleVRDPServer::VRDPCallbackVideoModeHint
 };
-#else
-int  (VBOXCALL *ConsoleVRDPServer::mpfnVRDPStartServer)     (IConsole *pConsole, IVRDPServer *pVRDPServer, HVRDPSERVER *phServer);
-int  (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSetFramebuffer)  (HVRDPSERVER hServer, IFramebuffer *pFramebuffer, uint32_t fFlags);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSetCallback)     (HVRDPSERVER hServer, VRDPSERVERCALLBACK *pcallback, void *pvUser);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPShutdownServer)  (HVRDPSERVER hServer);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSendUpdateBitmap)(HVRDPSERVER hServer, unsigned uScreenId, unsigned x, unsigned y, unsigned w, unsigned h);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSendResize)      (HVRDPSERVER hServer);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSendAudioSamples)(HVRDPSERVER hserver, void *pvSamples, uint32_t cSamples, VRDPAUDIOFORMAT format);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSendAudioVolume) (HVRDPSERVER hserver, uint16_t left, uint16_t right);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSendUSBRequest)  (HVRDPSERVER hserver, uint32_t u32ClientId, void *pvParms, uint32_t cbParms);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPSendUpdate)      (HVRDPSERVER hServer, unsigned uScreenId, void *pvUpdate, uint32_t cbUpdate);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPQueryInfo)       (HVRDPSERVER hserver, uint32_t index, void *pvBuffer, uint32_t cbBuffer, uint32_t *pcbOut);
-void (VBOXCALL *ConsoleVRDPServer::mpfnVRDPClipboard)       (HVRDPSERVER hserver, uint32_t u32Function, uint32_t u32Format, const void *pvData, uint32_t cbData, uint32_t *pcbActualRead);
-#endif /* VRDP_NO_COM */
-#endif /* VBOX_VRDP */
 
-#ifdef VRDP_NO_COM
 DECLCALLBACK(int)  ConsoleVRDPServer::VRDPCallbackQueryProperty (void *pvCallback, uint32_t index, void *pvBuffer, uint32_t cbBuffer, uint32_t *pcbOut)
 {
     ConsoleVRDPServer *server = static_cast <ConsoleVRDPServer *> (pvCallback);
@@ -649,19 +649,19 @@ DECLCALLBACK(int)  ConsoleVRDPServer::VRDPCallbackQueryProperty (void *pvCallbac
         {
             com::Bstr bstr;
             server->mConsole->getVRDPServer ()->COMGETTER(NetAddress) (bstr.asOutParam());
-            
+
             /* The server expects UTF8. */
             com::Utf8Str address = bstr;
 
             size_t cbAddress = address.length () + 1;
-            
+
             if (cbAddress >= 0x10000)
             {
                 /* More than 64K seems to be an  invalid address. */
                 rc = VERR_TOO_MUCH_DATA;
                 break;
             }
-            
+
             if ((size_t)cbBuffer >= cbAddress)
             {
                 if (cbAddress > 0)
@@ -990,7 +990,7 @@ DECLCALLBACK(void) ConsoleVRDPServer::VRDPCallbackInput (void *pvCallback, int t
                 server->m_InputSynch.fClientCapsLock   = (pInputSynch->uLockStatus & VRDP_INPUT_SYNCH_CAPITAL) != 0;
                 server->m_InputSynch.fClientScrollLock = (pInputSynch->uLockStatus & VRDP_INPUT_SYNCH_SCROLL) != 0;
 
-                /* The client initiated synchronization. Always make the guest to reflect the client state. 
+                /* The client initiated synchronization. Always make the guest to reflect the client state.
                  * Than means, when the guest changes the state itself, it is forced to return to the client
                  * state.
                  */
@@ -1019,7 +1019,7 @@ DECLCALLBACK(void) ConsoleVRDPServer::VRDPCallbackVideoModeHint (void *pvCallbac
 
     server->mConsole->getDisplay ()->SetVideoModeHint(cWidth, cHeight, cBitsPerPixel, uScreenId);
 }
-#endif /* VRDP_NO_COM */
+#endif /* VBOX_WITH_VRDP */
 
 ConsoleVRDPServer::ConsoleVRDPServer (Console *console)
 {
@@ -1040,10 +1040,9 @@ ConsoleVRDPServer::ConsoleVRDPServer (Console *console)
     mUSBBackends.event = 0;
 #endif
 
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     mhServer = 0;
 
-#ifdef VRDP_NO_COM
     m_fGuestWantsAbsolute = false;
     m_mousex = 0;
     m_mousey = 0;
@@ -1054,18 +1053,17 @@ ConsoleVRDPServer::ConsoleVRDPServer (Console *console)
     m_InputSynch.fGuestNumLock    = false;
     m_InputSynch.fGuestCapsLock   = false;
     m_InputSynch.fGuestScrollLock = false;
-    
+
     m_InputSynch.fClientNumLock    = false;
     m_InputSynch.fClientCapsLock   = false;
     m_InputSynch.fClientScrollLock = false;
-    
+
     memset (maFramebuffers, 0, sizeof (maFramebuffers));
 
     mConsoleCallback = new VRDPConsoleCallback(this);
     mConsoleCallback->AddRef();
     console->RegisterCallback(mConsoleCallback);
-#endif /* VRDP_NO_COM */
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 
     mAuthLibrary = 0;
 }
@@ -1074,7 +1072,7 @@ ConsoleVRDPServer::~ConsoleVRDPServer ()
 {
     Stop ();
 
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mConsoleCallback)
     {
         mConsole->UnregisterCallback(mConsoleCallback);
@@ -1083,7 +1081,7 @@ ConsoleVRDPServer::~ConsoleVRDPServer ()
     }
 
     unsigned i;
-    for (i = 0; i < ELEMENTS(maFramebuffers); i++)
+    for (i = 0; i < RT_ELEMENTS(maFramebuffers); i++)
     {
         if (maFramebuffers[i])
         {
@@ -1091,7 +1089,8 @@ ConsoleVRDPServer::~ConsoleVRDPServer ()
             maFramebuffers[i] = NULL;
         }
     }
-#endif /* VRDP_NO_COM */
+#endif /* VBOX_WITH_VRDP */
+
     if (RTCritSectIsInitialized (&mCritSect))
     {
         RTCritSectDelete (&mCritSect);
@@ -1102,7 +1101,7 @@ ConsoleVRDPServer::~ConsoleVRDPServer ()
 int ConsoleVRDPServer::Launch (void)
 {
     LogFlowMember(("ConsoleVRDPServer::Launch\n"));
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     int rc = VINF_SUCCESS;
     IVRDPServer *vrdpserver = mConsole->getVRDPServer ();
     Assert(vrdpserver);
@@ -1111,80 +1110,79 @@ int ConsoleVRDPServer::Launch (void)
     HRESULT rc2 = vrdpserver->COMGETTER(Enabled) (&vrdpEnabled);
     AssertComRC(rc2);
 
-    if (SUCCEEDED (rc2)
-        && vrdpEnabled
-        && loadVRDPLibrary ())
+    if (SUCCEEDED (rc2) && vrdpEnabled)
     {
-#ifdef VRDP_NO_COM
-        rc = mpfnVRDPCreateServer (&mCallbacks.header, this, (VRDPINTERFACEHDR **)&mpEntryPoints, &mhServer);
-#else
-        rc = mpfnVRDPStartServer(mConsole, vrdpserver, &mhServer);
-#endif /* VRDP_NO_COM */
-
-        if (VBOX_SUCCESS(rc))
+        if (loadVRDPLibrary ())
         {
+            rc = mpfnVRDPCreateServer (&mCallbacks.header, this, (VRDPINTERFACEHDR **)&mpEntryPoints, &mhServer);
+
+            if (RT_SUCCESS(rc))
+            {
 #ifdef VBOX_WITH_USB
-            remoteUSBThreadStart ();
+                remoteUSBThreadStart ();
 #endif /* VBOX_WITH_USB */
+            }
+            else
+                AssertMsgFailed(("Could not start VRDP server: rc = %Rrc\n", rc));
         }
         else
-            AssertMsgFailed(("Could not start VRDP server: rc = %Vrc\n", rc));
+        {
+            AssertMsgFailed(("Could not load the VRDP library\n"));
+            rc = VERR_FILE_NOT_FOUND;
+        }
     }
 #else
     int rc = VERR_NOT_SUPPORTED;
     LogRel(("VRDP: this version does not include the VRDP server.\n"));
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
     return rc;
 }
 
-#ifdef VRDP_NO_COM
 void ConsoleVRDPServer::EnableConnections (void)
 {
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPEnableConnections (mhServer, true);
     }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
+}
+
+void ConsoleVRDPServer::DisconnectClient (uint32_t u32ClientId, bool fReconnect)
+{
+#ifdef VBOX_WITH_VRDP
+    if (mpEntryPoints && mhServer)
+    {
+        mpEntryPoints->VRDPDisconnect (mhServer, u32ClientId, fReconnect);
+    }
+#endif /* VBOX_WITH_VRDP */
 }
 
 void ConsoleVRDPServer::MousePointerUpdate (const VRDPCOLORPOINTER *pPointer)
 {
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPColorPointer (mhServer, pPointer);
     }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 }
 
 void ConsoleVRDPServer::MousePointerHide (void)
 {
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPHidePointer (mhServer);
     }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 }
-#else
-void ConsoleVRDPServer::SetCallback (void)
-{
-#ifdef VBOX_VRDP
-    /* This is called after VM is created and allows the server to accept client connection. */
-    if (mhServer && mpfnVRDPSetCallback)
-    {
-        mpfnVRDPSetCallback (mhServer, mConsole->getVrdpServerCallback (), mConsole);
-    }
-#endif /* VBOX_VRDP */
-}
-#endif /* VRDP_NO_COM */
 
 void ConsoleVRDPServer::Stop (void)
 {
     Assert(VALID_PTR(this)); /** @todo r=bird: there are(/was) some odd cases where this buster was invalid on
                               * linux. Just remove this when it's 100% sure that problem has been fixed. */
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     if (mhServer)
     {
         HVRDPSERVER hServer = mhServer;
@@ -1192,16 +1190,12 @@ void ConsoleVRDPServer::Stop (void)
         /* Reset the handle to avoid further calls to the server. */
         mhServer = 0;
 
-#ifdef VRDP_NO_COM
         if (mpEntryPoints && hServer)
         {
             mpEntryPoints->VRDPDestroy (hServer);
         }
-#else
-        mpfnVRDPShutdownServer (hServer);
-#endif /* VRDP_NO_COM */
     }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 
 #ifdef VBOX_WITH_USB
     remoteUSBThreadStop ();
@@ -1275,7 +1269,7 @@ bool ConsoleVRDPServer::isRemoteUSBThreadRunning (void)
 void ConsoleVRDPServer::waitRemoteUSBThreadEvent (unsigned cMillies)
 {
     int rc = RTSemEventWait (mUSBBackends.event, cMillies);
-    Assert (VBOX_SUCCESS(rc) || rc == VERR_TIMEOUT);
+    Assert (RT_SUCCESS(rc) || rc == VERR_TIMEOUT);
     NOREF(rc);
 }
 
@@ -1283,21 +1277,21 @@ void ConsoleVRDPServer::remoteUSBThreadStart (void)
 {
     int rc = RTSemEventCreate (&mUSBBackends.event);
 
-    if (VBOX_FAILURE (rc))
+    if (RT_FAILURE (rc))
     {
         AssertFailed ();
         mUSBBackends.event = 0;
     }
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         rc = RTThreadCreate (&mUSBBackends.thread, threadRemoteUSB, this, 65536,
                              RTTHREADTYPE_VRDP_IO, RTTHREADFLAGS_WAITABLE, "remote usb");
     }
 
-    if (VBOX_FAILURE (rc))
+    if (RT_FAILURE (rc))
     {
-        LogRel(("Warning: could not start the remote USB thread, rc = %Vrc!!!\n", rc));
+        LogRel(("Warning: could not start the remote USB thread, rc = %Rrc!!!\n", rc));
         mUSBBackends.thread = NIL_RTTHREAD;
     }
     else
@@ -1305,7 +1299,7 @@ void ConsoleVRDPServer::remoteUSBThreadStart (void)
         /* Wait until the thread is ready. */
         rc = RTThreadUserWait (mUSBBackends.thread, 60000);
         AssertRC (rc);
-        Assert (mUSBBackends.fThreadRunning || VBOX_FAILURE (rc));
+        Assert (mUSBBackends.fThreadRunning || RT_FAILURE (rc));
     }
 }
 
@@ -1341,7 +1335,7 @@ VRDPAuthResult ConsoleVRDPServer::Authenticate (const Guid &uuid, VRDPAuthGuestJ
 
     memcpy (rawuuid, ((Guid &)uuid).ptr (), sizeof (rawuuid));
 
-    LogFlow(("ConsoleVRDPServer::Authenticate: uuid = %Vuuid, guestJudgement = %d, pszUser = %s, pszPassword = %s, pszDomain = %s, u32ClientId = %d\n",
+    LogFlow(("ConsoleVRDPServer::Authenticate: uuid = %RTuuid, guestJudgement = %d, pszUser = %s, pszPassword = %s, pszDomain = %s, u32ClientId = %d\n",
              rawuuid, guestJudgement, pszUser, pszPassword, pszDomain, u32ClientId));
 
     /*
@@ -1369,26 +1363,26 @@ VRDPAuthResult ConsoleVRDPServer::Authenticate (const Guid &uuid, VRDPAuthGuestJ
         LogRel(("VRDPAUTH: ConsoleVRDPServer::Authenticate: loading external authentication library '%ls'\n", authLibrary.raw()));
 
         int rc = RTLdrLoad (filename.raw(), &mAuthLibrary);
-        if (VBOX_FAILURE (rc))
-            LogRel(("VRDPAUTH: Failed to load external authentication library. Error code: %Vrc\n", rc));
+        if (RT_FAILURE (rc))
+            LogRel(("VRDPAUTH: Failed to load external authentication library. Error code: %Rrc\n", rc));
 
-        if (VBOX_SUCCESS (rc))
+        if (RT_SUCCESS (rc))
         {
             /* Get the entry point. */
             mpfnAuthEntry2 = NULL;
             int rc2 = RTLdrGetSymbol(mAuthLibrary, "VRDPAuth2", (void**)&mpfnAuthEntry2);
-            if (VBOX_FAILURE (rc2))
+            if (RT_FAILURE (rc2))
             {
-                LogRel(("VRDPAUTH: Could not resolve import '%s'. Error code: %Vrc\n", "VRDPAuth2", rc2));
+                LogRel(("VRDPAUTH: Could not resolve import '%s'. Error code: %Rrc\n", "VRDPAuth2", rc2));
                 rc = rc2;
             }
 
             /* Get the entry point. */
             mpfnAuthEntry = NULL;
             rc2 = RTLdrGetSymbol(mAuthLibrary, "VRDPAuth", (void**)&mpfnAuthEntry);
-            if (VBOX_FAILURE (rc2))
+            if (RT_FAILURE (rc2))
             {
-                LogRel(("VRDPAUTH: Could not resolve import '%s'. Error code: %Vrc\n", "VRDPAuth", rc2));
+                LogRel(("VRDPAUTH: Could not resolve import '%s'. Error code: %Rrc\n", "VRDPAuth", rc2));
                 rc = rc2;
             }
 
@@ -1399,7 +1393,7 @@ VRDPAuthResult ConsoleVRDPServer::Authenticate (const Guid &uuid, VRDPAuthGuestJ
             }
         }
 
-        if (VBOX_FAILURE (rc))
+        if (RT_FAILURE (rc))
         {
             mConsole->reportAuthLibraryError (filename.raw(), rc);
 
@@ -1449,7 +1443,7 @@ void ConsoleVRDPServer::AuthDisconnect (const Guid &uuid, uint32_t u32ClientId)
 
     memcpy (rawuuid, ((Guid &)uuid).ptr (), sizeof (rawuuid));
 
-    LogFlow(("ConsoleVRDPServer::AuthDisconnect: uuid = %Vuuid, u32ClientId = %d\n",
+    LogFlow(("ConsoleVRDPServer::AuthDisconnect: uuid = %RTuuid, u32ClientId = %d\n",
              rawuuid, u32ClientId));
 
     Assert (mAuthLibrary && (mpfnAuthEntry || mpfnAuthEntry2));
@@ -1527,7 +1521,7 @@ DECLCALLBACK(int) ConsoleVRDPServer::ClipboardServiceExtension (void *pvExtensio
 
     int rc = VINF_SUCCESS;
 
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
     ConsoleVRDPServer *pServer = static_cast <ConsoleVRDPServer *>(pvExtension);
 
     VBOXCLIPBOARDEXTPARMS *pParms = (VBOXCLIPBOARDEXTPARMS *)pvParms;
@@ -1536,13 +1530,12 @@ DECLCALLBACK(int) ConsoleVRDPServer::ClipboardServiceExtension (void *pvExtensio
     {
         case VBOX_CLIPBOARD_EXT_FN_SET_CALLBACK:
         {
-            pServer->mpfnClipboardCallback = (PFNVRDPCLIPBOARDEXTCALLBACK)pParms->pvData;
+            pServer->mpfnClipboardCallback = pParms->u.pfnCallback;
         } break;
 
         case VBOX_CLIPBOARD_EXT_FN_FORMAT_ANNOUNCE:
         {
             /* The guest announces clipboard formats. This must be delivered to all clients. */
-#ifdef VRDP_NO_COM
             if (mpEntryPoints && pServer->mhServer)
             {
                 mpEntryPoints->VRDPClipboard (pServer->mhServer,
@@ -1552,17 +1545,6 @@ DECLCALLBACK(int) ConsoleVRDPServer::ClipboardServiceExtension (void *pvExtensio
                                               0,
                                               NULL);
             }
-#else
-            if (mpfnVRDPClipboard)
-            {
-                mpfnVRDPClipboard (pServer->mhServer,
-                                   VRDP_CLIPBOARD_FUNCTION_FORMAT_ANNOUNCE,
-                                   pParms->u32Format,
-                                   NULL,
-                                   0,
-                                   NULL);
-            }
-#endif /* VRDP_NO_COM */
         } break;
 
         case VBOX_CLIPBOARD_EXT_FN_DATA_READ:
@@ -1571,91 +1553,54 @@ DECLCALLBACK(int) ConsoleVRDPServer::ClipboardServiceExtension (void *pvExtensio
              * with clipboard data. The server returns the data from the client that
              * announced the requested format most recently.
              */
-#ifdef VRDP_NO_COM
             if (mpEntryPoints && pServer->mhServer)
             {
                 mpEntryPoints->VRDPClipboard (pServer->mhServer,
                                               VRDP_CLIPBOARD_FUNCTION_DATA_READ,
                                               pParms->u32Format,
-                                              pParms->pvData,
+                                              pParms->u.pvData,
                                               pParms->cbData,
                                               &pParms->cbData);
             }
-#else
-            if (mpfnVRDPClipboard)
-            {
-                mpfnVRDPClipboard (pServer->mhServer,
-                                   VRDP_CLIPBOARD_FUNCTION_DATA_READ,
-                                   pParms->u32Format,
-                                   pParms->pvData,
-                                   pParms->cbData,
-                                   &pParms->cbData);
-            }
-#endif /* VRDP_NO_COM */
         } break;
 
         case VBOX_CLIPBOARD_EXT_FN_DATA_WRITE:
         {
-#ifdef VRDP_NO_COM
             if (mpEntryPoints && pServer->mhServer)
             {
                 mpEntryPoints->VRDPClipboard (pServer->mhServer,
                                               VRDP_CLIPBOARD_FUNCTION_DATA_WRITE,
                                               pParms->u32Format,
-                                              pParms->pvData,
+                                              pParms->u.pvData,
                                               pParms->cbData,
                                               NULL);
             }
-#else
-            if (mpfnVRDPClipboard)
-            {
-                mpfnVRDPClipboard (pServer->mhServer,
-                                   VRDP_CLIPBOARD_FUNCTION_DATA_WRITE,
-                                   pParms->u32Format,
-                                   pParms->pvData,
-                                   pParms->cbData,
-                                   NULL);
-            }
-#endif /* VRDP_NO_COM */
         } break;
 
         default:
             rc = VERR_NOT_SUPPORTED;
     }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 
     return rc;
 }
 
-#ifdef VRDP_NO_COM
 void ConsoleVRDPServer::ClipboardCreate (uint32_t u32ClientId)
-#else
-void ConsoleVRDPServer::ClipboardCreate (uint32_t u32ClientId, PFNVRDPCLIPBOARDCALLBACK *ppfn, void **ppv)
-#endif /* VRDP_NO_COM */
 {
     int rc = lockConsoleVRDPServer ();
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         if (mcClipboardRefs == 0)
         {
             rc = HGCMHostRegisterServiceExtension (&mhClipboard, "VBoxSharedClipboard", ClipboardServiceExtension, this);
 
-            if (VBOX_SUCCESS (rc))
+            if (RT_SUCCESS (rc))
             {
                 mcClipboardRefs++;
             }
         }
 
-#ifdef VRDP_NO_COM
-#else
-        if (VBOX_SUCCESS (rc))
-        {
-            *ppfn = ClipboardCallback;
-            *ppv = this;
-        }
-#endif /* VRDP_NO_COM */
-
         unlockConsoleVRDPServer ();
     }
 }
@@ -1664,7 +1609,7 @@ void ConsoleVRDPServer::ClipboardDelete (uint32_t u32ClientId)
 {
     int rc = lockConsoleVRDPServer ();
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         mcClipboardRefs--;
 
@@ -1680,11 +1625,7 @@ void ConsoleVRDPServer::ClipboardDelete (uint32_t u32ClientId)
 /* That is called on INPUT thread of the VRDP server.
  * The ConsoleVRDPServer keeps a list of created backend instances.
  */
-#ifdef VRDP_NO_COM
 void ConsoleVRDPServer::USBBackendCreate (uint32_t u32ClientId, void **ppvIntercept)
-#else
-void ConsoleVRDPServer::USBBackendCreate (uint32_t u32ClientId, PFNVRDPUSBCALLBACK *ppfn, void **ppv)
-#endif /* VRDP_NO_COM */
 {
 #ifdef VBOX_WITH_USB
     LogFlow(("ConsoleVRDPServer::USBBackendCreate: u32ClientId = %d\n", u32ClientId));
@@ -1699,7 +1640,7 @@ void ConsoleVRDPServer::USBBackendCreate (uint32_t u32ClientId, PFNVRDPUSBCALLBA
         /* Append the new instance in the list. */
         int rc = lockConsoleVRDPServer ();
 
-        if (VBOX_SUCCESS (rc))
+        if (RT_SUCCESS (rc))
         {
             pRemoteUSBBackend->pNext = mUSBBackends.pHead;
             if (mUSBBackends.pHead)
@@ -1715,17 +1656,13 @@ void ConsoleVRDPServer::USBBackendCreate (uint32_t u32ClientId, PFNVRDPUSBCALLBA
 
             unlockConsoleVRDPServer ();
 
-#ifdef VRDP_NO_COM
             if (ppvIntercept)
             {
                 *ppvIntercept = pRemoteUSBBackend;
             }
-#else
-            pRemoteUSBBackend->QueryVRDPCallbackPointer (ppfn, ppv);
-#endif /* VRDP_NO_COM */
         }
 
-        if (VBOX_FAILURE (rc))
+        if (RT_FAILURE (rc))
         {
             pRemoteUSBBackend->Release ();
         }
@@ -1743,7 +1680,7 @@ void ConsoleVRDPServer::USBBackendDelete (uint32_t u32ClientId)
     /* Find the instance. */
     int rc = lockConsoleVRDPServer ();
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         pRemoteUSBBackend = usbBackendFind (u32ClientId);
 
@@ -1772,7 +1709,7 @@ void *ConsoleVRDPServer::USBBackendRequestPointer (uint32_t u32ClientId, const G
     /* Find the instance. */
     int rc = lockConsoleVRDPServer ();
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         pRemoteUSBBackend = usbBackendFind (u32ClientId);
 
@@ -1812,7 +1749,7 @@ void ConsoleVRDPServer::USBBackendReleasePointer (const Guid *pGuid)
     /* Find the instance. */
     int rc = lockConsoleVRDPServer ();
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         pRemoteUSBBackend = usbBackendFindByUUID (pGuid);
 
@@ -1840,7 +1777,7 @@ RemoteUSBBackend *ConsoleVRDPServer::usbBackendGetNext (RemoteUSBBackend *pRemot
 
     int rc = lockConsoleVRDPServer ();
 
-    if (VBOX_SUCCESS (rc))
+    if (RT_SUCCESS (rc))
     {
         if (pRemoteUSBBackend == NULL)
         {
@@ -1943,38 +1880,27 @@ void ConsoleVRDPServer::usbBackendRemoveFromList (RemoteUSBBackend *pRemoteUSBBa
 
 void ConsoleVRDPServer::SendUpdate (unsigned uScreenId, void *pvUpdate, uint32_t cbUpdate) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPUpdate (mhServer, uScreenId, pvUpdate, cbUpdate);
     }
-#else
-    if (mpfnVRDPSendUpdate)
-        mpfnVRDPSendUpdate (mhServer, uScreenId, pvUpdate, cbUpdate);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
 void ConsoleVRDPServer::SendResize (void) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPResize (mhServer);
     }
-#else
-    if (mpfnVRDPSendResize)
-        mpfnVRDPSendResize (mhServer);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
 void ConsoleVRDPServer::SendUpdateBitmap (unsigned uScreenId, uint32_t x, uint32_t y, uint32_t w, uint32_t h) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     VRDPORDERHDR update;
     update.x = x;
     update.y = y;
@@ -1984,85 +1910,50 @@ void ConsoleVRDPServer::SendUpdateBitmap (unsigned uScreenId, uint32_t x, uint32
     {
         mpEntryPoints->VRDPUpdate (mhServer, uScreenId, &update, sizeof (update));
     }
-#else
-    if (mpfnVRDPSendUpdateBitmap)
-        mpfnVRDPSendUpdateBitmap (mhServer, uScreenId, x, y, w, h);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
-#ifdef VRDP_NO_COM
-#else
-void ConsoleVRDPServer::SetFramebuffer (IFramebuffer *framebuffer, uint32_t fFlags) const
-{
-#ifdef VBOX_VRDP
-    if (mpfnVRDPSetFramebuffer)
-        mpfnVRDPSetFramebuffer (mhServer, framebuffer, fFlags);
-#endif
-}
-#endif /* VRDP_NO_COM */
-
 void ConsoleVRDPServer::SendAudioSamples (void *pvSamples, uint32_t cSamples, VRDPAUDIOFORMAT format) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPAudioSamples (mhServer, pvSamples, cSamples, format);
     }
-#else
-    if (mpfnVRDPSendAudioSamples)
-        mpfnVRDPSendAudioSamples (mhServer, pvSamples, cSamples, format);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
 void ConsoleVRDPServer::SendAudioVolume (uint16_t left, uint16_t right) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPAudioVolume (mhServer, left, right);
     }
-#else
-    if (mpfnVRDPSendAudioVolume)
-        mpfnVRDPSendAudioVolume (mhServer, left, right);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
 void ConsoleVRDPServer::SendUSBRequest (uint32_t u32ClientId, void *pvParms, uint32_t cbParms) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPUSBRequest (mhServer, u32ClientId, pvParms, cbParms);
     }
-#else
-    if (mpfnVRDPSendUSBRequest)
-        mpfnVRDPSendUSBRequest (mhServer, u32ClientId, pvParms, cbParms);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
 void ConsoleVRDPServer::QueryInfo (uint32_t index, void *pvBuffer, uint32_t cbBuffer, uint32_t *pcbOut) const
 {
-#ifdef VBOX_VRDP
-#ifdef VRDP_NO_COM
+#ifdef VBOX_WITH_VRDP
     if (mpEntryPoints && mhServer)
     {
         mpEntryPoints->VRDPQueryInfo (mhServer, index, pvBuffer, cbBuffer, pcbOut);
     }
-#else
-    if (mpfnVRDPQueryInfo)
-        mpfnVRDPQueryInfo (mhServer, index, pvBuffer, cbBuffer, pcbOut);
-#endif /* VRDP_NO_COM */
 #endif
 }
 
-#ifdef VBOX_VRDP
+#ifdef VBOX_WITH_VRDP
 /* note: static function now! */
 bool ConsoleVRDPServer::loadVRDPLibrary (void)
 {
@@ -2070,9 +1961,9 @@ bool ConsoleVRDPServer::loadVRDPLibrary (void)
 
     if (!mVRDPLibrary)
     {
-        rc = RTLdrLoad("VBoxVRDP", &mVRDPLibrary);
+        rc = SUPR3HardenedLdrLoadAppPriv ("VBoxVRDP", &mVRDPLibrary);
 
-        if (VBOX_SUCCESS(rc))
+        if (RT_SUCCESS(rc))
         {
             LogFlow(("VRDPServer::loadLibrary(): successfully loaded VRDP library.\n"));
 
@@ -2084,37 +1975,20 @@ bool ConsoleVRDPServer::loadVRDPLibrary (void)
 
             #define DEFSYMENTRY(a) { #a, (void**)&mpfn##a }
 
-#ifdef VRDP_NO_COM
             static const struct SymbolEntry symbols[] =
             {
                 DEFSYMENTRY(VRDPCreateServer)
             };
-#else
-            static const struct SymbolEntry symbols[] =
-            {
-                DEFSYMENTRY(VRDPStartServer),
-                DEFSYMENTRY(VRDPSetCallback),
-                DEFSYMENTRY(VRDPShutdownServer),
-                DEFSYMENTRY(VRDPSendUpdate),
-                DEFSYMENTRY(VRDPSendUpdateBitmap),
-                DEFSYMENTRY(VRDPSendResize),
-                DEFSYMENTRY(VRDPSendAudioSamples),
-                DEFSYMENTRY(VRDPSendAudioVolume),
-                DEFSYMENTRY(VRDPSendUSBRequest),
-                DEFSYMENTRY(VRDPQueryInfo),
-                DEFSYMENTRY(VRDPClipboard)
-            };
-#endif /* VRDP_NO_COM */
 
             #undef DEFSYMENTRY
 
-            for (unsigned i = 0; i < ELEMENTS(symbols); i++)
+            for (unsigned i = 0; i < RT_ELEMENTS(symbols); i++)
             {
                 rc = RTLdrGetSymbol(mVRDPLibrary, symbols[i].name, symbols[i].ppfn);
 
                 AssertMsgRC(rc, ("Error resolving VRDP symbol %s\n", symbols[i].name));
 
-                if (VBOX_FAILURE(rc))
+                if (RT_FAILURE(rc))
                 {
                     break;
                 }
@@ -2122,13 +1996,13 @@ bool ConsoleVRDPServer::loadVRDPLibrary (void)
         }
         else
         {
-            LogFlow(("VRDPServer::loadLibrary(): failed to load VRDP library! VRDP not available.\n"));
+            LogRel(("VRDPServer::loadLibrary(): failed to load VRDP library! VRDP not available: rc = %Rrc\n", rc));
             mVRDPLibrary = NULL;
         }
     }
 
     // just to be safe
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
     {
         if (mVRDPLibrary)
         {
@@ -2139,7 +2013,7 @@ bool ConsoleVRDPServer::loadVRDPLibrary (void)
 
     return (mVRDPLibrary != NULL);
 }
-#endif /* VBOX_VRDP */
+#endif /* VBOX_WITH_VRDP */
 
 /*
  * IRemoteDisplayInfo implementation.
@@ -2147,6 +2021,8 @@ bool ConsoleVRDPServer::loadVRDPLibrary (void)
 // constructor / destructor
 /////////////////////////////////////////////////////////////////////////////
 
+DEFINE_EMPTY_CTOR_DTOR (RemoteDisplayInfo)
+
 HRESULT RemoteDisplayInfo::FinalConstruct()
 {
     return S_OK;
@@ -2154,8 +2030,7 @@ HRESULT RemoteDisplayInfo::FinalConstruct()
 
 void RemoteDisplayInfo::FinalRelease()
 {
-    if (isReady())
-        uninit ();
+    uninit ();
 }
 
 // public methods only for internal purposes
@@ -2166,16 +2041,19 @@ void RemoteDisplayInfo::FinalRelease()
  */
 HRESULT RemoteDisplayInfo::init (Console *aParent)
 {
-    LogFlowMember (("RemoteDisplayInfo::init (%p)\n", aParent));
+    LogFlowThisFunc (("aParent=%p\n", aParent));
 
     ComAssertRet (aParent, E_INVALIDARG);
 
-    AutoLock alock (this);
-    ComAssertRet (!isReady(), E_UNEXPECTED);
+    /* Enclose the state transition NotReady->InInit->Ready */
+    AutoInitSpan autoInitSpan (this);
+    AssertReturn (autoInitSpan.isOk(), E_FAIL);
 
-    mParent = aParent;
+    unconst (mParent) = aParent;
+
+    /* Confirm a successful initialization */
+    autoInitSpan.setSucceeded();
 
-    setReady (true);
     return S_OK;
 }
 
@@ -2185,14 +2063,14 @@ HRESULT RemoteDisplayInfo::init (Console *aParent)
  */
 void RemoteDisplayInfo::uninit()
 {
-    LogFlowMember (("RemoteDisplayInfo::uninit()\n"));
+    LogFlowThisFunc (("\n"));
 
-    AutoLock alock (this);
-    AssertReturn (isReady(), (void) 0);
+    /* Enclose the state transition Ready->InUninit->NotReady */
+    AutoUninitSpan autoUninitSpan (this);
+    if (autoUninitSpan.uninitDone())
+        return;
 
-    mParent.setNull();
-
-    setReady (false);
+    unconst (mParent).setNull();
 }
 
 // IRemoteDisplayInfo properties
@@ -2204,8 +2082,11 @@ void RemoteDisplayInfo::uninit()
         if (!a##_aName)                                                   \
             return E_POINTER;                                             \
                                                                           \
-        AutoLock alock (this);                                            \
-        CHECK_READY();                                                    \
+        AutoCaller autoCaller (this);                                     \
+        CheckComRCReturnRC (autoCaller.rc());                             \
+                                                                          \
+        /* todo: Not sure if a AutoReadLock would be sufficient. */       \
+        AutoWriteLock alock (this);                                       \
                                                                           \
         uint32_t value;                                                   \
         uint32_t cbOut = 0;                                               \
@@ -2224,8 +2105,11 @@ void RemoteDisplayInfo::uninit()
         if (!a##_aName)                                                   \
             return E_POINTER;                                             \
                                                                           \
-        AutoLock alock (this);                                            \
-        CHECK_READY();                                                    \
+        AutoCaller autoCaller (this);                                     \
+        CheckComRCReturnRC (autoCaller.rc());                             \
+                                                                          \
+        /* todo: Not sure if a AutoReadLock would be sufficient. */       \
+        AutoWriteLock alock (this);                                       \
                                                                           \
         _aType value;                                                     \
         uint32_t cbOut = 0;                                               \
@@ -2244,8 +2128,11 @@ void RemoteDisplayInfo::uninit()
         if (!a##_aName)                                                   \
             return E_POINTER;                                             \
                                                                           \
-        AutoLock alock (this);                                            \
-        CHECK_READY();                                                    \
+        AutoCaller autoCaller (this);                                     \
+        CheckComRCReturnRC (autoCaller.rc());                             \
+                                                                          \
+        /* todo: Not sure if a AutoReadLock would be sufficient. */       \
+        AutoWriteLock alock (this);                                       \
                                                                           \
         uint32_t cbOut = 0;                                               \
                                                                           \
@@ -2298,3 +2185,4 @@ IMPL_GETTER_SCALAR (ULONG,   EncryptionStyle,    VRDP_QI_ENCRYPTION_STYLE);
 
 #undef IMPL_GETTER_BSTR
 #undef IMPL_GETTER_SCALAR
+/* vi: set tabstop=4 shiftwidth=4 expandtab: */
