@@ -1,18 +1,24 @@
+/* $Id: ConsoleImpl.h 18645 2009-04-02 15:38:31Z vboxsync $ */
+
 /** @file
  *
  * VBox Console COM Class definition
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2008 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 #ifndef ____H_CONSOLEIMPL
@@ -36,6 +42,9 @@ class VMMDev;
 
 #include <VBox/vrdpapi.h>
 #include <VBox/pdmdrv.h>
+#ifdef VBOX_WITH_GUEST_PROPS
+# include <VBox/HostServices/GuestPropertySvc.h>  /* For the property notification callback */
+#endif
 
 struct VUSBIRHCONFIG;
 typedef struct VUSBIRHCONFIG *PVUSBIRHCONFIG;
@@ -51,7 +60,7 @@ typedef struct VUSBIRHCONFIG *PVUSBIRHCONFIG;
  *  not available (NULL), sets error info and returns returns E_ACCESSDENIED.
  *  The translatable error message is defined in null context.
  *
- *  Intended to used only within Console children (i,e. Keyboard, Mouse,
+ *  Intended to used only within Console children (i.e. Keyboard, Mouse,
  *  Display, etc.).
  *
  *  @param drv  driver pointer to check (compare it with NULL)
@@ -62,14 +71,6 @@ typedef struct VUSBIRHCONFIG *PVUSBIRHCONFIG;
             return setError (E_ACCESSDENIED, tr ("The console is not powered up")); \
     } while (0)
 
-/** @def VBOX_WITH_UNIXY_TAP_NETWORKING
- *  Unixy style TAP networking. This is defined in the Makefile since it's also
- *  used by NetworkAdapterImpl.h/cpp.
- */
-#ifdef __DOXYGEN__
-# define VBOX_WITH_UNIXY_TAP_NETWORKING
-#endif
-
 // Console
 ///////////////////////////////////////////////////////////////////////////////
 
@@ -113,30 +114,37 @@ public:
     STDMETHOD(COMGETTER(Mouse)) (IMouse **aMouse);
     STDMETHOD(COMGETTER(Display)) (IDisplay **aDisplay);
     STDMETHOD(COMGETTER(Debugger)) (IMachineDebugger **aDebugger);
-    STDMETHOD(COMGETTER(USBDevices)) (IUSBDeviceCollection **aUSBDevices);
-    STDMETHOD(COMGETTER(RemoteUSBDevices)) (IHostUSBDeviceCollection **aRemoteUSBDevices);
+    STDMETHOD(COMGETTER(USBDevices)) (ComSafeArrayOut (IUSBDevice *, aUSBDevices));
+    STDMETHOD(COMGETTER(RemoteUSBDevices)) (ComSafeArrayOut (IHostUSBDevice *, aRemoteUSBDevices));
     STDMETHOD(COMGETTER(RemoteDisplayInfo)) (IRemoteDisplayInfo **aRemoteDisplayInfo);
-    STDMETHOD(COMGETTER(SharedFolders)) (ISharedFolderCollection **aSharedFolders);
+    STDMETHOD(COMGETTER(SharedFolders)) (ComSafeArrayOut (ISharedFolder *, aSharedFolders));
 
     // IConsole methods
     STDMETHOD(PowerUp) (IProgress **aProgress);
+    STDMETHOD(PowerUpPaused) (IProgress **aProgress);
     STDMETHOD(PowerDown)();
+    STDMETHOD(PowerDownAsync) (IProgress **aProgress);
     STDMETHOD(Reset)();
     STDMETHOD(Pause)();
     STDMETHOD(Resume)();
     STDMETHOD(PowerButton)();
     STDMETHOD(SleepButton)();
+    STDMETHOD(GetPowerButtonHandled)(BOOL *aHandled);
+    STDMETHOD(GetGuestEnteredACPIMode)(BOOL *aEntered);
     STDMETHOD(SaveState) (IProgress **aProgress);
+    STDMETHOD(AdoptSavedState) (IN_BSTR aSavedStateFile);
     STDMETHOD(DiscardSavedState)();
     STDMETHOD(GetDeviceActivity) (DeviceType_T aDeviceType,
                                  DeviceActivity_T *aDeviceActivity);
-    STDMETHOD(AttachUSBDevice) (INPTR GUIDPARAM aId);
-    STDMETHOD(DetachUSBDevice) (INPTR GUIDPARAM aId, IUSBDevice **aDevice);
-    STDMETHOD(CreateSharedFolder) (INPTR BSTR aName, INPTR BSTR aHostPath);
-    STDMETHOD(RemoveSharedFolder) (INPTR BSTR aName);
-    STDMETHOD(TakeSnapshot) (INPTR BSTR aName, INPTR BSTR aDescription,
+    STDMETHOD(AttachUSBDevice) (IN_GUID aId);
+    STDMETHOD(DetachUSBDevice) (IN_GUID aId, IUSBDevice **aDevice);
+    STDMETHOD(FindUSBDeviceByAddress) (IN_BSTR aAddress, IUSBDevice **aDevice);
+    STDMETHOD(FindUSBDeviceById) (IN_GUID aId, IUSBDevice **aDevice);
+    STDMETHOD(CreateSharedFolder) (IN_BSTR aName, IN_BSTR aHostPath, BOOL aWritable);
+    STDMETHOD(RemoveSharedFolder) (IN_BSTR aName);
+    STDMETHOD(TakeSnapshot) (IN_BSTR aName, IN_BSTR aDescription,
                              IProgress **aProgress);
-    STDMETHOD(DiscardSnapshot) (INPTR GUIDPARAM aId, IProgress **aProgress);
+    STDMETHOD(DiscardSnapshot) (IN_GUID aId, IProgress **aProgress);
     STDMETHOD(DiscardCurrentState) (IProgress **aProgress);
     STDMETHOD(DiscardCurrentSnapshotAndState) (IProgress **aProgress);
     STDMETHOD(RegisterCallback) (IConsoleCallback *aCallback);
@@ -149,18 +157,18 @@ public:
      *  called only by the VM execution thread.
      */
 
-    Guest *getGuest() { return mGuest; }
-    Keyboard *getKeyboard() { return mKeyboard; }
-    Mouse *getMouse() { return mMouse; }
-    Display *getDisplay() { return mDisplay; }
-    MachineDebugger *getMachineDebugger() { return mDebugger; }
+    Guest *getGuest() const { return mGuest; }
+    Keyboard *getKeyboard() const { return mKeyboard; }
+    Mouse *getMouse() const { return mMouse; }
+    Display *getDisplay() const { return mDisplay; }
+    MachineDebugger *getMachineDebugger() const { return mDebugger; }
 
-    const ComPtr <IMachine> &machine() { return mMachine; }
+    const ComPtr <IMachine> &machine() const { return mMachine; }
 
     /** Method is called only from ConsoleVRDPServer */
-    IVRDPServer *getVRDPServer() { return mVRDPServer; }
+    IVRDPServer *getVRDPServer() const { return mVRDPServer; }
 
-    ConsoleVRDPServer *consoleVRDPServer() { return mConsoleVRDPServer; }
+    ConsoleVRDPServer *consoleVRDPServer() const { return mConsoleVRDPServer; }
 
     HRESULT updateMachineState (MachineState_T aMachineState);
 
@@ -170,25 +178,24 @@ public:
     HRESULT onNetworkAdapterChange (INetworkAdapter *aNetworkAdapter);
     HRESULT onSerialPortChange (ISerialPort *aSerialPort);
     HRESULT onParallelPortChange (IParallelPort *aParallelPort);
+    HRESULT onStorageControllerChange ();
     HRESULT onVRDPServerChange();
     HRESULT onUSBControllerChange();
     HRESULT onSharedFolderChange (BOOL aGlobal);
     HRESULT onUSBDeviceAttach (IUSBDevice *aDevice, IVirtualBoxErrorInfo *aError, ULONG aMaskedIfs);
-    HRESULT onUSBDeviceDetach (INPTR GUIDPARAM aId, IVirtualBoxErrorInfo *aError);
-
+    HRESULT onUSBDeviceDetach (IN_GUID aId, IVirtualBoxErrorInfo *aError);
+    HRESULT getGuestProperty (IN_BSTR aKey, BSTR *aValue, ULONG64 *aTimestamp, BSTR *aFlags);
+    HRESULT setGuestProperty (IN_BSTR aKey, IN_BSTR aValue, IN_BSTR aFlags);
+    HRESULT enumerateGuestProperties (IN_BSTR aPatterns, ComSafeArrayOut(BSTR, aNames), ComSafeArrayOut(BSTR, aValues), ComSafeArrayOut(ULONG64, aTimestamps), ComSafeArrayOut(BSTR, aFlags));
     VMMDev *getVMMDev() { return mVMMDev; }
     AudioSniffer *getAudioSniffer () { return mAudioSniffer; }
 
-#ifdef VRDP_NO_COM
     int VRDPClientLogon (uint32_t u32ClientId, const char *pszUser, const char *pszPassword, const char *pszDomain);
     void VRDPClientConnect (uint32_t u32ClientId);
     void VRDPClientDisconnect (uint32_t u32ClientId, uint32_t fu32Intercepted);
     void VRDPInterceptAudio (uint32_t u32ClientId);
     void VRDPInterceptUSB (uint32_t u32ClientId, void **ppvIntercept);
     void VRDPInterceptClipboard (uint32_t u32ClientId);
-#else
-    static VRDPSERVERCALLBACK *getVrdpServerCallback () { return &sVrdpServerCallback; };
-#endif /* VRDP_NO_COM */
 
     void processRemoteUSBDevices (uint32_t u32ClientId, VRDPUSBDEVICEDESC *pDevList, uint32_t cbDevList);
 
@@ -205,14 +212,14 @@ public:
     void onKeyboardLedsChange (bool fNumLock, bool fCapsLock, bool fScrollLock);
     void onUSBDeviceStateChange (IUSBDevice *aDevice, bool aAttached,
                                  IVirtualBoxErrorInfo *aError);
-    void onRuntimeError (BOOL aFatal, INPTR BSTR aErrorID, INPTR BSTR aMessage);
+    void onRuntimeError (BOOL aFatal, IN_BSTR aErrorID, IN_BSTR aMessage);
     HRESULT onShowWindow (BOOL aCheck, BOOL *aCanShow, ULONG64 *aWinId);
 
     static const PDMDRVREG DrvStatusReg;
 
     void reportAuthLibraryError (const char *filename, int rc)
     {
-        setError (E_FAIL, tr("Could not load the external authentication library '%s' (%Vrc)"), filename, rc);
+        setError (E_FAIL, tr("Could not load the external authentication library '%s' (%Rrc)"), filename, rc);
     }
 
     // for VirtualBoxSupportErrorInfoImpl
@@ -276,18 +283,24 @@ private:
      *      VMR3ReqCall (mpVM, ...
      *  </code>
      *
+     *  @note Temporarily locks the argument for writing.
+     *
      *  @sa SafeVMPtr, SafeVMPtrQuiet
      */
     typedef AutoVMCallerBase <false, false> AutoVMCaller;
 
     /**
      *  Same as AutoVMCaller but doesn't set extended error info on failure.
+     *
+     *  @note Temporarily locks the argument for writing.
      */
     typedef AutoVMCallerBase <true, false> AutoVMCallerQuiet;
 
     /**
      *  Same as AutoVMCaller but allows a null VM pointer (to trigger an error
      *  instead of assertion).
+     *
+     *  @note Temporarily locks the argument for writing.
      */
     typedef AutoVMCallerBase <false, true> AutoVMCallerWeak;
 
@@ -295,6 +308,8 @@ private:
      *  Same as AutoVMCaller but doesn't set extended error info on failure
      *  and allows a null VM pointer (to trigger an error instead of
      *  assertion).
+     *
+     *  @note Temporarily locks the argument for writing.
      */
     typedef AutoVMCallerBase <true, true> AutoVMCallerQuietWeak;
 
@@ -336,13 +351,15 @@ public:
      *      printf ("%p\n", pVM.raw());
      *  </code>
      *
+     *  @note Temporarily locks the argument for writing.
+     *
      *  @sa SafeVMPtrQuiet, AutoVMCaller
      */
     typedef SafeVMPtrBase <false> SafeVMPtr;
 
     /**
      *  A deviation of SaveVMPtr that doesn't set the error info on failure.
-     *  Intenede for pieces of code that don't need to return the VM access
+     *  Intended for pieces of code that don't need to return the VM access
      *  failure to the caller. The usage pattern is:
      *  <code>
      *      Console::SaveVMPtrQuiet pVM (mParent);
@@ -351,12 +368,27 @@ public:
      *      return S_OK;
      *  </code>
      *
+     *  @note Temporarily locks the argument for writing.
+     *
      *  @sa SafeVMPtr, AutoVMCaller
      */
     typedef SafeVMPtrBase <true> SafeVMPtrQuiet;
 
+    class SharedFolderData
+    {
+    public:
+        SharedFolderData() {}
+        SharedFolderData(Bstr aHostPath, BOOL aWritable)
+           : mHostPath (aHostPath)
+           , mWritable (aWritable) {}
+        SharedFolderData(const SharedFolderData& aThat)
+           : mHostPath (aThat.mHostPath)
+           , mWritable (aThat.mWritable) {}
+        Bstr mHostPath;
+        BOOL mWritable;
+    };
     typedef std::map <Bstr, ComObjPtr <SharedFolder> > SharedFolderMap;
-    typedef std::map <Bstr, Bstr> SharedFolderDataMap;
+    typedef std::map <Bstr, SharedFolderData> SharedFolderDataMap;
 
 private:
 
@@ -366,12 +398,15 @@ private:
     HRESULT addVMCaller (bool aQuiet = false, bool aAllowNullVM = false);
     void releaseVMCaller();
 
-    HRESULT powerDown();
+    HRESULT consoleInitReleaseLog (const ComPtr <IMachine> aMachine);
+
+    HRESULT powerUp (IProgress **aProgress, bool aPaused);
+    HRESULT powerDown (Progress *aProgress = NULL);
 
     HRESULT callTapSetupApplication(bool isStatic, RTFILE tapFD, Bstr &tapDevice,
                                     Bstr &tapSetupApplication);
-    HRESULT attachToHostInterface(INetworkAdapter *networkAdapter);
-    HRESULT detachFromHostInterface(INetworkAdapter *networkAdapter);
+    HRESULT attachToBridgedInterface(INetworkAdapter *networkAdapter);
+    HRESULT detachFromBridgedInterface(INetworkAdapter *networkAdapter);
     HRESULT powerDownHostInterfaces();
 
     HRESULT setMachineState (MachineState_T aMachineState, bool aUpdateServer = true);
@@ -380,16 +415,16 @@ private:
         return setMachineState (aMachineState, false /* aUpdateServer */);
     }
 
-    HRESULT findSharedFolder (const BSTR aName,
+    HRESULT findSharedFolder (CBSTR aName,
                               ComObjPtr <SharedFolder> &aSharedFolder,
                               bool aSetError = false);
 
     HRESULT fetchSharedFolders (BOOL aGlobal);
-    bool findOtherSharedFolder (INPTR BSTR aName,
+    bool findOtherSharedFolder (IN_BSTR aName,
                                 SharedFolderDataMap::const_iterator &aIt);
 
-    HRESULT createSharedFolder (INPTR BSTR aName, INPTR BSTR aHostPath);
-    HRESULT removeSharedFolder (INPTR BSTR aName);
+    HRESULT createSharedFolder (CBSTR aName, SharedFolderData aData);
+    HRESULT removeSharedFolder (CBSTR aName);
 
     static DECLCALLBACK(int) configConstructor(PVM pVM, void *pvConsole);
     static DECLCALLBACK(void) vmstateChangeCallback(PVM aVM, VMSTATE aState,
@@ -422,9 +457,9 @@ private:
                         const char *pszFormat, va_list args);
 
     static DECLCALLBACK(void)
-    setVMRuntimeErrorCallback (PVM pVM, void *pvUser, bool fFatal,
-                               const char *pszErrorID,
-                               const char *pszFormat, va_list args);
+    setVMRuntimeErrorCallback (PVM pVM, void *pvUser, uint32_t fFatal,
+                               const char *pszErrorId,
+                               const char *pszFormat, va_list va);
 
     HRESULT                     captureUSBDevices (PVM pVM);
     void                        detachAllUSBDevices (bool aDone);
@@ -440,18 +475,7 @@ private:
 
     int mcAudioRefs;
     volatile uint32_t mcVRDPClients;
-
-#ifdef VRDP_NO_COM
-#else
-    static DECLCALLBACK(int)    vrdp_ClientLogon (void *pvUser, uint32_t u32ClientId, const char *pszUser, const char *pszPassword, const char *pszDomain);
-    static DECLCALLBACK(void)   vrdp_ClientConnect (void *pvUser, uint32_t u32ClientId);
-    static DECLCALLBACK(void)   vrdp_ClientDisconnect (void *pvUser, uint32_t u32ClientId, uint32_t fu32Intercepted);
-    static DECLCALLBACK(void)   vrdp_InterceptAudio (void *pvUser, uint32_t u32ClientId);
-    static DECLCALLBACK(void)   vrdp_InterceptUSB (void *pvUser, uint32_t u32ClientId, PFNVRDPUSBCALLBACK *ppfn, void **ppv);
-    static DECLCALLBACK(void)   vrdp_InterceptClipboard (void *pvUser, uint32_t u32ClientId, PFNVRDPCLIPBOARDCALLBACK *ppfn, void **ppv);
-
-    static VRDPSERVERCALLBACK   sVrdpServerCallback;
-#endif /* VRDP_NO_COM */
+    uint32_t mu32SingleRDPClientId; /* The id of a connected client in the single connection mode. */
 
     static const char *sSSMConsoleUnit;
     static uint32_t sSSMConsoleVer;
@@ -461,6 +485,16 @@ private:
     static DECLCALLBACK(void)   saveStateFileExec (PSSMHANDLE pSSM, void *pvUser);
     static DECLCALLBACK(int)    loadStateFileExec (PSSMHANDLE pSSM, void *pvUser, uint32_t u32Version);
 
+#ifdef VBOX_WITH_GUEST_PROPS
+    static DECLCALLBACK(int)    doGuestPropNotification (void *pvExtension, uint32_t,
+                                                         void *pvParms, uint32_t cbParms);
+    HRESULT doEnumerateGuestProperties (CBSTR aPatterns,
+                                        ComSafeArrayOut(BSTR, aNames),
+                                        ComSafeArrayOut(BSTR, aValues),
+                                        ComSafeArrayOut(ULONG64, aTimestamps),
+                                        ComSafeArrayOut(BSTR, aFlags));
+#endif
+
     bool mSavedStateDataLoaded : 1;
 
     const ComPtr <IMachine> mMachine;
@@ -494,6 +528,8 @@ private:
     RTSEMEVENT mVMZeroCallersSem;
     /** true when Console has entered the mpVM destruction phase */
     bool mVMDestroying : 1;
+    /** true when power down is initiated by vmstateChangeCallback (EMT) */
+    bool mVMPoweredOff : 1;
 
     /** The current DVD drive state in the VM.
      * This does not have to match the state maintained in the DVD. */
@@ -507,10 +543,12 @@ private:
 
     PPDMLED     mapFDLeds[2];
     PPDMLED     mapIDELeds[4];
+    PPDMLED     mapSATALeds[30];
+    PPDMLED     mapSCSILeds[16];
     PPDMLED     mapNetworkLeds[8];
     PPDMLED     mapSharedFolderLed;
-    PPDMLED     mapUSBLed;
-#ifdef VBOX_WITH_UNIXY_TAP_NETWORKING
+    PPDMLED     mapUSBLed[2];
+#if !defined(VBOX_WITH_NETFLT) && defined(RT_OS_LINUX)
     Utf8Str     maTAPDeviceName[8];
     RTFILE      maTapFD[8];
 #endif
@@ -565,3 +603,4 @@ private:
 };
 
 #endif // ____H_CONSOLEIMPL
+/* vi: set tabstop=4 shiftwidth=4 expandtab: */
