@@ -1,45 +1,54 @@
+/* $Id: MachineImpl.h 18520 2009-03-30 08:45:50Z vboxsync $ */
+
 /** @file
  *
  * VirtualBox COM class declaration
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2009 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 #ifndef ____H_MACHINEIMPL
 #define ____H_MACHINEIMPL
 
 #include "VirtualBoxBase.h"
-#include "VirtualBoxXMLUtil.h"
 #include "ProgressImpl.h"
 #include "SnapshotImpl.h"
 #include "VRDPServerImpl.h"
 #include "DVDDriveImpl.h"
 #include "FloppyDriveImpl.h"
 #include "HardDiskAttachmentImpl.h"
-#include "Collection.h"
 #include "NetworkAdapterImpl.h"
 #include "AudioAdapterImpl.h"
 #include "SerialPortImpl.h"
 #include "ParallelPortImpl.h"
 #include "BIOSSettingsImpl.h"
+#include "StorageControllerImpl.h"
+#ifdef VBOX_WITH_RESOURCE_USAGE_API
+#include "PerformanceImpl.h"
+#endif /* VBOX_WITH_RESOURCE_USAGE_API */
 
 // generated header
 #include "SchemaDefs.h"
 
 #include <VBox/types.h>
-#include <VBox/cfgldr.h>
+
 #include <iprt/file.h>
 #include <iprt/thread.h>
+#include <iprt/time.h>
 
 #include <list>
 
@@ -68,7 +77,6 @@ class SessionMachine;
 
 class ATL_NO_VTABLE Machine :
     public VirtualBoxBaseWithChildrenNEXT,
-    public VirtualBoxXMLUtil,
     public VirtualBoxSupportErrorInfoImpl <Machine, IMachine>,
     public VirtualBoxSupportTranslation <Machine>,
     public IMachine
@@ -77,25 +85,27 @@ class ATL_NO_VTABLE Machine :
 
 public:
 
+    enum InstanceType { IsMachine, IsSessionMachine, IsSnapshotMachine };
+
     /**
-     *  Internal machine data.
+     * Internal machine data.
      *
-     *  Only one instance of this data exists per every machine --
-     *  it is shared by the Machine, SessionMachine and all SnapshotMachine
-     *  instances associated with the given machine using the util::Shareable
-     *  template through the mData variable.
+     * Only one instance of this data exists per every machine -- it is shared
+     * by the Machine, SessionMachine and all SnapshotMachine instances
+     * associated with the given machine using the util::Shareable template
+     * through the mData variable.
      *
-     *  @note |const| members are persistent during lifetime so can be
-     *  accessed without locking.
+     * @note |const| members are persistent during lifetime so can be
+     * accessed without locking.
      *
-     *  @note There is no need to lock anything inside init() or uninit()
-     *  methods, because they are always serialized (see AutoCaller).
+     * @note There is no need to lock anything inside init() or uninit()
+     * methods, because they are always serialized (see AutoCaller).
      */
     struct Data
     {
         /**
-         *  Data structure to hold information about sessions opened for the
-         *  given machine.
+         * Data structure to hold information about sessions opened for the
+         * given machine.
          */
         struct Session
         {
@@ -111,9 +121,9 @@ public:
             ComObjPtr <Progress> mProgress;
 
             /**
-             *  PID of the session object that must be passed to openSession()
-             *  to finalize the openRemoteSession() request
-             *  (i.e., PID of the process created by openRemoteSession())
+             * PID of the session object that must be passed to openSession() to
+             * finalize the openRemoteSession() request (i.e., PID of the
+             * process created by openRemoteSession())
              */
             RTPROCESS mPid;
 
@@ -123,8 +133,16 @@ public:
             /** Session type string (for indirect sessions) */
             Bstr mType;
 
-            /** Sesison machine object */
+            /** Session machine object */
             ComObjPtr <SessionMachine> mMachine;
+
+            /**
+             * Successfully locked media list. The 2nd value in the pair is true
+             * if the medium is locked for writing and false if locked for
+             * reading.
+             */
+            typedef std::list <std::pair <ComPtr <IMedium>, bool > > LockedMedia;
+            LockedMedia mLockedMedia;
         };
 
         Data();
@@ -136,15 +154,18 @@ public:
         Bstr mConfigFile;
         Bstr mConfigFileFull;
 
+        Utf8Str mSettingsFileVersion;
+
         BOOL mAccessible;
         com::ErrorInfo mAccessError;
 
         MachineState_T mMachineState;
-        LONG64 mLastStateChange;
+        RTTIMESPEC mLastStateChange;
 
+        /* Note: These are guarded by VirtualBoxBase::stateLockHandle() */
         uint32_t mMachineStateDeps;
-        RTSEMEVENT mZeroMachineStateDepsSem;
-        BOOL mWaitingStateDeps;
+        RTSEMEVENTMULTI mMachineStateDepsSem;
+        uint32_t mMachineStateChangePending;
 
         BOOL mCurrentStateModified;
 
@@ -154,6 +175,7 @@ public:
 
         ComObjPtr <Snapshot> mFirstSnapshot;
         ComObjPtr <Snapshot> mCurrentSnapshot;
+
     };
 
     /**
@@ -215,31 +237,57 @@ public:
      */
     struct HWData
     {
+        /**
+         * Data structure to hold information about a guest property.
+         */
+        struct GuestProperty {
+            /** Property name */
+            Bstr mName;
+            /** Property value */
+            Bstr mValue;
+            /** Property timestamp */
+            ULONG64 mTimestamp;
+            /** Property flags */
+            ULONG mFlags;
+        };
+
         HWData();
         ~HWData();
 
         bool operator== (const HWData &that) const;
 
+        Bstr           mHWVersion;
         ULONG          mMemorySize;
+        ULONG          mMemoryBalloonSize;
+        ULONG          mStatisticsUpdateInterval;
         ULONG          mVRAMSize;
         ULONG          mMonitorCount;
-        TriStateBool_T mHWVirtExEnabled;
+        TSBool_T       mHWVirtExEnabled;
+        BOOL           mHWVirtExNestedPagingEnabled;
+        BOOL           mHWVirtExVPIDEnabled;
+        BOOL           mAccelerate3DEnabled;
+        BOOL           mPAEEnabled;
+        ULONG          mCPUCount;
 
         DeviceType_T   mBootOrder [SchemaDefs::MaxBootPosition];
 
         typedef std::list <ComObjPtr <SharedFolder> > SharedFolderList;
         SharedFolderList mSharedFolders;
         ClipboardMode_T mClipboardMode;
+        typedef std::list <GuestProperty> GuestPropertyList;
+        GuestPropertyList mGuestProperties;
+        BOOL           mPropertyServiceActive;
+        Bstr           mGuestPropertyNotificationPatterns;
     };
 
     /**
      *  Hard disk data.
      *
      *  The usage policy is the same as for HWData, but a separate structure
-     *  is necessarym because hard disk data requires different procedures when
+     *  is necessary because hard disk data requires different procedures when
      *  taking or discarding snapshots, etc.
      *
-     *  The data variable is |mHWData|.
+     *  The data variable is |mHDData|.
      */
     struct HDData
     {
@@ -248,14 +296,8 @@ public:
 
         bool operator== (const HDData &that) const;
 
-        typedef std::list <ComObjPtr <HardDiskAttachment> > HDAttachmentList;
-        HDAttachmentList mHDAttachments;
-
-        /**
-         *  Right after Machine::fixupHardDisks(true): |true| if hard disks
-         *  were actually changed, |false| otherwise
-         */
-        bool mHDAttachmentsChanged;
+        typedef std::list< ComObjPtr<HardDiskAttachment> > AttachmentList;
+        AttachmentList mAttachments;
     };
 
     enum StateDependency
@@ -296,7 +338,7 @@ public:
 
         AutoStateDependency (Machine *aThat)
             : mThat (aThat), mRC (S_OK)
-            , mMachineState (MachineState_InvalidMachineState)
+            , mMachineState (MachineState_Null)
             , mRegistered (FALSE)
         {
             Assert (aThat);
@@ -423,8 +465,9 @@ public:
     enum InitMode { Init_New, Init_Existing, Init_Registered };
 
     // public initializer/uninitializer for internal purposes only
-    HRESULT init (VirtualBox *aParent, const BSTR aConfigFile,
-                  InitMode aMode, const BSTR aName = NULL,
+    HRESULT init (VirtualBox *aParent, CBSTR aConfigFile,
+                  InitMode aMode, CBSTR aName = NULL,
+                  GuestOSType *aOsType = NULL,
                   BOOL aNameSync = TRUE, const Guid *aId = NULL);
     void uninit();
 
@@ -433,137 +476,231 @@ public:
     STDMETHOD(COMGETTER(Accessible)) (BOOL *aAccessible);
     STDMETHOD(COMGETTER(AccessError)) (IVirtualBoxErrorInfo **aAccessError);
     STDMETHOD(COMGETTER(Name))(BSTR *aName);
-    STDMETHOD(COMSETTER(Name))(INPTR BSTR aName);
+    STDMETHOD(COMSETTER(Name))(IN_BSTR aName);
     STDMETHOD(COMGETTER(Description))(BSTR *aDescription);
-    STDMETHOD(COMSETTER(Description))(INPTR BSTR aDescription);
-    STDMETHOD(COMGETTER(Id))(GUIDPARAMOUT aId);
+    STDMETHOD(COMSETTER(Description))(IN_BSTR aDescription);
+    STDMETHOD(COMGETTER(Id))(OUT_GUID aId);
     STDMETHOD(COMGETTER(OSTypeId)) (BSTR *aOSTypeId);
-    STDMETHOD(COMSETTER(OSTypeId)) (INPTR BSTR aOSTypeId);
+    STDMETHOD(COMSETTER(OSTypeId)) (IN_BSTR aOSTypeId);
+    STDMETHOD(COMGETTER(HardwareVersion))(BSTR *aVersion);
+    STDMETHOD(COMSETTER(HardwareVersion))(IN_BSTR aVersion);
     STDMETHOD(COMGETTER(MemorySize))(ULONG *memorySize);
     STDMETHOD(COMSETTER(MemorySize))(ULONG memorySize);
+    STDMETHOD(COMGETTER(CPUCount))(ULONG *cpuCount);
+    STDMETHOD(COMSETTER(CPUCount))(ULONG cpuCount);
+    STDMETHOD(COMGETTER(MemoryBalloonSize))(ULONG *memoryBalloonSize);
+    STDMETHOD(COMSETTER(MemoryBalloonSize))(ULONG memoryBalloonSize);
+    STDMETHOD(COMGETTER(StatisticsUpdateInterval))(ULONG *statisticsUpdateInterval);
+    STDMETHOD(COMSETTER(StatisticsUpdateInterval))(ULONG statisticsUpdateInterval);
     STDMETHOD(COMGETTER(VRAMSize))(ULONG *memorySize);
     STDMETHOD(COMSETTER(VRAMSize))(ULONG memorySize);
     STDMETHOD(COMGETTER(MonitorCount))(ULONG *monitorCount);
     STDMETHOD(COMSETTER(MonitorCount))(ULONG monitorCount);
+    STDMETHOD(COMGETTER(Accelerate3DEnabled))(BOOL *enabled);
+    STDMETHOD(COMSETTER(Accelerate3DEnabled))(BOOL enabled);
     STDMETHOD(COMGETTER(BIOSSettings))(IBIOSSettings **biosSettings);
-    STDMETHOD(COMGETTER(HWVirtExEnabled))(TriStateBool_T *enabled);
-    STDMETHOD(COMSETTER(HWVirtExEnabled))(TriStateBool_T enabled);
+    STDMETHOD(COMGETTER(HWVirtExEnabled))(TSBool_T *enabled);
+    STDMETHOD(COMSETTER(HWVirtExEnabled))(TSBool_T enabled);
+    STDMETHOD(COMGETTER(HWVirtExNestedPagingEnabled))(BOOL *enabled);
+    STDMETHOD(COMSETTER(HWVirtExNestedPagingEnabled))(BOOL enabled);
+    STDMETHOD(COMGETTER(HWVirtExVPIDEnabled))(BOOL *enabled);
+    STDMETHOD(COMSETTER(HWVirtExVPIDEnabled))(BOOL enabled);
+    STDMETHOD(COMGETTER(PAEEnabled))(BOOL *enabled);
+    STDMETHOD(COMSETTER(PAEEnabled))(BOOL enabled);
     STDMETHOD(COMGETTER(SnapshotFolder))(BSTR *aSavedStateFolder);
-    STDMETHOD(COMSETTER(SnapshotFolder))(INPTR BSTR aSavedStateFolder);
-    STDMETHOD(COMGETTER(HardDiskAttachments))(IHardDiskAttachmentCollection **attachments);
+    STDMETHOD(COMSETTER(SnapshotFolder))(IN_BSTR aSavedStateFolder);
+    STDMETHOD(COMGETTER(HardDiskAttachments))(ComSafeArrayOut (IHardDiskAttachment *, aAttachments));
     STDMETHOD(COMGETTER(VRDPServer))(IVRDPServer **vrdpServer);
     STDMETHOD(COMGETTER(DVDDrive))(IDVDDrive **dvdDrive);
     STDMETHOD(COMGETTER(FloppyDrive))(IFloppyDrive **floppyDrive);
     STDMETHOD(COMGETTER(AudioAdapter))(IAudioAdapter **audioAdapter);
-    STDMETHOD(COMGETTER(USBController))(IUSBController * *a_ppUSBController);
-    STDMETHOD(COMGETTER(SettingsFilePath))(BSTR *filePath);
-    STDMETHOD(COMGETTER(SettingsModified))(BOOL *modified);
-    STDMETHOD(COMGETTER(SessionState))(SessionState_T *aSessionState);
-    STDMETHOD(COMGETTER(SessionType))(BSTR *aSessionType);
-    STDMETHOD(COMGETTER(SessionPid))(ULONG *aSessionPid);
-    STDMETHOD(COMGETTER(State))(MachineState_T *machineState);
-    STDMETHOD(COMGETTER(LastStateChange))(LONG64 *aLastStateChange);
+    STDMETHOD(COMGETTER(USBController)) (IUSBController * *aUSBController);
+    STDMETHOD(COMGETTER(SettingsFilePath)) (BSTR *aFilePath);
+    STDMETHOD(COMGETTER(SettingsFileVersion)) (BSTR *aSettingsFileVersion);
+    STDMETHOD(COMGETTER(SettingsModified)) (BOOL *aModified);
+    STDMETHOD(COMGETTER(SessionState)) (SessionState_T *aSessionState);
+    STDMETHOD(COMGETTER(SessionType)) (BSTR *aSessionType);
+    STDMETHOD(COMGETTER(SessionPid)) (ULONG *aSessionPid);
+    STDMETHOD(COMGETTER(State)) (MachineState_T *machineState);
+    STDMETHOD(COMGETTER(LastStateChange)) (LONG64 *aLastStateChange);
     STDMETHOD(COMGETTER(StateFilePath)) (BSTR *aStateFilePath);
     STDMETHOD(COMGETTER(LogFolder)) (BSTR *aLogFolder);
     STDMETHOD(COMGETTER(CurrentSnapshot)) (ISnapshot **aCurrentSnapshot);
     STDMETHOD(COMGETTER(SnapshotCount)) (ULONG *aSnapshotCount);
     STDMETHOD(COMGETTER(CurrentStateModified))(BOOL *aCurrentStateModified);
-    STDMETHOD(COMGETTER(SharedFolders)) (ISharedFolderCollection **aSharedFolders);
+    STDMETHOD(COMGETTER(SharedFolders)) (ComSafeArrayOut (ISharedFolder *, aSharedFolders));
     STDMETHOD(COMGETTER(ClipboardMode)) (ClipboardMode_T *aClipboardMode);
     STDMETHOD(COMSETTER(ClipboardMode)) (ClipboardMode_T aClipboardMode);
+    STDMETHOD(COMGETTER(GuestPropertyNotificationPatterns)) (BSTR *aPattern);
+    STDMETHOD(COMSETTER(GuestPropertyNotificationPatterns)) (IN_BSTR aPattern);
+    STDMETHOD(COMGETTER(StorageControllers)) (ComSafeArrayOut(IStorageController *, aStorageControllers));
 
     // IMachine methods
     STDMETHOD(SetBootOrder)(ULONG aPosition, DeviceType_T aDevice);
     STDMETHOD(GetBootOrder)(ULONG aPosition, DeviceType_T *aDevice);
-    STDMETHOD(AttachHardDisk)(INPTR GUIDPARAM aId, DiskControllerType_T aCtl, LONG aDev);
-    STDMETHOD(GetHardDisk)(DiskControllerType_T aCtl, LONG aDev, IHardDisk **aHardDisk);
-    STDMETHOD(DetachHardDisk) (DiskControllerType_T aCtl, LONG aDev);
+    STDMETHOD(AttachHardDisk)(IN_GUID aId, IN_BSTR aControllerName,
+                              LONG aControllerPort, LONG aDevice);
+    STDMETHOD(GetHardDisk)(IN_BSTR aControllerName, LONG aControllerPort, LONG aDevice,
+                           IHardDisk **aHardDisk);
+    STDMETHOD(DetachHardDisk)(IN_BSTR aControllerName, LONG aControllerPort, LONG aDevice);
     STDMETHOD(GetSerialPort) (ULONG slot, ISerialPort **port);
     STDMETHOD(GetParallelPort) (ULONG slot, IParallelPort **port);
     STDMETHOD(GetNetworkAdapter) (ULONG slot, INetworkAdapter **adapter);
-    STDMETHOD(GetNextExtraDataKey)(INPTR BSTR aKey, BSTR *aNextKey, BSTR *aNextValue);
-    STDMETHOD(GetExtraData)(INPTR BSTR aKey, BSTR *aValue);
-    STDMETHOD(SetExtraData)(INPTR BSTR aKey, INPTR BSTR aValue);
+    STDMETHOD(GetNextExtraDataKey)(IN_BSTR aKey, BSTR *aNextKey, BSTR *aNextValue);
+    STDMETHOD(GetExtraData)(IN_BSTR aKey, BSTR *aValue);
+    STDMETHOD(SetExtraData)(IN_BSTR aKey, IN_BSTR aValue);
     STDMETHOD(SaveSettings)();
+    STDMETHOD(SaveSettingsWithBackup) (BSTR *aBakFileName);
     STDMETHOD(DiscardSettings)();
     STDMETHOD(DeleteSettings)();
-    STDMETHOD(GetSnapshot) (INPTR GUIDPARAM aId, ISnapshot **aSnapshot);
-    STDMETHOD(FindSnapshot) (INPTR BSTR aName, ISnapshot **aSnapshot);
-    STDMETHOD(SetCurrentSnapshot) (INPTR GUIDPARAM aId);
-    STDMETHOD(CreateSharedFolder) (INPTR BSTR aName, INPTR BSTR aHostPath);
-    STDMETHOD(RemoveSharedFolder) (INPTR BSTR aName);
+    STDMETHOD(Export)(IAppliance *aAppliance, IVirtualSystemDescription **aDescription);
+    STDMETHOD(GetSnapshot) (IN_GUID aId, ISnapshot **aSnapshot);
+    STDMETHOD(FindSnapshot) (IN_BSTR aName, ISnapshot **aSnapshot);
+    STDMETHOD(SetCurrentSnapshot) (IN_GUID aId);
+    STDMETHOD(CreateSharedFolder) (IN_BSTR aName, IN_BSTR aHostPath, BOOL aWritable);
+    STDMETHOD(RemoveSharedFolder) (IN_BSTR aName);
     STDMETHOD(CanShowConsoleWindow) (BOOL *aCanShow);
     STDMETHOD(ShowConsoleWindow) (ULONG64 *aWinId);
+    STDMETHOD(GetGuestProperty) (IN_BSTR aName, BSTR *aValue, ULONG64 *aTimestamp, BSTR *aFlags);
+    STDMETHOD(GetGuestPropertyValue) (IN_BSTR aName, BSTR *aValue);
+    STDMETHOD(GetGuestPropertyTimestamp) (IN_BSTR aName, ULONG64 *aTimestamp);
+    STDMETHOD(SetGuestProperty) (IN_BSTR aName, IN_BSTR aValue, IN_BSTR aFlags);
+    STDMETHOD(SetGuestPropertyValue) (IN_BSTR aName, IN_BSTR aValue);
+    STDMETHOD(EnumerateGuestProperties) (IN_BSTR aPattern, ComSafeArrayOut(BSTR, aNames), ComSafeArrayOut(BSTR, aValues), ComSafeArrayOut(ULONG64, aTimestamps), ComSafeArrayOut(BSTR, aFlags));
+    STDMETHOD(GetHardDiskAttachmentsOfController)(IN_BSTR aName, ComSafeArrayOut (IHardDiskAttachment *, aAttachments));
+    STDMETHOD(AddStorageController) (IN_BSTR aName, StorageBus_T aConnectionType, IStorageController **controller);
+    STDMETHOD(RemoveStorageController (IN_BSTR aName));
+    STDMETHOD(GetStorageControllerByName (IN_BSTR aName, IStorageController **storageController));
 
     // public methods only for internal purposes
 
+    InstanceType type() const { return mType; }
+
     /// @todo (dmik) add lock and make non-inlined after revising classes
     //  that use it. Note: they should enter Machine lock to keep the returned
     //  information valid!
     bool isRegistered() { return !!mData->mRegistered; }
 
-    ComObjPtr <SessionMachine> sessionMachine();
+    // unsafe inline public methods for internal purposes only (ensure there is
+    // a caller and a read lock before calling them!)
 
     /**
-     *  Returns the VirtualBox object this machine belongs to.
+     * Returns the VirtualBox object this machine belongs to.
      *
-     *  @note This method doesn't check this object's readiness as it is
-     *  intended to be used only by Machine children where it is guaranteed
-     *  that this object still exists in memory.
+     * @note This method doesn't check this object's readiness. Intended to be
+     * used by ready Machine children (whose readiness is bound to the parent's
+     * one) or after doing addCaller() manually.
      */
     const ComObjPtr <VirtualBox, ComWeakRef> &virtualBox() const { return mParent; }
 
     /**
-     *  Returns this machine's name.
+     * Returns this machine ID.
      *
-     *  @note This method doesn't check this object's readiness as it is
-     *  intended to be used only after adding a caller to this object (that
-     *  guarantees that the object is ready or at least limited).
+     * @note This method doesn't check this object's readiness. Intended to be
+     * used by ready Machine children (whose readiness is bound to the parent's
+     * one) or after adding a caller manually.
      */
-    const Guid &uuid() const { return mData->mUuid; }
+    const Guid &id() const { return mData->mUuid; }
 
     /**
-     *  Returns this machine's name.
+     * Returns the snapshot ID this machine represents or an empty UUID if this
+     * instance is not SnapshotMachine.
      *
-     *  @note This method doesn't lock this object or check its readiness as
-     *  it is intended to be used only after adding a caller to this object
-     *  (that guarantees that the object is ready) and locking it for reading.
+     * @note This method doesn't check this object's readiness. Intended to be
+     * used by ready Machine children (whose readiness is bound to the parent's
+     * one) or after adding a caller manually.
+     */
+    inline const Guid &snapshotId() const;
+
+    /**
+     * Returns this machine's full settings file path.
+     *
+     * @note This method doesn't lock this object or check its readiness.
+     * Intended to be used only after doing addCaller() manually and locking it
+     * for reading.
+     */
+    const Bstr &settingsFileFull() const { return mData->mConfigFileFull; }
+
+    /**
+     * Returns this machine name.
+     *
+     * @note This method doesn't lock this object or check its readiness.
+     * Intended to be used only after doing addCaller() manually and locking it
+     * for reading.
      */
     const Bstr &name() const { return mUserData->mName; }
 
     // callback handlers
     virtual HRESULT onDVDDriveChange() { return S_OK; }
     virtual HRESULT onFloppyDriveChange() { return S_OK; }
-    virtual HRESULT onNetworkAdapterChange(INetworkAdapter *networkAdapter) { return S_OK; }
-    virtual HRESULT onSerialPortChange(ISerialPort *serialPort) { return S_OK; }
-    virtual HRESULT onParallelPortChange(IParallelPort *ParallelPort) { return S_OK; }
+    virtual HRESULT onNetworkAdapterChange(INetworkAdapter * /* networkAdapter */) { return S_OK; }
+    virtual HRESULT onSerialPortChange(ISerialPort * /* serialPort */) { return S_OK; }
+    virtual HRESULT onParallelPortChange(IParallelPort * /* parallelPort */) { return S_OK; }
     virtual HRESULT onVRDPServerChange() { return S_OK; }
     virtual HRESULT onUSBControllerChange() { return S_OK; }
+    virtual HRESULT onStorageControllerChange() { return S_OK; }
     virtual HRESULT onSharedFolderChange() { return S_OK; }
 
-    HRESULT saveRegistryEntry (CFGNODE aEntryNode);
+    HRESULT saveRegistryEntry (settings::Key &aEntryNode);
 
     int calculateFullPath (const char *aPath, Utf8Str &aResult);
     void calculateRelativePath (const char *aPath, Utf8Str &aResult);
 
     void getLogFolder (Utf8Str &aLogFolder);
 
-    bool isDVDImageUsed (const Guid &aId, ResourceUsage_T aUsage);
-    bool isFloppyImageUsed (const Guid &aId, ResourceUsage_T aUsage);
-
     HRESULT openSession (IInternalSessionControl *aControl);
     HRESULT openRemoteSession (IInternalSessionControl *aControl,
-                               INPTR BSTR aType, INPTR BSTR aEnvironment,
+                               IN_BSTR aType, IN_BSTR aEnvironment,
                                Progress *aProgress);
     HRESULT openExistingSession (IInternalSessionControl *aControl);
 
+#if defined (RT_OS_WINDOWS)
+
+    bool isSessionOpen (ComObjPtr <SessionMachine> &aMachine,
+                        ComPtr <IInternalSessionControl> *aControl = NULL,
+                        HANDLE *aIPCSem = NULL, bool aAllowClosing = false);
+    bool isSessionSpawning (RTPROCESS *aPID = NULL);
+
+    bool isSessionOpenOrClosing (ComObjPtr <SessionMachine> &aMachine,
+                                 ComPtr <IInternalSessionControl> *aControl = NULL,
+                                 HANDLE *aIPCSem = NULL)
+    { return isSessionOpen (aMachine, aControl, aIPCSem, true /* aAllowClosing */); }
+
+#elif defined (RT_OS_OS2)
+
+    bool isSessionOpen (ComObjPtr <SessionMachine> &aMachine,
+                        ComPtr <IInternalSessionControl> *aControl = NULL,
+                        HMTX *aIPCSem = NULL, bool aAllowClosing = false);
+
+    bool isSessionSpawning (RTPROCESS *aPID = NULL);
+
+    bool isSessionOpenOrClosing (ComObjPtr <SessionMachine> &aMachine,
+                                 ComPtr <IInternalSessionControl> *aControl = NULL,
+                                 HMTX *aIPCSem = NULL)
+    { return isSessionOpen (aMachine, aControl, aIPCSem, true /* aAllowClosing */); }
+
+#else
+
+    bool isSessionOpen (ComObjPtr <SessionMachine> &aMachine,
+                        ComPtr <IInternalSessionControl> *aControl = NULL,
+                        bool aAllowClosing = false);
+    bool isSessionSpawning();
+
+    bool isSessionOpenOrClosing (ComObjPtr <SessionMachine> &aMachine,
+                                 ComPtr <IInternalSessionControl> *aControl = NULL)
+    { return isSessionOpen (aMachine, aControl, true /* aAllowClosing */); }
+
+#endif
+
+    bool checkForSpawnFailure();
+
     HRESULT trySetRegistered (BOOL aRegistered);
 
-    HRESULT getSharedFolder (const BSTR aName,
+    HRESULT getSharedFolder (CBSTR aName,
                              ComObjPtr <SharedFolder> &aSharedFolder,
                              bool aSetError = false)
     {
-        AutoLock alock (this);
+        AutoWriteLock alock (this);
         return findSharedFolder (aName, aSharedFolder, aSetError);
     }
 
@@ -577,8 +714,6 @@ public:
 
 protected:
 
-    enum InstanceType { IsMachine, IsSessionMachine, IsSnapshotMachine };
-
     HRESULT registeredInit();
 
     HRESULT checkStateDependency (StateDependency aDepType);
@@ -588,79 +723,85 @@ protected:
     HRESULT initDataAndChildObjects();
     void uninitDataAndChildObjects();
 
-    void checkStateDependencies (AutoLock &aLock);
+    void ensureNoStateDependencies();
 
     virtual HRESULT setMachineState (MachineState_T aMachineState);
 
-    HRESULT findSharedFolder (const BSTR aName,
+    HRESULT findSharedFolder (CBSTR aName,
                               ComObjPtr <SharedFolder> &aSharedFolder,
                               bool aSetError = false);
 
     HRESULT loadSettings (bool aRegistered);
-    HRESULT loadSnapshot (CFGNODE aNode, const Guid &aCurSnapshotId,
+    HRESULT loadSnapshot (const settings::Key &aNode, const Guid &aCurSnapshotId,
                           Snapshot *aParentSnapshot);
-    HRESULT loadHardware (CFGNODE aNode);
-    HRESULT loadHardDisks (CFGNODE aNode, bool aRegistered,
+    HRESULT loadHardware (const settings::Key &aNode);
+    HRESULT loadStorageControllers (const settings::Key &aNode, bool aRegistered,
                            const Guid *aSnapshotId = NULL);
+    HRESULT loadStorageDevices (ComObjPtr<StorageController> aStorageController,
+                                const settings::Key &aNode, bool aRegistered,
+                                const Guid *aSnapshotId = NULL);
 
-    HRESULT openConfigLoader (CFGHANDLE *aLoader, bool aIsNew = false);
-    HRESULT closeConfigLoader (CFGHANDLE aLoader, bool aSaveBeforeClose);
-
-    HRESULT findSnapshotNode (Snapshot *aSnapshot, CFGNODE aMachineNode,
-                              CFGNODE *aSnapshotsNode, CFGNODE *aSnapshotNode);
+    HRESULT findSnapshotNode (Snapshot *aSnapshot, settings::Key &aMachineNode,
+                              settings::Key *aSnapshotsNode,
+                              settings::Key *aSnapshotNode);
 
     HRESULT findSnapshot (const Guid &aId, ComObjPtr <Snapshot> &aSnapshot,
                           bool aSetError = false);
-    HRESULT findSnapshot (const BSTR aName, ComObjPtr <Snapshot> &aSnapshot,
+    HRESULT findSnapshot (IN_BSTR aName, ComObjPtr <Snapshot> &aSnapshot,
                           bool aSetError = false);
 
-    HRESULT findHardDiskAttachment (const ComObjPtr <HardDisk> &aHd,
-                                    ComObjPtr <Machine> *aMachine,
-                                    ComObjPtr <Snapshot> *aSnapshot,
-                                    ComObjPtr <HardDiskAttachment> *aHda);
+    HRESULT getStorageControllerByName(CBSTR aName,
+                                       ComObjPtr <StorageController> &aStorageController,
+                                       bool aSetError = false);
 
-    HRESULT prepareSaveSettings (bool &aRenamed, bool &aNew);
-    HRESULT saveSettings (bool aMarkCurStateAsModified = true,
-                          bool aInformCallbacksAnyway = false);
+    HRESULT getHardDiskAttachmentsOfController(CBSTR aName,
+                                               HDData::AttachmentList &aAttachments);
 
     enum
     {
-        // ops for #saveSnapshotSettings()
+        /* flags for #saveSettings() */
+        SaveS_ResetCurStateModified = 0x01,
+        SaveS_InformCallbacksAnyway = 0x02,
+        /* ops for #saveSnapshotSettings() */
         SaveSS_NoOp = 0x00, SaveSS_AddOp = 0x01,
         SaveSS_UpdateAttrsOp = 0x02, SaveSS_UpdateAllOp = 0x03,
         SaveSS_OpMask = 0xF,
-        // flags for #saveSnapshotSettings()
-        SaveSS_UpdateCurStateModified = 0x40,
-        SaveSS_UpdateCurrentId = 0x80,
-        // flags for #saveStateSettings()
+        /* flags for #saveSnapshotSettings() */
+        SaveSS_CurStateModified = 0x40,
+        SaveSS_CurrentId = 0x80,
+        /* flags for #saveStateSettings() */
         SaveSTS_CurStateModified = 0x20,
         SaveSTS_StateFilePath = 0x40,
         SaveSTS_StateTimeStamp = 0x80,
     };
 
+    HRESULT prepareSaveSettings (bool &aRenamed, bool &aNew);
+    HRESULT saveSettings (int aFlags = 0);
+
     HRESULT saveSnapshotSettings (Snapshot *aSnapshot, int aOpFlags);
-    HRESULT saveSnapshotSettingsWorker (CFGNODE aMachineNode,
+    HRESULT saveSnapshotSettingsWorker (settings::Key &aMachineNode,
                                         Snapshot *aSnapshot, int aOpFlags);
 
-    HRESULT saveSnapshot (CFGNODE aNode, Snapshot *aSnapshot, bool aAttrsOnly);
-    HRESULT saveHardware (CFGNODE aNode);
-    HRESULT saveHardDisks (CFGNODE aNode);
+    HRESULT saveSnapshot (settings::Key &aNode, Snapshot *aSnapshot, bool aAttrsOnly);
+    HRESULT saveHardware (settings::Key &aNode);
+    HRESULT saveStorageControllers (settings::Key &aNode);
+    HRESULT saveStorageDevices (ComObjPtr<StorageController> aStorageController,
+                                settings::Key &aNode);
 
     HRESULT saveStateSettings (int aFlags);
 
-    HRESULT wipeOutImmutableDiffs();
-
-    HRESULT fixupHardDisks (bool aCommit);
-
-    HRESULT createSnapshotDiffs (const Guid *aSnapshotId,
-                                 const Bstr &aFolder,
-                                 const ComObjPtr <Progress> &aProgress,
+    HRESULT createImplicitDiffs (const Bstr &aFolder,
+                                 ComObjPtr <Progress> &aProgress,
                                  bool aOnline);
-    HRESULT deleteSnapshotDiffs (const ComObjPtr <Snapshot> &aSnapshot);
+    HRESULT deleteImplicitDiffs();
+
+    void fixupHardDisks(bool aCommit, bool aOnline = false);
 
     HRESULT lockConfig();
+public: /* To have it available in VirtualBox::UnregisterMachine. */
     HRESULT unlockConfig();
 
+protected:
     /** @note This method is not thread safe */
     BOOL isConfigLocked()
     {
@@ -672,9 +813,14 @@ protected:
     bool isModified();
     bool isReallyModified (bool aIgnoreUserData = false);
     void rollback (bool aNotify);
-    HRESULT commit();
+    void commit();
     void copyFrom (Machine *aThat);
 
+#ifdef VBOX_WITH_RESOURCE_USAGE_API
+    void registerMetrics (PerformanceCollector *aCollector, Machine *aMachine, RTPROCESS pid);
+    void unregisterMetrics (PerformanceCollector *aCollector, Machine *aMachine);
+#endif /* VBOX_WITH_RESOURCE_USAGE_API */
+
     const InstanceType mType;
 
     const ComObjPtr <Machine, ComWeakRef> mPeer;
@@ -704,6 +850,9 @@ protected:
     const ComObjPtr <NetworkAdapter>
         mNetworkAdapters [SchemaDefs::NetworkAdapterCount];
 
+    typedef std::list< ComObjPtr<StorageController> > StorageControllerList;
+    Backupable<StorageControllerList> mStorageControllers;
+
     friend class SessionMachine;
     friend class SnapshotMachine;
 };
@@ -749,45 +898,48 @@ public:
     HRESULT init (Machine *aMachine);
     void uninit() { uninit (Uninit::Unexpected); }
 
-    // AutoLock::Lockable interface
-    AutoLock::Handle *lockHandle() const;
+    // util::Lockable interface
+    RWLockHandle *lockHandle() const;
 
     // IInternalMachineControl methods
     STDMETHOD(UpdateState)(MachineState_T machineState);
     STDMETHOD(GetIPCId)(BSTR *id);
     STDMETHOD(RunUSBDeviceFilters) (IUSBDevice *aUSBDevice, BOOL *aMatched, ULONG *aMaskedIfs);
-    STDMETHOD(CaptureUSBDevice) (INPTR GUIDPARAM aId);
-    STDMETHOD(DetachUSBDevice) (INPTR GUIDPARAM aId, BOOL aDone);
+    STDMETHOD(CaptureUSBDevice) (IN_GUID aId);
+    STDMETHOD(DetachUSBDevice) (IN_GUID aId, BOOL aDone);
     STDMETHOD(AutoCaptureUSBDevices)();
     STDMETHOD(DetachAllUSBDevices)(BOOL aDone);
     STDMETHOD(OnSessionEnd)(ISession *aSession, IProgress **aProgress);
     STDMETHOD(BeginSavingState) (IProgress *aProgress, BSTR *aStateFilePath);
     STDMETHOD(EndSavingState) (BOOL aSuccess);
+    STDMETHOD(AdoptSavedState) (IN_BSTR aSavedStateFile);
     STDMETHOD(BeginTakingSnapshot) (IConsole *aInitiator,
-                                    INPTR BSTR aName, INPTR BSTR aDescription,
+                                    IN_BSTR aName, IN_BSTR aDescription,
                                     IProgress *aProgress, BSTR *aStateFilePath,
                                     IProgress **aServerProgress);
     STDMETHOD(EndTakingSnapshot) (BOOL aSuccess);
-    STDMETHOD(DiscardSnapshot) (IConsole *aInitiator, INPTR GUIDPARAM aId,
+    STDMETHOD(DiscardSnapshot) (IConsole *aInitiator, IN_GUID aId,
                                MachineState_T *aMachineState, IProgress **aProgress);
     STDMETHOD(DiscardCurrentState) (
         IConsole *aInitiator, MachineState_T *aMachineState, IProgress **aProgress);
     STDMETHOD(DiscardCurrentSnapshotAndState) (
         IConsole *aInitiator, MachineState_T *aMachineState, IProgress **aProgress);
+    STDMETHOD(PullGuestProperties) (ComSafeArrayOut(BSTR, aNames), ComSafeArrayOut(BSTR, aValues),
+              ComSafeArrayOut(ULONG64, aTimestamps), ComSafeArrayOut(BSTR, aFlags));
+    STDMETHOD(PushGuestProperties) (ComSafeArrayIn(IN_BSTR, aNames), ComSafeArrayIn(IN_BSTR, aValues),
+              ComSafeArrayIn(ULONG64, aTimestamps), ComSafeArrayIn(IN_BSTR, aFlags));
+    STDMETHOD(PushGuestProperty) (IN_BSTR aName, IN_BSTR aValue,
+                                  ULONG64 aTimestamp, IN_BSTR aFlags);
+    STDMETHOD(LockMedia)() { return lockMedia(); }
 
     // public methods only for internal purposes
 
     bool checkForDeath();
 
-#if defined (RT_OS_WINDOWS)
-    HANDLE ipcSem() { return mIPCSem; }
-#elif defined (RT_OS_OS2)
-    HMTX ipcSem() { return mIPCSem; }
-#endif
-
     HRESULT onDVDDriveChange();
     HRESULT onFloppyDriveChange();
     HRESULT onNetworkAdapterChange(INetworkAdapter *networkAdapter);
+    HRESULT onStorageControllerChange();
     HRESULT onSerialPortChange(ISerialPort *serialPort);
     HRESULT onParallelPortChange(IParallelPort *parallelPort);
     HRESULT onVRDPServerChange();
@@ -795,7 +947,7 @@ public:
     HRESULT onUSBDeviceAttach (IUSBDevice *aDevice,
                                IVirtualBoxErrorInfo *aError,
                                ULONG aMaskedIfs);
-    HRESULT onUSBDeviceDetach (INPTR GUIDPARAM aId,
+    HRESULT onUSBDeviceDetach (IN_GUID aId,
                                IVirtualBoxErrorInfo *aError);
     HRESULT onSharedFolderChange();
 
@@ -805,7 +957,7 @@ private:
 
     struct SnapshotData
     {
-        SnapshotData() : mLastState (MachineState_InvalidMachineState) {}
+        SnapshotData() : mLastState (MachineState_Null) {}
 
         MachineState_T mLastState;
 
@@ -819,7 +971,8 @@ private:
         Bstr mStateFilePath;
     };
 
-    struct Uninit {
+    struct Uninit
+    {
         enum Reason { Unexpected, Abnormal, Normal };
     };
 
@@ -843,20 +996,32 @@ private:
     void discardSnapshotHandler (DiscardSnapshotTask &aTask);
     void discardCurrentStateHandler (DiscardCurrentStateTask &aTask);
 
+    HRESULT lockMedia();
+    void unlockMedia();
+
     HRESULT setMachineState (MachineState_T aMachineState);
     HRESULT updateMachineStateOnClient();
 
     SnapshotData mSnapshotData;
 
-    /** interprocess semaphore handle (id) for this machine */
-#if defined(RT_OS_WINDOWS)
+    /** interprocess semaphore handle for this machine */
+#if defined (RT_OS_WINDOWS)
     HANDLE mIPCSem;
     Bstr mIPCSemName;
-#elif defined(RT_OS_OS2)
+    friend bool Machine::isSessionOpen (ComObjPtr <SessionMachine> &aMachine,
+                                        ComPtr <IInternalSessionControl> *aControl,
+                                        HANDLE *aIPCSem, bool aAllowClosing);
+#elif defined (RT_OS_OS2)
     HMTX mIPCSem;
     Bstr mIPCSemName;
-#elif defined(VBOX_WITH_SYS_V_IPC_SESSION_WATCHER)
+    friend bool Machine::isSessionOpen (ComObjPtr <SessionMachine> &aMachine,
+                                        ComPtr <IInternalSessionControl> *aControl,
+                                        HMTX *aIPCSem, bool aAllowClosing);
+#elif defined (VBOX_WITH_SYS_V_IPC_SESSION_WATCHER)
     int mIPCSem;
+# ifdef VBOX_WITH_NEW_SYS_V_KEYGEN
+    Bstr mIPCKey;
+# endif /*VBOX_WITH_NEW_SYS_V_KEYGEN */
 #else
 # error "Port me!"
 #endif
@@ -901,18 +1066,24 @@ public:
 
     // public initializer/uninitializer for internal purposes only
     HRESULT init (SessionMachine *aSessionMachine,
-                  INPTR GUIDPARAM aSnapshotId, INPTR BSTR aStateFilePath);
-    HRESULT init (Machine *aMachine, CFGNODE aHWNode, CFGNODE aHDAsNode,
-                  INPTR GUIDPARAM aSnapshotId, INPTR BSTR aStateFilePath);
+                  IN_GUID aSnapshotId, IN_BSTR aStateFilePath);
+    HRESULT init (Machine *aMachine,
+                  const settings::Key &aHWNode, const settings::Key &aHDAsNode,
+                  IN_GUID aSnapshotId, IN_BSTR aStateFilePath);
     void uninit();
 
-    // AutoLock::Lockable interface
-    AutoLock::Handle *lockHandle() const;
+    // util::Lockable interface
+    RWLockHandle *lockHandle() const;
 
     // public methods only for internal purposes
 
     HRESULT onSnapshotChange (Snapshot *aSnapshot);
 
+    // unsafe inline public methods for internal purposes only (ensure there is
+    // a caller and a read lock before calling them!)
+
+    const Guid &snapshotId() const { return mSnapshotId; }
+
 private:
 
     Guid mSnapshotId;
@@ -920,6 +1091,14 @@ private:
     friend class Snapshot;
 };
 
+// third party methods that depend on SnapshotMachine definiton
+
+inline const Guid &Machine::snapshotId() const
+{
+    return mType != IsSnapshotMachine ? Guid::Empty :
+                    static_cast <const SnapshotMachine *> (this)->snapshotId();
+}
+
 ////////////////////////////////////////////////////////////////////////////////
 
 /**
@@ -937,6 +1116,5 @@ inline Machine *Machine::machine()
     return this;
 }
 
-COM_DECL_READONLY_ENUM_AND_COLLECTION (Machine)
-
 #endif // ____H_MACHINEIMPL
+/* vi: set tabstop=4 shiftwidth=4 expandtab: */
