@@ -1,18 +1,31 @@
-/* $Id: strformat-vbox.cpp 23517 2007-08-07 17:07:59Z umoeller $ */
+/* $Id: strformat-vbox.cpp 14061 2008-11-10 23:12:54Z vboxsync $ */
 /** @file
- * innotek Portable Runtime - VBox String Formatter extensions.
+ * IPRT - VBox String Formatter extensions.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * The contents of this file may alternatively be used under the terms
+ * of the Common Development and Distribution License Version 1.0
+ * (CDDL) only, as it comes in the "COPYING.CDDL" file of the
+ * VirtualBox OSE distribution, in which case the provisions of the
+ * CDDL are applicable instead of those of the GPL.
+ *
+ * You may elect to license modified versions of this file under the
+ * terms and conditions of either the GPL or the CDDL or both.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 /** @page pg_rt_str_format_vbox    The VBox String Format Extensions
@@ -39,10 +52,11 @@
  *      - \%Vhxs - Takes a pointer to the memory to be displayed as a hex string,
  *        i.e. a series of space separated bytes formatted as two digit hex value.
  *        Use the width to specify the length. Default length is 16 bytes.
- *      - \%VGp  - Guest context physical address.
- *      - \%VGv  - Guest context virtual address.
- *      - \%VHp  - Host context physical address.
- *      - \%VHv  - Host context virtual address.
+ *      - \%VGp   - Guest context physical address.
+ *      - \%VGv   - Guest context virtual address.
+ *      - \%VRv   - 32 bits guest context virtual address.
+ *      - \%VHp   - Host context physical address.
+ *      - \%VHv   - Host context virtual address.
  *      - \%VI[8|16|32|64] - Signed integer value of the specifed bit count.
  *      - \%VU[8|16|32|64] - Unsigned integer value of the specifed bit count.
  *      - \%VX[8|16|32|64] - Hexadecimal integer value of the specifed bit count.
@@ -129,11 +143,11 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
              * Current thread.
              */
             case 't':
-#ifdef IN_RING3                         /* we don't want this anywhere else yet. */
+#if defined(IN_RING3) && !defined(RT_MINI)  /* we don't want this anywhere else yet. */
                 return RTStrFormat(pfnOutput, pvArgOutput, NULL, 0, "%#x\n", RTThreadSelf());
-#else /* !IN_RING3 */
+#else /* !IN_RING3 || RT_MINI */
                 return pfnOutput(pvArgOutput, "0xffffffff", sizeof("0xffffffff") - 1);
-#endif /* !IN_RING3 */
+#endif /* !IN_RING3 || RT_MINI */
 
             case 'h':
             {
@@ -158,11 +172,12 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                                  */
                                 case 'd':
                                 {
+                                    size_t  cch = 0;
+                                    int     off = 0;
+
                                     if (cchPrecision <= 0)
                                         cchPrecision = 16;
 
-                                    int cch = 0;
-                                    int off = 0;
                                     while (off < cchWidth)
                                     {
                                         int i;
@@ -175,7 +190,7 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
 
                                         cch += pfnOutput(pvArgOutput, " ", 1);
 
-                                        for (int i = 0; i < cchPrecision && off + i < cchWidth; i++)
+                                        for (i = 0; i < cchPrecision && off + i < cchWidth; i++)
                                         {
                                             uint8_t u8 = pu8[i];
                                             cch += pfnOutput(pvArgOutput, u8 < 127 && u8 >= 32 ? (const char *)&u8 : ".", 1);
@@ -195,7 +210,7 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                                 {
                                     if (cchWidth-- > 0)
                                     {
-                                        int cch = RTStrFormat(pfnOutput, pvArgOutput, NULL, 0, "%02x", *pu8++);
+                                        size_t cch = RTStrFormat(pfnOutput, pvArgOutput, NULL, 0, "%02x", *pu8++);
                                         for (; cchWidth > 0; cchWidth--, pu8++)
                                             cch += RTStrFormat(pfnOutput, pvArgOutput, NULL, 0, " %02x", *pu8);
                                         return cch;
@@ -261,6 +276,35 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                 break;
             }
 
+            /*
+             * Raw mode types.
+             */
+            case 'R':
+            {
+                /*
+                 * Raw mode types.
+                 */
+                ch = *(*ppszFormat)++;
+                switch (ch)
+                {
+                    case 'v':   /* Virtual address. */
+                    {
+                        RTRCPTR RCPtr = va_arg(*pArgs, RTRCPTR);
+                        switch (sizeof(RCPtr))
+                        {
+                            case 4: return RTStrFormat(pfnOutput, pvArgOutput, NULL, 0, "%08x", RCPtr);
+                            default: AssertFailed(); return 0;
+                        }
+                    }
+
+                    default:
+                        AssertMsgFailed(("Invalid status code format type '%%VR%c%.10s'!\n", ch, *ppszFormat));
+                        break;
+                }
+                break;
+            }
+
+
             /*
              * Host types.
              */
@@ -310,6 +354,9 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                  * Figure out number of bits.
                  */
                 int64_t i64 = 0;
+                char szNum[64];
+                int cch;
+
                 ch = *(*ppszFormat)++;
                 if (ch == '8')
                     i64 = (int8_t)va_arg(*pArgs, int);
@@ -340,8 +387,7 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                 else
                     AssertMsgFailed(("Invalid format %%VI%c.10s\n", ch, *ppszFormat));
 
-                char szNum[64];
-                int cch = RTStrFormatNumber(szNum, i64, 10, cchWidth, cchPrecision, fFlags | RTSTR_F_VALSIGNED);
+                cch = RTStrFormatNumber(szNum, i64, 10, cchWidth, cchPrecision, fFlags | RTSTR_F_VALSIGNED);
                 Assert(cch < (int)sizeof(szNum));
                 return pfnOutput(pvArgOutput, szNum, cch);
             }
@@ -359,6 +405,9 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                  * Figure out number of bits and read the value.
                  */
                 uint64_t u64 = 0;
+                char szNum[64];
+                int cch;
+
                 ch = *(*ppszFormat)++;
                 if (ch == '8')
                     u64 = (uint8_t)va_arg(*pArgs, unsigned);
@@ -389,8 +438,7 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                 else
                     AssertMsgFailed(("Invalid format %%VI%c.10s\n", ch, *ppszFormat));
 
-                char szNum[64];
-                int cch = RTStrFormatNumber(szNum, u64, iBase, cchWidth, cchPrecision, fFlags);
+                cch = RTStrFormatNumber(szNum, u64, iBase, cchWidth, cchPrecision, fFlags);
                 Assert(cch < (int)sizeof(szNum));
                 return pfnOutput(pvArgOutput, szNum, cch);
             }
@@ -403,8 +451,10 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                     &&  (*ppszFormat)[1] == 'i'
                     &&  (*ppszFormat)[2] == 'd')
                 {
-                    (*ppszFormat) += 3;
                     PRTUUID pUuid = va_arg(*pArgs, PRTUUID);
+                    static const char szNull[] = "<NULL>";
+
+                    (*ppszFormat) += 3;
                     if (VALID_PTR(pUuid))
                     {
                         /* cannot call RTUuidToStr because of GC/R0. */
@@ -413,8 +463,8 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                                            pUuid->Gen.u32TimeLow,
                                            pUuid->Gen.u16TimeMid,
                                            pUuid->Gen.u16TimeHiAndVersion,
-                                           pUuid->Gen.u16ClockSeq & 0xff,
-                                           pUuid->Gen.u16ClockSeq >> 8,
+                                           pUuid->Gen.u8ClockSeqHiAndReserved,
+                                           pUuid->Gen.u8ClockSeqLow,
                                            pUuid->Gen.au8Node[0],
                                            pUuid->Gen.au8Node[1],
                                            pUuid->Gen.au8Node[2],
@@ -423,7 +473,6 @@ size_t rtstrFormatVBox(PFNRTSTROUTPUT pfnOutput, void *pvArgOutput, const char *
                                            pUuid->Gen.au8Node[5]);
                     }
 
-                    static const char szNull[] = "<NULL>";
                     return pfnOutput(pvArgOutput, szNull, sizeof(szNull) - 1);
                 }
                 /* fall thru */
