@@ -1,18 +1,31 @@
-/* $Id: memobj-r0drv-nt.cpp 24320 2007-09-10 15:19:44Z sandervl $ */
+/* $Id: memobj-r0drv-nt.cpp 14824 2008-11-30 07:52:59Z vboxsync $ */
 /** @file
- * innotek Portable Runtime - Ring-0 Memory Objects, NT.
+ * IPRT - Ring-0 Memory Objects, NT.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * The contents of this file may alternatively be used under the terms
+ * of the Common Development and Distribution License Version 1.0
+ * (CDDL) only, as it comes in the "COPYING.CDDL" file of the
+ * VirtualBox OSE distribution, in which case the provisions of the
+ * CDDL are applicable instead of those of the GPL.
+ *
+ * You may elect to license modified versions of this file under the
+ * terms and conditions of either the GPL or the CDDL or both.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 
@@ -60,6 +73,8 @@ typedef struct RTR0MEMOBJNT
     /** Used MmAllocatePagesForMdl(). */
     bool                fAllocatedPagesForMdl;
 #endif
+    /** Pointer returned by MmSecureVirtualMemory */
+    PVOID               pvSecureMem;
     /** The number of PMDLs (memory descriptor lists) in the array. */
     uint32_t            cMdls;
     /** Array of MDL pointers. (variable size) */
@@ -83,6 +98,11 @@ int rtR0MemObjNativeFree(RTR0MEMOBJ pMem)
                 Assert(pMemNt->Core.pv && pMemNt->cMdls == 1 && pMemNt->apMdls[0]);
                 MmUnmapLockedPages(pMemNt->Core.pv, pMemNt->apMdls[0]);
                 pMemNt->Core.pv = NULL;
+                if (pMemNt->pvSecureMem)
+                {
+                    MmUnsecureVirtualMemory(pMemNt->pvSecureMem);
+                    pMemNt->pvSecureMem = NULL;
+                }
 
                 MmFreePagesFromMdl(pMemNt->apMdls[0]);
                 ExFreePool(pMemNt->apMdls[0]);
@@ -132,6 +152,11 @@ int rtR0MemObjNativeFree(RTR0MEMOBJ pMem)
             break;
 
         case RTR0MEMOBJTYPE_LOCK:
+            if (pMemNt->pvSecureMem)
+            {
+                MmUnsecureVirtualMemory(pMemNt->pvSecureMem);
+                pMemNt->pvSecureMem = NULL;
+            }
             for (uint32_t i = 0; i < pMemNt->cMdls; i++)
             {
                 MmUnlockPages(pMemNt->apMdls[i]);
@@ -284,6 +309,8 @@ int rtR0MemObjNativeAllocLow(PPRTR0MEMOBJINTERNAL ppMem, size_t cb, bool fExecut
             }
             __except(EXCEPTION_EXECUTE_HANDLER)
             {
+                NTSTATUS rcNt = GetExceptionCode();
+                Log(("rtR0MemObjNativeAllocLow: Exception Code %#x\n", rcNt));
                 /* nothing */
             }
         }
@@ -476,7 +503,7 @@ int rtR0MemObjNativeEnterPhys(PPRTR0MEMOBJINTERNAL ppMem, RTHCPHYS Phys, size_t
  * @param ppMem     Where to store the memory object pointer.
  * @param pv        First page.
  * @param cb        Number of bytes.
- * @param Task      The task \a pv and \a cb refers to.
+ * @param R0Process The process \a pv and \a cb refers to.
  */
 static int rtR0MemObjNtLock(PPRTR0MEMOBJINTERNAL ppMem, void *pv, size_t cb, RTR0PROCESS R0Process)
 {
@@ -522,6 +549,7 @@ static int rtR0MemObjNtLock(PPRTR0MEMOBJINTERNAL ppMem, void *pv, size_t cb, RTR
         __try
         {
             MmProbeAndLockPages(pMdl, R0Process == NIL_RTR0PROCESS ? KernelMode : UserMode, IoModifyAccess);
+
             pMemNt->apMdls[iMdl] = pMdl;
             pMemNt->cMdls++;
         }
@@ -532,6 +560,17 @@ static int rtR0MemObjNtLock(PPRTR0MEMOBJINTERNAL ppMem, void *pv, size_t cb, RTR
             break;
         }
 
+        if (R0Process != NIL_RTR0PROCESS )
+        {
+            /* Make sure the user process can't change the allocation. */
+            pMemNt->pvSecureMem = MmSecureVirtualMemory(pv, cb, PAGE_READWRITE);
+            if (!pMemNt->pvSecureMem)
+            {
+                rc = VERR_NO_MEMORY;
+                break;
+            }
+        }
+
         /* next */
         cbTotal += cbCur;
         pb      += cbCur;
@@ -553,8 +592,14 @@ static int rtR0MemObjNtLock(PPRTR0MEMOBJINTERNAL ppMem, void *pv, size_t cb, RTR
         IoFreeMdl(pMemNt->apMdls[iMdl]);
         pMemNt->apMdls[iMdl] = NULL;
     }
+    if (pMemNt->pvSecureMem)
+    {
+        MmUnsecureVirtualMemory(pMemNt->pvSecureMem);
+        pMemNt->pvSecureMem = NULL;
+    }
+
     rtR0MemObjDelete(&pMemNt->Core);
-    return VERR_LOCK_FAILED;
+    return rc;
 }
 
 
@@ -657,9 +702,13 @@ static int rtR0MemObjNtMap(PPRTR0MEMOBJINTERNAL ppMem, RTR0MEMOBJ pMemToMap, voi
         }
         __except(EXCEPTION_EXECUTE_HANDLER)
         {
+            NTSTATUS rcNt = GetExceptionCode();
+            Log(("rtR0MemObjNtMap: Exception Code %#x\n", rcNt));
+
             /* nothing */
             rc = VERR_MAP_FAILED;
         }
+
     }
     else
     {
@@ -696,8 +745,10 @@ static int rtR0MemObjNtMap(PPRTR0MEMOBJINTERNAL ppMem, RTR0MEMOBJ pMemToMap, voi
 }
 
 
-int rtR0MemObjNativeMapKernel(PPRTR0MEMOBJINTERNAL ppMem, RTR0MEMOBJ pMemToMap, void *pvFixed, size_t uAlignment, unsigned fProt)
+int rtR0MemObjNativeMapKernel(PPRTR0MEMOBJINTERNAL ppMem, RTR0MEMOBJ pMemToMap, void *pvFixed, size_t uAlignment,
+                              unsigned fProt, size_t offSub, size_t cbSub)
 {
+    AssertMsgReturn(!offSub && !cbSub, ("%#x %#x\n", offSub, cbSub), VERR_NOT_SUPPORTED);
     return rtR0MemObjNtMap(ppMem, pMemToMap, pvFixed, uAlignment, fProt, NIL_RTR0PROCESS);
 }
 
