@@ -1,18 +1,22 @@
-/* $Id: DBGFBp.cpp 23517 2007-08-07 17:07:59Z umoeller $ */
+/* $Id: DBGFBp.cpp 13818 2008-11-04 22:59:47Z vboxsync $ */
 /** @file
- * VMM DBGF - Debugger Facility, Breakpoint Management.
+ * DBGF - Debugger Facility, Breakpoint Management.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 
@@ -51,6 +55,7 @@ static int dbgfR3BpInt3Disarm(PVM pVM, PDBGFBP pBp);
 __END_DECLS
 
 
+
 /**
  * Initialize the breakpoint stuff.
  *
@@ -63,16 +68,16 @@ int dbgfR3BpInit(PVM pVM)
      * Init structures.
      */
     unsigned i;
-    for (i = 0; i < ELEMENTS(pVM->dbgf.s.aHwBreakpoints); i++)
+    for (i = 0; i < RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints); i++)
     {
         pVM->dbgf.s.aHwBreakpoints[i].iBp = i;
         pVM->dbgf.s.aHwBreakpoints[i].enmType = DBGFBPTYPE_FREE;
         pVM->dbgf.s.aHwBreakpoints[i].u.Reg.iReg = i;
     }
 
-    for (i = 0; i < ELEMENTS(pVM->dbgf.s.aBreakpoints); i++)
+    for (i = 0; i < RT_ELEMENTS(pVM->dbgf.s.aBreakpoints); i++)
     {
-        pVM->dbgf.s.aBreakpoints[i].iBp = i + ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
+        pVM->dbgf.s.aBreakpoints[i].iBp = i + RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
         pVM->dbgf.s.aBreakpoints[i].enmType = DBGFBPTYPE_FREE;
     }
 
@@ -105,14 +110,14 @@ static PDBGFBP dbgfR3BpAlloc(PVM pVM, DBGFBPTYPE enmType)
     switch (enmType)
     {
         case DBGFBPTYPE_REG:
-            cBps = ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
+            cBps = RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
             paBps = &pVM->dbgf.s.aHwBreakpoints[0];
             pcBpsCur = &pVM->dbgf.s.cHwBreakpoints;
             break;
 
         case DBGFBPTYPE_INT3:
         case DBGFBPTYPE_REM:
-            cBps = ELEMENTS(pVM->dbgf.s.aBreakpoints);
+            cBps = RT_ELEMENTS(pVM->dbgf.s.aBreakpoints);
             paBps = &pVM->dbgf.s.aBreakpoints[0];
             pcBpsCur = &pVM->dbgf.s.cBreakpoints;
             break;
@@ -151,12 +156,12 @@ static PDBGFBP dbgfR3BpGet(PVM pVM, RTUINT iBp)
 {
     /* Find it. */
     PDBGFBP pBp;
-    if (iBp < ELEMENTS(pVM->dbgf.s.aHwBreakpoints))
+    if (iBp < RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints))
         pBp = &pVM->dbgf.s.aHwBreakpoints[iBp];
     else
     {
-        iBp -= ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
-        if (iBp >= ELEMENTS(pVM->dbgf.s.aBreakpoints))
+        iBp -= RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
+        if (iBp >= RT_ELEMENTS(pVM->dbgf.s.aBreakpoints))
             return NULL;
         pBp = &pVM->dbgf.s.aBreakpoints[iBp];
     }
@@ -200,13 +205,13 @@ static PDBGFBP dbgfR3BpGetByAddr(PVM pVM, DBGFBPTYPE enmType, RTGCUINTPTR GCPtr)
     switch (enmType)
     {
         case DBGFBPTYPE_REG:
-            cBps = ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
+            cBps = RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints);
             paBps = &pVM->dbgf.s.aHwBreakpoints[0];
             break;
 
         case DBGFBPTYPE_INT3:
         case DBGFBPTYPE_REM:
-            cBps = ELEMENTS(pVM->dbgf.s.aBreakpoints);
+            cBps = RT_ELEMENTS(pVM->dbgf.s.aBreakpoints);
             paBps = &pVM->dbgf.s.aBreakpoints[0];
             break;
 
@@ -276,17 +281,17 @@ static void dbgfR3BpFree(PVM pVM, PDBGFBP pBp)
  * @param   piBp        Where to store the breakpoint id. (optional)
  * @thread  Any thread.
  */
-DBGFR3DECL(int) DBGFR3BpSet(PVM pVM, PCDBGFADDRESS pAddress, uint64_t iHitTrigger, uint64_t iHitDisable, PRTUINT piBp)
+VMMR3DECL(int) DBGFR3BpSet(PVM pVM, PCDBGFADDRESS pAddress, uint64_t iHitTrigger, uint64_t iHitDisable, PRTUINT piBp)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpSetInt3, 5, pVM, pAddress, &iHitTrigger, &iHitDisable, piBp);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpSetInt3, 5, pVM, pAddress, &iHitTrigger, &iHitDisable, piBp);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpSet: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpSet: returns %Rrc\n", rc));
     return rc;
 }
 
@@ -326,7 +331,7 @@ static DECLCALLBACK(int) dbgfR3BpSetInt3(PVM pVM, PCDBGFADDRESS pAddress, uint64
         int rc = VINF_SUCCESS;
         if (!pBp->fEnabled)
             rc = dbgfR3BpInt3Arm(pVM, pBp);
-        if (VBOX_SUCCESS(rc))
+        if (RT_SUCCESS(rc))
         {
             rc = VINF_DBGF_BP_ALREADY_EXIST;
             if (piBp)
@@ -350,7 +355,7 @@ static DECLCALLBACK(int) dbgfR3BpSetInt3(PVM pVM, PCDBGFADDRESS pAddress, uint64
      * Now ask REM to set the breakpoint.
      */
     int rc = dbgfR3BpInt3Arm(pVM, pBp);
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
     {
         if (piBp)
             *piBp = pBp->iBp;
@@ -377,7 +382,7 @@ static int dbgfR3BpInt3Arm(PVM pVM, PDBGFBP pBp)
      * Save current byte and write int3 instruction.
      */
     int rc = MMR3ReadGCVirt(pVM, &pBp->u.Int3.bOrg, pBp->GCPtr, 1);
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
     {
         static const uint8_t s_bInt3 = 0xcc;
         rc = MMR3WriteGCVirt(pVM, pBp->GCPtr, &s_bInt3, 1);
@@ -424,18 +429,18 @@ static int dbgfR3BpInt3Disarm(PVM pVM, PDBGFBP pBp)
  * @param   piBp        Where to store the breakpoint id. (optional)
  * @thread  Any thread.
  */
-DBGFR3DECL(int) DBGFR3BpSetReg(PVM pVM, PCDBGFADDRESS pAddress, uint64_t iHitTrigger, uint64_t iHitDisable,
+VMMR3DECL(int) DBGFR3BpSetReg(PVM pVM, PCDBGFADDRESS pAddress, uint64_t iHitTrigger, uint64_t iHitDisable,
                                uint8_t fType, uint8_t cb, PRTUINT piBp)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpSetReg, 7, pVM, pAddress, &iHitTrigger, &iHitDisable, fType, cb, piBp);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpSetReg, 7, pVM, pAddress, &iHitTrigger, &iHitDisable, fType, cb, piBp);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpSetReg: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpSetReg: returns %Rrc\n", rc));
     return rc;
 
 }
@@ -508,7 +513,7 @@ static DECLCALLBACK(int) dbgfR3BpSetReg(PVM pVM, PCDBGFADDRESS pAddress, uint64_
         int rc = VINF_SUCCESS;
         if (!pBp->fEnabled)
             rc = dbgfR3BpRegArm(pVM, pBp);
-        if (VBOX_SUCCESS(rc))
+        if (RT_SUCCESS(rc))
         {
             rc = VINF_DBGF_BP_ALREADY_EXIST;
             if (piBp)
@@ -535,7 +540,7 @@ static DECLCALLBACK(int) dbgfR3BpSetReg(PVM pVM, PCDBGFADDRESS pAddress, uint64_
      * Arm the breakpoint.
      */
     int rc = dbgfR3BpRegArm(pVM, pBp);
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
     {
         if (piBp)
             *piBp = pBp->iBp;
@@ -590,17 +595,17 @@ static int dbgfR3BpRegDisarm(PVM pVM, PDBGFBP pBp)
  * @param   piBp        Where to store the breakpoint id. (optional)
  * @thread  Any thread.
  */
-DBGFR3DECL(int) DBGFR3BpSetREM(PVM pVM, PCDBGFADDRESS pAddress, uint64_t iHitTrigger, uint64_t iHitDisable, PRTUINT piBp)
+VMMR3DECL(int) DBGFR3BpSetREM(PVM pVM, PCDBGFADDRESS pAddress, uint64_t iHitTrigger, uint64_t iHitDisable, PRTUINT piBp)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpSetREM, 5, pVM, pAddress, &iHitTrigger, &iHitDisable, piBp);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpSetREM, 5, pVM, pAddress, &iHitTrigger, &iHitDisable, piBp);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpSetREM: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpSetREM: returns %Rrc\n", rc));
     return rc;
 }
 
@@ -642,7 +647,7 @@ static DECLCALLBACK(int) dbgfR3BpSetREM(PVM pVM, PCDBGFADDRESS pAddress, uint64_
         int rc = VINF_SUCCESS;
         if (!pBp->fEnabled)
             rc = REMR3BreakpointSet(pVM, pBp->GCPtr);
-        if (VBOX_SUCCESS(rc))
+        if (RT_SUCCESS(rc))
         {
             rc = VINF_DBGF_BP_ALREADY_EXIST;
             if (piBp)
@@ -666,7 +671,7 @@ static DECLCALLBACK(int) dbgfR3BpSetREM(PVM pVM, PCDBGFADDRESS pAddress, uint64_
      * Now ask REM to set the breakpoint.
      */
     int rc = REMR3BreakpointSet(pVM, pAddress->FlatPtr);
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
     {
         if (piBp)
             *piBp = pBp->iBp;
@@ -686,17 +691,17 @@ static DECLCALLBACK(int) dbgfR3BpSetREM(PVM pVM, PCDBGFADDRESS pAddress, uint64_
  * @param   iBp         The id of the breakpoint which should be removed (cleared).
  * @thread  Any thread.
  */
-DBGFR3DECL(int) DBGFR3BpClear(PVM pVM, RTUINT iBp)
+VMMR3DECL(int) DBGFR3BpClear(PVM pVM, RTUINT iBp)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpClear, 2, pVM, iBp);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpClear, 2, pVM, iBp);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpClear: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpClear: returns %Rrc\n", rc));
     return rc;
 }
 
@@ -763,17 +768,17 @@ static DECLCALLBACK(int) dbgfR3BpClear(PVM pVM, RTUINT iBp)
  * @param   iBp         The id of the breakpoint which should be enabled.
  * @thread  Any thread.
  */
-DBGFR3DECL(int) DBGFR3BpEnable(PVM pVM, RTUINT iBp)
+VMMR3DECL(int) DBGFR3BpEnable(PVM pVM, RTUINT iBp)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpEnable, 2, pVM, iBp);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpEnable, 2, pVM, iBp);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpEnable: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpEnable: returns %Rrc\n", rc));
     return rc;
 }
 
@@ -825,7 +830,7 @@ static DECLCALLBACK(int) dbgfR3BpEnable(PVM pVM, RTUINT iBp)
             AssertMsgFailed(("Invalid enmType=%d!\n", pBp->enmType));
             return VERR_INTERNAL_ERROR;
     }
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
         pBp->fEnabled = false;
 
     return rc;
@@ -840,17 +845,17 @@ static DECLCALLBACK(int) dbgfR3BpEnable(PVM pVM, RTUINT iBp)
  * @param   iBp         The id of the breakpoint which should be disabled.
  * @thread  Any thread.
  */
-DBGFR3DECL(int) DBGFR3BpDisable(PVM pVM, RTUINT iBp)
+VMMR3DECL(int) DBGFR3BpDisable(PVM pVM, RTUINT iBp)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpDisable, 2, pVM, iBp);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpDisable, 2, pVM, iBp);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpDisable: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpDisable: returns %Rrc\n", rc));
     return rc;
 }
 
@@ -916,17 +921,17 @@ static DECLCALLBACK(int) dbgfR3BpDisable(PVM pVM, RTUINT iBp)
  * @param   pvUser      The user argument to pass to the callback.
  * @thread  Any thread but the callback will be called from EMT.
  */
-DBGFR3DECL(int) DBGFR3BpEnum(PVM pVM, PFNDBGFBPENUM pfnCallback, void *pvUser)
+VMMR3DECL(int) DBGFR3BpEnum(PVM pVM, PFNDBGFBPENUM pfnCallback, void *pvUser)
 {
     /*
      * This must be done in EMT.
      */
     PVMREQ pReq;
-    int rc = VMR3ReqCall(pVM, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpEnum, 3, pVM, pfnCallback, pvUser);
-    if (VBOX_SUCCESS(rc))
+    int rc = VMR3ReqCall(pVM, VMREQDEST_ANY, &pReq, RT_INDEFINITE_WAIT, (PFNRT)dbgfR3BpEnum, 3, pVM, pfnCallback, pvUser);
+    if (RT_SUCCESS(rc))
         rc = pReq->iStatus;
     VMR3ReqFree(pReq);
-    LogFlow(("DBGFR3BpClear: returns %Vrc\n", rc));
+    LogFlow(("DBGFR3BpClear: returns %Rrc\n", rc));
     return rc;
 }
 
@@ -952,22 +957,22 @@ static DECLCALLBACK(int) dbgfR3BpEnum(PVM pVM, PFNDBGFBPENUM pfnCallback, void *
      * Enumerate the hardware breakpoints.
      */
     unsigned i;
-    for (i = 0; i < ELEMENTS(pVM->dbgf.s.aHwBreakpoints); i++)
+    for (i = 0; i < RT_ELEMENTS(pVM->dbgf.s.aHwBreakpoints); i++)
         if (pVM->dbgf.s.aHwBreakpoints[i].enmType != DBGFBPTYPE_FREE)
         {
             int rc = pfnCallback(pVM, pvUser, &pVM->dbgf.s.aHwBreakpoints[i]);
-            if (VBOX_FAILURE(rc))
+            if (RT_FAILURE(rc))
                 return rc;
         }
 
     /*
      * Enumerate the other breakpoints.
      */
-    for (i = 0; i < ELEMENTS(pVM->dbgf.s.aBreakpoints); i++)
+    for (i = 0; i < RT_ELEMENTS(pVM->dbgf.s.aBreakpoints); i++)
         if (pVM->dbgf.s.aBreakpoints[i].enmType != DBGFBPTYPE_FREE)
         {
             int rc = pfnCallback(pVM, pvUser, &pVM->dbgf.s.aBreakpoints[i]);
-            if (VBOX_FAILURE(rc))
+            if (RT_FAILURE(rc))
                 return rc;
         }
 
