@@ -1,18 +1,22 @@
-/* $Id: PDMThread.cpp 24047 2007-08-30 21:57:57Z bird $ */
+/* $Id: PDMThread.cpp 15540 2008-12-15 20:01:28Z vboxsync $ */
 /** @file
  * PDM Thread - VM Thread Management.
  */
 
 /*
- * Copyright (C) 2007 innotek GmbH
+ * Copyright (C) 2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 /*******************************************************************************
@@ -57,6 +61,8 @@ DECLINLINE(bool) pdmR3AtomicCmpXchgState(PPDMTHREAD pThread, PDMTHREADSTATE enmN
  */
 static DECLCALLBACK(int) pdmR3ThreadWakeUp(PPDMTHREAD pThread)
 {
+    RTSemEventMultiSignal(pThread->Internal.s.SleepEvent);
+
     int rc;
     switch (pThread->Internal.s.enmType)
     {
@@ -137,37 +143,43 @@ static int pdmR3ThreadInit(PVM pVM, PPPDMTHREAD ppThread, size_t cbStack, RTTHRE
     int rc = RTSemEventMultiCreate(&pThread->Internal.s.BlockEvent);
     if (RT_SUCCESS(rc))
     {
-        /*
-         * Create the thread and wait for it to initialize.
-         * The newly created thread will set the PDMTHREAD::Thread member.
-         */
-        RTTHREAD Thread;
-        rc = RTThreadCreate(&Thread, pdmR3ThreadMain, pThread, cbStack, enmType, RTTHREADFLAGS_WAITABLE, pszName);
+        rc = RTSemEventMultiCreate(&pThread->Internal.s.SleepEvent);
         if (RT_SUCCESS(rc))
         {
-            rc = RTThreadUserWait(Thread, 60*1000);
-            if (    RT_SUCCESS(rc)
-                &&  pThread->enmState != PDMTHREADSTATE_SUSPENDED)
-                rc = VERR_INTERNAL_ERROR;
+            /*
+             * Create the thread and wait for it to initialize.
+             * The newly created thread will set the PDMTHREAD::Thread member.
+             */
+            RTTHREAD Thread;
+            rc = RTThreadCreate(&Thread, pdmR3ThreadMain, pThread, cbStack, enmType, RTTHREADFLAGS_WAITABLE, pszName);
             if (RT_SUCCESS(rc))
             {
-                /*
-                 * Insert it into the thread list.
-                 */
-                pThread->Internal.s.pNext = NULL;
-                if (pVM->pdm.s.pThreadsTail)
-                    pVM->pdm.s.pThreadsTail->Internal.s.pNext = pThread;
-                else
-                    pVM->pdm.s.pThreads = pThread;
-                pVM->pdm.s.pThreadsTail = pThread;
+                rc = RTThreadUserWait(Thread, 60*1000);
+                if (    RT_SUCCESS(rc)
+                    &&  pThread->enmState != PDMTHREADSTATE_SUSPENDED)
+                    rc = VERR_INTERNAL_ERROR;
+                if (RT_SUCCESS(rc))
+                {
+                    /*
+                     * Insert it into the thread list.
+                     */
+                    pThread->Internal.s.pNext = NULL;
+                    if (pVM->pdm.s.pThreadsTail)
+                        pVM->pdm.s.pThreadsTail->Internal.s.pNext = pThread;
+                    else
+                        pVM->pdm.s.pThreads = pThread;
+                    pVM->pdm.s.pThreadsTail = pThread;
 
-                rc = RTThreadUserReset(Thread);
-                AssertRC(rc);
-                return rc;
+                    rc = RTThreadUserReset(Thread);
+                    AssertRC(rc);
+                    return rc;
+                }
+
+                /* bailout */
+                RTThreadWait(Thread, 60*1000, NULL);
             }
-
-            /* bailout */
-            RTThreadWait(Thread, 60*1000, NULL);
+            RTSemEventMultiDestroy(pThread->Internal.s.SleepEvent);
+            pThread->Internal.s.SleepEvent = NIL_RTSEMEVENTMULTI;
         }
         RTSemEventMultiDestroy(pThread->Internal.s.BlockEvent);
         pThread->Internal.s.BlockEvent = NIL_RTSEMEVENTMULTI;
@@ -200,11 +212,12 @@ int pdmR3ThreadCreateDevice(PVM pVM, PPDMDEVINS pDevIns, PPPDMTHREAD ppThread, v
     int rc = pdmR3ThreadNew(pVM, ppThread);
     if (RT_SUCCESS(rc))
     {
-        (*ppThread)->pvUser = pvUser;
-        (*ppThread)->Internal.s.enmType = PDMTHREADTYPE_DEVICE;
-        (*ppThread)->u.Dev.pDevIns = pDevIns;
-        (*ppThread)->u.Dev.pfnThread = pfnThread;
-        (*ppThread)->u.Dev.pfnWakeUp = pfnWakeUp;
+        PPDMTHREAD pThread = *ppThread;
+        pThread->pvUser = pvUser;
+        pThread->Internal.s.enmType = PDMTHREADTYPE_DEVICE;
+        pThread->u.Dev.pDevIns = pDevIns;
+        pThread->u.Dev.pfnThread = pfnThread;
+        pThread->u.Dev.pfnWakeUp = pfnWakeUp;
         rc = pdmR3ThreadInit(pVM, ppThread, cbStack, enmType, pszName);
     }
     return rc;
@@ -232,11 +245,12 @@ int pdmR3ThreadCreateUsb(PVM pVM, PPDMUSBINS pUsbIns, PPPDMTHREAD ppThread, void
     int rc = pdmR3ThreadNew(pVM, ppThread);
     if (RT_SUCCESS(rc))
     {
-        (*ppThread)->pvUser = pvUser;
-        (*ppThread)->Internal.s.enmType = PDMTHREADTYPE_USB;
-        (*ppThread)->u.Usb.pUsbIns = pUsbIns;
-        (*ppThread)->u.Usb.pfnThread = pfnThread;
-        (*ppThread)->u.Usb.pfnWakeUp = pfnWakeUp;
+        PPDMTHREAD pThread = *ppThread;
+        pThread->pvUser = pvUser;
+        pThread->Internal.s.enmType = PDMTHREADTYPE_USB;
+        pThread->u.Usb.pUsbIns = pUsbIns;
+        pThread->u.Usb.pfnThread = pfnThread;
+        pThread->u.Usb.pfnWakeUp = pfnWakeUp;
         rc = pdmR3ThreadInit(pVM, ppThread, cbStack, enmType, pszName);
     }
     return rc;
@@ -264,11 +278,12 @@ int pdmR3ThreadCreateDriver(PVM pVM, PPDMDRVINS pDrvIns, PPPDMTHREAD ppThread, v
     int rc = pdmR3ThreadNew(pVM, ppThread);
     if (RT_SUCCESS(rc))
     {
-        (*ppThread)->pvUser = pvUser;
-        (*ppThread)->Internal.s.enmType = PDMTHREADTYPE_DRIVER;
-        (*ppThread)->u.Drv.pDrvIns = pDrvIns;
-        (*ppThread)->u.Drv.pfnThread = pfnThread;
-        (*ppThread)->u.Drv.pfnWakeUp = pfnWakeUp;
+        PPDMTHREAD pThread = *ppThread;
+        pThread->pvUser = pvUser;
+        pThread->Internal.s.enmType = PDMTHREADTYPE_DRIVER;
+        pThread->u.Drv.pDrvIns = pDrvIns;
+        pThread->u.Drv.pfnThread = pfnThread;
+        pThread->u.Drv.pfnWakeUp = pfnWakeUp;
         rc = pdmR3ThreadInit(pVM, ppThread, cbStack, enmType, pszName);
     }
     return rc;
@@ -289,16 +304,17 @@ int pdmR3ThreadCreateDriver(PVM pVM, PPDMDRVINS pDrvIns, PPPDMTHREAD ppThread, v
  * @param   enmType     See RTThreadCreate.
  * @param   pszName     See RTThreadCreate.
  */
-PDMR3DECL(int) PDMR3ThreadCreate(PVM pVM, PPPDMTHREAD ppThread, void *pvUser, PFNPDMTHREADINT pfnThread,
+VMMR3DECL(int) PDMR3ThreadCreate(PVM pVM, PPPDMTHREAD ppThread, void *pvUser, PFNPDMTHREADINT pfnThread,
                                  PFNPDMTHREADWAKEUPINT pfnWakeUp, size_t cbStack, RTTHREADTYPE enmType, const char *pszName)
 {
     int rc = pdmR3ThreadNew(pVM, ppThread);
     if (RT_SUCCESS(rc))
     {
-        (*ppThread)->pvUser = pvUser;
-        (*ppThread)->Internal.s.enmType = PDMTHREADTYPE_INTERNAL;
-        (*ppThread)->u.Int.pfnThread = pfnThread;
-        (*ppThread)->u.Int.pfnWakeUp = pfnWakeUp;
+        PPDMTHREAD pThread = *ppThread;
+        pThread->pvUser = pvUser;
+        pThread->Internal.s.enmType = PDMTHREADTYPE_INTERNAL;
+        pThread->u.Int.pfnThread = pfnThread;
+        pThread->u.Int.pfnWakeUp = pfnWakeUp;
         rc = pdmR3ThreadInit(pVM, ppThread, cbStack, enmType, pszName);
     }
     return rc;
@@ -319,16 +335,17 @@ PDMR3DECL(int) PDMR3ThreadCreate(PVM pVM, PPPDMTHREAD ppThread, void *pvUser, PF
  * @param   enmType     See RTThreadCreate.
  * @param   pszName     See RTThreadCreate.
  */
-PDMR3DECL(int) PDMR3ThreadCreateExternal(PVM pVM, PPPDMTHREAD ppThread, void *pvUser, PFNPDMTHREADEXT pfnThread,
+VMMR3DECL(int) PDMR3ThreadCreateExternal(PVM pVM, PPPDMTHREAD ppThread, void *pvUser, PFNPDMTHREADEXT pfnThread,
                                          PFNPDMTHREADWAKEUPEXT pfnWakeUp, size_t cbStack, RTTHREADTYPE enmType, const char *pszName)
 {
     int rc = pdmR3ThreadNew(pVM, ppThread);
     if (RT_SUCCESS(rc))
     {
-        (*ppThread)->pvUser = pvUser;
-        (*ppThread)->Internal.s.enmType = PDMTHREADTYPE_EXTERNAL;
-        (*ppThread)->u.Ext.pfnThread = pfnThread;
-        (*ppThread)->u.Ext.pfnWakeUp = pfnWakeUp;
+        PPDMTHREAD pThread = *ppThread;
+        pThread->pvUser = pvUser;
+        pThread->Internal.s.enmType = PDMTHREADTYPE_EXTERNAL;
+        pThread->u.Ext.pfnThread = pfnThread;
+        pThread->u.Ext.pfnWakeUp = pfnWakeUp;
         rc = pdmR3ThreadInit(pVM, ppThread, cbStack, enmType, pszName);
     }
     return rc;
@@ -347,7 +364,7 @@ PDMR3DECL(int) PDMR3ThreadCreateExternal(PVM pVM, PPPDMTHREAD ppThread, void *pv
  * @param   pRcThread       Where to store the thread exit code. Optional.
  * @thread  The emulation thread (EMT).
  */
-PDMR3DECL(int) PDMR3ThreadDestroy(PPDMTHREAD pThread, int *pRcThread)
+VMMR3DECL(int) PDMR3ThreadDestroy(PPDMTHREAD pThread, int *pRcThread)
 {
     /*
      * Assert sanity.
@@ -430,7 +447,13 @@ PDMR3DECL(int) PDMR3ThreadDestroy(PPDMTHREAD pThread, int *pRcThread)
         }
         pThread->Internal.s.pNext = NULL;
 
-        /* free it */
+        /* free the resources */
+        RTSemEventMultiDestroy(pThread->Internal.s.BlockEvent);
+        pThread->Internal.s.BlockEvent = NIL_RTSEMEVENTMULTI;
+
+        RTSemEventMultiDestroy(pThread->Internal.s.SleepEvent);
+        pThread->Internal.s.SleepEvent = NIL_RTSEMEVENTMULTI;
+
         MMR3HeapFree(pThread);
     }
     else if (RT_SUCCESS(rc))
@@ -602,7 +625,7 @@ static void pdmR3ThreadBailMeOut(PPDMTHREAD pThread)
  *          On failure, terminate the thread.
  * @param   pThread     The PDM thread.
  */
-PDMR3DECL(int) PDMR3ThreadIAmSuspending(PPDMTHREAD pThread)
+VMMR3DECL(int) PDMR3ThreadIAmSuspending(PPDMTHREAD pThread)
 {
     /*
      * Assert sanity.
@@ -643,14 +666,14 @@ PDMR3DECL(int) PDMR3ThreadIAmSuspending(PPDMTHREAD pThread)
  * Called by the PDM thread in response to a resuming state.
  *
  * The purpose of this API is to tell the PDMR3ThreadResume caller that
- * the the PDM thread has successfully resumed. It will also do the
+ * the PDM thread has successfully resumed. It will also do the
  * state transition from the resuming to the running state.
  *
  * @returns VBox status code.
  *          On failure, terminate the thread.
  * @param   pThread     The PDM thread.
  */
-PDMR3DECL(int) PDMR3ThreadIAmRunning(PPDMTHREAD pThread)
+VMMR3DECL(int) PDMR3ThreadIAmRunning(PPDMTHREAD pThread)
 {
     /*
      * Assert sanity.
@@ -675,6 +698,37 @@ PDMR3DECL(int) PDMR3ThreadIAmRunning(PPDMTHREAD pThread)
 }
 
 
+/**
+ * Called by the PDM thread instead of RTThreadSleep.
+ *
+ * The difference is that the sleep will be interrupted on state change. The
+ * thread must be in the running state, otherwise it will return immediately.
+ *
+ * @returns VBox status code.
+ * @retval  VINF_SUCCESS on success or state change.
+ * @retval  VERR_INTERRUPTED on signal or APC.
+ *
+ * @param   pThread     The PDM thread.
+ * @param   cMillies    The number of milliseconds to sleep.
+ */
+VMMR3DECL(int) PDMR3ThreadSleep(PPDMTHREAD pThread, unsigned cMillies)
+{
+    /*
+     * Assert sanity.
+     */
+    AssertReturn(pThread->enmState > PDMTHREADSTATE_INVALID && pThread->enmState < PDMTHREADSTATE_TERMINATED, VERR_INTERNAL_ERROR);
+    AssertReturn(pThread->Thread == RTThreadSelf(), VERR_INTERNAL_ERROR);
+
+    /*
+     * Reset the event semaphore, check the state and sleep.
+     */
+    RTSemEventMultiReset(pThread->Internal.s.SleepEvent);
+    if (pThread->enmState != PDMTHREADSTATE_RUNNING)
+        return VINF_SUCCESS;
+    return RTSemEventMultiWaitNoResume(pThread->Internal.s.SleepEvent, cMillies);
+}
+
+
 /**
  * The PDM thread function.
  *
@@ -753,6 +807,9 @@ static DECLCALLBACK(int) pdmR3ThreadMain(RTTHREAD Thread, void *pvUser)
             break;
     }
 
+    if (RT_FAILURE(rc))
+        LogRel(("PDMThread: Thread '%s' (%RTthrd) quit unexpectedly with rc=%Rrc.\n", RTThreadGetName(Thread), Thread, rc));
+
     /*
      * Advance the state to terminating and then on to terminated.
      */
@@ -829,7 +886,7 @@ static void pdmR3ThreadBailOut(PPDMTHREAD pThread)
  * @returns VBox status code.
  * @param   pThread     The PDM thread.
  */
-PDMR3DECL(int) PDMR3ThreadSuspend(PPDMTHREAD pThread)
+VMMR3DECL(int) PDMR3ThreadSuspend(PPDMTHREAD pThread)
 {
     /*
      * Assert sanity.
@@ -923,7 +980,7 @@ int pdmR3ThreadSuspendAll(PVM pVM)
  * @returns VBox status code.
  * @param   pThread     The PDM thread.
  */
-PDMR3DECL(int) PDMR3ThreadResume(PPDMTHREAD pThread)
+VMMR3DECL(int) PDMR3ThreadResume(PPDMTHREAD pThread)
 {
     /*
      * Assert sanity.
