@@ -1,18 +1,22 @@
-/* $Id: TMInternal.h 23517 2007-08-07 17:07:59Z umoeller $ */
+/* $Id: TMInternal.h 13742 2008-11-03 12:19:40Z vboxsync $ */
 /** @file
  * TM - Internal header file.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 #ifndef ___TMInternal_h
@@ -20,6 +24,7 @@
 
 #include <VBox/cdefs.h>
 #include <VBox/types.h>
+#include <iprt/time.h>
 #include <iprt/timer.h>
 #include <VBox/stam.h>
 
@@ -125,7 +130,7 @@ typedef struct TMTIMER
             /** Callback. */
             R3PTRTYPE(PFNTMTIMERDEV)    pfnTimer;
             /** Device instance. */
-            R3PTRTYPE(PPDMDEVINS)       pDevIns;
+            PPDMDEVINSR3                pDevIns;
         } Dev;
 
         /** TMTIMERTYPE_DRV. */
@@ -143,7 +148,7 @@ typedef struct TMTIMER
             /** Callback. */
             R3PTRTYPE(PFNTMTIMERINT)    pfnTimer;
             /** User argument. */
-            R3PTRTYPE(void *)           pvUser;
+            RTR3PTR                     pvUser;
         } Internal;
 
         /** TMTIMERTYPE_EXTERNAL. */
@@ -152,7 +157,7 @@ typedef struct TMTIMER
             /** Callback. */
             R3PTRTYPE(PFNTMTIMEREXT)    pfnTimer;
             /** User data. */
-            R3PTRTYPE(void *)           pvUser;
+            RTR3PTR                     pvUser;
         } External;
     } u;
 
@@ -171,15 +176,15 @@ typedef struct TMTIMER
     /** Pointer to the previous timer in the list of all created timers. (TM::pTimers) */
     PTMTIMERR3              pBigPrev;
     /** Pointer to the timer description. */
-    HCPTRTYPE(const char *) pszDesc;
+    R3PTRTYPE(const char *) pszDesc;
     /** Pointer to the VM the timer belongs to - R3 Ptr. */
     PVMR3                   pVMR3;
     /** Pointer to the VM the timer belongs to - R0 Ptr. */
     PVMR0                   pVMR0;
-    /** Pointer to the VM the timer belongs to - GC Ptr. */
-    PVMGC                   pVMGC;
-#if HC_ARCH_BITS == 64 && GC_ARCH_BITS == 32
-    RTGCPTR                 padding0; /**< pad structure to multiple of 8 bytes. */
+    /** Pointer to the VM the timer belongs to - RC Ptr. */
+    PVMRC                   pVMRC;
+#if HC_ARCH_BITS == 64
+    RTRCPTR                 padding0; /**< pad structure to multiple of 8 bytes. */
 #endif
 } TMTIMER;
 
@@ -281,9 +286,6 @@ typedef struct TM
      * See TM2VM(). */
     RTUINT                      offVM;
 
-    /** Flag indicating that the host TSC is suitable for use in AMD-V and VT-x mode.
-     * Config variable: MaybeUseOffsettedHostTSC (boolean) */
-    bool                        fMaybeUseOffsettedHostTSC;
     /** CPU timestamp ticking enabled indicator (bool). (RDTSC) */
     bool                        fTSCTicking;
     /** Set if we fully virtualize the TSC, i.e. intercept all rdtsc instructions.
@@ -294,6 +296,16 @@ typedef struct TM
      * ticking. fTSCVirtualized = false implies fTSCUseRealTSC = true.
      * Config variable: TSCUseRealTSC (bool) */
     bool                        fTSCUseRealTSC;
+    /** Flag indicating that the host TSC is suitable for use in AMD-V and VT-x mode.
+     * Config variable: MaybeUseOffsettedHostTSC (boolean) */
+    bool                        fMaybeUseOffsettedHostTSC;
+    /** Whether the TSC is tied to the execution of code.
+     * Config variable: TSCTiedToExecution (bool) */
+    bool                        fTSCTiedToExecution;
+    /** Modifier for fTSCTiedToExecution which pauses the TSC while halting if true.
+     * Config variable: TSCNotTiedToHalt (bool) */
+    bool                        fTSCNotTiedToHalt;
+    bool                        afAlignment0[6]; /**< alignment padding */
     /** The offset between the host TSC and the Guest TSC.
      * Only valid if fTicking is set and and fTSCUseRealTSC is clear. */
     uint64_t                    u64TSCOffset;
@@ -330,11 +342,20 @@ typedef struct TM
      * This handles TSC drift on SMP systems and expired interval.
      * This is a valid range u64NanoTS to u64NanoTS + 1000000000 (ie. 1sec). */
     uint64_t volatile           u64VirtualRawPrev;
-    /** The number of times we've had to resort to 1ns walking. */
-    uint32_t volatile           c1nsVirtualRawSteps;
-    /** Number of times u64VirtualRawPrev has been considered bad. */
-    uint32_t volatile           cVirtualRawBadRawPrev;
-
+    /** The ring-3 data structure for the RTTimeNanoTS workers used by tmVirtualGetRawNanoTS. */
+    RTTIMENANOTSDATAR3          VirtualGetRawDataR3;
+    /** The ring-0 data structure for the RTTimeNanoTS workers used by tmVirtualGetRawNanoTS. */
+    RTTIMENANOTSDATAR0          VirtualGetRawDataR0;
+    /** The ring-0 data structure for the RTTimeNanoTS workers used by tmVirtualGetRawNanoTS. */
+    RTTIMENANOTSDATARC          VirtualGetRawDataRC;
+    /** Pointer to the ring-3 tmVirtualGetRawNanoTS worker function. */
+    R3PTRTYPE(PFNTIMENANOTSINTERNAL) pfnVirtualGetRawR3;
+    /** Pointer to the ring-3 tmVirtualGetRawNanoTS worker function. */
+    R0PTRTYPE(PFNTIMENANOTSINTERNAL) pfnVirtualGetRawR0;
+    /** Pointer to the ring-3 tmVirtualGetRawNanoTS worker function. */
+    RCPTRTYPE(PFNTIMENANOTSINTERNAL) pfnVirtualGetRawRC;
+    /** Alignment. */
+    RTRCPTR                     AlignmentRCPtr;
     /** The guest virtual timer synchronous time when fVirtualSyncTicking is cleared. */
     uint64_t volatile           u64VirtualSync;
     /** The offset of the timer synchronous virtual clock (TMCLOCK_VIRTUAL_SYNC) relative
@@ -377,11 +398,11 @@ typedef struct TM
     R3PTRTYPE(PTMTIMERQUEUE)    paTimerQueuesR3;
     /** Timer queues for the different clock types - R0 Ptr */
     R0PTRTYPE(PTMTIMERQUEUE)    paTimerQueuesR0;
-    /** Timer queues for the different clock types - GC Ptr */
-    GCPTRTYPE(PTMTIMERQUEUE)    paTimerQueuesGC;
+    /** Timer queues for the different clock types - RC Ptr */
+    RCPTRTYPE(PTMTIMERQUEUE)    paTimerQueuesRC;
 
-    /** Pointer to our GC mapping of the GIP. */
-    GCPTRTYPE(void *)           pvGIPGC;
+    /** Pointer to our RC mapping of the GIP. */
+    RCPTRTYPE(void *)           pvGIPRC;
     /** Pointer to our R3 mapping of the GIP. */
     R3PTRTYPE(void *)           pvGIPR3;
 
@@ -399,7 +420,7 @@ typedef struct TM
      * This timer will do freqent check on pending queue schedulations and
      * raise VM_FF_TIMER to pull EMTs attention to them.
      */
-    HCPTRTYPE(PRTTIMER)         pTimer;
+    R3PTRTYPE(PRTTIMER)         pTimer;
     /** Interval in milliseconds of the pTimer timer. */
     uint32_t                    u32TimerMillies;
 
@@ -414,13 +435,11 @@ typedef struct TM
     /** @} */
     /** tmSchedule
      * @{ */
-    STAMPROFILE                 StatScheduleOneGC;
-    STAMPROFILE                 StatScheduleOneR0;
+    STAMPROFILE                 StatScheduleOneRZ;
     STAMPROFILE                 StatScheduleOneR3;
     STAMCOUNTER                 StatScheduleSetFF;
     STAMCOUNTER                 StatPostponedR3;
-    STAMCOUNTER                 StatPostponedR0;
-    STAMCOUNTER                 StatPostponedGC;
+    STAMCOUNTER                 StatPostponedRZ;
     /** @} */
     /** Read the time
      * @{ */
@@ -440,14 +459,12 @@ typedef struct TM
     /** @} */
     /** TMTimerSet
      * @{ */
-    STAMPROFILE                 StatTimerSetGC;
-    STAMPROFILE                 StatTimerSetR0;
+    STAMPROFILE                 StatTimerSetRZ;
     STAMPROFILE                 StatTimerSetR3;
     /** @} */
     /** TMTimerStop
      * @{ */
-    STAMPROFILE                 StatTimerStopGC;
-    STAMPROFILE                 StatTimerStopR0;
+    STAMPROFILE                 StatTimerStopRZ;
     STAMPROFILE                 StatTimerStopR3;
     /** @} */
     /** VirtualSync - Running and Catching Up
@@ -466,17 +483,47 @@ typedef struct TM
     /** The timer callback. */
     STAMCOUNTER                 StatTimerCallbackSetFF;
 
+    /** @name Reasons for refusing TSC offsetting in TMCpuTickCanUseRealTSC.
+     * @{ */
+    STAMCOUNTER                 StatTSCNotFixed;
+    STAMCOUNTER                 StatTSCNotTicking;
+    STAMCOUNTER                 StatTSCCatchupLE010;
+    STAMCOUNTER                 StatTSCCatchupLE025;
+    STAMCOUNTER                 StatTSCCatchupLE100;
+    STAMCOUNTER                 StatTSCCatchupOther;
+    STAMCOUNTER                 StatTSCWarp;
+    STAMCOUNTER                 StatTSCSyncNotTicking;
+    /** @} */
 } TM;
 /** Pointer to TM VM instance data. */
 typedef TM *PTM;
 
+/**
+ * TM VMCPU Instance data.
+ * Changes to this must checked against the padding of the tm union in VM!
+ */
+typedef struct TMCPU
+{
+    /** Offset to the VMCPU structure.
+     * See TMCPU2VM(). */
+    RTUINT                      offVMCPU;
+} TMCPU;
+/** Pointer to TM VMCPU instance data. */
+typedef TMCPU *PTMCPU;
 
-const char *tmTimerState(TMTIMERSTATE enmState);
-void tmTimerQueueSchedule(PVM pVM, PTMTIMERQUEUE pQueue);
+const char             *tmTimerState(TMTIMERSTATE enmState);
+void                    tmTimerQueueSchedule(PVM pVM, PTMTIMERQUEUE pQueue);
 #ifdef VBOX_STRICT
-void tmTimerQueuesSanityChecks(PVM pVM, const char *pszWhere);
+void                    tmTimerQueuesSanityChecks(PVM pVM, const char *pszWhere);
 #endif
 
+int                     tmCpuTickPause(PVM pVM);
+int                     tmCpuTickResume(PVM pVM);
+
+DECLEXPORT(void)        tmVirtualNanoTSBad(PRTTIMENANOTSDATA pData, uint64_t u64NanoTS, uint64_t u64DeltaPrev, uint64_t u64PrevNanoTS);
+DECLEXPORT(uint64_t)    tmVirtualNanoTSRediscover(PRTTIMENANOTSDATA pData);
+
+
 /** @} */
 
 __END_DECLS
