@@ -1,18 +1,22 @@
-/* $Id: TRPMGCHandlers.cpp 23517 2007-08-07 17:07:59Z umoeller $ */
+/* $Id: TRPMGCHandlers.cpp 18770 2009-04-06 15:00:15Z vboxsync $ */
 /** @file
  * TRPM - Guest Context Trap Handlers, CPP part
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 
@@ -43,11 +47,21 @@
 #include <iprt/asm.h>
 #include <iprt/assert.h>
 
+/*******************************************************************************
+*   Defined Constants And Macros                                               *
+*******************************************************************************/
 /* still here. MODR/M byte parsing */
 #define X86_OPCODE_MODRM_MOD_MASK       0xc0
 #define X86_OPCODE_MODRM_REG_MASK       0x38
 #define X86_OPCODE_MODRM_RM_MASK        0x07
 
+/** @todo fix/remove/permanent-enable this when DIS/PATM handles invalid lock sequences. */
+#define DTRACE_EXPERIMENT
+
+
+/*******************************************************************************
+*   Structures and Typedefs                                                    *
+*******************************************************************************/
 /** Pointer to a readonly hypervisor trap record. */
 typedef const struct TRPMGCHYPER *PCTRPMGCHYPER;
 
@@ -74,7 +88,7 @@ typedef struct TRPMGCHYPER
      * @param   pRegFrame       The register frame.
      * @param   uUser           The user argument.
      */
-    DECLCALLBACKMEMBER(int, pfnHandler)(PVM pVM, PCPUMCTXCORE pRegFrame, uintptr_t uUser);
+    DECLRCCALLBACKMEMBER(int, pfnHandler, (PVM pVM, PCPUMCTXCORE pRegFrame, uintptr_t uUser));
     /** Whatever the handler desires to put here. */
     uintptr_t uUser;
 } TRPMGCHYPER;
@@ -111,7 +125,8 @@ __END_DECLS
  * Will reset the trap if it's not a guest trap or the trap
  * is already handled. Will process resume guest FFs.
  *
- * @returns rc.
+ * @returns rc, can be adjusted if its VINF_SUCCESS or something really bad
+ *          happened.
  * @param   pVM         VM handle.
  * @param   rc          The VBox status code to return.
  * @param   pRegFrame   Pointer to the register frame for the trap.
@@ -128,7 +143,7 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
 
 #ifdef VBOX_HIGH_RES_TIMERS_HACK
     /*
-     * Occationally we should poll timers.
+     * We should poll the timers occationally.
      * We must *NOT* do this too frequently as it adds a significant overhead
      * and it'll kill us if the trap load is high. (See #1354.)
      * (The heuristic is not very intelligent, we should really check trap
@@ -140,7 +155,7 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
         if (!(++s_iTimerPoll & 0xf))
         {
             uint64_t cTicks = TMTimerPoll(pVM); NOREF(cTicks);
-            Log2(("TMTimerPoll at %VGv returned %RX64 (VM_FF_TIMER=%d)\n", pRegFrame->eip, cTicks, VM_FF_ISPENDING(pVM, VM_FF_TIMER)));
+            Log2(("TMTimerPoll at %08RX32 returned %RX64 (VM_FF_TIMER=%d)\n", pRegFrame->eip, cTicks, VM_FF_ISPENDING(pVM, VM_FF_TIMER)));
         }
     }
     else
@@ -150,7 +165,7 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
     /* Clear pending inhibit interrupt state if required. (necessary for dispatching interrupts later on) */
     if (VM_FF_ISSET(pVM, VM_FF_INHIBIT_INTERRUPTS))
     {
-        Log2(("VM_FF_INHIBIT_INTERRUPTS at %VGv successor %VGv\n", pRegFrame->eip, EMGetInhibitInterruptsPC(pVM)));
+        Log2(("VM_FF_INHIBIT_INTERRUPTS at %08RX32 successor %RGv\n", pRegFrame->eip, EMGetInhibitInterruptsPC(pVM)));
         if (pRegFrame->eip != EMGetInhibitInterruptsPC(pVM))
         {
             /** @note we intentionally don't clear VM_FF_INHIBIT_INTERRUPTS here if the eip is the same as the inhibited instr address.
@@ -167,10 +182,14 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
      * Or pending (A)PIC interrupt? Windows XP will crash if we delay APIC interrupts.
      */
     if (    rc == VINF_SUCCESS
-        &&  VM_FF_ISPENDING(pVM, VM_FF_TO_R3 | VM_FF_TIMER | VM_FF_INTERRUPT_APIC | VM_FF_INTERRUPT_PIC | VM_FF_PGM_SYNC_CR3 | VM_FF_PGM_SYNC_CR3_NON_GLOBAL | VM_FF_REQUEST))
+        &&  VM_FF_ISPENDING(pVM, VM_FF_TO_R3 | VM_FF_TIMER | VM_FF_INTERRUPT_APIC | VM_FF_INTERRUPT_PIC | VM_FF_REQUEST
+                               | VM_FF_PGM_SYNC_CR3 | VM_FF_PGM_SYNC_CR3_NON_GLOBAL | VM_FF_PGM_NO_MEMORY))
     {
+        /* The out of memory condition naturally outrang the others. */
+        if (RT_UNLIKELY(VM_FF_ISPENDING(pVM, VM_FF_PGM_NO_MEMORY)))
+            rc = VINF_EM_NO_MEMORY;
         /* Pending Ring-3 action. */
-        if (VM_FF_ISPENDING(pVM, VM_FF_TO_R3))
+        else if (VM_FF_ISPENDING(pVM, VM_FF_TO_R3))
         {
             VM_FF_CLEAR(pVM, VM_FF_TO_R3);
             rc = VINF_EM_RAW_TO_R3;
@@ -186,9 +205,9 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
         {
             uint8_t u8Interrupt;
             rc = PDMGetInterrupt(pVM, &u8Interrupt);
-            Log(("trpmGCExitTrap: u8Interrupt=%d (%#x) rc=%Vrc\n", u8Interrupt, u8Interrupt, rc));
-            AssertFatalMsgRC(rc, ("PDMGetInterrupt failed with %Vrc\n", rc));
-            rc = TRPMForwardTrap(pVM, pRegFrame, (uint32_t)u8Interrupt, 0, TRPM_TRAP_NO_ERRORCODE, TRPM_HARDWARE_INT);
+            Log(("trpmGCExitTrap: u8Interrupt=%d (%#x) rc=%Rrc\n", u8Interrupt, u8Interrupt, rc));
+            AssertFatalMsgRC(rc, ("PDMGetInterrupt failed with %Rrc\n", rc));
+            rc = TRPMForwardTrap(pVM, pRegFrame, (uint32_t)u8Interrupt, 0, TRPM_TRAP_NO_ERRORCODE, TRPM_HARDWARE_INT, uOldActiveVector);
             /* can't return if successful */
             Assert(rc != VINF_SUCCESS);
 
@@ -204,7 +223,6 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
         }
         /*
          * Try sync CR3?
-         * This ASSUMES that the MOV CRx, x emulation doesn't return with VINF_PGM_SYNC_CR3. (a bit hackish)
          */
         else if (VM_FF_ISPENDING(pVM, VM_FF_PGM_SYNC_CR3 | VM_FF_PGM_SYNC_CR3_NON_GLOBAL))
 #if 1
@@ -220,7 +238,7 @@ static int trpmGCExitTrap(PVM pVM, int rc, PCPUMCTXCORE pRegFrame)
     AssertMsg(     rc != VINF_SUCCESS
               ||   (   pRegFrame->eflags.Bits.u1IF
                     && ( pRegFrame->eflags.Bits.u2IOPL < (unsigned)(pRegFrame->ss & X86_SEL_RPL) || pRegFrame->eflags.Bits.u1VM))
-              , ("rc = %VGv\neflags=%RX32 ss=%RTsel IOPL=%d\n", rc, pRegFrame->eflags.u32, pRegFrame->ss, pRegFrame->eflags.Bits.u2IOPL));
+              , ("rc=%Rrc\neflags=%RX32 ss=%RTsel IOPL=%d\n", rc, pRegFrame->eflags.u32, pRegFrame->ss, pRegFrame->eflags.Bits.u2IOPL));
     return rc;
 }
 
@@ -240,7 +258,7 @@ DECLASM(int) TRPMGCTrap01Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
     RTGCUINTREG uDr6 = ASMGetAndClearDR6();
     PVM pVM = TRPM2VM(pTrpm);
-    LogFlow(("TRPMGCTrap01Handler: cs:eip=%04x:%08x uDr6=%RTreg\n", pRegFrame->cs, pRegFrame->eip, uDr6));
+    LogFlow(("TRPMGC01: cs:eip=%04x:%08x uDr6=%RTreg\n", pRegFrame->cs, pRegFrame->eip, uDr6));
 
     /*
      * We currently don't make sure of the X86_DR7_GD bit, but
@@ -262,11 +280,12 @@ DECLASM(int) TRPMGCTrap01Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
     if (rc == VINF_EM_RAW_GUEST_TRAP)
         CPUMSetGuestDR6(pVM, uDr6);
 
-    return trpmGCExitTrap(pVM, rc, pRegFrame);
+    rc = trpmGCExitTrap(pVM, rc, pRegFrame);
+    Log6(("TRPMGC01: %Rrc (%04x:%08x %RTreg)\n", rc, pRegFrame->cs, pRegFrame->eip, uDr6));
+    return rc;
 }
 
 
-
 /**
  * NMI handler, for when we are using NMIs to debug things.
  *
@@ -300,7 +319,7 @@ DECLASM(int) TRPMGCTrap02Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
  */
 DECLASM(int) TRPMGCTrap03Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
-    LogFlow(("TRPMGCTrap03Handler: cs:eip=%04x:%08x\n", pRegFrame->cs, pRegFrame->eip));
+    LogFlow(("TRPMGC03: %04x:%08x\n", pRegFrame->cs, pRegFrame->eip));
     PVM pVM = TRPM2VM(pTrpm);
     int rc;
 
@@ -312,11 +331,18 @@ DECLASM(int) TRPMGCTrap03Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
     {
         rc = PATMHandleInt3PatchTrap(pVM, pRegFrame);
         if (rc == VINF_SUCCESS || rc == VINF_EM_RAW_EMULATE_INSTR || rc == VINF_PATM_PATCH_INT3 || rc == VINF_PATM_DUPLICATE_FUNCTION)
-            return trpmGCExitTrap(pVM, rc, pRegFrame);
+        {
+            rc = trpmGCExitTrap(pVM, rc, pRegFrame);
+            Log6(("TRPMGC03: %Rrc (%04x:%08x) (PATM)\n", rc, pRegFrame->cs, pRegFrame->eip));
+            return rc;
+        }
     }
     rc = DBGFGCTrap03Handler(pVM, pRegFrame);
+
     /* anything we should do with this? Schedule it in GC? */
-    return trpmGCExitTrap(pVM, rc, pRegFrame);
+    rc = trpmGCExitTrap(pVM, rc, pRegFrame);
+    Log6(("TRPMGC03: %Rrc (%04x:%08x)\n", rc, pRegFrame->cs, pRegFrame->eip));
+    return rc;
 }
 
 
@@ -333,11 +359,10 @@ DECLASM(int) TRPMGCTrap03Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
  */
 DECLASM(int) TRPMGCTrap06Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
+    LogFlow(("TRPMGC06: %04x:%08x efl=%x\n", pRegFrame->cs, pRegFrame->eip, pRegFrame->eflags.u32));
     PVM pVM = TRPM2VM(pTrpm);
     int rc;
 
-    LogFlow(("TRPMGCTrap06Handler %VGv eflags=%x\n", pRegFrame->eip, pRegFrame->eflags.u32));
-
     if (CPUMGetGuestCPL(pVM, pRegFrame) == 0)
     {
         /*
@@ -345,45 +370,81 @@ DECLASM(int) TRPMGCTrap06Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
          */
         RTGCPTR PC;
         rc = SELMValidateAndConvertCSAddr(pVM, pRegFrame->eflags, pRegFrame->ss, pRegFrame->cs, &pRegFrame->csHid, (RTGCPTR)pRegFrame->eip, &PC);
-        if (VBOX_FAILURE(rc))
+        if (RT_FAILURE(rc))
         {
-            Log(("TRPMGCTrap06Handler: Failed to convert %RTsel:%RX32 (cpl=%d) - rc=%Vrc !!\n", pRegFrame->cs, pRegFrame->eip, pRegFrame->ss & X86_SEL_RPL, rc));
-            return trpmGCExitTrap(pVM, VINF_EM_RAW_GUEST_TRAP, pRegFrame);
+            Log(("TRPMGCTrap06Handler: Failed to convert %RTsel:%RX32 (cpl=%d) - rc=%Rrc !!\n", pRegFrame->cs, pRegFrame->eip, pRegFrame->ss & X86_SEL_RPL, rc));
+            rc = trpmGCExitTrap(pVM, VINF_EM_RAW_GUEST_TRAP, pRegFrame);
+            Log6(("TRPMGC06: %Rrc (%04x:%08x) (SELM)\n", rc, pRegFrame->cs, pRegFrame->eip));
+            return rc;
         }
 
         DISCPUSTATE Cpu;
         uint32_t    cbOp;
         rc = EMInterpretDisasOneEx(pVM, (RTGCUINTPTR)PC, pRegFrame, &Cpu, &cbOp);
-        if (VBOX_FAILURE(rc))
-            return trpmGCExitTrap(pVM, VINF_EM_RAW_EMULATE_INSTR, pRegFrame);
+        if (RT_FAILURE(rc))
+        {
+            rc = trpmGCExitTrap(pVM, VINF_EM_RAW_EMULATE_INSTR, pRegFrame);
+            Log6(("TRPMGC06: %Rrc (%04x:%08x) (EM)\n", rc, pRegFrame->cs, pRegFrame->eip));
+            return rc;
+        }
 
-        if (    PATMIsPatchGCAddr(pVM, (RTGCPTR)pRegFrame->eip)
-            &&  Cpu.pCurInstr->opcode == OP_ILLUD2)
+        /*
+         * UD2 in a patch?
+         */
+        if (    Cpu.pCurInstr->opcode == OP_ILLUD2
+            &&  PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip))
         {
             rc = PATMGCHandleIllegalInstrTrap(pVM, pRegFrame);
-            if (rc == VINF_SUCCESS || rc == VINF_EM_RAW_EMULATE_INSTR || rc == VINF_PATM_DUPLICATE_FUNCTION || rc == VINF_PATM_PENDING_IRQ_AFTER_IRET || rc == VINF_EM_RESCHEDULE)
-                return trpmGCExitTrap(pVM, rc, pRegFrame);
+            /** @todo  These tests are completely unnecessary, should just follow the
+             *         flow and return at the end of the function. */
+            if (    rc == VINF_SUCCESS
+                ||  rc == VINF_EM_RAW_EMULATE_INSTR
+                ||  rc == VINF_PATM_DUPLICATE_FUNCTION
+                ||  rc == VINF_PATM_PENDING_IRQ_AFTER_IRET
+                ||  rc == VINF_EM_RESCHEDULE)
+            {
+                rc = trpmGCExitTrap(pVM, rc, pRegFrame);
+                Log6(("TRPMGC06: %Rrc (%04x:%08x) (PATM)\n", rc, pRegFrame->cs, pRegFrame->eip));
+                return rc;
+            }
         }
-        else
-        /** Note: monitor causes an #UD exception instead of #GP when not executed in ring 0. */
-        if (Cpu.pCurInstr->opcode == OP_MONITOR)
+        /*
+         * Speed up dtrace and don't entrust invalid lock sequences to the recompiler.
+         */
+        else if (Cpu.prefix & PREFIX_LOCK)
+        {
+            Log(("TRPMGCTrap06Handler: pc=%08x op=%d\n", pRegFrame->eip, Cpu.pCurInstr->opcode));
+#ifdef DTRACE_EXPERIMENT /** @todo fix/remove/permanent-enable this when DIS/PATM handles invalid lock sequences. */
+            Assert(!PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip));
+            rc = TRPMForwardTrap(pVM, pRegFrame, 0x6, 0, TRPM_TRAP_NO_ERRORCODE, TRPM_TRAP, 0x6);
+            Assert(rc == VINF_EM_RAW_GUEST_TRAP);
+#else
+            rc = VINF_EM_RAW_EMULATE_INSTR;
+#endif
+        }
+        /*
+         * Handle MONITOR - it causes an #UD exception instead of #GP when not executed in ring 0.
+         */
+        else if (Cpu.pCurInstr->opcode == OP_MONITOR)
         {
             uint32_t cbIgnored;
             rc = EMInterpretInstructionCPU(pVM, &Cpu, pRegFrame, PC, &cbIgnored);
-            if (RT_LIKELY(VBOX_SUCCESS(rc)))
+            if (RT_LIKELY(RT_SUCCESS(rc)))
                 pRegFrame->eip += Cpu.opsize;
         }
+        /* Never generate a raw trap here; it might be an instruction, that requires emulation. */
         else
-            /* Never generate a raw trap here; it might be an instruction, that requires emulation. */
             rc = VINF_EM_RAW_EMULATE_INSTR;
     }
     else
     {
-        rc = TRPMForwardTrap(pVM, pRegFrame, 0x6, 0, TRPM_TRAP_NO_ERRORCODE, TRPM_TRAP);
+        rc = TRPMForwardTrap(pVM, pRegFrame, 0x6, 0, TRPM_TRAP_NO_ERRORCODE, TRPM_TRAP, 0x6);
         Assert(rc == VINF_EM_RAW_GUEST_TRAP);
     }
 
-    return trpmGCExitTrap(pVM, rc, pRegFrame);
+    rc = trpmGCExitTrap(pVM, rc, pRegFrame);
+    Log6(("TRPMGC06: %Rrc (%04x:%08x)\n", rc, pRegFrame->cs, pRegFrame->eip));
+    return rc;
 }
 
 
@@ -402,10 +463,12 @@ DECLASM(int) TRPMGCTrap06Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
  */
 DECLASM(int) TRPMGCTrap07Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
+    LogFlow(("TRPMGC07: %04x:%08x\n", pRegFrame->cs, pRegFrame->eip));
     PVM pVM = TRPM2VM(pTrpm);
 
-    LogFlow(("TRPMTrap07HandlerGC: eip=%VGv\n", pRegFrame->eip));
-    return CPUMHandleLazyFPU(pVM);
+    int rc = CPUMHandleLazyFPU(pVM, VMMGetCpu(pVM));
+    Log6(("TRPMGC07: %Rrc (%04x:%08x)\n", rc, pRegFrame->cs, pRegFrame->eip));
+    return rc; /** @todo call trpmGCExitTrap! (after 2.2.0) */
 }
 
 
@@ -422,7 +485,7 @@ DECLASM(int) TRPMGCTrap07Handler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
  */
 DECLASM(int) TRPMGCTrap0bHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
-    LogFlow(("TRPMGCTrap0bHandler: eip=%VGv\n", pRegFrame->eip));
+    LogFlow(("TRPMGC0b: %04x:%08x\n", pRegFrame->cs, pRegFrame->eip));
     PVM pVM = TRPM2VM(pTrpm);
 
     /*
@@ -430,9 +493,11 @@ DECLASM(int) TRPMGCTrap0bHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
      * XXX note: this code may cause \#PF (trap e) or \#GP (trap d) while
      * accessing user code. need to handle it somehow in future!
      */
-    uint8_t *pu8Code;
-    if (SELMValidateAndConvertCSAddr(pVM, pRegFrame->eflags, pRegFrame->ss, pRegFrame->cs, &pRegFrame->csHid, (RTGCPTR)pRegFrame->eip, (PRTGCPTR)&pu8Code) == VINF_SUCCESS)
+    RTGCPTR GCPtr;
+    if (SELMValidateAndConvertCSAddr(pVM, pRegFrame->eflags, pRegFrame->ss, pRegFrame->cs, &pRegFrame->csHid, (RTGCPTR)pRegFrame->eip, &GCPtr) == VINF_SUCCESS)
     {
+        uint8_t *pu8Code = (uint8_t *)(uintptr_t)GCPtr;
+
         /*
          * First skip possible instruction prefixes, such as:
          *      OS, AS
@@ -492,6 +557,7 @@ DECLASM(int) TRPMGCTrap0bHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
              * starting from the instruction which caused the trap.
              */
             pTrpm->uActiveVector = ~0;
+            Log6(("TRPMGC0b: %Rrc (%04x:%08x) (CG)\n", VINF_EM_RAW_RING_SWITCH, pRegFrame->cs, pRegFrame->eip));
             return VINF_EM_RAW_RING_SWITCH;
         }
     }
@@ -499,6 +565,7 @@ DECLASM(int) TRPMGCTrap0bHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
     /*
      * Pass trap 0b as is to the recompiler in all other cases.
      */
+    Log6(("TRPMGC0b: %Rrc (%04x:%08x)\n", VINF_EM_RAW_GUEST_TRAP, pRegFrame->cs, pRegFrame->eip));
     return VINF_EM_RAW_GUEST_TRAP;
 }
 
@@ -534,7 +601,7 @@ static int trpmGCTrap0dHandlerRing0(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
         case OP_INT:
         {
             Assert(pCpu->param1.flags & USE_IMMEDIATE8);
-            Assert(!(PATMIsPatchGCAddr(pVM, PC)));
+            Assert(!(PATMIsPatchGCAddr(pVM, (RTRCPTR)PC)));
             if (pCpu->param1.parval == 3)
             {
                 /* Int 3 replacement patch? */
@@ -544,8 +611,8 @@ static int trpmGCTrap0dHandlerRing0(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
                     return trpmGCExitTrap(pVM, VINF_SUCCESS, pRegFrame);
                 }
             }
-            rc = TRPMForwardTrap(pVM, pRegFrame, (uint32_t)pCpu->param1.parval, pCpu->opsize, TRPM_TRAP_NO_ERRORCODE, TRPM_SOFTWARE_INT);
-            if (VBOX_SUCCESS(rc) && rc != VINF_EM_RAW_GUEST_TRAP)
+            rc = TRPMForwardTrap(pVM, pRegFrame, (uint32_t)pCpu->param1.parval, pCpu->opsize, TRPM_TRAP_NO_ERRORCODE, TRPM_SOFTWARE_INT, 0xd);
+            if (RT_SUCCESS(rc) && rc != VINF_EM_RAW_GUEST_TRAP)
                 return trpmGCExitTrap(pVM, VINF_SUCCESS, pRegFrame);
 
             pVM->trpm.s.uActiveVector = (pVM->trpm.s.uActiveErrorCode & X86_TRAP_ERR_SEL_MASK) >> X86_TRAP_ERR_SEL_SHIFT;
@@ -562,7 +629,7 @@ static int trpmGCTrap0dHandlerRing0(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
 
         case OP_HLT:
             /* If it's in patch code, defer to ring-3. */
-            if (PATMIsPatchGCAddr(pVM, PC))
+            if (PATMIsPatchGCAddr(pVM, (RTRCPTR)PC))
                 break;
 
             pRegFrame->eip += pCpu->opsize;
@@ -578,18 +645,22 @@ static int trpmGCTrap0dHandlerRing0(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
         case OP_MOV_CR:
         case OP_MOV_DR:
             /* We can safely emulate control/debug register move instructions in patched code. */
-            if (    !PATMIsPatchGCAddr(pVM, PC)
-                &&  !CSAMIsKnownDangerousInstr(pVM, PC))
+            if (    !PATMIsPatchGCAddr(pVM, (RTRCPTR)PC)
+                &&  !CSAMIsKnownDangerousInstr(pVM, (RTRCPTR)PC))
                 break;
         case OP_INVLPG:
         case OP_LLDT:
         case OP_STI:
-        case OP_RDTSC:
+        case OP_RDTSC:  /* just in case */
+        case OP_RDPMC:
         case OP_CLTS:
+        case OP_WBINVD: /* nop */
+        case OP_RDMSR:
+        case OP_WRMSR:
         {
             uint32_t cbIgnored;
             rc = EMInterpretInstructionCPU(pVM, pCpu, pRegFrame, PC, &cbIgnored);
-            if (VBOX_SUCCESS(rc))
+            if (RT_SUCCESS(rc))
                 pRegFrame->eip += pCpu->opsize;
             else if (rc == VERR_EM_INTERPRETER)
                 rc = VINF_EM_RAW_EXCEPTION_PRIVILEGED;
@@ -621,22 +692,6 @@ static int trpmGCTrap0dHandlerRing3(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
 
     switch (pCpu->pCurInstr->opcode)
     {
-        /*
-         * STI and CLI are I/O privileged, i.e. if IOPL
-         */
-        case OP_STI:
-        case OP_CLI:
-        {
-            uint32_t efl = CPUMRawGetEFlags(pVM, pRegFrame);
-            if (X86_EFL_GET_IOPL(efl) >= (unsigned)(pRegFrame->ss & X86_SEL_RPL))
-            {
-                LogFlow(("trpmGCTrap0dHandlerRing3: CLI/STI -> REM\n"));
-                return trpmGCExitTrap(pVM, VINF_EM_RESCHEDULE_REM, pRegFrame);
-            }
-            LogFlow(("trpmGCTrap0dHandlerRing3: CLI/STI -> #GP(0)\n"));
-            break;
-        }
-
         /*
          * INT3 and INT xx are ring-switching.
          * (The shadow IDT will have set the entries to DPL=0, that's why we're here.)
@@ -651,8 +706,8 @@ static int trpmGCTrap0dHandlerRing3(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
         case OP_INT:
         {
             Assert(pCpu->param1.flags & USE_IMMEDIATE8);
-            rc = TRPMForwardTrap(pVM, pRegFrame, (uint32_t)pCpu->param1.parval, pCpu->opsize, TRPM_TRAP_NO_ERRORCODE, TRPM_SOFTWARE_INT);
-            if (VBOX_SUCCESS(rc) && rc != VINF_EM_RAW_GUEST_TRAP)
+            rc = TRPMForwardTrap(pVM, pRegFrame, (uint32_t)pCpu->param1.parval, pCpu->opsize, TRPM_TRAP_NO_ERRORCODE, TRPM_SOFTWARE_INT, 0xd);
+            if (RT_SUCCESS(rc) && rc != VINF_EM_RAW_GUEST_TRAP)
                 return trpmGCExitTrap(pVM, VINF_SUCCESS, pRegFrame);
 
             pVM->trpm.s.uActiveVector = (pVM->trpm.s.uActiveErrorCode & X86_TRAP_ERR_SEL_MASK) >> X86_TRAP_ERR_SEL_SHIFT;
@@ -677,18 +732,35 @@ static int trpmGCTrap0dHandlerRing3(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
             return trpmGCExitTrap(pVM, VINF_EM_RAW_RING_SWITCH, pRegFrame);
 
         /*
-         * Handle virtualized TSC reads.
+         * Handle virtualized TSC & PMC reads, just in case.
          */
         case OP_RDTSC:
+        case OP_RDPMC:
         {
             uint32_t cbIgnored;
             rc = EMInterpretInstructionCPU(pVM, pCpu, pRegFrame, PC, &cbIgnored);
-            if (VBOX_SUCCESS(rc))
+            if (RT_SUCCESS(rc))
                 pRegFrame->eip += pCpu->opsize;
             else if (rc == VERR_EM_INTERPRETER)
                 rc = VINF_EM_RAW_EXCEPTION_PRIVILEGED;
             return trpmGCExitTrap(pVM, rc, pRegFrame);
         }
+
+        /*
+         * STI and CLI are I/O privileged, i.e. if IOPL
+         */
+        case OP_STI:
+        case OP_CLI:
+        {
+            uint32_t efl = CPUMRawGetEFlags(pVM, pRegFrame);
+            if (X86_EFL_GET_IOPL(efl) >= (unsigned)(pRegFrame->ss & X86_SEL_RPL))
+            {
+                LogFlow(("trpmGCTrap0dHandlerRing3: CLI/STI -> REM\n"));
+                return trpmGCExitTrap(pVM, VINF_EM_RESCHEDULE_REM, pRegFrame);
+            }
+            LogFlow(("trpmGCTrap0dHandlerRing3: CLI/STI -> #GP(0)\n"));
+            break;
+        }
     }
 
     /*
@@ -698,6 +770,30 @@ static int trpmGCTrap0dHandlerRing3(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
 }
 
 
+/**
+ * Emulates RDTSC for the \#GP handler.
+ *
+ * @returns VINF_SUCCESS or VINF_EM_RAW_EMULATE_INSTR.
+ *
+ * @param   pVM         Pointer to the shared VM structure.
+ * @param   pRegFrame   Pointer to the registre frame for the trap.
+ *                      This will be updated on successful return.
+ */
+DECLINLINE(int) trpmGCTrap0dHandlerRdTsc(PVM pVM, PCPUMCTXCORE pRegFrame)
+{
+    STAM_COUNTER_INC(&pVM->trpm.s.StatTrap0dRdTsc);
+
+    if (CPUMGetGuestCR4(pVM) & X86_CR4_TSD)
+        return trpmGCExitTrap(pVM, VINF_EM_RAW_EMULATE_INSTR, pRegFrame); /* will trap (optimize later). */
+
+    uint64_t uTicks = TMCpuTickGet(pVM);
+    pRegFrame->eax = uTicks;
+    pRegFrame->edx = uTicks >> 32;
+    pRegFrame->eip += 2;
+    return trpmGCExitTrap(pVM, VINF_SUCCESS, pRegFrame);
+}
+
+
 /**
  * \#GP (General Protection Fault) handler.
  *
@@ -711,45 +807,51 @@ static int trpmGCTrap0dHandlerRing3(PVM pVM, PCPUMCTXCORE pRegFrame, PDISCPUSTAT
  */
 static int trpmGCTrap0dHandler(PVM pVM, PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
-    LogFlow(("trpmGCTrap0dHandler: cs:eip=%RTsel:%VGv uErr=%RX32\n", pRegFrame->ss, pRegFrame->eip, pTrpm->uActiveErrorCode));
+    LogFlow(("trpmGCTrap0dHandler: cs:eip=%RTsel:%08RX32 uErr=%RGv\n", pRegFrame->ss, pRegFrame->eip, pTrpm->uActiveErrorCode));
 
-#if 0 /* not right for iret. Shouldn't really be needed as SELMValidateAndConvertCSAddr deals with invalid cs. */
     /*
-     * Filter out selector problems first as these may mean that the
-     * instruction isn't safe to read. If we're here because CS is NIL
-     * the flattening of cs:eip will deal with that.
-     */
-    if (    !(pTrpm->uActiveErrorCode & (X86_TRAP_ERR_IDT | X86_TRAP_ERR_EXTERNAL))
-        &&  (pTrpm->uActiveErrorCode & X86_TRAP_ERR_SEL_MASK))
-    {
-        /* It's a guest trap. */
-        return trpmGCExitTrap(pVM, VINF_EM_RAW_GUEST_TRAP, pRegFrame);
-    }
-#endif
-
-    STAM_PROFILE_ADV_START(&pVM->trpm.s.StatTrap0dDisasm, a);
-    /*
-     * Decode the instruction.
+     * Convert and validate CS.
      */
+    STAM_PROFILE_START(&pVM->trpm.s.StatTrap0dDisasm, a);
     RTGCPTR PC;
-    int rc = SELMValidateAndConvertCSAddr(pVM, pRegFrame->eflags, pRegFrame->ss, pRegFrame->cs, &pRegFrame->csHid, (RTGCPTR)pRegFrame->eip, &PC);
-    if (VBOX_FAILURE(rc))
+    uint32_t cBits;
+    int rc = SELMValidateAndConvertCSAddrGCTrap(pVM, pRegFrame->eflags, pRegFrame->ss, pRegFrame->cs,
+                                                (RTGCPTR)pRegFrame->eip, &PC, &cBits);
+    if (RT_FAILURE(rc))
     {
-        Log(("trpmGCTrap0dHandler: Failed to convert %RTsel:%RX32 (cpl=%d) - rc=%Vrc !!\n",
+        Log(("trpmGCTrap0dHandler: Failed to convert %RTsel:%RX32 (cpl=%d) - rc=%Rrc !!\n",
              pRegFrame->cs, pRegFrame->eip, pRegFrame->ss & X86_SEL_RPL, rc));
-        STAM_PROFILE_ADV_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
+        STAM_PROFILE_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
         return trpmGCExitTrap(pVM, VINF_EM_RAW_EMULATE_INSTR, pRegFrame);
     }
 
+    /*
+     * Optimize RDTSC traps.
+     * Some guests (like Solaris) are using RDTSC all over the place and
+     * will end up trapping a *lot* because of that.
+     */
+    if (   !pRegFrame->eflags.Bits.u1VM
+        && ((uint8_t *)PC)[0] == 0x0f
+        && ((uint8_t *)PC)[1] == 0x31)
+    {
+        STAM_PROFILE_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
+        return trpmGCTrap0dHandlerRdTsc(pVM, pRegFrame);
+    }
+
+    /*
+     * Disassemble the instruction.
+     */
     DISCPUSTATE Cpu;
     uint32_t    cbOp;
-    rc = EMInterpretDisasOneEx(pVM, (RTGCUINTPTR)PC, pRegFrame, &Cpu, &cbOp);
-    if (VBOX_FAILURE(rc))
+    rc = DISCoreOneEx((RTGCUINTPTR)PC, cBits == 32 ? CPUMODE_32BIT : cBits == 16 ? CPUMODE_16BIT : CPUMODE_64BIT,
+                      NULL, NULL, &Cpu, &cbOp);
+    if (RT_FAILURE(rc))
     {
-        STAM_PROFILE_ADV_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
+        AssertMsgFailed(("DISCoreOneEx failed to PC=%RGv rc=%Rrc\n", PC, rc));
+        STAM_PROFILE_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
         return trpmGCExitTrap(pVM, VINF_EM_RAW_EMULATE_INSTR, pRegFrame);
     }
-    STAM_PROFILE_ADV_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
+    STAM_PROFILE_STOP(&pVM->trpm.s.StatTrap0dDisasm, a);
 
     /*
      * Deal with I/O port access.
@@ -761,7 +863,6 @@ static int trpmGCTrap0dHandler(PVM pVM, PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
         return trpmGCExitTrap(pVM, rc, pRegFrame);
     }
 
-
     /*
      * Deal with Ring-0 (privileged instructions)
      */
@@ -777,20 +878,20 @@ static int trpmGCTrap0dHandler(PVM pVM, PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 
     /*
      * Deal with v86 code.
+     *
+     * We always set IOPL to zero which makes e.g. pushf fault in V86
+     * mode. The guest might use IOPL=3 and therefore not expect a #GP.
+     * Simply fall back to the recompiler to emulate this instruction if
+     * that's the case. To get the correct we must use CPUMRawGetEFlags.
      */
-
-    /* We always set IOPL to zero which makes e.g. pushf fault in V86 mode. The guest might use IOPL=3 and therefor not expect a #GP.
-     * Simply fall back to the recompiler to emulate this instruction.
-     */
-    /* Retrieve the eflags including the virtualized bits. */
-    /** @note hackish as the cpumctxcore structure doesn't contain the right value */
     X86EFLAGS eflags;
-    eflags.u32 = CPUMRawGetEFlags(pVM, pRegFrame);
+    eflags.u32 = CPUMRawGetEFlags(pVM, pRegFrame); /* Get the correct value. */
+    Log3(("TRPM #GP V86: cs:eip=%04x:%08x IOPL=%d efl=%08x\n", pRegFrame->cs, pRegFrame->eip, eflags.Bits.u2IOPL, eflags.u));
     if (eflags.Bits.u2IOPL != 3)
     {
         Assert(eflags.Bits.u2IOPL == 0);
 
-        int rc = TRPMForwardTrap(pVM, pRegFrame, 0xD, 0, TRPM_TRAP_HAS_ERRORCODE, TRPM_TRAP);
+        int rc = TRPMForwardTrap(pVM, pRegFrame, 0xD, 0, TRPM_TRAP_HAS_ERRORCODE, TRPM_TRAP, 0xd);
         Assert(rc == VINF_EM_RAW_GUEST_TRAP);
         return trpmGCExitTrap(pVM, rc, pRegFrame);
     }
@@ -811,7 +912,7 @@ static int trpmGCTrap0dHandler(PVM pVM, PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
  */
 DECLASM(int) TRPMGCTrap0dHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
-    LogFlow(("TRPMGCTrap0dHandler: eip=%RGv\n", pRegFrame->eip));
+    LogFlow(("TRPMGC0d: %04x:%08x err=%x\n", pRegFrame->cs, pRegFrame->eip, (uint32_t)pTrpm->uActiveErrorCode));
     PVM pVM = TRPM2VM(pTrpm);
 
     int rc = trpmGCTrap0dHandler(pVM, pTrpm, pRegFrame);
@@ -819,7 +920,7 @@ DECLASM(int) TRPMGCTrap0dHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
     {
         case VINF_EM_RAW_GUEST_TRAP:
         case VINF_EM_RAW_EXCEPTION_PRIVILEGED:
-            if (PATMIsPatchGCAddr(pVM, (RTGCPTR)pRegFrame->eip))
+            if (PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip))
                 rc = VINF_PATM_PATCH_TRAP_GP;
             break;
 
@@ -827,12 +928,14 @@ DECLASM(int) TRPMGCTrap0dHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
             Assert(TRPMHasTrap(pVM));
             /* no break; */
         case VINF_PGM_SYNC_CR3: /** @todo Check this with Sander. */
+        case VINF_EM_RAW_EMULATE_INSTR:
         case VINF_IOM_HC_IOPORT_READ:
         case VINF_IOM_HC_IOPORT_WRITE:
         case VINF_IOM_HC_MMIO_WRITE:
         case VINF_IOM_HC_MMIO_READ:
         case VINF_IOM_HC_MMIO_READ_WRITE:
         case VINF_PATM_PATCH_INT3:
+        case VINF_EM_NO_MEMORY:
         case VINF_EM_RAW_TO_R3:
         case VINF_EM_RAW_TIMER_PENDING:
         case VINF_EM_PENDING_REQUEST:
@@ -841,12 +944,14 @@ DECLASM(int) TRPMGCTrap0dHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
             break;
 
         default:
-            AssertMsg(PATMIsPatchGCAddr(pVM, (RTGCPTR)pRegFrame->eip) == false, ("return code %d\n", rc));
+            AssertMsg(PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip) == false, ("return code %d\n", rc));
             break;
         }
+    Log6(("TRPMGC0d: %Rrc (%04x:%08x)\n", rc, pRegFrame->cs, pRegFrame->eip));
     return rc;
 }
 
+
 /**
  * \#PF (Page Fault) handler.
  *
@@ -863,55 +968,58 @@ DECLASM(int) TRPMGCTrap0dHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
  */
 DECLASM(int) TRPMGCTrap0eHandler(PTRPM pTrpm, PCPUMCTXCORE pRegFrame)
 {
-    LogBird(("TRPMGCTrap0eHandler: eip=%RGv\n", pRegFrame->eip));
+    LogFlow(("TRPMGC0e: %04x:%08x err=%x cr2=%08x\n", pRegFrame->cs, pRegFrame->eip, (uint32_t)pTrpm->uActiveErrorCode, (uint32_t)pTrpm->uActiveCR2));
     PVM pVM = TRPM2VM(pTrpm);
 
+
     /*
      * This is all PGM stuff.
      */
     int rc = PGMTrap0eHandler(pVM, pTrpm->uActiveErrorCode, pRegFrame, (RTGCPTR)pTrpm->uActiveCR2);
-
     switch (rc)
     {
-    case VINF_EM_RAW_EMULATE_INSTR:
-    case VINF_EM_RAW_EMULATE_INSTR_PD_FAULT:
-    case VINF_EM_RAW_EMULATE_INSTR_GDT_FAULT:
-    case VINF_EM_RAW_EMULATE_INSTR_TSS_FAULT:
-    case VINF_EM_RAW_EMULATE_INSTR_LDT_FAULT:
-    case VINF_EM_RAW_EMULATE_INSTR_IDT_FAULT:
-        if (PATMIsPatchGCAddr(pVM, (RTGCPTR)pRegFrame->eip))
-            rc = VINF_PATCH_EMULATE_INSTR;
-        break;
+        case VINF_EM_RAW_EMULATE_INSTR:
+        case VINF_EM_RAW_EMULATE_INSTR_PD_FAULT:
+        case VINF_EM_RAW_EMULATE_INSTR_GDT_FAULT:
+        case VINF_EM_RAW_EMULATE_INSTR_TSS_FAULT:
+        case VINF_EM_RAW_EMULATE_INSTR_LDT_FAULT:
+        case VINF_EM_RAW_EMULATE_INSTR_IDT_FAULT:
+            if (PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip))
+                rc = VINF_PATCH_EMULATE_INSTR;
+            break;
 
-    case VINF_EM_RAW_GUEST_TRAP:
-        if (PATMIsPatchGCAddr(pVM, (RTGCPTR)pRegFrame->eip))
-            return VINF_PATM_PATCH_TRAP_PF;
+        case VINF_EM_RAW_GUEST_TRAP:
+            if (PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip))
+                return VINF_PATM_PATCH_TRAP_PF;
 
-        rc = TRPMForwardTrap(pVM, pRegFrame, 0xE, 0, TRPM_TRAP_HAS_ERRORCODE, TRPM_TRAP);
-        Assert(rc == VINF_EM_RAW_GUEST_TRAP);
-        break;
+            rc = TRPMForwardTrap(pVM, pRegFrame, 0xE, 0, TRPM_TRAP_HAS_ERRORCODE, TRPM_TRAP, 0xe);
+            Assert(rc == VINF_EM_RAW_GUEST_TRAP);
+            break;
 
-    case VINF_EM_RAW_INTERRUPT_PENDING:
-        Assert(TRPMHasTrap(pVM));
-        /* no break; */
-    case VINF_IOM_HC_MMIO_READ:
-    case VINF_IOM_HC_MMIO_WRITE:
-    case VINF_IOM_HC_MMIO_READ_WRITE:
-    case VINF_PATM_HC_MMIO_PATCH_READ:
-    case VINF_PATM_HC_MMIO_PATCH_WRITE:
-    case VINF_SUCCESS:
-    case VINF_EM_RAW_TO_R3:
-    case VINF_EM_PENDING_REQUEST:
-    case VINF_EM_RAW_TIMER_PENDING:
-    case VINF_CSAM_PENDING_ACTION:
-    case VINF_PGM_SYNC_CR3: /** @todo Check this with Sander. */
-        break;
+        case VINF_EM_RAW_INTERRUPT_PENDING:
+            Assert(TRPMHasTrap(pVM));
+            /* no break; */
+        case VINF_IOM_HC_MMIO_READ:
+        case VINF_IOM_HC_MMIO_WRITE:
+        case VINF_IOM_HC_MMIO_READ_WRITE:
+        case VINF_PATM_HC_MMIO_PATCH_READ:
+        case VINF_PATM_HC_MMIO_PATCH_WRITE:
+        case VINF_SUCCESS:
+        case VINF_EM_RAW_TO_R3:
+        case VINF_EM_PENDING_REQUEST:
+        case VINF_EM_RAW_TIMER_PENDING:
+        case VINF_EM_NO_MEMORY:
+        case VINF_CSAM_PENDING_ACTION:
+        case VINF_PGM_SYNC_CR3: /** @todo Check this with Sander. */
+            break;
 
-    default:
-        AssertMsg(PATMIsPatchGCAddr(pVM, (RTGCPTR)pRegFrame->eip) == false, ("Patch address for return code %d. eip=%08x\n", rc, pRegFrame->eip));
-        break;
+        default:
+            AssertMsg(PATMIsPatchGCAddr(pVM, (RTRCPTR)pRegFrame->eip) == false, ("Patch address for return code %d. eip=%08x\n", rc, pRegFrame->eip));
+            break;
     }
-    return trpmGCExitTrap(pVM, rc, pRegFrame);
+    rc = trpmGCExitTrap(pVM, rc, pRegFrame);
+    Log6(("TRPMGC0e: %Rrc (%04x:%08x)\n", rc, pRegFrame->cs, pRegFrame->eip));
+    return rc;
 }
 
 
@@ -1034,7 +1142,7 @@ DECLCALLBACK(int) trpmGCTrapInGeneric(PVM pVM, PCPUMCTXCORE pRegFrame, uintptr_t
          * Check that there is still some stack left, if not we'll flag
          * a guru meditation (the alternative is a triple fault).
          */
-        RTGCUINTPTR cbStackUsed = (RTGCUINTPTR)VMMGetStackGC(pVM) - pRegFrame->esp;
+        RTRCUINTPTR cbStackUsed = (RTRCUINTPTR)VMMGetStackRC(pVM) - pRegFrame->esp;
         if (cbStackUsed > VMM_STACK_SIZE - _1K)
         {
             LogRel(("trpmGCTrapInGeneric: ran out of stack: esp=#x cbStackUsed=%#x\n", pRegFrame->esp, cbStackUsed));
