@@ -1,18 +1,22 @@
-/* $Id: VMMTests.cpp 23517 2007-08-07 17:07:59Z umoeller $ */
+/* $Id: VMMTests.cpp 16861 2009-02-17 16:37:45Z vboxsync $ */
 /** @file
  * VMM - The Virtual Machine Monitor Core, Tests.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 //#define NO_SUPCALLR0VMM
@@ -51,21 +55,25 @@
  */
 static int vmmR3DoGCTest(PVM pVM, VMMGCOPERATION enmTestcase, unsigned uVariation)
 {
-    RTGCPTR GCPtrEP;
-    int rc = PDMR3GetSymbolGC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &GCPtrEP);
-    if (VBOX_FAILURE(rc))
+    RTRCPTR RCPtrEP;
+    int rc = PDMR3LdrGetSymbolRC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &RCPtrEP);
+    if (RT_FAILURE(rc))
         return rc;
 
     CPUMHyperSetCtxCore(pVM, NULL);
-    memset(pVM->vmm.s.pbHCStack, 0xaa, VMM_STACK_SIZE);
-    CPUMSetHyperESP(pVM, pVM->vmm.s.pbGCStackBottom); /* Clear the stack. */
+    memset(pVM->vmm.s.pbEMTStackR3, 0xaa, VMM_STACK_SIZE);
+    CPUMSetHyperESP(pVM, pVM->vmm.s.pbEMTStackBottomRC); /* Clear the stack. */
     CPUMPushHyper(pVM, uVariation);
     CPUMPushHyper(pVM, enmTestcase);
-    CPUMPushHyper(pVM, pVM->pVMGC);
-    CPUMPushHyper(pVM, 3 * sizeof(RTGCPTR));  /* stack frame size */
-    CPUMPushHyper(pVM, GCPtrEP);                /* what to call */
-    CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnGCCallTrampoline);
-    return SUPCallVMMR0(pVM->pVMR0, VMMR0_DO_RAW_RUN, NULL);
+    CPUMPushHyper(pVM, pVM->pVMRC);
+    CPUMPushHyper(pVM, 3 * sizeof(RTRCPTR));    /* stack frame size */
+    CPUMPushHyper(pVM, RCPtrEP);                /* what to call */
+    CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnCallTrampolineRC);
+    Assert(CPUMGetHyperCR3(pVM) && CPUMGetHyperCR3(pVM) == PGMGetHyperCR3(pVM));
+    rc = SUPCallVMMR0Fast(pVM->pVMR0, VMMR0_DO_RAW_RUN, 0);
+    if (RT_LIKELY(rc == VINF_SUCCESS))
+        rc = pVM->vmm.s.iLastGZRc;
+    return rc;
 }
 
 
@@ -85,25 +93,28 @@ static int vmmR3DoTrapTest(PVM pVM, uint8_t u8Trap, unsigned uVariation, int rcE
 {
     RTPrintf("VMM: testing 0%x / %d - %s\n", u8Trap, uVariation, pszDesc);
 
-    RTGCPTR GCPtrEP;
-    int rc = PDMR3GetSymbolGC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &GCPtrEP);
-    if (VBOX_FAILURE(rc))
+    RTRCPTR RCPtrEP;
+    int rc = PDMR3LdrGetSymbolRC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &RCPtrEP);
+    if (RT_FAILURE(rc))
         return rc;
 
     CPUMHyperSetCtxCore(pVM, NULL);
-    memset(pVM->vmm.s.pbHCStack, 0xaa, VMM_STACK_SIZE);
-    CPUMSetHyperESP(pVM, pVM->vmm.s.pbGCStackBottom); /* Clear the stack. */
+    memset(pVM->vmm.s.pbEMTStackR3, 0xaa, VMM_STACK_SIZE);
+    CPUMSetHyperESP(pVM, pVM->vmm.s.pbEMTStackBottomRC); /* Clear the stack. */
     CPUMPushHyper(pVM, uVariation);
     CPUMPushHyper(pVM, u8Trap + VMMGC_DO_TESTCASE_TRAP_FIRST);
-    CPUMPushHyper(pVM, pVM->pVMGC);
-    CPUMPushHyper(pVM, 3 * sizeof(RTGCPTR));  /* stack frame size */
-    CPUMPushHyper(pVM, GCPtrEP);                /* what to call */
-    CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnGCCallTrampoline);
-    rc = SUPCallVMMR0(pVM->pVMR0, VMMR0_DO_RAW_RUN, NULL);
+    CPUMPushHyper(pVM, pVM->pVMRC);
+    CPUMPushHyper(pVM, 3 * sizeof(RTRCPTR));    /* stack frame size */
+    CPUMPushHyper(pVM, RCPtrEP);                /* what to call */
+    CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnCallTrampolineRC);
+    Assert(CPUMGetHyperCR3(pVM) && CPUMGetHyperCR3(pVM) == PGMGetHyperCR3(pVM));
+    rc = SUPCallVMMR0Fast(pVM->pVMR0, VMMR0_DO_RAW_RUN, 0);
+    if (RT_LIKELY(rc == VINF_SUCCESS))
+        rc = pVM->vmm.s.iLastGZRc;
     bool fDump = false;
     if (rc != rcExpect)
     {
-        RTPrintf("VMM: FAILURE - rc=%Vrc expected %Vrc\n", rc, rcExpect);
+        RTPrintf("VMM: FAILURE - rc=%Rrc expected %Rrc\n", rc, rcExpect);
         if (rc != VERR_NOT_IMPLEMENTED)
             fDump = true;
     }
@@ -118,13 +129,13 @@ static int vmmR3DoTrapTest(PVM pVM, uint8_t u8Trap, unsigned uVariation, int rcE
     }
     else if (pszFaultEIP)
     {
-        RTGCPTR GCPtrFault;
-        int rc2 = PDMR3GetSymbolGC(pVM, VMMGC_MAIN_MODULE_NAME, pszFaultEIP, &GCPtrFault);
-        if (VBOX_FAILURE(rc2))
-            RTPrintf("VMM: FAILURE - Failed to resolve symbol '%s', %Vrc!\n", pszFaultEIP, rc);
-        else if (GCPtrFault != CPUMGetHyperEIP(pVM))
+        RTRCPTR RCPtrFault;
+        int rc2 = PDMR3LdrGetSymbolRC(pVM, VMMGC_MAIN_MODULE_NAME, pszFaultEIP, &RCPtrFault);
+        if (RT_FAILURE(rc2))
+            RTPrintf("VMM: FAILURE - Failed to resolve symbol '%s', %Rrc!\n", pszFaultEIP, rc);
+        else if (RCPtrFault != CPUMGetHyperEIP(pVM))
         {
-            RTPrintf("VMM: FAILURE - EIP=%VGv expected %VGv (%s)\n", CPUMGetHyperEIP(pVM), GCPtrFault, pszFaultEIP);
+            RTPrintf("VMM: FAILURE - EIP=%08RX32 expected %RRv (%s)\n", CPUMGetHyperEIP(pVM), RCPtrFault, pszFaultEIP);
             fDump = true;
         }
     }
@@ -173,11 +184,11 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
     /*
      * Setup stack for calling VMMGCEntry().
      */
-    RTGCPTR GCPtrEP;
-    int rc = PDMR3GetSymbolGC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &GCPtrEP);
-    if (VBOX_SUCCESS(rc))
+    RTRCPTR RCPtrEP;
+    int rc = PDMR3LdrGetSymbolRC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &RCPtrEP);
+    if (RT_SUCCESS(rc))
     {
-        RTPrintf("VMM: VMMGCEntry=%VGv\n", GCPtrEP);
+        RTPrintf("VMM: VMMGCEntry=%RRv\n", RCPtrEP);
 
         /*
          * Test various crashes which we must be able to recover from.
@@ -191,7 +202,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         bool f;
         rc = CFGMR3QueryBool(CFGMR3GetRoot(pVM), "DoubleFault", &f);
 #if !defined(DEBUG_bird)
-        if (VBOX_SUCCESS(rc) && f)
+        if (RT_SUCCESS(rc) && f)
 #endif
         {
             /* see tripple fault warnings in SELM and VMMGC.cpp. */
@@ -215,7 +226,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         rc = vmmR3DoGCTest(pVM, VMMGC_DO_TESTCASE_NOP, 0);
         if (rc != VINF_SUCCESS)
         {
-            RTPrintf("VMM: Nop test failed, rc=%Vrc not VINF_SUCCESS\n", rc);
+            RTPrintf("VMM: Nop test failed, rc=%Rrc not VINF_SUCCESS\n", rc);
             return rc;
         }
 
@@ -229,20 +240,20 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         rc = vmmR3DoGCTest(pVM, VMMGC_DO_TESTCASE_NOP, 0);
         if (rc != VINF_SUCCESS)
         {
-            RTPrintf("VMM: DR0=0x10000 test failed with rc=%Vrc!\n", rc);
+            RTPrintf("VMM: DR0=0x10000 test failed with rc=%Rrc!\n", rc);
             return rc;
         }
 
         /* a bad one at VMMGCEntry */
         RTPrintf("VMM: testing hardware bp at VMMGCEntry (hit)\n");
-        DBGFR3AddrFromFlat(pVM, &Addr, GCPtrEP);
+        DBGFR3AddrFromFlat(pVM, &Addr, RCPtrEP);
         RTUINT iBp1;
         rc = DBGFR3BpSetReg(pVM, &Addr, 0,  ~(uint64_t)0, X86_DR7_RW_EO, 1, &iBp1);
         AssertReleaseRC(rc);
         rc = vmmR3DoGCTest(pVM, VMMGC_DO_TESTCASE_NOP, 0);
         if (rc != VINF_EM_DBG_HYPER_BREAKPOINT)
         {
-            RTPrintf("VMM: DR1=VMMGCEntry test failed with rc=%Vrc! expected VINF_EM_RAW_BREAKPOINT_HYPER\n", rc);
+            RTPrintf("VMM: DR1=VMMGCEntry test failed with rc=%Rrc! expected VINF_EM_RAW_BREAKPOINT_HYPER\n", rc);
             return rc;
         }
 
@@ -252,7 +263,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         rc = VMMR3ResumeHyper(pVM);
         if (rc != VINF_SUCCESS)
         {
-            RTPrintf("VMM: failed to resume on hyper breakpoint, rc=%Vrc\n", rc);
+            RTPrintf("VMM: failed to resume on hyper breakpoint, rc=%Rrc\n", rc);
             return rc;
         }
 
@@ -261,7 +272,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         rc = vmmR3DoGCTest(pVM, VMMGC_DO_TESTCASE_NOP, 0);
         if (rc != VINF_EM_DBG_HYPER_BREAKPOINT)
         {
-            RTPrintf("VMM: DR1=VMMGCEntry test failed with rc=%Vrc! expected VINF_EM_RAW_BREAKPOINT_HYPER\n", rc);
+            RTPrintf("VMM: DR1=VMMGCEntry test failed with rc=%Rrc! expected VINF_EM_RAW_BREAKPOINT_HYPER\n", rc);
             return rc;
         }
 
@@ -274,7 +285,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
             rc = VMMR3ResumeHyper(pVM);
             if (rc != VINF_EM_DBG_HYPER_STEPPED)
             {
-                RTPrintf("\nVMM: failed to step on hyper breakpoint, rc=%Vrc\n", rc);
+                RTPrintf("\nVMM: failed to step on hyper breakpoint, rc=%Rrc\n", rc);
                 return rc;
             }
             RTGCUINTREG Pc = CPUMGetHyperEIP(pVM);
@@ -289,8 +300,8 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         RTPrintf("ok\n");
 
         /* done, clear it */
-        if (    VBOX_FAILURE(DBGFR3BpClear(pVM, iBp0))
-            ||  VBOX_FAILURE(DBGFR3BpClear(pVM, iBp1)))
+        if (    RT_FAILURE(DBGFR3BpClear(pVM, iBp0))
+            ||  RT_FAILURE(DBGFR3BpClear(pVM, iBp1)))
         {
             RTPrintf("VMM: Failed to clear breakpoints!\n");
             return VERR_GENERAL_FAILURE;
@@ -298,7 +309,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         rc = vmmR3DoGCTest(pVM, VMMGC_DO_TESTCASE_NOP, 0);
         if (rc != VINF_SUCCESS)
         {
-            RTPrintf("VMM: NOP failed, rc=%Vrc\n", rc);
+            RTPrintf("VMM: NOP failed, rc=%Rrc\n", rc);
             return rc;
         }
 
@@ -312,7 +323,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
             rc = vmmR3DoGCTest(pVM, VMMGC_DO_TESTCASE_INTERRUPT_MASKING, 0);
             if (rc != VINF_SUCCESS)
             {
-                RTPrintf("VMM: Interrupt masking failed: rc=%Vrc\n", rc);
+                RTPrintf("VMM: Interrupt masking failed: rc=%Rrc\n", rc);
                 return rc;
             }
             uint64_t Ticks = ASMReadTSC() - StartTick;
@@ -324,14 +335,14 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
          * Interrupt forwarding.
          */
         CPUMHyperSetCtxCore(pVM, NULL);
-        CPUMSetHyperESP(pVM, pVM->vmm.s.pbGCStackBottom); /* Clear the stack. */
+        CPUMSetHyperESP(pVM, pVM->vmm.s.pbEMTStackBottomRC); /* Clear the stack. */
         CPUMPushHyper(pVM, 0);
         CPUMPushHyper(pVM, VMMGC_DO_TESTCASE_HYPER_INTERRUPT);
-        CPUMPushHyper(pVM, pVM->pVMGC);
-        CPUMPushHyper(pVM, 3 * sizeof(RTGCPTR));  /* stack frame size */
-        CPUMPushHyper(pVM, GCPtrEP);                /* what to call */
-        CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnGCCallTrampoline);
-        Log(("trampoline=%x\n", pVM->vmm.s.pfnGCCallTrampoline));
+        CPUMPushHyper(pVM, pVM->pVMRC);
+        CPUMPushHyper(pVM, 3 * sizeof(RTRCPTR));    /* stack frame size */
+        CPUMPushHyper(pVM, RCPtrEP);                /* what to call */
+        CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnCallTrampolineRC);
+        Log(("trampoline=%x\n", pVM->vmm.s.pfnCallTrampolineRC));
 
         /*
          * Switch and do da thing.
@@ -340,12 +351,15 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         i = 0;
         uint64_t    tsBegin = RTTimeNanoTS();
         uint64_t    TickStart = ASMReadTSC();
+        Assert(CPUMGetHyperCR3(pVM) && CPUMGetHyperCR3(pVM) == PGMGetHyperCR3(pVM));
         do
         {
-            rc = SUPCallVMMR0(pVM->pVMR0, VMMR0_DO_RAW_RUN, NULL);
-            if (VBOX_FAILURE(rc))
+            rc = SUPCallVMMR0Fast(pVM->pVMR0, VMMR0_DO_RAW_RUN, 0);
+            if (RT_LIKELY(rc == VINF_SUCCESS))
+                rc = pVM->vmm.s.iLastGZRc;
+            if (RT_FAILURE(rc))
             {
-                Log(("VMM: GC returned fatal %Vra in iteration %d\n", rc, i));
+                Log(("VMM: GC returned fatal %Rra in iteration %d\n", rc, i));
                 VMMR3FatalDump(pVM, rc);
                 return rc;
             }
@@ -381,23 +395,26 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         uint64_t TickMin = ~0;
         tsBegin = RTTimeNanoTS();
         TickStart = ASMReadTSC();
+        Assert(CPUMGetHyperCR3(pVM) && CPUMGetHyperCR3(pVM) == PGMGetHyperCR3(pVM));
         for (i = 0; i < 1000000; i++)
         {
             CPUMHyperSetCtxCore(pVM, NULL);
-            CPUMSetHyperESP(pVM, pVM->vmm.s.pbGCStackBottom); /* Clear the stack. */
+            CPUMSetHyperESP(pVM, pVM->vmm.s.pbEMTStackBottomRC); /* Clear the stack. */
             CPUMPushHyper(pVM, 0);
             CPUMPushHyper(pVM, VMMGC_DO_TESTCASE_NOP);
-            CPUMPushHyper(pVM, pVM->pVMGC);
-            CPUMPushHyper(pVM, 3 * sizeof(RTGCPTR));    /* stack frame size */
-            CPUMPushHyper(pVM, GCPtrEP);                /* what to call */
-            CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnGCCallTrampoline);
+            CPUMPushHyper(pVM, pVM->pVMRC);
+            CPUMPushHyper(pVM, 3 * sizeof(RTRCPTR));    /* stack frame size */
+            CPUMPushHyper(pVM, RCPtrEP);                /* what to call */
+            CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnCallTrampolineRC);
 
             uint64_t TickThisStart = ASMReadTSC();
-            rc = SUPCallVMMR0(pVM->pVMR0, VMMR0_DO_RAW_RUN, NULL);
+            rc = SUPCallVMMR0Fast(pVM->pVMR0, VMMR0_DO_RAW_RUN, 0);
+            if (RT_LIKELY(rc == VINF_SUCCESS))
+                rc = pVM->vmm.s.iLastGZRc;
             uint64_t TickThisElapsed = ASMReadTSC() - TickThisStart;
-            if (VBOX_FAILURE(rc))
+            if (RT_FAILURE(rc))
             {
-                Log(("VMM: GC returned fatal %Vra in iteration %d\n", rc, i));
+                Log(("VMM: GC returned fatal %Rra in iteration %d\n", rc, i));
                 VMMR3FatalDump(pVM, rc);
                 return rc;
             }
@@ -420,7 +437,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
         rc = VINF_SUCCESS;
     }
     else
-        AssertMsgFailed(("Failed to resolved VMMGC.gc::VMMGCEntry(), rc=%Vrc\n", rc));
+        AssertMsgFailed(("Failed to resolved VMMGC.gc::VMMGCEntry(), rc=%Rrc\n", rc));
 #endif
     return rc;
 }
@@ -432,7 +449,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
             int rc = SELMR3GetShadowSelectorInfo(pVM, pHyperCtx->reg, &selInfo);        \
             AssertRC(rc);                                                               \
                                                                                         \
-            pHyperCtx->reg##Hid.u32Base              = selInfo.GCPtrBase;               \
+            pHyperCtx->reg##Hid.u64Base              = selInfo.GCPtrBase;               \
             pHyperCtx->reg##Hid.u32Limit             = selInfo.cbLimit;                 \
             pHyperCtx->reg##Hid.Attr.n.u1Present     = selInfo.Raw.Gen.u1Present;       \
             pHyperCtx->reg##Hid.Attr.n.u1DefBig      = selInfo.Raw.Gen.u1DefBig;        \
@@ -440,7 +457,7 @@ VMMR3DECL(int) VMMDoTest(PVM pVM)
             pHyperCtx->reg##Hid.Attr.n.u4Type        = selInfo.Raw.Gen.u4Type;          \
             pHyperCtx->reg##Hid.Attr.n.u2Dpl         = selInfo.Raw.Gen.u2Dpl;           \
             pHyperCtx->reg##Hid.Attr.n.u1DescType    = selInfo.Raw.Gen.u1DescType;      \
-            pHyperCtx->reg##Hid.Attr.n.u1Reserved    = selInfo.Raw.Gen.u1Reserved;      \
+            pHyperCtx->reg##Hid.Attr.n.u1Long        = selInfo.Raw.Gen.u1Long;          \
         }
 
 /* execute the switch. */
@@ -464,13 +481,19 @@ VMMR3DECL(int) VMMDoHwAccmTest(PVM pVM)
     VM_FF_CLEAR(pVM, VM_FF_SELM_SYNC_TSS);
 
     /* Enable mapping of the hypervisor into the shadow page table. */
-    PGMR3ChangeShwPDMappings(pVM, true);
+    uint32_t cb;
+    rc = PGMR3MappingsSize(pVM, &cb);
+    AssertRCReturn(rc, rc);
+
+    /* Pretend the mappings are now fixed; to force a refresh of the reserved PDEs. */
+    rc = PGMR3MappingsFix(pVM, MM_HYPER_AREA_ADDRESS, cb);
+    AssertRCReturn(rc, rc);
 
     CPUMQueryHyperCtxPtr(pVM, &pHyperCtx);
 
     pHyperCtx->cr0 = X86_CR0_PE | X86_CR0_WP | X86_CR0_PG | X86_CR0_TS | X86_CR0_ET | X86_CR0_NE | X86_CR0_MP;
     pHyperCtx->cr4 = X86_CR4_PGE | X86_CR4_OSFSXR | X86_CR4_OSXMMEEXCPT;
-    PGMChangeMode(pVM, pHyperCtx->cr0, pHyperCtx->cr4, 0);
+    PGMChangeMode(pVM, pHyperCtx->cr0, pHyperCtx->cr4, pHyperCtx->msrEFER);
     PGMSyncCR3(pVM, pHyperCtx->cr0, CR3Phys, pHyperCtx->cr4, true);
 
     VM_FF_CLEAR(pVM, VM_FF_TO_R3);
@@ -480,11 +503,11 @@ VMMR3DECL(int) VMMDoHwAccmTest(PVM pVM)
     /*
      * Setup stack for calling VMMGCEntry().
      */
-    RTGCPTR GCPtrEP;
-    rc = PDMR3GetSymbolGC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &GCPtrEP);
-    if (VBOX_SUCCESS(rc))
+    RTRCPTR RCPtrEP;
+    rc = PDMR3LdrGetSymbolRC(pVM, VMMGC_MAIN_MODULE_NAME, "VMMGCEntry", &RCPtrEP);
+    if (RT_SUCCESS(rc))
     {
-        RTPrintf("VMM: VMMGCEntry=%VGv\n", GCPtrEP);
+        RTPrintf("VMM: VMMGCEntry=%RRv\n", RCPtrEP);
 
         CPUMQueryHyperCtxPtr(pVM, &pHyperCtx);
 
@@ -509,16 +532,16 @@ VMMR3DECL(int) VMMDoHwAccmTest(PVM pVM)
         {
             CPUMHyperSetCtxCore(pVM, NULL);
 
-            CPUMSetHyperESP(pVM, pVM->vmm.s.pbGCStackBottom); /* Clear the stack. */
+            CPUMSetHyperESP(pVM, pVM->vmm.s.pbEMTStackBottomRC); /* Clear the stack. */
             CPUMPushHyper(pVM, 0);
             CPUMPushHyper(pVM, VMMGC_DO_TESTCASE_HWACCM_NOP);
-            CPUMPushHyper(pVM, pVM->pVMGC);
-            CPUMPushHyper(pVM, 3 * sizeof(RTGCPTR));    /* stack frame size */
-            CPUMPushHyper(pVM, GCPtrEP);                /* what to call */
-            CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnGCCallTrampoline);
+            CPUMPushHyper(pVM, pVM->pVMRC);
+            CPUMPushHyper(pVM, 3 * sizeof(RTRCPTR));    /* stack frame size */
+            CPUMPushHyper(pVM, RCPtrEP);                /* what to call */
+            CPUMSetHyperEIP(pVM, pVM->vmm.s.pfnCallTrampolineRC);
 
             CPUMQueryHyperCtxPtr(pVM, &pHyperCtx);
-            CPUMQueryGuestCtxPtr(pVM, &pGuestCtx);
+            pGuestCtx = CPUMQueryGuestCtxPtr(pVM);
 
             /* Copy the hypervisor context to make sure we have a valid guest context. */
             *pGuestCtx = *pHyperCtx;
@@ -528,11 +551,11 @@ VMMR3DECL(int) VMMDoHwAccmTest(PVM pVM)
             VM_FF_CLEAR(pVM, VM_FF_TIMER);
 
             uint64_t TickThisStart = ASMReadTSC();
-            rc = SUPCallVMMR0(pVM->pVMR0, VMMR0_DO_HWACC_RUN, NULL);
+            rc = SUPCallVMMR0Fast(pVM->pVMR0, VMMR0_DO_HWACC_RUN, 0);
             uint64_t TickThisElapsed = ASMReadTSC() - TickThisStart;
-            if (VBOX_FAILURE(rc))
+            if (RT_FAILURE(rc))
             {
-                Log(("VMM: R0 returned fatal %Vrc in iteration %d\n", rc, i));
+                Log(("VMM: R0 returned fatal %Rrc in iteration %d\n", rc, i));
                 VMMR3FatalDump(pVM, rc);
                 return rc;
             }
@@ -555,7 +578,7 @@ VMMR3DECL(int) VMMDoHwAccmTest(PVM pVM)
         rc = VINF_SUCCESS;
     }
     else
-        AssertMsgFailed(("Failed to resolved VMMGC.gc::VMMGCEntry(), rc=%Vrc\n", rc));
+        AssertMsgFailed(("Failed to resolved VMMGC.gc::VMMGCEntry(), rc=%Rrc\n", rc));
 
     return rc;
 }
