@@ -1,18 +1,22 @@
-/* $Id: VMReq.cpp 24696 2007-09-21 07:56:41Z sunlover $ */
+/* $Id: VMReq.cpp 13858 2008-11-05 13:45:41Z vboxsync $ */
 /** @file
  * VM - Virtual Machine
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 
@@ -24,6 +28,7 @@
 #include <VBox/vmm.h>
 #include "VMInternal.h"
 #include <VBox/vm.h>
+#include <VBox/uvm.h>
 
 #include <VBox/err.h>
 #include <VBox/param.h>
@@ -39,7 +44,7 @@
 /*******************************************************************************
 *   Internal Functions                                                         *
 *******************************************************************************/
-static int  vmR3ReqProcessOne(PVM pVM, PVMREQ pReq);
+static int  vmR3ReqProcessOneU(PUVM pUVM, PVMREQ pReq);
 
 
 /**
@@ -55,6 +60,7 @@ static int  vmR3ReqProcessOne(PVM pVM, PVMREQ pReq);
  * @returns VERR_TIMEOUT if cMillies was reached without the packet being completed.
  *
  * @param   pVM             The VM handle.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
  * @param   ppReq           Where to store the pointer to the request.
  *                          This will be NULL or a valid request pointer not matter what happends.
  * @param   cMillies        Number of milliseconds to wait for the request to
@@ -65,11 +71,45 @@ static int  vmR3ReqProcessOne(PVM pVM, PVMREQ pReq);
  *                          Not possible to pass 64-bit arguments!
  * @param   ...             Function arguments.
  */
-VMR3DECL(int) VMR3ReqCall(PVM pVM, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFunction, unsigned cArgs, ...)
+VMMR3DECL(int) VMR3ReqCall(PVM pVM, VMREQDEST enmDest, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFunction, unsigned cArgs, ...)
 {
     va_list va;
     va_start(va, cArgs);
-    int rc = VMR3ReqCallV(pVM, ppReq, cMillies, VMREQFLAGS_VBOX_STATUS, pfnFunction, cArgs, va);
+    int rc = VMR3ReqCallVU(pVM->pUVM, enmDest, ppReq, cMillies, VMREQFLAGS_VBOX_STATUS, pfnFunction, cArgs, va);
+    va_end(va);
+    return rc;
+}
+
+
+/**
+ * Allocate and queue a call request to a void function.
+ *
+ * If it's desired to poll on the completion of the request set cMillies
+ * to 0 and use VMR3ReqWait() to check for completation. In the other case
+ * use RT_INDEFINITE_WAIT.
+ * The returned request packet must be freed using VMR3ReqFree().
+ *
+ * @returns VBox status code.
+ *          Will not return VERR_INTERRUPTED.
+ * @returns VERR_TIMEOUT if cMillies was reached without the packet being completed.
+ *
+ * @param   pUVM            Pointer to the user mode VM structure.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
+ * @param   ppReq           Where to store the pointer to the request.
+ *                          This will be NULL or a valid request pointer not matter what happends.
+ * @param   cMillies        Number of milliseconds to wait for the request to
+ *                          be completed. Use RT_INDEFINITE_WAIT to only
+ *                          wait till it's completed.
+ * @param   pfnFunction     Pointer to the function to call.
+ * @param   cArgs           Number of arguments following in the ellipsis.
+ *                          Not possible to pass 64-bit arguments!
+ * @param   ...             Function arguments.
+ */
+VMMR3DECL(int) VMR3ReqCallVoidU(PUVM pUVM, VMREQDEST enmDest, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFunction, unsigned cArgs, ...)
+{
+    va_list va;
+    va_start(va, cArgs);
+    int rc = VMR3ReqCallVU(pUVM, enmDest, ppReq, cMillies, VMREQFLAGS_VOID, pfnFunction, cArgs, va);
     va_end(va);
     return rc;
 }
@@ -88,6 +128,7 @@ VMR3DECL(int) VMR3ReqCall(PVM pVM, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFu
  * @returns VERR_TIMEOUT if cMillies was reached without the packet being completed.
  *
  * @param   pVM             The VM handle.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
  * @param   ppReq           Where to store the pointer to the request.
  *                          This will be NULL or a valid request pointer not matter what happends.
  * @param   cMillies        Number of milliseconds to wait for the request to
@@ -98,11 +139,11 @@ VMR3DECL(int) VMR3ReqCall(PVM pVM, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFu
  *                          Not possible to pass 64-bit arguments!
  * @param   ...             Function arguments.
  */
-VMR3DECL(int) VMR3ReqCallVoid(PVM pVM, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFunction, unsigned cArgs, ...)
+VMMR3DECL(int) VMR3ReqCallVoid(PVM pVM, VMREQDEST enmDest, PVMREQ *ppReq, unsigned cMillies, PFNRT pfnFunction, unsigned cArgs, ...)
 {
     va_list va;
     va_start(va, cArgs);
-    int rc = VMR3ReqCallV(pVM, ppReq, cMillies, VMREQFLAGS_VOID, pfnFunction, cArgs, va);
+    int rc = VMR3ReqCallVU(pVM->pUVM, enmDest, ppReq, cMillies, VMREQFLAGS_VOID, pfnFunction, cArgs, va);
     va_end(va);
     return rc;
 }
@@ -121,6 +162,7 @@ VMR3DECL(int) VMR3ReqCallVoid(PVM pVM, PVMREQ *ppReq, unsigned cMillies, PFNRT p
  * @returns VERR_TIMEOUT if cMillies was reached without the packet being completed.
  *
  * @param   pVM             The VM handle.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
  * @param   ppReq           Where to store the pointer to the request.
  *                          This will be NULL or a valid request pointer not matter what happends, unless fFlags
  *                          contains VMREQFLAGS_NO_WAIT when it will be optional and always NULL.
@@ -133,11 +175,47 @@ VMR3DECL(int) VMR3ReqCallVoid(PVM pVM, PVMREQ *ppReq, unsigned cMillies, PFNRT p
  *                          Not possible to pass 64-bit arguments!
  * @param   ...             Function arguments.
  */
-VMR3DECL(int) VMR3ReqCallEx(PVM pVM, PVMREQ *ppReq, unsigned cMillies, unsigned fFlags, PFNRT pfnFunction, unsigned cArgs, ...)
+VMMR3DECL(int) VMR3ReqCallEx(PVM pVM, VMREQDEST enmDest, PVMREQ *ppReq, unsigned cMillies, unsigned fFlags, PFNRT pfnFunction, unsigned cArgs, ...)
 {
     va_list va;
     va_start(va, cArgs);
-    int rc = VMR3ReqCallV(pVM, ppReq, cMillies, fFlags, pfnFunction, cArgs, va);
+    int rc = VMR3ReqCallVU(pVM->pUVM, enmDest, ppReq, cMillies, fFlags, pfnFunction, cArgs, va);
+    va_end(va);
+    return rc;
+}
+
+
+/**
+ * Allocate and queue a call request to a void function.
+ *
+ * If it's desired to poll on the completion of the request set cMillies
+ * to 0 and use VMR3ReqWait() to check for completation. In the other case
+ * use RT_INDEFINITE_WAIT.
+ * The returned request packet must be freed using VMR3ReqFree().
+ *
+ * @returns VBox status code.
+ *          Will not return VERR_INTERRUPTED.
+ * @returns VERR_TIMEOUT if cMillies was reached without the packet being completed.
+ *
+ * @param   pUVM            Pointer to the user mode VM structure.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
+ * @param   ppReq           Where to store the pointer to the request.
+ *                          This will be NULL or a valid request pointer not matter what happends, unless fFlags
+ *                          contains VMREQFLAGS_NO_WAIT when it will be optional and always NULL.
+ * @param   cMillies        Number of milliseconds to wait for the request to
+ *                          be completed. Use RT_INDEFINITE_WAIT to only
+ *                          wait till it's completed.
+ * @param   fFlags          A combination of the VMREQFLAGS values.
+ * @param   pfnFunction     Pointer to the function to call.
+ * @param   cArgs           Number of arguments following in the ellipsis.
+ *                          Not possible to pass 64-bit arguments!
+ * @param   ...             Function arguments.
+ */
+VMMR3DECL(int) VMR3ReqCallU(PUVM pUVM, VMREQDEST enmDest, PVMREQ *ppReq, unsigned cMillies, unsigned fFlags, PFNRT pfnFunction, unsigned cArgs, ...)
+{
+    va_list va;
+    va_start(va, cArgs);
+    int rc = VMR3ReqCallVU(pUVM, enmDest, ppReq, cMillies, fFlags, pfnFunction, cArgs, va);
     va_end(va);
     return rc;
 }
@@ -155,7 +233,8 @@ VMR3DECL(int) VMR3ReqCallEx(PVM pVM, PVMREQ *ppReq, unsigned cMillies, unsigned
  *          Will not return VERR_INTERRUPTED.
  * @returns VERR_TIMEOUT if cMillies was reached without the packet being completed.
  *
- * @param   pVM             The VM handle.
+ * @param   pUVM            Pointer to the user mode VM structure.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
  * @param   ppReq           Where to store the pointer to the request.
  *                          This will be NULL or a valid request pointer not matter what happends, unless fFlags
  *                          contains VMREQFLAGS_NO_WAIT when it will be optional and always NULL.
@@ -168,35 +247,31 @@ VMR3DECL(int) VMR3ReqCallEx(PVM pVM, PVMREQ *ppReq, unsigned cMillies, unsigned
  *                          Stuff which differs in size from uintptr_t is gonna make trouble, so don't try!
  * @param   Args            Argument vector.
  */
-VMR3DECL(int) VMR3ReqCallV(PVM pVM, PVMREQ *ppReq, unsigned cMillies, unsigned fFlags, PFNRT pfnFunction, unsigned cArgs, va_list Args)
+VMMR3DECL(int) VMR3ReqCallVU(PUVM pUVM, VMREQDEST enmDest, PVMREQ *ppReq, unsigned cMillies, unsigned fFlags, PFNRT pfnFunction, unsigned cArgs, va_list Args)
 {
     LogFlow(("VMR3ReqCallV: cMillies=%d fFlags=%#x pfnFunction=%p cArgs=%d\n", cMillies, fFlags, pfnFunction, cArgs));
 
     /*
-     * Check input.
+     * Validate input.
      */
-    if (!pfnFunction || !pVM || (fFlags & ~(VMREQFLAGS_RETURN_MASK | VMREQFLAGS_NO_WAIT)))
-    {
-        AssertFailed();
-        return VERR_INVALID_PARAMETER;
-    }
+    AssertPtrReturn(pfnFunction, VERR_INVALID_POINTER);
+    AssertPtrReturn(pUVM, VERR_INVALID_POINTER);
+    AssertReturn(!(fFlags & ~(VMREQFLAGS_RETURN_MASK | VMREQFLAGS_NO_WAIT)), VERR_INVALID_PARAMETER);
     if (!(fFlags & VMREQFLAGS_NO_WAIT) || ppReq)
     {
-        Assert(ppReq);
+        AssertPtrReturn(ppReq, VERR_INVALID_POINTER);
         *ppReq = NULL;
     }
     PVMREQ pReq = NULL;
-    if (cArgs * sizeof(uintptr_t) > sizeof(pReq->u.Internal.aArgs))
-    {
-        AssertMsgFailed(("cArg=%d\n", cArgs));
-        return VERR_TOO_MUCH_DATA;
-    }
+    AssertMsgReturn(cArgs * sizeof(uintptr_t) <= sizeof(pReq->u.Internal.aArgs),
+                    ("cArg=%d\n", cArgs),
+                    VERR_TOO_MUCH_DATA);
 
     /*
      * Allocate request
      */
-    int rc = VMR3ReqAlloc(pVM, &pReq, VMREQTYPE_INTERNAL);
-    if (VBOX_FAILURE(rc))
+    int rc = VMR3ReqAllocU(pUVM, &pReq, VMREQTYPE_INTERNAL, enmDest);
+    if (RT_FAILURE(rc))
         return rc;
 
     /*
@@ -212,7 +287,7 @@ VMR3DECL(int) VMR3ReqCallV(PVM pVM, PVMREQ *ppReq, unsigned cMillies, unsigned f
      * Queue the request and return.
      */
     rc = VMR3ReqQueue(pReq, cMillies);
-    if (    VBOX_FAILURE(rc)
+    if (    RT_FAILURE(rc)
         && rc != VERR_TIMEOUT)
     {
         VMR3ReqFree(pReq);
@@ -221,10 +296,10 @@ VMR3DECL(int) VMR3ReqCallV(PVM pVM, PVMREQ *ppReq, unsigned cMillies, unsigned f
     if (!(fFlags & VMREQFLAGS_NO_WAIT))
     {
         *ppReq = pReq;
-        LogFlow(("VMR3ReqCallV: returns %Vrc *ppReq=%p\n", rc, pReq));
+        LogFlow(("VMR3ReqCallV: returns %Rrc *ppReq=%p\n", rc, pReq));
     }
     else
-        LogFlow(("VMR3ReqCallV: returns %Vrc\n", rc));
+        LogFlow(("VMR3ReqCallV: returns %Rrc\n", rc));
     Assert(rc != VERR_INTERRUPTED);
     return rc;
 }
@@ -259,7 +334,7 @@ static void vmr3ReqJoinFreeSub(volatile PVMREQ *ppHead, PVMREQ pList)
 /**
  * Joins the list pList with whatever is linked up at *pHead.
  */
-static void vmr3ReqJoinFree(PVMINT pVMInt, PVMREQ pList)
+static void vmr3ReqJoinFree(PVMINTUSERPERVM pVMInt, PVMREQ pList)
 {
     /*
      * Split the list if it's too long.
@@ -271,15 +346,15 @@ static void vmr3ReqJoinFree(PVMINT pVMInt, PVMREQ pList)
         if (cReqs++ > 25)
         {
             const uint32_t i = pVMInt->iReqFree;
-            vmr3ReqJoinFreeSub(&pVMInt->apReqFree[(i + 2) % ELEMENTS(pVMInt->apReqFree)], pTail->pNext);
+            vmr3ReqJoinFreeSub(&pVMInt->apReqFree[(i + 2) % RT_ELEMENTS(pVMInt->apReqFree)], pTail->pNext);
 
             pTail->pNext = NULL;
-            vmr3ReqJoinFreeSub(&pVMInt->apReqFree[(i + 2 + (i == pVMInt->iReqFree)) % ELEMENTS(pVMInt->apReqFree)], pTail->pNext);
+            vmr3ReqJoinFreeSub(&pVMInt->apReqFree[(i + 2 + (i == pVMInt->iReqFree)) % RT_ELEMENTS(pVMInt->apReqFree)], pTail->pNext);
             return;
         }
         pTail = pTail->pNext;
     }
-    vmr3ReqJoinFreeSub(&pVMInt->apReqFree[(pVMInt->iReqFree + 2) % ELEMENTS(pVMInt->apReqFree)], pList);
+    vmr3ReqJoinFreeSub(&pVMInt->apReqFree[(pVMInt->iReqFree + 2) % RT_ELEMENTS(pVMInt->apReqFree)], pList);
 }
 
 
@@ -294,29 +369,48 @@ static void vmr3ReqJoinFree(PVMINT pVMInt, PVMREQ pList)
  * @param   pVM             VM handle.
  * @param   ppReq           Where to store the pointer to the allocated packet.
  * @param   enmType         Package type.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
  */
-VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
+VMMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType, VMREQDEST enmDest)
+{
+    return VMR3ReqAllocU(pVM->pUVM, ppReq, enmType, enmDest);
+}
+
+
+/**
+ * Allocates a request packet.
+ *
+ * The caller allocates a request packet, fills in the request data
+ * union and queues the request.
+ *
+ * @returns VBox status code.
+ *
+ * @param   pUVM            Pointer to the user mode VM structure.
+ * @param   ppReq           Where to store the pointer to the allocated packet.
+ * @param   enmType         Package type.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
+ */
+VMMR3DECL(int) VMR3ReqAllocU(PUVM pUVM, PVMREQ *ppReq, VMREQTYPE enmType, VMREQDEST enmDest)
 {
     /*
      * Validate input.
      */
-    if (    enmType < VMREQTYPE_INVALID
-        ||  enmType > VMREQTYPE_MAX)
-    {
-        AssertMsgFailed(("Invalid package type %d valid range %d-%d inclusivly.\n",
-                         enmType, VMREQTYPE_INVALID + 1, VMREQTYPE_MAX - 1));
-        return VERR_VM_REQUEST_INVALID_TYPE;
-    }
+    AssertMsgReturn(enmType > VMREQTYPE_INVALID && enmType < VMREQTYPE_MAX,
+                    ("Invalid package type %d valid range %d-%d inclusivly.\n",
+                     enmType, VMREQTYPE_INVALID + 1, VMREQTYPE_MAX - 1),
+                    VERR_VM_REQUEST_INVALID_TYPE);
+    AssertPtrReturn(ppReq, VERR_INVALID_POINTER);
+    AssertMsgReturn(enmDest == VMREQDEST_ANY || enmDest == VMREQDEST_BROADCAST || (unsigned)enmDest < pUVM->pVM->cCPUs, ("Invalid destination %d (max=%d)\n", enmDest, pUVM->pVM->cCPUs), VERR_INVALID_PARAMETER);
 
     /*
      * Try get a recycled packet.
      * While this could all be solved with a single list with a lock, it's a sport
      * of mine to avoid locks.
      */
-    int cTries = ELEMENTS(pVM->vm.s.apReqFree) * 2;
+    int cTries = RT_ELEMENTS(pUVM->vm.s.apReqFree) * 2;
     while (--cTries >= 0)
     {
-        PVMREQ volatile *ppHead = &pVM->vm.s.apReqFree[ASMAtomicIncU32(&pVM->vm.s.iReqFree) % ELEMENTS(pVM->vm.s.apReqFree)];
+        PVMREQ volatile *ppHead = &pUVM->vm.s.apReqFree[ASMAtomicIncU32(&pUVM->vm.s.iReqFree) % RT_ELEMENTS(pUVM->vm.s.apReqFree)];
 #if 0 /* sad, but this won't work safely because the reading of pReq->pNext. */
         PVMREQ pNext = NULL;
         PVMREQ pReq = *ppHead;
@@ -336,11 +430,11 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
             if (    pNext
                 &&  !ASMAtomicCmpXchgPtr((void * volatile *)ppHead, pNext, NULL))
             {
-                STAM_COUNTER_INC(&pVM->vm.s.StatReqAllocRaces);
-                vmr3ReqJoinFree(&pVM->vm.s, pReq->pNext);
+                STAM_COUNTER_INC(&pUVM->vm.s.StatReqAllocRaces);
+                vmr3ReqJoinFree(&pUVM->vm.s, pReq->pNext);
             }
 #endif
-            ASMAtomicDecU32(&pVM->vm.s.cReqFree);
+            ASMAtomicDecU32(&pUVM->vm.s.cReqFree);
 
             /*
              * Make sure the event sem is not signaled.
@@ -354,11 +448,11 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
                      * This shall not happen, but if it does we'll just destroy
                      * the semaphore and create a new one.
                      */
-                    AssertMsgFailed(("rc=%Vrc from RTSemEventWait(%#x).\n", rc, pReq->EventSem));
+                    AssertMsgFailed(("rc=%Rrc from RTSemEventWait(%#x).\n", rc, pReq->EventSem));
                     RTSemEventDestroy(pReq->EventSem);
                     rc = RTSemEventCreate(&pReq->EventSem);
                     AssertRC(rc);
-                    if (VBOX_FAILURE(rc))
+                    if (RT_FAILURE(rc))
                         return rc;
                 }
                 pReq->fEventSemClear = true;
@@ -371,15 +465,16 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
              */
             Assert(pReq->enmType == VMREQTYPE_INVALID);
             Assert(pReq->enmState == VMREQSTATE_FREE);
-            Assert(pReq->pVM == pVM);
+            Assert(pReq->pUVM == pUVM);
             ASMAtomicXchgSize(&pReq->pNext, NULL);
             pReq->enmState = VMREQSTATE_ALLOCATED;
             pReq->iStatus  = VERR_VM_REQUEST_STATUS_STILL_PENDING;
             pReq->fFlags   = VMREQFLAGS_VBOX_STATUS;
             pReq->enmType  = enmType;
+            pReq->enmDest  = enmDest;
 
             *ppReq = pReq;
-            STAM_COUNTER_INC(&pVM->vm.s.StatReqAllocRecycled);
+            STAM_COUNTER_INC(&pUVM->vm.s.StatReqAllocRecycled);
             LogFlow(("VMR3ReqAlloc: returns VINF_SUCCESS *ppReq=%p recycled\n", pReq));
             return VINF_SUCCESS;
         }
@@ -388,7 +483,7 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
     /*
      * Ok allocate one.
      */
-    PVMREQ pReq = (PVMREQ)MMR3HeapAlloc(pVM, MM_TAG_VM_REQ, sizeof(*pReq));
+    PVMREQ pReq = (PVMREQ)MMR3HeapAllocU(pUVM, MM_TAG_VM_REQ, sizeof(*pReq));
     if (!pReq)
         return VERR_NO_MEMORY;
 
@@ -397,7 +492,7 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
      */
     int rc = RTSemEventCreate(&pReq->EventSem);
     AssertRC(rc);
-    if (VBOX_FAILURE(rc))
+    if (RT_FAILURE(rc))
     {
         MMR3HeapFree(pReq);
         return rc;
@@ -407,15 +502,16 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
      * Initialize the packet and return it.
      */
     pReq->pNext    = NULL;
-    pReq->pVM      = pVM;
+    pReq->pUVM     = pUVM;
     pReq->enmState = VMREQSTATE_ALLOCATED;
     pReq->iStatus  = VERR_VM_REQUEST_STATUS_STILL_PENDING;
     pReq->fEventSemClear = true;
     pReq->fFlags   = VMREQFLAGS_VBOX_STATUS;
     pReq->enmType  = enmType;
+    pReq->enmDest  = enmDest;
 
     *ppReq = pReq;
-    STAM_COUNTER_INC(&pVM->vm.s.StatReqAllocNew);
+    STAM_COUNTER_INC(&pUVM->vm.s.StatReqAllocNew);
     LogFlow(("VMR3ReqAlloc: returns VINF_SUCCESS *ppReq=%p new\n", pReq));
     return VINF_SUCCESS;
 }
@@ -429,7 +525,7 @@ VMR3DECL(int) VMR3ReqAlloc(PVM pVM, PVMREQ *ppReq, VMREQTYPE enmType)
  * @param   pReq            Package to free.
  * @remark  The request packet must be in allocated or completed state!
  */
-VMR3DECL(int) VMR3ReqFree(PVMREQ pReq)
+VMMR3DECL(int) VMR3ReqFree(PVMREQ pReq)
 {
     /*
      * Ignore NULL (all free functions should do this imho).
@@ -437,8 +533,6 @@ VMR3DECL(int) VMR3ReqFree(PVMREQ pReq)
     if (!pReq)
         return VINF_SUCCESS;
 
-    STAM_COUNTER_INC(&pReq->pVM->vm.s.StatReqFree);
-
     /*
      * Check packet state.
      */
@@ -459,11 +553,13 @@ VMR3DECL(int) VMR3ReqFree(PVMREQ pReq)
     pReq->iStatus  = VERR_VM_REQUEST_STATUS_FREED;
     pReq->enmType  = VMREQTYPE_INVALID;
 
-    PVM pVM = pReq->pVM;
-    if (pVM->vm.s.cReqFree < 128)
+    PUVM pUVM = pReq->pUVM;
+    STAM_COUNTER_INC(&pUVM->vm.s.StatReqFree);
+
+    if (pUVM->vm.s.cReqFree < 128)
     {
-        ASMAtomicIncU32(&pVM->vm.s.cReqFree);
-        PVMREQ volatile *ppHead = &pVM->vm.s.apReqFree[ASMAtomicIncU32(&pVM->vm.s.iReqFree) % ELEMENTS(pVM->vm.s.apReqFree)];
+        ASMAtomicIncU32(&pUVM->vm.s.cReqFree);
+        PVMREQ volatile *ppHead = &pUVM->vm.s.apReqFree[ASMAtomicIncU32(&pUVM->vm.s.iReqFree) % RT_ELEMENTS(pUVM->vm.s.apReqFree)];
         PVMREQ pNext;
         do
         {
@@ -473,7 +569,7 @@ VMR3DECL(int) VMR3ReqFree(PVMREQ pReq)
     }
     else
     {
-        STAM_COUNTER_INC(&pReq->pVM->vm.s.StatReqFreeOverflow);
+        STAM_COUNTER_INC(&pReq->pUVM->vm.s.StatReqFreeOverflow);
         RTSemEventDestroy(pReq->EventSem);
         MMR3HeapFree(pReq);
     }
@@ -499,35 +595,86 @@ VMR3DECL(int) VMR3ReqFree(PVMREQ pReq)
  *                          be completed. Use RT_INDEFINITE_WAIT to only
  *                          wait till it's completed.
  */
-VMR3DECL(int) VMR3ReqQueue(PVMREQ pReq, unsigned cMillies)
+VMMR3DECL(int) VMR3ReqQueue(PVMREQ pReq, unsigned cMillies)
 {
     LogFlow(("VMR3ReqQueue: pReq=%p cMillies=%d\n", pReq, cMillies));
     /*
      * Verify the supplied package.
      */
-    if (pReq->enmState != VMREQSTATE_ALLOCATED)
-    {
-        AssertMsgFailed(("Invalid state %d\n", pReq->enmState));
-        return VERR_VM_REQUEST_STATE;
-    }
-    if (   !pReq->pVM
-        ||  pReq->pNext
-        ||  !pReq->EventSem)
-    {
-        AssertMsgFailed(("Invalid request package! Anyone cooking their own packages???\n"));
-        return VERR_VM_REQUEST_INVALID_PACKAGE;
-    }
-    if (    pReq->enmType < VMREQTYPE_INVALID
-        || pReq->enmType > VMREQTYPE_MAX)
-    {
-        AssertMsgFailed(("Invalid package type %d valid range %d-%d inclusivly. This was verified on alloc too...\n",
-                         pReq->enmType, VMREQTYPE_INVALID + 1, VMREQTYPE_MAX - 1));
-        return VERR_VM_REQUEST_INVALID_TYPE;
-    }
+    AssertMsgReturn(pReq->enmState == VMREQSTATE_ALLOCATED, ("%d\n", pReq->enmState), VERR_VM_REQUEST_STATE);
+    AssertMsgReturn(    VALID_PTR(pReq->pUVM)
+                    &&  !pReq->pNext
+                    &&  pReq->EventSem != NIL_RTSEMEVENT,
+                    ("Invalid request package! Anyone cooking their own packages???\n"),
+                    VERR_VM_REQUEST_INVALID_PACKAGE);
+    AssertMsgReturn(    pReq->enmType > VMREQTYPE_INVALID
+                    &&  pReq->enmType < VMREQTYPE_MAX,
+                    ("Invalid package type %d valid range %d-%d inclusivly. This was verified on alloc too...\n",
+                     pReq->enmType, VMREQTYPE_INVALID + 1, VMREQTYPE_MAX - 1),
+                    VERR_VM_REQUEST_INVALID_TYPE);
 
-    int rc = VINF_SUCCESS;
-    if (pReq->pVM->NativeThreadEMT != RTThreadNativeSelf())
+    /*
+     * Are we the EMT or not?
+     * Also, store pVM (and fFlags) locally since pReq may be invalid after queuing it.
+     */
+    int     rc      = VINF_SUCCESS;
+    PUVM    pUVM    = ((VMREQ volatile *)pReq)->pUVM;                 /* volatile paranoia */
+    PUVMCPU pUVMCPU = (PUVMCPU)RTTlsGet(pUVM->vm.s.idxTLS);
+
+    if (pReq->enmDest == VMREQDEST_BROADCAST)
     {
+        unsigned fFlags = ((VMREQ volatile *)pReq)->fFlags;     /* volatile paranoia */
+
+        for (unsigned i=0;i<pUVM->pVM->cCPUs;i++)
+        {
+            if (   !pUVMCPU
+                ||  pUVMCPU->idCpu != i)
+            {
+                /*
+                 * Insert it.
+                 */
+                pReq->enmState = VMREQSTATE_QUEUED;
+                PVMREQ pNext;
+                do
+                {
+                    pNext = pUVM->aCpus[i].vm.s.pReqs;
+                    pReq->pNext = pNext;
+                } while (!ASMAtomicCmpXchgPtr((void * volatile *)&pUVM->aCpus[i].vm.s.pReqs, (void *)pReq, (void *)pNext));
+
+                /*
+                 * Notify EMT.
+                 */
+                if (pUVM->pVM)
+                    VMCPU_FF_SET(pUVM->pVM, VM_FF_REQUEST, i);
+                /* @todo: VMR3NotifyFFU*/
+                AssertFailed();
+                VMR3NotifyFFU(pUVM, false);
+
+                /*
+                 * Wait and return.
+                 */
+                if (!(fFlags & VMREQFLAGS_NO_WAIT))
+                    rc = VMR3ReqWait(pReq, cMillies);
+                LogFlow(("VMR3ReqQueue: returns %Rrc\n", rc));
+            }
+            else
+            {
+                /*
+                 * The requester was EMT, just execute it.
+                 */
+                pReq->enmState = VMREQSTATE_QUEUED;
+                rc = vmR3ReqProcessOneU(pUVM, pReq);
+                LogFlow(("VMR3ReqQueue: returns %Rrc (processed)\n", rc));
+            }
+        } /* for each VMCPU */
+    }
+    else
+    if (    pReq->enmDest  != VMREQDEST_ANY  /* for a specific VMCPU? */
+        &&  pUVMCPU->idCpu != (unsigned)pReq->enmDest)
+    {
+        RTCPUID  idTarget = (RTCPUID)pReq->enmDest;
+        unsigned fFlags = ((VMREQ volatile *)pReq)->fFlags;     /* volatile paranoia */
+
         /*
          * Insert it.
          */
@@ -535,31 +682,66 @@ VMR3DECL(int) VMR3ReqQueue(PVMREQ pReq, unsigned cMillies)
         PVMREQ pNext;
         do
         {
-            pNext = pReq->pVM->vm.s.pReqs;
+            pNext = pUVM->aCpus[idTarget].vm.s.pReqs;
             pReq->pNext = pNext;
-        } while (!ASMAtomicCmpXchgPtr((void * volatile *)&pReq->pVM->vm.s.pReqs, (void *)pReq, (void *)pNext));
+        } while (!ASMAtomicCmpXchgPtr((void * volatile *)&pUVM->aCpus[idTarget].vm.s.pReqs, (void *)pReq, (void *)pNext));
 
         /*
          * Notify EMT.
          */
-        VM_FF_SET(pReq->pVM, VM_FF_REQUEST);
-        VMR3NotifyFF(pReq->pVM, false);
+        if (pUVM->pVM)
+            VMCPU_FF_SET(pUVM->pVM, VM_FF_REQUEST, idTarget);
+        /* @todo: VMR3NotifyFFU*/
+        AssertFailed();
+        VMR3NotifyFFU(pUVM, false);
 
         /*
          * Wait and return.
          */
-        if (!(pReq->fFlags & VMREQFLAGS_NO_WAIT))
+        if (!(fFlags & VMREQFLAGS_NO_WAIT))
             rc = VMR3ReqWait(pReq, cMillies);
-        LogFlow(("VMR3ReqQueue: returns %Vrc\n", rc));
+        LogFlow(("VMR3ReqQueue: returns %Rrc\n", rc));
+    }
+    else if (    pReq->enmDest == VMREQDEST_ANY
+             &&  !pUVMCPU /* only EMT threads have a valid pointer stored in the TLS slot. */)
+    {
+        unsigned fFlags = ((VMREQ volatile *)pReq)->fFlags;     /* volatile paranoia */
+
+        /*
+         * Insert it.
+         */
+        pReq->enmState = VMREQSTATE_QUEUED;
+        PVMREQ pNext;
+        do
+        {
+            pNext = pUVM->vm.s.pReqs;
+            pReq->pNext = pNext;
+        } while (!ASMAtomicCmpXchgPtr((void * volatile *)&pUVM->vm.s.pReqs, (void *)pReq, (void *)pNext));
+
+        /*
+         * Notify EMT.
+         */
+        if (pUVM->pVM)
+            VM_FF_SET(pUVM->pVM, VM_FF_REQUEST);
+        VMR3NotifyFFU(pUVM, false);
+
+        /*
+         * Wait and return.
+         */
+        if (!(fFlags & VMREQFLAGS_NO_WAIT))
+            rc = VMR3ReqWait(pReq, cMillies);
+        LogFlow(("VMR3ReqQueue: returns %Rrc\n", rc));
     }
     else
     {
+        Assert(pUVMCPU);
+
         /*
          * The requester was EMT, just execute it.
          */
         pReq->enmState = VMREQSTATE_QUEUED;
-        rc = vmR3ReqProcessOne(pReq->pVM, pReq);
-        LogFlow(("VMR3ReqQueue: returns %Vrc (processed)\n", rc));
+        rc = vmR3ReqProcessOneU(pUVM, pReq);
+        LogFlow(("VMR3ReqQueue: returns %Rrc (processed)\n", rc));
     }
     return rc;
 }
@@ -575,38 +757,35 @@ VMR3DECL(int) VMR3ReqQueue(PVMREQ pReq, unsigned cMillies)
  * @param   cMillies        Number of milliseconds to wait.
  *                          Use RT_INDEFINITE_WAIT to only wait till it's completed.
  */
-VMR3DECL(int) VMR3ReqWait(PVMREQ pReq, unsigned cMillies)
+VMMR3DECL(int) VMR3ReqWait(PVMREQ pReq, unsigned cMillies)
 {
     LogFlow(("VMR3ReqWait: pReq=%p cMillies=%d\n", pReq, cMillies));
 
     /*
      * Verify the supplied package.
      */
-    if (    pReq->enmState != VMREQSTATE_QUEUED
-        &&  pReq->enmState != VMREQSTATE_PROCESSING
-        &&  pReq->enmState != VMREQSTATE_COMPLETED)
-    {
-        AssertMsgFailed(("Invalid state %d\n", pReq->enmState));
-        return VERR_VM_REQUEST_STATE;
-    }
-    if (    !pReq->pVM
-        ||  !pReq->EventSem)
-    {
-        AssertMsgFailed(("Invalid request package! Anyone cooking their own packages???\n"));
-        return VERR_VM_REQUEST_INVALID_PACKAGE;
-    }
-    if (    pReq->enmType < VMREQTYPE_INVALID
-        ||  pReq->enmType > VMREQTYPE_MAX)
-    {
-        AssertMsgFailed(("Invalid package type %d valid range %d-%d inclusivly. This was verified on alloc and queue too...\n",
-                         pReq->enmType, VMREQTYPE_INVALID + 1, VMREQTYPE_MAX - 1));
-        return VERR_VM_REQUEST_INVALID_TYPE;
-    }
+    AssertMsgReturn(    pReq->enmState == VMREQSTATE_QUEUED
+                    ||  pReq->enmState == VMREQSTATE_PROCESSING
+                    ||  pReq->enmState == VMREQSTATE_COMPLETED,
+                    ("Invalid state %d\n", pReq->enmState),
+                    VERR_VM_REQUEST_STATE);
+    AssertMsgReturn(    VALID_PTR(pReq->pUVM)
+                    &&  pReq->EventSem != NIL_RTSEMEVENT,
+                    ("Invalid request package! Anyone cooking their own packages???\n"),
+                    VERR_VM_REQUEST_INVALID_PACKAGE);
+    AssertMsgReturn(    pReq->enmType > VMREQTYPE_INVALID
+                    &&  pReq->enmType < VMREQTYPE_MAX,
+                    ("Invalid package type %d valid range %d-%d inclusivly. This was verified on alloc too...\n",
+                     pReq->enmType, VMREQTYPE_INVALID + 1, VMREQTYPE_MAX - 1),
+                    VERR_VM_REQUEST_INVALID_TYPE);
 
     /*
      * Check for deadlock condition
      */
-    AssertMsg(!VMMR3LockIsOwner(pReq->pVM), ("Waiting for EMT to process a request, but we own the global VM lock!?!?!?!\n"));
+    PUVM pUVM = pReq->pUVM;
+    NOREF(pUVM);
+    AssertMsg(!pUVM->pVM || !VMMR3LockIsOwner(pUVM->pVM),
+              ("Waiting for EMT to process a request, but we own the global VM lock!?!?!?!\n"));
 
     /*
      * Wait on the package.
@@ -623,11 +802,11 @@ VMR3DECL(int) VMR3ReqWait(PVMREQ pReq, unsigned cMillies)
         } while (   pReq->enmState != VMREQSTATE_COMPLETED
                  && pReq->enmState != VMREQSTATE_INVALID);
     }
-    if (VBOX_SUCCESS(rc))
+    if (RT_SUCCESS(rc))
         ASMAtomicXchgSize(&pReq->fEventSemClear, true);
     if (pReq->enmState == VMREQSTATE_COMPLETED)
         rc = VINF_SUCCESS;
-    LogFlow(("VMR3ReqWait: returns %Vrc\n", rc));
+    LogFlow(("VMR3ReqWait: returns %Rrc\n", rc));
     Assert(rc != VERR_INTERRUPTED);
     return rc;
 }
@@ -636,16 +815,17 @@ VMR3DECL(int) VMR3ReqWait(PVMREQ pReq, unsigned cMillies)
 /**
  * Process pending request(s).
  *
- * This function is called from a forced action handler in
- * the EMT.
+ * This function is called from a forced action handler in the EMT
+ * or from one of the EMT loops.
  *
  * @returns VBox status code.
  *
- * @param   pVM         VM handle.
+ * @param   pUVM            Pointer to the user mode VM structure.
+ * @param   enmDest         Destination of the request packet (global or per VCPU).
  */
-VMR3DECL(int) VMR3ReqProcess(PVM pVM)
+VMMR3DECL(int) VMR3ReqProcessU(PUVM pUVM, VMREQDEST enmDest)
 {
-    LogFlow(("VMR3ReqProcess: (enmVMState=%d)\n", pVM->enmVMState));
+    LogFlow(("VMR3ReqProcessU: (enmVMState=%d) enmDest=%d\n", pUVM->pVM ? pUVM->pVM->enmVMState : VMSTATE_CREATING, enmDest));
 
     /*
      * Process loop.
@@ -659,8 +839,21 @@ VMR3DECL(int) VMR3ReqProcess(PVM pVM)
         /*
          * Get pending requests.
          */
-        VM_FF_CLEAR(pVM, VM_FF_REQUEST);
-        PVMREQ pReqs = (PVMREQ)ASMAtomicXchgPtr((void * volatile *)&pVM->vm.s.pReqs, NULL);
+        void *volatile *ppReqs;
+        if (enmDest == VMREQDEST_ANY)
+        {
+            ppReqs = (void * volatile *)&pUVM->vm.s.pReqs;
+            if (RT_LIKELY(pUVM->pVM))
+                VM_FF_CLEAR(pUVM->pVM, VM_FF_REQUEST);
+        }
+        else
+        {
+            ppReqs = (void * volatile *)&pUVM->aCpus[enmDest].vm.s.pReqs;
+            if (RT_LIKELY(pUVM->pVM))
+                VMCPU_FF_CLEAR(pUVM->pVM, enmDest, VM_FF_REQUEST);
+        }
+
+        PVMREQ pReqs = (PVMREQ)ASMAtomicXchgPtr(ppReqs, NULL);
         if (!pReqs)
             break;
 
@@ -674,7 +867,7 @@ VMR3DECL(int) VMR3ReqProcess(PVM pVM)
         while (pReq)
         {
             Assert(pReq->enmState == VMREQSTATE_QUEUED);
-            Assert(pReq->pVM == pVM);
+            Assert(pReq->pUVM == pUVM);
             PVMREQ pCur = pReq;
             pReq = pReq->pNext;
             pCur->pNext = pReqs;
@@ -696,7 +889,7 @@ VMR3DECL(int) VMR3ReqProcess(PVM pVM)
             pReq->pNext = NULL;
 
             /* Process the request */
-            int rc2 = vmR3ReqProcessOne(pVM, pReq);
+            int rc2 = vmR3ReqProcessOneU(pUVM, pReq);
 
             /*
              * The status code handling extremely important yet very fragile. Should probably
@@ -710,7 +903,7 @@ VMR3DECL(int) VMR3ReqProcess(PVM pVM)
         }
     }
 
-    LogFlow(("VMR3ReqProcess: returns %Vrc (enmVMState=%d)\n", rc, pVM->enmVMState));
+    LogFlow(("VMR3ReqProcess: returns %Rrc (enmVMState=%d)\n", rc, pUVM->pVM ? pUVM->pVM->enmVMState : VMSTATE_CREATING));
     return rc;
 }
 
@@ -723,7 +916,7 @@ VMR3DECL(int) VMR3ReqProcess(PVM pVM)
  * @param   pVM         VM handle.
  * @param   pReq        Request packet to process.
  */
-static int  vmR3ReqProcessOne(PVM pVM, PVMREQ pReq)
+static int  vmR3ReqProcessOneU(PUVM pUVM, PVMREQ pReq)
 {
     LogFlow(("vmR3ReqProcessOne: pReq=%p type=%d fFlags=%#x\n", pReq, pReq->enmType, pReq->fFlags));
 
@@ -839,18 +1032,18 @@ static int  vmR3ReqProcessOne(PVM pVM, PVMREQ pReq)
     if (pReq->fFlags & VMREQFLAGS_NO_WAIT)
     {
         /* Free the packet, nobody is waiting. */
-        LogFlow(("vmR3ReqProcessOne: Completed request %p: rcReq=%Vrc rcRet=%Vrc - freeing it\n",
+        LogFlow(("vmR3ReqProcessOne: Completed request %p: rcReq=%Rrc rcRet=%Rrc - freeing it\n",
                  pReq, rcReq, rcRet));
         VMR3ReqFree(pReq);
     }
     else
     {
         /* Notify the waiter and him free up the packet. */
-        LogFlow(("vmR3ReqProcessOne: Completed request %p: rcReq=%Vrc rcRet=%Vrc - notifying waiting thread\n",
+        LogFlow(("vmR3ReqProcessOne: Completed request %p: rcReq=%Rrc rcRet=%Rrc - notifying waiting thread\n",
                  pReq, rcReq, rcRet));
         ASMAtomicXchgSize(&pReq->fEventSemClear, false);
         int rc2 = RTSemEventSignal(pReq->EventSem);
-        if (VBOX_FAILURE(rc2))
+        if (RT_FAILURE(rc2))
         {
             AssertRC(rc2);
             rcRet = rc2;
