@@ -1,18 +1,22 @@
-/* $Id: tstAsmStructs.cpp 23517 2007-08-07 17:07:59Z umoeller $ */
+/* $Id: tstAsmStructs.cpp 15236 2008-12-10 09:50:10Z vboxsync $ */
 /** @file
  * Testcase for checking offsets in the assembly structures shared with C/C++.
  */
 
 /*
- * Copyright (C) 2006-2007 innotek GmbH
+ * Copyright (C) 2006-2007 Sun Microsystems, Inc.
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
  * you can redistribute it and/or modify it under the terms of the GNU
- * General Public License as published by the Free Software Foundation,
- * in version 2 as it comes in the "COPYING" file of the VirtualBox OSE
- * distribution. VirtualBox OSE is distributed in the hope that it will
- * be useful, but WITHOUT ANY WARRANTY of any kind.
+ * General Public License (GPL) as published by the Free Software
+ * Foundation, in version 2 as it comes in the "COPYING" file of the
+ * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
+ * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
+ *
+ * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
+ * Clara, CA 95054 USA or visit http://www.sun.com if you need
+ * additional information or have any questions.
  */
 
 /*******************************************************************************
@@ -22,6 +26,7 @@
 #include "CPUMInternal.h"
 #include <VBox/trpm.h>
 #include "TRPMInternal.h"
+#include "HWACCMInternal.h"
 #include "../VMMSwitcher/VMMSwitcher.h"
 #include "VMMInternal.h"
 #include <VBox/vm.h>
